/*
** Arc and wedge stuff.
**
** These work by generating regions that describe the desired shape and then
** calling the appropriate region drawing routine.  Not the quickest way to
** do it by any means, but fairly easy to implement.
**
** For arcs/wedges of less than 180 degrees of coverage this first builds an
** oval shaped region, then builds wedge-shaped (half a trapezoid) regions (2
** of 'em), and then trims the oval with these two regions.  For arcs/wedges
** of more than 180 degrees we build two oval regions, build the two trimming
** regions, trim the ovals with one each, then merge the results.
*/

/*#define DEBUG                 /* DEFINED for debugging statements. */

#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include "../line/cdline.h"
#include "../rounds/cdoval.h"
#include "cdarc.h"
#include <alloca.h>

/* Build a line for trimming arc-shaped regions with. */

static void
lineforarc(angle, inclusive, rect, p1, p2)
short angle;                    /* Angle to build line for. */
short inclusive;                /* TRUE to include angle itself in line. */
Rect *rect;                     /* Bounding rectangle. */
short *p1;                      /* Where to build line control points. */
short *p2;                      /* Where to build line-shaped region. */
{
    short midx, midy;
    short quadrant;
    short remaindr;
    Fixed slope;
    short delta;
    Rect line;
    RgnPtr lineRgn;
    long divmod;

    divmod = DivModS(angle, 90); /* Native divide instruction, */
    quadrant = LoWord(divmod);  /* pick up quotient */
    remaindr = HiWord(divmod);  /* and remainder. */
    line = *rect;
    midx = (rect->edge.right + rect->edge.left) >> 1;
    midy = (rect->edge.bottom + rect->edge.top) >> 1;
    slope = SlopeFromAngle(angle);
    if (!(quadrant & 1))        /* Figure out proper slope. */
        slope = -slope;
    delta = Fix2Int(slope * (rect->edge.right - midx));
    line.edge.right = midx + delta;
    line.edge.left = midx - delta;

    if (inclusive) {
        switch (quadrant) {
          case 0:
            OffsetRect(&line, -1, -1);
            break;
          case 1:
            OffsetRect(&line, 1, -1);
            break;
          case 2:
            OffsetRect(&line, 1, 1);
            break;
          case 3:
            OffsetRect(&line, -1, 1);
            break;
        }
    }
    _BuildLine(&line, !(quadrant & 1), p1);
    _BuildLRgn(&line, !(quadrant & 1), p1, p2);
    lineRgn = (RgnPtr) p2;
    lineRgn->rgnBBox = *rect;
    if (!remaindr) {            /* Handle degenerate line cases. */
        switch (quadrant) {
          case 0:               /* 0 degrees. */
            lineRgn->rgnBBox.edge.right = midx;
            break;

          case 1:               /* 90 degrees. */
            lineRgn->rgnBBox.edge.bottom = midy;
            break;

          case 2:               /* 180 degrees. */
            lineRgn->rgnBBox.edge.left = midx;
            break;

          case 3:               /* 270 degrees. */
            lineRgn->rgnBBox.edge.top = midy;
            break;
        }
        lineRgn->rgnSize = sizeof(RegionHdr); /* For purely rectangular rgn. */
    } else {                    /* Smear the line region out into a wedge */
        short i, j, *p;         /* shaped region for trimming purposes. */

        p = &lineRgn->rgnData;
        if (quadrant == 1 || quadrant == 2) {
            i = rect->edge.right;
            j = rect->edge.left;
            while (*p != END_TAG) { /* Do each scanline of the region. */
                if (p[1] < j)   /* Force region into conformance. */
                    p[1] = j;
                if (p[1] > i)
                    p[1] = i;
                p[2] = i;       /* Stretch width by resetting the */
                p += 4;         /* ending X coordinate. */
            }
        } else {
            i = rect->edge.left;
            j = rect->edge.right;
            while (*p != END_TAG) { /* Do each scanline of the region. */
                if (p[2] > j)   /* Force region into conformance. */
                    p[2] = j;
                if (p[2] < i)
                    p[2] = i;
                p[1] = i;       /* Stretch width by resetting the */
                p += 4;         /* starting X coordinate. */
            }
        }
    }
}

/*
** doArc draws arcs with the specified drawing verb (via regions).
*/

void
_DrawArc(rect, frame, verb, startAngle, arcAngle, arg)
const Rect *rect;
short frame;
void (*verb)();                 /* Function to call with the result. */
short startAngle, arcAngle;
void * const arg;               /* Optional argument to the verb function. */
{
    short *p1;                  /* For creating arc points in. */
    short *p2;                  /* For creating dummy oval region in. */
    long size1, size2;          /* Sizes of above areas. */
    short *p3;                  /* For creating line points in. */
    short *p4;                  /* For creating dummy line region in. */
    short *p5;                  /* For creating line points in. */
    short *p6;                  /* For creating dummy line region in. */
    short endangle;
    RgnPtr ovlRgn, line1Rgn, line2Rgn;
    RgnHandle resultRgn, tempRgn;

    if (EmptyRect(rect))
        return;
    if (arcAngle < 0) {         /* Fold arguments to positive angles. */
        startAngle += arcAngle;
        arcAngle = -arcAngle;
    }
    startAngle = (short) startAngle % 360;
    if (startAngle < 0)
        startAngle += 360;
    endangle = (short) (startAngle + arcAngle + 180) % 360;
                                /* Now build the oval region. */
    size1 = (1 + rect->edge.bottom - rect->edge.top) * sizeof(short); /* Num */
    p1 = alloca(size1);         /* of arc points (maximum one per scanline). */
    _BuildArc(rect, p1);
    size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short); /* Size of */
    p2 = alloca(size2);         /* dummy region. */
    _BuildORgn(rect, p1, p2);
    ovlRgn = (RgnPtr) p2;
    resultRgn = NewRgn();
    CopyRgn(&ovlRgn, resultRgn);
    if (frame) {
        tempRgn = NewRgn();
        CopyRgn(resultRgn, tempRgn);
        InsetRgn(tempRgn, thePort->pnSize.h, thePort->pnSize.v);
        DiffRgn(resultRgn, tempRgn, resultRgn);
        DisposeRgn(tempRgn);
    }

    if (arcAngle < 360) {       /* Now build the first line region. */
        size1 = (1 + rect->edge.bottom - rect->edge.top) * sizeof(short);
        p3 = alloca(size1);     /* # of line points (max one per scan). */
        size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short); /* Size of */
        p4 = alloca(size2);     /* dummy region. */
        p5 = alloca(size1);     /* Ditto for 2nd set of line control pts */
        p6 = alloca(size2);     /* and its region. */
        lineforarc(startAngle, 1, rect, p3, p4);
        line1Rgn = (RgnPtr) p4;
        lineforarc(endangle, 0, rect, p5, p6);
        line2Rgn = (RgnPtr) p6;

        if (arcAngle > 180) {
            tempRgn = NewRgn();
            CopyRgn(resultRgn, tempRgn);
            DiffRgn(tempRgn, &line2Rgn, tempRgn);
            DiffRgn(resultRgn, &line1Rgn, resultRgn);
            UnionRgn(tempRgn, resultRgn, resultRgn);
            DisposeRgn(tempRgn);
        } else {
            DiffRgn(resultRgn, &line1Rgn, resultRgn);
            DiffRgn(resultRgn, &line2Rgn, resultRgn);
            if (arcAngle < 90) { /* Trim off long points that may hang out */
                Rect trimmer;   /* as artifacts of especially acute angles. */
                short quad1, quad2, midx, midy, didlx, didly;

                tempRgn = NewRgn();
                quad1 = (short) startAngle / 90; /* Find out if we trim */
                quad2 = (short) endangle / 90; /* with a quarter-section */
                quad2 = (quad1 == quad2); /* or a half-section. */
                trimmer = *rect;
                midx = (short) (trimmer.edge.right + trimmer.edge.left) / 2;
                midy = (short) (trimmer.edge.bottom + trimmer.edge.top) / 2;
                didlx = midx + ((trimmer.edge.right - trimmer.edge.left) & 1);
                didly = midy + ((trimmer.edge.bottom - trimmer.edge.top) & 1);
                switch (quad1) {
                  case 0:
                    trimmer.edge.left = midx;
                    if (quad2)
                        trimmer.edge.bottom = didly;
                    break;
                  case 1:
                    trimmer.edge.top = midy;
                    if (quad2)
                        trimmer.edge.left = midx;
                    break;
                  case 2:
                    trimmer.edge.right = didlx;
                    if (quad2)
                        trimmer.edge.top = midy;
                    break;
                  case 3:
                    trimmer.edge.bottom = didly;
                    if (quad2)
                        trimmer.edge.right = didlx;
                    break;
                }
                RectRgn(tempRgn, &trimmer);
                SectRgn(resultRgn, tempRgn, resultRgn);
                DisposeRgn(tempRgn);
            }
        }
        (*verb)(resultRgn, arg);
    } else                      /* Full circle. */
        (*verb)(resultRgn, arg);
    DisposeRgn(resultRgn);
}
