/*
** Program to take a binary file and build a single data item
** out of it.  The item will be named, and in the data segment
** unless -t is given, in which case it will end up in the text
** segment.  (This is for read-only items.)  The label is limited
** to 8 characters, and should include a leading underscore if the
** label is to be referred to from C.  If no explicit label is
** supplied, the filename will be used as a label (with an underscore
** prepended).
**
** Usage:  coffwrap [-t] [-d] [-n label] infile outfile
**
**	V1.0	Jim Cathey	9/3/92
**		Original version, based on something by Mark Conrad.
*/

#define VERSION "1.0"

#include <stdio.h>
#include <a.out.h>
#include <fcntl.h>

typedef unsigned char uchar;
typedef unsigned short ushort;

extern int errno;
int fd_out, fd_in, i;
long fsize, lseek();
uchar *malloc(), *bp;
char *progname, *mysprintf();

uchar text[SYMNMLEN] = {".text"};
uchar data[SYMNMLEN] = {".data"};
uchar bss[SYMNMLEN] = {".bss"};
uchar label[SYMNMLEN];
int segment = 2;		/* Data segment == 2, Text segment == 1. */

main(argc, argv)
int argc;
char *argv[];
{
    char *argp, buf[80];

    progname = argv[0];
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'd':		/* Use Data segment (the default). */
		segment = 2;
		break;

	      case 't':		/* Use Text segment instead of Data segment. */
		segment = 1;
		break;

	      case 'n':		/* Label name. */
		if (!*argp)
		    --argc, argp = *++argv;
		strncpy(label, argp, sizeof label);
		break;

	      default:
		usage(progname);
		break;
	    }
	} else {
	    break;
	}
    }
    if (argc < 2) {
	usage();
    }
    if (!label[0]) {		/* Supply default label? */
	strncpy(&label[1], argv[0], sizeof label - 1);
	label[0] = '_';
    }
			/* Ready the binary file for conversion to COFF/68k */
    if ((fd_in = open(argv[0], O_RDONLY, 0)) < 0) {
	mysprintf(buf, "%s: Open failure on file %s, error: %d\n", progname,
		  argv[0], errno);
	write(2, buf, strlen(buf));
	exit(2);
    }
    if (!(bp = malloc((fsize = lseek(fd_in, 0L, 2))))) {
	memerror();
    }
    lseek(fd_in, 0L, 0);	/* Rewind file to the beginning. */
    if (read(fd_in, bp, fsize) != fsize) {
	mysprintf(buf, "%s: Read error on file %s, error: %d\n", progname,
		  argv[0], errno);
	write(2, buf, strlen(buf));
	exit(2);
    }
    close(fd_in);
    if ((fd_out = open(argv[1], O_CREAT | O_TRUNC | O_WRONLY, 0640)) < 0) {
	mysprintf(buf, "%s: Can't create file %s, error: %d\n", progname,
		  argv[1], errno);
	write(2, buf, strlen(buf));
	exit(4);
    }
    build_filehdr();
    build_texthdr();
    build_datahdr();
    build_bsshdr();
    build_data();
    build_syms();
    close(fd_out);
    return 0;
}

build_filehdr()
{
    struct filehdr *fhp;

    if (!(fhp = (struct filehdr *) malloc(sizeof (struct filehdr)))) {
	memerror();
    }
    fhp->f_magic = 0x0195;	/* Initialize file header. */
    fhp->f_nscns = 3;
    fhp->f_timdat = time();	/* Date & time. */
    fhp->f_symptr = (long) (sizeof (struct filehdr)
			    + (3 * (sizeof (struct scnhdr))) + fsize);
    fhp->f_nsyms = 7;
    fhp->f_opthdr = 0;
    fhp->f_flags = 0x020c;

    if ((i = write(fd_out, fhp, sizeof (struct filehdr)))
	!= sizeof (struct filehdr)) {
	wrerror();
    }
}

build_texthdr()
{
    struct scnhdr *thp;

    if (!(thp = (struct scnhdr *) malloc(sizeof (struct scnhdr)))) {
	memerror();
    }
    memcpy(&thp->s_name[0], &text[0], sizeof text);
    thp->s_paddr = 0;		/* Initialize section header. */
    thp->s_vaddr = 0;
    if (segment == 1) {
	thp->s_size = fsize;
	thp->s_scnptr = (long) (sizeof (struct filehdr)
				+ (3 * (sizeof (struct scnhdr))));;
    } else {
	thp->s_size = 0;
	thp->s_scnptr = 0;
    }
    thp->s_relptr = 0;
    thp->s_lnnoptr = 0;
    thp->s_nreloc = 0;
    thp->s_nlnno = 0;
    thp->s_flags = 0x20;

    if ((i = write(fd_out, thp, sizeof (struct scnhdr)))
	!= sizeof (struct scnhdr)) {
	wrerror();
    }
}

build_datahdr()
{
    struct scnhdr *dhp;

    if (!(dhp = (struct scnhdr *) malloc(sizeof (struct scnhdr)))) {
	memerror();
    }
    memcpy(&dhp->s_name[0], &data[0], sizeof data);
    dhp->s_paddr = 0;
    dhp->s_vaddr = 0;
    if (segment == 2) {
	dhp->s_size = fsize;
	dhp->s_scnptr = (long) (sizeof (struct filehdr)
				+ (3 * (sizeof (struct scnhdr))));;
    } else {
	dhp->s_size = 0;
	dhp->s_scnptr = 0;
    }
    dhp->s_relptr = 0;
    dhp->s_lnnoptr = 0;
    dhp->s_nreloc = 0;
    dhp->s_nlnno = 0;
    dhp->s_flags = 0x40;

    if ((i = write(fd_out, dhp, sizeof (struct scnhdr)))
	!= sizeof (struct scnhdr)) {
	wrerror();
    }
}

build_bsshdr()
{
    struct scnhdr *bhp;

    if (!(bhp = (struct scnhdr *) malloc(sizeof (struct scnhdr)))) {
	memerror();
    }
    memcpy(&bhp->s_name[0], &bss[0], sizeof bss);
    bhp->s_paddr = fsize;
    bhp->s_vaddr = fsize;
    bhp->s_size = 0;
    bhp->s_scnptr = 0;
    bhp->s_relptr = 0;
    bhp->s_lnnoptr = 0;
    bhp->s_nreloc = 0;
    bhp->s_nlnno = 0;
    bhp->s_flags = 0x80;

    if ((i = write(fd_out, bhp, sizeof (struct scnhdr)))
	!= sizeof (struct scnhdr)) {
	wrerror();
    }
}

build_data()
{
    if (write(fd_out, bp, fsize) != fsize) {
	wrerror();
    }
}

build_syms()
{
    struct syment *shp;
    union auxent *ahp;


    if (!(shp = (struct syment *) malloc(sizeof (struct syment)))) {
	memerror();
    }
    if (!(ahp = (union auxent *) malloc(sizeof (union auxent)))) {
	memerror();
    }
    memset(ahp, 0, sizeof (union auxent));
    memcpy(&shp->n_name[0], &text[0], sizeof text); /* .text symbol */
    shp->n_value = 0;
    shp->n_scnum = 0x0001;
    shp->n_type = 0;
    shp->n_sclass = 0x03;
    shp->n_numaux = 0x01;
    if ((i = write(fd_out, shp, sizeof (struct syment)))
	!= sizeof (struct syment)) {
	wrerror();
    }
    if ((i = write(fd_out, ahp, sizeof (union auxent)))
	!= sizeof (union auxent)) {
	wrerror();
    }
    memcpy(&shp->n_name[0], &data[0], sizeof data); /* .data symbol */
    shp->n_value = 0;
    shp->n_scnum = 0x0002;
    shp->n_type = 0;
    shp->n_sclass = 0x03;
    shp->n_numaux = 0x01;
    if ((i = write(fd_out, shp, sizeof (struct syment)))
	!= sizeof (struct syment)) {
	wrerror();
    }
    ahp->x_scn.x_scnlen = fsize;
    if ((i = write(fd_out, ahp, sizeof (union auxent)))
	!= sizeof (union auxent)) {
	wrerror();
    }
    memcpy(&shp->n_name[0], &bss[0], sizeof bss); /* .bss symbol */
    shp->n_value = fsize;
    shp->n_scnum = 0x0003;
    shp->n_type = 0;
    shp->n_sclass = 0x03;
    shp->n_numaux = 0x01;
    if ((i = write(fd_out, shp, sizeof (struct syment)))
	!= sizeof (struct syment)) {
	wrerror();
    }
    ahp->x_scn.x_scnlen = 0;
    if ((i = write(fd_out, ahp, sizeof (union auxent)))
	!= sizeof (union auxent)) {
	wrerror();
    }
    memcpy(&shp->n_name[0], &label[0], sizeof label); /* The symbolic name. */
    shp->n_value = 0;
    shp->n_scnum = segment;
    shp->n_type = 0;
    shp->n_sclass = 0x02;
    shp->n_numaux = 0;
    if ((i = write(fd_out, shp, sizeof (struct syment)))
	!= sizeof (struct syment)) {
	wrerror();
    }
}

memerror()
{
    char buf[80];

    mysprintf(buf, "%s: Memory allocation error: %d\n", progname, errno);
    write(2, buf, strlen(buf));
    exit(3);
}

wrerror()
{
    char buf[80];

    mysprintf(buf, "%s: File write error: %d\n", progname, errno);
    write(2, buf, strlen(buf));
    exit(4);
}

usage()
{
    char buf[80];

    mysprintf(buf, "%s: Version %s\n", progname, VERSION);
    write(1, buf, strlen(buf));
    mysprintf(buf, "Usage: %s [-t] [-d] [-n label] in_file out_file\n", progname);
    write(1, buf, strlen(buf));
    mysprintf(buf, "\twhere in_file = a binary file, and out_file = a 68k coff object file.\n");
    write(1, buf, strlen(buf));
    mysprintf(buf, "\tThe -t puts data into the text segment instead of the data segment,\n");
    write(1, buf, strlen(buf));
    mysprintf(buf, "\t(which is what -d does).  If no label is supplied with -n,\n");
    write(1, buf, strlen(buf));
    mysprintf(buf, "\tone will be made out of an underscore and the input filename.\n");
    write(1, buf, strlen(buf));
    exit(1);
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %[1..9]s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.  Returns pointer to the NULL.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {            /* While there's some format string. */
	if (*p == '%') {    /* Is it a format control? */
	    p++;            /* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    if (ndigits)
		while (*p2 && ndigits--) /* Copy the inclusion. */
		    *q++ = *p2++;
	    else
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
