/*
** getcwd -- get current working directory name (POSIX and SVID compatible)
**
** last edit: 21-Sep-1987 D A Gwyn
**
** This public-domain getcwd() routine can be used to replace the UNIX
** System V library routine (which uses popen() to capture the output of
** the "pwd" command).  Once that is done, "pwd" can be reimplemented as
** just puts(getcwd()).
**
** This implementation depends on every directory having entries for
** "." and "..".  It also depends on the internals of the <dirent.h>
** data structures to some degree.
**
** I considered using chdir() to ascend the hierarchy, followed by a
** final chdir() to the path being returned by getcwd() to restore the
** location, but decided that error recovery was too difficult that way.
** The algorithm I settled on was inspired by my rewrite of the "pwd"
** utility, combined with the dotdots[] array trick from the SVR2 shell.
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#ifdef FIO
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <errno.h>

typedef void *pointer;		/* (void *) if you have it */

extern void free();
extern pointer malloc();
extern int fstat(), stat();

#ifndef NAME_MAX
#define	NAME_MAX (255)		/* maximum directory entry size */
#endif

char *
getcwd(buf, size)		/* returns pointer to CWD pathname */
char *buf;			/* where to put name (NULL to malloc) */
int size;			/* size of buf[] or malloc()ed memory */
{
    static char	dotdots[] =
"../../../../../../../../../../../../../../../../../../../../../../../../../..";
    char *dotdot;		/* -> dotdots[.], right to left */
#ifdef FIO
    register FILE *fp;
    struct direct fpdir;
    struct direct *dir = &fpdir;
#else
    DIR *dirp;			/* -> parent directory stream */
    struct dirent *dir;		/* -> directory entry */
#endif
    struct stat	stat1, stat2;	/* info from stat() */
    struct stat	*d = &stat1;	/* -> info about "." */
    struct stat	*dd = &stat2;	/* -> info about ".." */
    register char *buffer;	/* local copy of buf, or malloc()ed */
    char *bufend;		/* -> buffer[size] */
    register char *endp;	/* -> end of reversed string */
    register char *dname;	/* entry name ("" for root) */
    int serrno = errno;		/* save entry errno */

    if (size == 0) {
	errno = EINVAL;		/* invalid argument */
	return NULL;
    }

    if ((buffer = buf) == NULL	/* wants us to malloc() the string */
	&& (buffer = (char *)malloc((unsigned)size)) == NULL) {
	errno = ENOMEM;		/* cannot malloc() specified size */
	return NULL;
    }

    if (stat(".", dd))
	goto error;		/* Prime the pump. */

    endp = buffer;		/* Initially, an empty string. */
    bufend = &buffer[size];

    for (dotdot = &dotdots[sizeof(dotdots)]; dotdot != dotdots;) {
	dotdot -= 3;		/* include one more "/.." section */
				/* (first time is actually "..") */

		/*
		** swap stat() info buffers
		*/
	{
	    register struct stat *temp = d;
	    d = dd;		/* new current dir is old parent dir */
	    dd = temp;
	}

#ifdef FIO
	if ((fp = fopen(dotdot, "r")) == NULL)
	    goto error;
	if (fstat(fileno(fp), dd))
#else
	    if ((dirp = opendir(dotdot)) == NULL)
		goto error;
	if (fstat(dirp->dd_fd, dd))
#endif
	    {
		serrno = errno;	/* set by fstat() */
#ifdef FIO
		(void) fclose(fp);
#else
		(void)closedir(dirp);
#endif
		errno = serrno;	/* in case closedir() clobbered it */
		goto error;
	    }
	if ( d->st_dev == dd->st_dev ) {
			/*
			** not crossing a mount point
			*/
	    if ( d->st_ino == dd->st_ino ) {
				/*
				** root directory
				*/
		dname = "";
		goto append;
	    }

	    do {
#ifdef FIO
		if (fread(&fpdir, sizeof(struct direct), 1, fp) != 1)
#else
		    if ((dir = readdir(dirp)) == NULL)
#endif
			{
#ifdef FIO
			    (void) fclose(fp);
#else
			    (void) closedir(dirp);
#endif
			    errno = ENOENT; /* missing entry */
			    goto error;
			}
	    } while (dir->d_ino != d->st_ino);
	} else {
			/*
			** crossing a mount point
			*/
	    struct stat	t;	/* info re. test entry */
	    char name[sizeof(dotdots) + 1 + NAME_MAX];

	    (void)strcpy(name, dotdot);
	    dname = &name[strlen(name)];
	    *dname++ = '/';

	    do {
#ifdef FIO
		if (fread(&fpdir, sizeof(struct direct), 1, fp) != 1)
#else
		    if ((dir = readdir(dirp)) == NULL)
#endif
			{
#ifdef FIO
			    (void)fclose(fp);
#else
			    (void)closedir( dirp );
#endif
			    errno = ENOENT; /* missing entry */
			    goto error;
			}
		(void)strcpy( dname, dir->d_name );
				/* must fit if NAME_MAX is not a lie */
	    } while (stat(name, &t) != 0
		     || t.st_ino != d->st_ino
		     || t.st_dev != d->st_dev);
	}
	dname = dir->d_name;

		/*
		** append "/" and reversed dname string onto buffer
		*/
      append:
#ifdef FIO
	(void)fclose(fp);
#else
	(void)closedir(dirp);
#endif
	if (endp != buffer	/* avoid trailing / in final name */
	    || dname[0] == '\0') /* but allow "/" when CWD is root */
	    *endp++ = '/';
	{
	    register char *app;	/* traverses dname string */

	    for (app = dname; *app != '\0'; ++app)
		;
	    if (app - dname >= bufend - endp) {
		errno = ERANGE;	/* won't fit allotted space */
		goto error;
	    }
	    while (app != dname)
		*endp++ = *--app;
	}
	if (dname[0] == '\0') {	/* reached root; wrap it up */
	    register char *startp;	/* -> buffer[.] */

	    *endp = '\0';	/* plant null terminator */

				/* straighten out reversed pathname string */
	    for (startp = buffer; --endp > startp; ++startp) {
		char temp = *endp;

		*endp = *startp;
		*startp = temp;
	    }
	    errno = serrno;	/* restore entry errno */
	    return buffer;
	}
    }
    errno = ENOMEM;		/* actually, algorithm failure */
  error:
    if (buf == NULL)
	free((pointer)buffer);
    return NULL;
}
