/*$
**
**  A quick hack of the X-windows program 'maze' for ISC's windowman
**
**     ===== H. Marc Lewis, Spokane, March-1992 =====
**
**  V1.2 5/17/93 DH
**
**  Improved error handling in check_for_key() so loss of the
**  connection to windowman doesn't cause core dump.
**
**  Respond to reset key from ISC keyboards by exiting.
**
**  Fixed core dump that happened when connection to windowman was lost
**  and call to GetHdwType() failed.
**
**  Fixed potential core dump if screen size was too big for statically
**  allocated arrays.
**
**  Removed an unnecessary CheckKey() during startup that didn't check
**  its return code and would cause a hang if the window connection
**  had gone down already (since you have only one chance to respond
**  to the connection lost error). Any extraneous key-up events are thrown
**  away in check_for_key().
*/

#include <isc/window.h>
#include <isc/keycodes.h>
#include <stdio.h>

#define SOLID		(Pattern *)0
#define GRAY		(Pattern *)7
#define BLANK		(Pattern *)(-1)
#define SPEED		2
#define PAUSE		3
#define OPTIONS		1

#define MIN_MAZE_SIZE	3
#define MAX_MAZE_SIZE_X	205
#define MAX_MAZE_SIZE_Y	205

#define MOVE_LIST_SIZE  (MAX_MAZE_SIZE_X * MAX_MAZE_SIZE_Y)

#define WALL_TOP	0x8000
#define WALL_RIGHT	0x4000
#define WALL_BOTTOM	0x2000
#define WALL_LEFT	0x1000

#define DOOR_IN_TOP	0x800
#define DOOR_IN_RIGHT	0x400
#define DOOR_IN_BOTTOM	0x200
#define DOOR_IN_LEFT	0x100
#define DOOR_IN_ANY	0xF00

#define DOOR_OUT_TOP	0x80
#define DOOR_OUT_RIGHT	0x40
#define DOOR_OUT_BOTTOM	0x20
#define DOOR_OUT_LEFT	0x10

#define START_SQUARE	0x2
#define END_SQUARE	0x1

#define SQ_SIZE_X	10
#define SQ_SIZE_Y       10

#define NUM_RANDOM      100

#define	BORDERWIDTH     2
#define	border_x        (5)
#define	border_y        (5)
#define	MIN_W	        200
#define	MIN_H	        200

#define	get_random(x)	(rand() % (x))

int trapqfd;
int wmfd;
int wid;
short flashing=1, first_color, second_color, third_color;

static int logo_x, logo_y;

static unsigned short maze[MAX_MAZE_SIZE_X][MAX_MAZE_SIZE_Y];

static struct {
    unsigned char x;
    unsigned char y;
    unsigned char dir;
} move_list[MOVE_LIST_SIZE],
  save_path[MOVE_LIST_SIZE],
  path[MOVE_LIST_SIZE];

static int maze_size_x, maze_size_y;
static int sqnum, cur_sq_x, cur_sq_y, path_length;
static int start_x, start_y, start_dir, end_x, end_y, end_dir;
static short keydelay = 30;
static short pause = 3;
static short last_pause = 4;
static short last_speed;
static short show_backtrace = 1;
static short Width;
static short Height;
static HdwrType hwinfo;

static int pulldnevent(EventRecord *ep);
static int windowevent(EventRecord *ep);
void draw_solid_square(int i, int j,
		       int dir, Pattern *pat);
void draw_wall(int i, int j, int dir);
void enter_square(int n);

int x = 0, y = 0, restart = 0, stop = 1, state = 1;

main(int argc, char *argv[])
{
    short l,r,t,b;
    char *cp;
    Rect win, erase;

    if (argc > 1) {
	keydelay = atoi(argv[1]);
	if (keydelay < 1)
	    keydelay = 1;
	if (keydelay > 500)
	    keydelay = 500;
    }
    /* Initialize graphics toolbox */
    if ((trapqfd = InitGraf()) < 0) {
 	exit(1);
    }
    /* Establish connection to a window manager */
    if ((wmfd = ConnectGraf(0)) < 0) {
	cp = "Cannot connect to Window Manager\n";
	write(2, cp, strlen(cp));
	exit(2);
    }
    /*	Create a window */
    GetHdwrType(&hwinfo);
    if (hwinfo.hdots == 0)	/* Lost connection. */
	exit(3);
    SetRect(&win, 0, 40, hwinfo.hdots, hwinfo.vdots);
    Width = hwinfo.hdots - 10;
    Height = hwinfo.vdots - 50;
    wid = NewWindow(&win,"maze v1.2, by HML",1,documentProc,-1,hasGoAway,wmfd);
    erase = win;
    OffsetRect(&erase,-erase.edge.left,-erase.edge.top);
    SetPort(wid);
    Event(menuEvt,(void (*)())pulldnevent);
    Event(windowEvt,(void (*)())windowevent);
    NewMenu(OPTIONS,"Options");
    AppendMenu(OPTIONS,"Flash endpoints;Show backtrace");
    CheckItem(OPTIONS,1,1);
    CheckItem(OPTIONS,2,1);
    NewMenu(SPEED,"Speed");
    AppendMenu(SPEED,"Very Slow;Slow;Normal;Fast;Very Fast");
    CheckItem(SPEED,3,1);
    last_speed = 3;
    NewMenu(PAUSE,"Pause");
    AppendMenu(PAUSE,"None;0.5 sec.;1.0 sec.;1.5 sec.;2.0 sec.;2.5 sec.");
    CheckItem(PAUSE,4,1);
    last_pause = 4;
    srand(getpid());
    set_maze_sizes(Width,Height);
    set_colors();
    for (;;) {
	initialize_maze();
	create_maze();
	draw_maze_border();
	solve_maze();
	ForeColor(1);
	BackColor(0);
	EraseRect(&erase);
    }
}


initialize_maze()         /* draw the surrounding wall and start/end squares */
{
    int i, j, wall;

    /* initialize all squares */
    for (i=0; i<maze_size_x; i++) {
	for (j=0; j<maze_size_y; j++) {
	    maze[i][j] = 0;
	}
    }

    /* top wall */
    for (i=0; i<maze_size_x; i++) {
	maze[i][0] |= WALL_TOP;
    }

    /* right wall */
    for (j=0; j<maze_size_y; j++) {
	maze[maze_size_x-1][j] |= WALL_RIGHT;
    }

    /* bottom wall */
    for (i=0; i<maze_size_x; i++) {
	maze[i][maze_size_y-1] |= WALL_BOTTOM;
    }

    /* left wall */
    for (j=0; j<maze_size_y; j++) {
	maze[0][j] |= WALL_LEFT;
    }

    /* set start square */
    wall = get_random(4);
    switch (wall) {

      case 0:
	i = get_random(maze_size_x);
	j = 0;
	break;

      case 1:
	i = maze_size_x - 1;
	j = get_random(maze_size_y);
	break;

      case 2:
	i = get_random(maze_size_x);
	j = maze_size_y - 1;
	break;

      case 3:
	i = 0;
	j = get_random(maze_size_y);
	break;
    }
    maze[i][j] |= START_SQUARE;
    maze[i][j] |= (DOOR_IN_TOP >> wall);
    maze[i][j] &= ~(WALL_TOP >> wall);
    cur_sq_x = i;
    cur_sq_y = j;
    start_x = i;
    start_y = j;
    start_dir = wall;
    sqnum = 0;

    /* set end square */
    wall = (wall + 2) % 4;
    switch (wall) {
      case 0:
	i = get_random(maze_size_x);
	j = 0;
	break;
      case 1:
	i = maze_size_x - 1;
	j = get_random(maze_size_y);
	break;
      case 2:
	i = get_random(maze_size_x);
	j = maze_size_y - 1;
	break;
      case 3:
	i = 0;
	j = get_random(maze_size_y);
	break;
    }
    maze[i][j] |= END_SQUARE;
    maze[i][j] |= (DOOR_OUT_TOP >> wall);
    maze[i][j] &= ~(WALL_TOP >> wall);
    end_x = i;
    end_y = j;
    end_dir = wall;
    check_for_key();
}



draw_maze_border()           /* draw the maze outline */
{
    short i, j;
    short trace_color;
    short false_color;

    trace_color = 8;
    false_color = 8;

    PenNormal();
    for (i = 0; i < maze_size_x; i++) {
	if (maze[i][0] & WALL_TOP) {
	    MoveTo(border_x + SQ_SIZE_X * i, border_y);
	    Line(SQ_SIZE_X, 0);
	}
	if ((maze[i][maze_size_y - 1] & WALL_BOTTOM)) {
	    MoveTo(border_x + SQ_SIZE_X * i,
		   border_y + SQ_SIZE_Y * (maze_size_y));
	    Line(SQ_SIZE_X, 0);
	}
    }
    check_for_key();
    for (j = 0; j < maze_size_y; j++) {
	if (maze[maze_size_x - 1][j] & WALL_RIGHT) {
	    MoveTo(border_x + SQ_SIZE_X * maze_size_x,
		   border_y + SQ_SIZE_Y * j);
	    Line(0, SQ_SIZE_Y);
	}
	if (maze[0][j] & WALL_LEFT) {
	    MoveTo(border_x, border_y + SQ_SIZE_Y * j);
	    Line(0, SQ_SIZE_Y);
	}
    }
    check_for_key();
    trace_color = different_color(trace_color);
    if ((trace_color & 1) == 0)		/* make it ODD */
	++trace_color;
    ForeColor(trace_color);
    first_color = theForeColor;
    second_color = third_color = -1;
    false_color = different_color(false_color);
    if (false_color & 1)		/* make it EVEN */
	false_color = --false_color & 0xF;
    BackColor(false_color);
    draw_solid_square(start_x, start_y, start_dir,SOLID);
    draw_solid_square(end_x, end_y, end_dir,SOLID);
    if (flashing) {
	j = trace_color;
	for (i=0; i<10; ++i) {
	    j ^= trace_color;
	    ForeColor(j);
	    draw_solid_square(start_x, start_y, start_dir,SOLID);
	    draw_solid_square(end_x, end_y, end_dir,SOLID);
	    FlushWindow();
	    nap(100);
	}
	check_for_key();
    }
}


set_maze_sizes(width, height)
{
    if ((maze_size_x = width / SQ_SIZE_X) > MAX_MAZE_SIZE_X)
	exit(4);		/* Screen too big. */
    if ((maze_size_y = height / SQ_SIZE_Y) > MAX_MAZE_SIZE_Y)
	exit(4);		/* Screen too big. */
}



create_maze()     /* create a maze layout given the initialized maze */
{
    int i, newdoor;

    for (;;) {
	move_list[sqnum].x = cur_sq_x;
	move_list[sqnum].y = cur_sq_y;
	move_list[sqnum].dir = newdoor;
	while ((newdoor = choose_door()) == -1) { /* pick a door */
	    if (backup() == -1) { /* no more doors ... backup */
		return; /* done ... return */
	    }
	}

	/* mark the out door */
	maze[cur_sq_x][cur_sq_y] |= (DOOR_OUT_TOP >> newdoor);

	switch (newdoor) {
	  case 0:
	    cur_sq_y--;
	    break;
	  case 1:
	    cur_sq_x++;
	    break;
	  case 2:
	    cur_sq_y++;
	    break;
	  case 3:
	    cur_sq_x--;
	    break;
	}
	sqnum++;

	/* mark the in door */
	maze[cur_sq_x][cur_sq_y] |= (DOOR_IN_TOP >> ((newdoor+2)%4));

	/* if end square set path length and save path */
	if (maze[cur_sq_x][cur_sq_y] & END_SQUARE) {
	    path_length = sqnum;
	    for (i=0; i<path_length; i++) {
		save_path[i].x = move_list[i].x;
		save_path[i].y = move_list[i].y;
		save_path[i].dir = move_list[i].dir;
	    }
	}
    }
}



backup()                                                  /* back up a move */
{
    sqnum--;
    cur_sq_x = move_list[sqnum].x;
    cur_sq_y = move_list[sqnum].y;
    return(sqnum);
}


choose_door()
{				/* pick a new path */
    unsigned short *p, current;
    int	num_candidates;
    int candidates[3];

    num_candidates = 0;

    p = &maze[cur_sq_x][cur_sq_y];
    current = *p;
  topwall:
    /* top wall */
    if (current & DOOR_IN_TOP)
	goto rightwall;
    if (current & DOOR_OUT_TOP)
	goto rightwall;
    if (current & WALL_TOP)
	goto rightwall;
    if (maze[cur_sq_x][cur_sq_y - 1] & DOOR_IN_ANY) {
	*p = (current |= WALL_TOP);
	maze[cur_sq_x][cur_sq_y - 1] |= WALL_BOTTOM;
	draw_wall(cur_sq_x, cur_sq_y, 0);
	goto rightwall;
    }
    candidates[num_candidates++] = 0;

  rightwall:
    /* right wall */
    if (current & DOOR_IN_RIGHT)
	goto bottomwall;
    if (current & DOOR_OUT_RIGHT)
	goto bottomwall;
    if (current & WALL_RIGHT)
	goto bottomwall;
    if (maze[cur_sq_x + 1][cur_sq_y] & DOOR_IN_ANY) {
	*p = (current |= WALL_RIGHT);
	maze[cur_sq_x + 1][cur_sq_y] |= WALL_LEFT;
	draw_wall(cur_sq_x, cur_sq_y, 1);
	goto bottomwall;
    }
    candidates[num_candidates++] = 1;

  bottomwall:
    /* bottom wall */
    if (current & DOOR_IN_BOTTOM)
	goto leftwall;
    if (current & DOOR_OUT_BOTTOM)
	goto leftwall;
    if (current & WALL_BOTTOM)
	goto leftwall;
    if (maze[cur_sq_x][cur_sq_y + 1] & DOOR_IN_ANY) {
	*p = (current |= WALL_BOTTOM);
	maze[cur_sq_x][cur_sq_y + 1] |= WALL_TOP;
	draw_wall(cur_sq_x, cur_sq_y, 2);
	goto leftwall;
    }
    candidates[num_candidates++] = 2;

  leftwall:
    /* left wall */
    if (current & DOOR_IN_LEFT)
	goto donewall;
    if (current & DOOR_OUT_LEFT)
	goto donewall;
    if (current & WALL_LEFT)
	goto donewall;
    if (maze[cur_sq_x - 1][cur_sq_y] & DOOR_IN_ANY) {
	*p = (current |= WALL_LEFT);
	maze[cur_sq_x - 1][cur_sq_y] |= WALL_RIGHT;
	draw_wall(cur_sq_x, cur_sq_y, 3);
	goto donewall;
    }
    candidates[num_candidates++] = 3;

  donewall:
    if (num_candidates == 0)
	return(-1);
    if (num_candidates == 1)
	return(candidates[0]);
    return(candidates[get_random(num_candidates)]);
}

/* Draw a single wall. */
void
draw_wall(int i, int j, int dir)
{
    switch (dir) {
      case 0:
	MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * j);
	Line(SQ_SIZE_X, 0);
	break;

      case 1:
	MoveTo(border_x + SQ_SIZE_X * (i + 1), border_y + SQ_SIZE_Y * j);
	Line(0, SQ_SIZE_Y);
	break;

      case 2:
	MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * (j + 1));
	Line(SQ_SIZE_X, 0);
	break;

      case 3:
	MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * j);
	Line(0, SQ_SIZE_Y);
	break;
    }
}

solve_maze()        /* solve it with graphical feedback */
{
    short i;
    EventRecord	*ep;
    int	ch;
    short every = 0;

    /* plug up the surrounding wall */
    maze[start_x][start_y] |= (WALL_TOP >> start_dir);
    maze[end_x][end_y] |= (WALL_TOP >> end_dir);

    /* initialize search path */
    i = 0;
    path[i].x = end_x;
    path[i].y = end_y;
    path[i].dir = -1;

    /* do it */
    for (;;) {
	if (++path[i].dir >= 4) {
	    i--;
	    draw_solid_square((int)(path[i].x), (int)(path[i].y),
			      (int)(path[i].dir),
			      (show_backtrace) ? GRAY : BLANK);
	}
	else if (!(maze[path[i].x][path[i].y] & (WALL_TOP >> path[i].dir))
		 && ((i == 0) || ((path[i].dir !=
				(int)(path[i-1].dir+2)%4)))) {
	    enter_square(i);
	    i++;
	    if (maze[path[i].x][path[i].y] & START_SQUARE) {
		FlushWindow();
		for (i=0; i<pause; ++i) {
		    nap(500);
		    if (check_for_key())
			return;
		}
		return;
	    }
	}
	if (++every >= keydelay) {
	    every = 0;
	    if (check_for_key())
		return;
	}
    }
}


/* move into a neighboring square */
void
enter_square(int n)
{
    draw_solid_square((int)path[n].x, (int)path[n].y,
		      (int)path[n].dir,SOLID);

    path[n+1].dir = -1;
    switch (path[n].dir) {
      case 0:
	path[n+1].x = path[n].x;
	path[n+1].y = path[n].y - 1;
	break;
      case 1:
	path[n+1].x = path[n].x + 1;
	path[n+1].y = path[n].y;
	break;
      case 2:
	path[n+1].x = path[n].x;
	path[n+1].y = path[n].y + 1;
	break;
      case 3:
	path[n+1].x = path[n].x - 1;
	path[n+1].y = path[n].y;
	break;
    }
}



void
draw_solid_square(int i, int j, int dir, Pattern *pat) /* draw a solid square
							  in a square */
{
    Rect sq;

    switch (dir) {
      case 0:
	SetRect(&sq,
		border_x + 3 + SQ_SIZE_X * i,
		border_y - 3 + SQ_SIZE_Y * j,
		border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X - 6,
		border_y - 3 + SQ_SIZE_Y * j + SQ_SIZE_Y);
	break;
      case 1:
	SetRect(&sq,
		border_x + 3 + SQ_SIZE_X * i,
		border_y + 3 + SQ_SIZE_Y * j,
		border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X,
		border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y - 6);
	break;
      case 2:
	SetRect(&sq,
		border_x + 3 + SQ_SIZE_X * i,
		border_y + 3 + SQ_SIZE_Y * j,
		border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X - 6,
		border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y);
	break;
      case 3:
	SetRect(&sq,
		border_x - 3 + SQ_SIZE_X * i,
		border_y + 3 + SQ_SIZE_Y * j,
		border_x - 3 + SQ_SIZE_X * i + SQ_SIZE_X,
		border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y - 6);
	break;
    }
    if (pat == (Pattern *)BLANK) {
	short save = theBackColor;
	BackColor(0);
	EraseRect(&sq);
	BackColor(save);
    }
    else
	FillRect(&sq,pat);
}


different_color(short old)
{
    short cl;

    cl = old;
    while (cl == old)
	cl = (rand() % 8) + 8;
    return(cl);
}

even_color()
{
    short cl;

    cl = (rand() % 8) + 8;
    if (cl & 1)
	--cl;
    return(cl & 0xF);
}


set_colors()
{
    SetRGBColor(8,3072,65535,38912);
    SetRGBColor(9,0,45056,0);
    SetRGBColor(10,22528,47104,65535);
    SetRGBColor(11,0,0,65535);
    SetRGBColor(12,65535,35840,58368);
    SetRGBColor(13,65535,0,0);
    SetRGBColor(14,63488,65535,19456);
    SetRGBColor(15,41984,0,65535);
}

check_for_key()
{
    EventRecord	*ep;
    int	ch;
    int i;

    i = CheckEvent(anyWindow, 1<<keyEvt);

    if (i == -1)
	exit(5);
    else if (i == 0)
	return(0);
    else {
	ep = GetEvent(anyWindow, 1<<keyEvt);

	if (ep == (EventRecord *) -1 || ep == NULL)
	    exit(5);
	if (ep->message & F_KEYUP)
	    return(0);		/* throw away key up events */
	if (ep->who != wid) {
	    return(0);
	}
	ch = ep->message & F_KEYMASK;
	if (ch == K_ESC || ch == K_RESET)
	    exit(0);
	return(1);
    }
}


/* ----------------------------- */
/*  Handle MENU PullDown events  */
/* ----------------------------- */

static int
pulldnevent(EventRecord *ep)
{
    int menuid, menuitem;
    int epmsg;

    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;
    epmsg = ep->message;
    switch(menuid) {

      case SPEED:
	CheckItem(SPEED,last_speed,0);
	switch(menuitem) {

	  case  1:	/* very slow */
	    CheckItem(SPEED,1,1);
	    keydelay = 1;
	    last_speed = 1;
	    break;

	  case  2:	/* slow */
	    CheckItem(SPEED,2,1);
	    keydelay = 5;
	    last_speed = 2;
	    break;

	  case  3:	/* normal */
	    CheckItem(SPEED,3,1);
	    keydelay = 10;
	    last_speed = 3;
	    break;

	  case  4:	/* fast */
	    CheckItem(SPEED,4,1);
	    keydelay = 30;
	    last_speed = 4;
	    break;

	  case  5:	/* very fast */
	    CheckItem(SPEED,5,1);
	    keydelay = 300;
	    last_speed = 5;
	    break;
	}
	break;

      case PAUSE:
	CheckItem(PAUSE,last_pause,0);
	switch(menuitem) {

	  case  1:	/* none */
	    CheckItem(PAUSE,1,1);
	    pause = 0;
	    last_pause = 1;
	    break;

	  case  2:	/* 0.5 sec */
	    CheckItem(PAUSE,2,1);
	    pause = 1;
	    last_pause = 2;
	    break;

	  case  3:	/* 1.0 sec */
	    CheckItem(PAUSE,3,1);
	    pause = 2;
	    last_pause = 3;
	    break;

	  case  4:	/* 1.5 sec */
	    CheckItem(PAUSE,4,1);
	    pause = 3;
	    last_pause = 4;
	    break;

	  case  5:	/* 2.0 sec */
	    CheckItem(PAUSE,5,1);
	    pause = 4;
	    last_pause = 5;
	    break;

	  case  6:	/* 2.5 sec */
	    CheckItem(PAUSE,6,1);
	    pause = 5;
	    last_pause = 6;
	    break;
	}
	break;

      case OPTIONS:
	switch(menuitem) {

	  case  1:	/* flashing option */
	    if (flashing)
		CheckItem(OPTIONS,1,0);
	    else
		CheckItem(OPTIONS,1,1);
	    flashing = !flashing;
	    break;

	  case  2:	/* show backtrace */
	    if (show_backtrace)
		CheckItem(OPTIONS,2,0);
	    else
		CheckItem(OPTIONS,2,1);
	    show_backtrace = !show_backtrace;
	    break;
	}
	break;
    }
    EndMenu(epmsg);
    return(0);
}


/* --------------------------------- */
/*  Handle Window Events, like exit  */
/* --------------------------------- */

int
windowevent(EventRecord *ep)
{
    Rect window;

    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	exit(0);

      default:
	return(0);
    }

}
