
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include <isc/dlgman.h>
#include <moview.h>

/*
 *  === MUCH OF THIS CODE STOLEN FROM line.c IN THE PROGRAM "less" ===
 *  === THIS COMMENT BLOCK IS FROM THE ORIGINAL AUTHOR ===
 * A ridiculously complex state machine takes care of backspaces 
 * when in BS_SPECIAL mode.  The complexity arises from the attempt
 * to deal with all cases, especially involving long lines with underlining,
 * boldfacing or whatever.  There are still some cases which will break it.
 *
 * There are four states:
 *	LN_NORMAL is the normal state (not in underline mode).
 *	LN_UNDERLINE means we are in underline mode.  We expect to get
 *		either a sequence like "_\bX" or "X\b_" to continue
 *		underline mode, or anything else to end underline mode.
 *	LN_BOLDFACE means we are in boldface mode.  We expect to get sequences
 *		like "X\bX\b...X\bX" to continue boldface mode, or anything
 *		else to end boldface mode.
 *	LN_UL_X means we are one character after LN_UNDERLINE
 *		(we have gotten the '_' in "_\bX" or the 'X' in "X\b_").
 *	LN_UL_XB means we are one character after LN_UL_X 
 *		(we have gotten the backspace in "_\bX" or "X\b_";
 *		we expect one more ordinary character, 
 *		which will put us back in state LN_UNDERLINE).
 *	LN_BO_X means we are one character after LN_BOLDFACE
 *		(we have gotten the 'X' in "X\bX").
 *	LN_BO_XB means we are one character after LN_BO_X
 *		(we have gotten the backspace in "X\bX";
 *		we expect one more 'X' which will put us back
 *		in LN_BOLDFACE).
 */
static int ln_state;		/* Currently in normal/underline/bold/etc mode? */
#define	LN_NORMAL	0	/* Not in underline, boldface or whatever mode */
#define	LN_UNDERLINE	1	/* In underline, need next char */
#define	LN_UL_X		2	/* In underline, got char, need \b */
#define	LN_UL_XB	3	/* In underline, got char & \b, need one more */
#define	LN_BOLDFACE	4	/* In boldface, need next char */
#define	LN_BO_X		5	/* In boldface, got char, need \b */
#define	LN_BO_XB	6	/* In boldface, got char & \b, need same char */

/*
 * Append a character to the line buffer.
 * Handle underlining, boldfacing, etc.
 */


unsigned char *pappend(register unsigned char *, register int);

unsigned char *
pappend(unsigned char *curr, int c)
{
	if (c == '\0')
	{
		/*
		 * Terminate any special modes, if necessary.
		 * Append a '\0' to the end of the line.
		 */
		switch (ln_state)
		{
		case LN_UL_X:
			curr[0] = curr[-1];
			curr[-1] = NM_CHAR;
			curr++;
			break;
		case LN_BO_X:
			curr[0] = curr[-1];
			curr[-1] = NM_CHAR;
			curr++;
			break;
		case LN_UL_XB:
		case LN_UNDERLINE:
			*curr++ = NM_CHAR;
			break;
		case LN_BO_XB:
		case LN_BOLDFACE:
			*curr++ = NM_CHAR;
			break;
		}
		ln_state = LN_NORMAL;
		*curr = '\0';
		return curr;
	}

	/*
	 * Advance the state machine.
	 */
	switch (ln_state)
	{
	case LN_NORMAL:
		if (curr[-1] != '\b')
			break;
		
		if (c == curr[-2])
			goto enter_boldface;
		if (c == '_' || curr[-2] == '_')
			goto enter_underline;
		curr -= 2;
		break;
		
enter_boldface:
		/*
		 * We have "X\bX" (including the current char).
		 * Switch into boldface mode.
		 */
		curr[-1] = curr[-2];
		curr[-2] = BO_CHAR;
		curr++;
			
		goto ln_bo_xb_case;

enter_underline:
		/*
		 * We have either "_\bX" or "X\b_" (including
		 * the current char).  Switch into underline mode.
		 */
		curr[-1] = curr[-2];
		curr[-2] = UL_CHAR;
		curr++;
		
		goto ln_ul_xb_case;
		/*NOTREACHED*/
	case LN_UL_XB:
		/*
		 * Termination of a sequence "_\bX" or "X\b_".
		 */
		if (c != '_' && curr[-2] != '_' && c == curr[-2])
		{
			/*
			 * We seem to have run on from underlining
			 * into boldfacing - this is a nasty fix, but
			 * until this whole routine is rewritten as a
			 * real DFA, ...  well ...
			 */
			curr[0] = curr[-2];
			curr[-2] = NM_CHAR;
			curr[-1] = BO_CHAR;
			curr += 2; /* char & non-existent backspace */
			ln_state = LN_BO_XB;
			goto ln_bo_xb_case;
		}
ln_ul_xb_case:
		if (c == '_')
			c = curr[-2];
		curr -= 2;
		ln_state = LN_UNDERLINE;
		break;
	case LN_BO_XB:
		/*
		 * Termination of a sequnce "X\bX".
		 */
		if (c != curr[-2] && (c == '_' || curr[-2] == '_'))
		{
			/*
			 * We seem to have run on from
			 * boldfacing into underlining.
			 */
			curr[0] = curr[-2];
			curr[-2] = NM_CHAR;
			curr[-1] = UL_CHAR;
			curr += 2; /* char & non-existent backspace */
			ln_state = LN_UL_XB;
			goto ln_ul_xb_case;
		}
ln_bo_xb_case:
		curr -= 2;
		ln_state = LN_BOLDFACE;
		break;
	case LN_UNDERLINE:
		ln_state = LN_UL_X;
		break;
	case LN_BOLDFACE:
		if (c == '\b')
		{
			ln_state = LN_BO_XB;
			break;
		}
		ln_state = LN_BO_X;
		break;
	case LN_UL_X:
		if (c == '\b')
			ln_state = LN_UL_XB;
		else
		{
			/*
			 * Exit underline mode.
			 * We have to shuffle the chars a bit
			 * to make this work.
			 */
			curr[0] = curr[-1];
			curr[-1] = NM_CHAR;
			curr++;
			ln_state = LN_NORMAL;
		} 
		break;
	case LN_BO_X:
		if (c == '\b')
			ln_state = LN_BO_XB;
		else
		{
			/*
			 * Exit boldface mode.
			 * We have to shuffle the chars a bit
			 * to make this work.
			 */
			curr[0] = curr[-1];
			curr[-1] = NM_CHAR;
			curr++;
			ln_state = LN_NORMAL;
		} 
		break;
	}
	
	/*
	 * Ordinary character.  Just put it in the buffer.
	 */
	*curr++ = c;
	return curr;
}
