/*
** Function to dump out a boot protocol frame.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include "protos.h"

struct bootmsg {
    struct etheraddr dstaddr;	/* Common part of boot message. */
    struct etheraddr srcaddr;
    short protocol;
    short framelen;
    short format;
    unsigned char clientid[8];
    unsigned char serverid[8];
    union {
	struct {		/* Boot request variant. */
	    short boottype;
	    char partnum[16];
	    long serial;
	    short systype;
	    long memsize;
	    unsigned char montype;
	    unsigned char kbdtype;
	} info;
	struct {		/* Data request variant. */
	    long fileoffset;
	} dreq;
	struct {		/* Data transfer variant. */
	    long fileoffset;
	    short bytecount;
	    char data[1500];
	} xfer;
    } u;
};

/* The file header for a boot file. */

struct boothdr {
    unsigned short magic;
    unsigned long length;
    unsigned long loadaddr;
    unsigned long startaddr;
};

/*
** Convert a packed system name to something we can print.
*/

char *
nametostr(name, which)
unsigned char *name;
int which;			/* There are two output buffers. */
{
    static char buf[2][17];

    strncpy(buf[which], name, 8);
    if (*name == 0xFF) {	/* Malformed name? */
	strcpy(buf[which], "<<unknown>>");
    }
    return buf[which];
}

/*
** Dump boot protocol request.
*/

dumpboot(buffer, byte_count, dumpdata)
void *buffer;
int byte_count;
int dumpdata;			/* TRUE to dump data portion of packets. */
{
    int temp;
    struct bootmsg *msgp;
    struct boothdr *bhp;

    msgp = (struct bootmsg *) buffer;
    switch (msgp->format) {
      case 0x00:		/* Initiate a boot. */
	printf("\tBOOT\tWorkstation '%s' looking for%s server\n", 
	       nametostr(msgp->clientid, 0),
	       (msgp->u.info.boottype ? " Alternate" : ""));
	temp = ((msgp->u.info.serial >> 27) & 0x1F) + '@';
	printf("\t\tPROM: %.16s, Serial No. %c%08ld\n",
	       msgp->u.info.partnum, temp, msgp->u.info.serial & 0x7FFFFFFL);
	printf("\t\tMem size: %d bytes, Station type %d\n",
	       msgp->u.info.memsize, msgp->u.info.systype);
	fputs("\t\tMonitor type:   ", stdout);
	switch (msgp->u.info.montype) {
	  case 0:		/* 9" Black & White. */
	    fputs("9\" Black & White.", stdout);
	    break;

	  case 1:		/* 14" Black & White. */
	    fputs("14\" Black & White.", stdout);
	    break;

	  case 2:		/* 14" Color. */
	    fputs("14\" Color.", stdout);
	    break;

	  default:		/* Unknown. */
	    printf("Unknown (%d).", msgp->u.info.montype);
	    break;
	}
	putc('\n', stdout);
	fputs("\t\tKeyboard type:  ", stdout);
	switch (msgp->u.info.kbdtype) {
	  case 0:		/* None. */
	    fputs("None.", stdout);
	    break;

	  case 1:		/* PC AT keyboard. */
	    fputs("PC AT.", stdout);
	    break;

	  case 2:		/* ISC Wincom keyboard. */
	    fputs("ISC Wincom.", stdout);
	    break;

	  case 3:		/* Both. */
	    fputs("PC AT and ISC Wincom both.", stdout);
	    break;

	  default:		/* Unknown. */
	    printf("Unknown (%d).", msgp->u.info.kbdtype);
	    break;
	}
	putc('\n', stdout);
	break;

      case 0x01:		/* Acknowledge a boot. */
	printf("\tBOOT\tServer '%s' is able to boot '%s'\n",
	       nametostr(msgp->serverid, 0), nametostr(msgp->clientid, 1));
	break;

      case 0x02:		/* Boot data request. */
	printf("\tBOOT\tWorkstation '%s' requests data from '%s' at offset %d\n",
	       nametostr(msgp->clientid, 0), nametostr(msgp->serverid, 1),
	       msgp->u.dreq.fileoffset);
	break;

      case 0x03:		/* Boot data. */
	printf("\tBOOT\tServer '%s' supplies %d bytes to '%s' at "
	       "offset %d\n",
	       nametostr(msgp->serverid, 0), msgp->u.xfer.bytecount,
	       nametostr(msgp->clientid, 1), msgp->u.xfer.fileoffset);
	if (!msgp->u.dreq.fileoffset /* First block, contains prog hdr. */
	    && msgp->u.xfer.bytecount > sizeof(struct boothdr)) {
	    bhp = (struct boothdr *) msgp->u.xfer.data;
	    if (bhp->magic == 0x0197) {
		printf("\t  (Program %d bytes starting at $%08X, exec "
		       "at $%08X)\n", bhp->length, bhp->loadaddr,
		       bhp->startaddr);
	    }
	}
	if (!msgp->u.xfer.bytecount)
	    fputs("\t  (Boot done!)\n", stdout);
	if (dumpdata)
	    hd((unsigned char *) bhp, msgp->u.xfer.bytecount, 0);
	break;


      case 0x10:		/* Initiate a dump. */
	printf("\tDUMP\tWorkstation '%s' looking for server\n",
	       nametostr(msgp->clientid, 0));
	break;

      case 0x11:		/* Acknowledge a dump. */
	printf("\tDUMP\tServer '%s' is able to dump '%s'\n",
	       nametostr(msgp->serverid, 0), nametostr(msgp->clientid, 1));
	break;

      case 0x12:		/* Dump data. */
	printf("\tDUMP\tWorkstation '%s' supplies %d bytes to '%s' at offset %d\n",
	       nametostr(msgp->clientid, 0), msgp->u.xfer.bytecount,
	       nametostr(msgp->serverid, 1), msgp->u.dreq.fileoffset);
	if (!msgp->u.xfer.bytecount)
	    fputs("\t  (Dump done!)\n", stdout);
	if (dumpdata) {
	    bhp = (struct boothdr *) msgp->u.xfer.data;
	    hd((unsigned char *) bhp, msgp->u.xfer.bytecount, 0);
	}
	break;

      case 0x13:		/* Dump data acknowledge. */
	printf("\tDUMP\tServer '%s' acknowledges data from '%s' at "
	       "offset %d\n", nametostr(msgp->serverid, 0),
	       nametostr(msgp->clientid, 1), msgp->u.xfer.fileoffset);
	break;
    }
    return (char *) &msgp->u.xfer.data - (char *) buffer;
}
