/*
** Dump X.25/DNET (netman) protocol packet.
*/

#include <stdio.h>
#include <malloc.h>
#include <dnix/cheapdrv.h>
#include <sys/types.h>
#include <net/x25.h>
#include "protos.h"

enum l3type {			/* Classes of Level 3 packets. */
    callsetup, supervisory, confirmation, data,
    flowcontrol, diagnostic, interrupt, registration, unknown
};

struct dnetmsg {
    struct etheraddr dstaddr;
    struct etheraddr srcaddr;
    unsigned short protocol;

    unsigned short length;
    unsigned char l1resp;
    unsigned char l2cmd;
};

extern int packnum;
char l6name[32];

dumpdnet(buffer, byte_count, dumpdata)
void *buffer;
int byte_count;
int dumpdata;			/* TRUE to dump data portion of packets. */
{
    struct dnetmsg *dnp;
    unsigned char *cp, *cp2;
    char *text, buf2[32], buf3[32], line[132], *ptype;
    int len;
    int rseq;			/* Receive sequence #, N(R) from X.25 doc. */
    int sseq;			/* Send sequence #, N(S) from X.25 doc. */
    int pf;			/* Poll/Final bit from X.25 doc. */
    int qbit;
    int mbit;
    unsigned char l2cmd;
    int l3infoknown;		/* 1 == DATA, 2 == INRQ. */

    dnp = (struct dnetmsg *) buffer;
    fputs("\tDNET ", stdout);
    ptype = "PF";
    if (dnp->l1resp & 2) {
	fputs("A\t", stdout);
    } else {
	fputs("B\t", stdout);
    }
    cp = (unsigned char *) &dnp[1]; /* Point to L3 block. */
    len = dnp->length - 4;	/* Length of L3 block. */
    if (len < 0)
	len = 0;
    if ((dnp->l2cmd & ~X25_POLL) == X25_MUI) { /* Mode-independent */
	fputs("MUI ", stdout);
	switch (*cp) {
	  case '?':
	    printf("Search for service '%.*s'\n", len, &cp[1]);
	    cp2 = &cp[len];
	    break;
	  case '!':
	    cp2 = cp;		/* Move past the name to the address. */
	    while (*cp2++)
		;
	    cp2++;
	    printf("Can supply service '%.*s'\n", len, &cp[1]);
	    if (rseq = *cp2++) { /* Valid address? */
		fputs("\t\t    on DTE ID $", stdout);
		while (rseq > 0) {
		    printf("%02X", *cp2++);
		    rseq -= 2;
		}
		fputs("\n", stdout);
	    }
	    break;
	  default:
	    printf("<<Unknown -- $%02X>>", *cp);
	    cp2 = &cp[1];
	    break;
	}
	return cp2 - buffer;
    }
    line[0] = '\0';		/* Clear any existing line. */

    strcat(line, "L2 ");
    cp = (unsigned char *) &dnp->l2cmd;

    sseq = rseq = -1;
    l2cmd = *cp;
    pf = (l2cmd & X25_POLL);	/* Split out poll/final bit. */

    if (!(l2cmd & 0x01)) {	/* Information frame (Data packet). */
	strcat(line, "INFO ");
	ptype = "P";
	sseq = (l2cmd >> 1) & 0x07;
	rseq = l2cmd >> 5;
    } else {			/* A Supervisory or Unnumbered frame. */
	if ((l2cmd & 0x03) == 0x01) { /* Supervisory frame. */
	    rseq = l2cmd >> 5;
	    switch (l2cmd & 0x0F) {
	      case X25_RR:	/* Receiver Ready */
		strcat(line, "RR   ");
		break;
	      case X25_RNR:	/* Receiver Not Ready */
		strcat(line, "RNR  ");
		break;
	      case X25_REJ:	/* Reject */
		strcat(line, "REJ  ");
		break;
	      default:
		sprintf(buf2, "-%02X- ", *cp);
		strcat(line, buf2);
		break;
	    }
	} else {		/* Unnumbered (non-sequenced) frame. */
	    switch (l2cmd & ~X25_POLL) {
	      case X25_SABM:	/* Set Async Balanced Mode .*/
		strcat(line, "SABM ");
		ptype = "P";
		break;
	      case X25_UA:	/* Unnumbered Acknowledgement. */
		strcat(line, "UA   ");
		ptype = "F";
		break;
	      case X25_FRMR:	/* Frame Reject. */
		strcat(line, "FRMR ");
		ptype = "F";
		break;
	      case X25_DISC:	/* Disconnect Command. */
		strcat(line, "DISC ");
		ptype = "P";
		break;
	      case X25_DM:	/* Disconnected Mode. */
		strcat(line, "DM   ");
		ptype = "F";
		break;
	      default:
		sprintf(buf2, "-%02X- ", *cp);
		strcat(line, buf2);
		break;
	    }
	}
    }

    sprintf(buf2, "%-2s         ", pf ? ptype : "- ");
    strcat(line, buf2);
    if (sseq < 0)
	strcat(line, "         ");
    else {
	sprintf(buf2, "N(S): %1d  ", sseq);
	strcat(line, buf2);
    }
    if (rseq < 0)
	strcat(line, "         ");
    else {
	sprintf(buf2, "N(R): %1d  ", rseq);
	strcat(line, buf2);
    }
    if (l2cmd == X25_FRMR) {	/* Frame reject. */
	sprintf(buf2, "$%02X $%02X $%02X", cp[1], cp[2], cp[3]);
	strcat(line, buf2);
	len -= 3;
	cp += 3;
    }
    puts(line);

    l3infoknown = 0;		/* Made TRUE if we try a L6 disassembler */
    qbit = 0;
    cp++;			/* Move past L2 command. */
    if (len && !(l2cmd & 0x01)) { /* L2 INFO frame, decode L3 data in it. */
	if (*cp & 0x80)
	    qbit++;
	rseq = dumpl3(cp, len, buffer, &l3infoknown, &mbit);
	len -= rseq;
	cp += rseq;
    }

    if (dumpdata) {
	if (l3infoknown) {	/* Try for L6? */
	    strcpy(buf2, "WH_");
	    strcpy(buf3, buf2);
	    makerawname((unsigned char *) &dnp->dstaddr, &buf2[3]);
	    makerawname((unsigned char *) &dnp->srcaddr, &buf3[3]);
	}
	switch (l3infoknown) {
	  case 1:		/* DATA. */
	    if (!strcmp(buf2, l6name)) {
		printf("\tWINDOWMAN Commands:\n");
		cp += len - dumpwmcmd(cp, len, qbit);
	    } else if (buf2[1] = 'X',!strcmp(&buf2[1], l6name)) {
		printf("\tRX A to B Data: ");
		cp += len - dumprxatobcmd(cp, len, qbit, mbit);
	    } else if (buf3[1] = 'X',!strcmp(&buf3[1], l6name)) {
		printf("\tRX B to A Data: ");
		cp += len - dumprxbtoacmd(cp, len, qbit, mbit);
	    } else if (buf2[1] = 'L',!strcmp(&buf2[1], l6name)) {
		printf("\tNCU A to B Data: ");
		cp += len - dumpncuatobcmd(cp, len, qbit, mbit);
	    } else if (buf3[1] = 'L',!strcmp(&buf3[1], l6name)) {
		printf("\tNCU B to A Data: ");
		cp += len - dumpncubtoacmd(cp, len, qbit, mbit);
	    } else {
		goto hdump;
	    }
	    break;

	  case 2:		/* INRQ. */
	    if (!strcmp(buf3, l6name)) {
		printf("\tWINDOWMAN Out-of-band Response:\n");
		cp += len - dumpwmint(cp, len);
	    } else if (buf2[1] = 'X',!strcmp(&buf2[1], l6name)) {
		printf("\tRX A to B Interrupt: ");
		cp += len - dumprxatobint(cp, len);
	    } else if (buf3[1] = 'X',!strcmp(&buf3[1], l6name)) {
		printf("\tRX B to A Interrupt: ");
		cp += len - dumprxbtoaint(cp, len);
	    } else {
		goto hdump;
	    }
	    break;

	  default:
hdump:
	    hd(cp, len, 0);
	    cp += len;
	}
    } else if (l3infoknown == 2) {
	hd(cp, len, 0);		/* Show INTR data anyway. */
	cp += len;
    }
    return cp - buffer;		/* Return size of L2/L3 header. */
}

/*
** Dump out the X.25 Level 3 header.
*/

dumpl3(cp, len, buffer, known, mbitp)
unsigned char *cp, *buffer;
int len, *known, *mbitp;
{
    int pktlen;
    int gfi;			/* General (packet) Format Indicator. */
    int sn;			/* Short (or long) numbering. */
    int pti;			/* Packet Type Indicator. */
    int class;			/* Class of L3 packet (implicit hdr length). */
    char buf2[80], line[256], *cause(), *diag();
    int rseq;			/* Receive sequence #, N(R) from X.25 doc. */
    int sseq;			/* Send sequence #, N(S) from X.25 doc. */
    int ps, pr, mbit, qbit, dbit, channel;
    char *name;

    pktlen = len;
    strcpy(line, "\t\tL3 ");
    if (len < 3) {		/* Too short for anything? */
	strcat(line, "(PACKET TOO SHORT)");
	puts(line);
	return 0;
    }
    qbit = dbit = 0;
    gfi = cp[0] >> 4;		/* General Format Indicator. */
    if (cp[0] & 0x40)
	dbit = 1;
    if (cp[0] & 0x80)
	qbit = 1;
    sn = gfi & 3;
    channel = ((cp[0] & 0x0F) << 8) | cp[1]; /* Channel number */
    pti = cp[2];

    if (name = findoverride(packnum)) {
	addsname(buffer, channel, (unsigned char *) name);
    }
    ps = pr = -1;
    if (!(pti & 0x01)) {	/* Information frame (Data packet). */
	strcat(line, "INFO ");
	class = data;
	if (sn == 1) {		/* Short numbering. */
	    ps = (pti >> 1) & 0x07;
	    pr = pti >> 5;
	    mbit = (pti >> 4) & 0x01;
	} else {		/* Long numbering. */
	    ps = pti >> 1;
	    pr = cp[3] >> 1;
	    mbit = cp[3] & 0x01;
	}
	*mbitp = mbit;
    } else {			/* Supervisory Frame or Unnumbered Frame. */
	if ((pti & 0x03) == 0x01 && pti != 0xF1) { /* S-frame */
	    class = flowcontrol;
	    switch (pti & 0x0F) {
	      case X25_RR:	/* Receiver Ready */
		strcat(line, "RR   ");
		break;
	      case X25_RNR:	/* Receiver Not Ready */
		strcat(line, "RNR  ");
		break;
	      case X25_REJ:	/* Reject */
		strcat(line, "REJ  ");
		break;
	      default:
		sprintf(buf2, "-%02X- ", pti);
		strcat(line, buf2);
		break;
	    }
	} else {		/* Unnumbered. */
	    switch (pti) {
	      case X25_CARQ:	/* Call Request */
		strcat(line, "CARQ ");
		class = callsetup;
		break;
	      case X25_CACO: /* Call Confirm */
		strcat(line, "CACO ");
		class = callsetup;
		break;
	      case X25_CLRQ: /* Clear Request */
		strcat(line, "CLRQ ");
		class = supervisory;
		break;
	      case X25_CLCO: /* Clear Confirm */
		strcat(line, "CLCO ");
		class = confirmation;
		break;
	      case X25_INRQ: /* Interrupt request */
		strcat(line, "INRQ ");
		class = interrupt;
		break;
	      case X25_INCO: /* Interrupt confirmation */
		strcat(line, "INCO ");
		class = confirmation;
		break;
	      case X25_RSTR: /* Restart indication/request */
		strcat(line, "RSTR ");
		class = supervisory;
		break;
	      case X25_RSTC: /* Restart confirmation */
		strcat(line, "RSTC ");
		class = confirmation;
		break;
	      case X25_RESR: /* Reset Request */
		strcat(line, "RESR ");
		class = supervisory;
		break;
	      case X25_RESC: /* Reset Confirm */
		strcat(line, "RESC ");
		class = confirmation;
		break;
	      case 0xF1:	/* Diagnostic. */
		strcat(line, "DIAG ");
		class = diagnostic;
		break;
	      case 0xF3:	/* Registration request. */
		strcat(line, "REGR ");
		class = registration;
		break;
	      case 0xF7:	/* Registration confirmation. */
		strcat(line, "REGC ");
		class = registration;
		break;
	      default:
		sprintf(buf2, "-%02X-", pti);
		strcat(line, buf2);
		class = unknown;
		break;
	    }
	}
    }
    switch (class) {
      case callsetup:		/* CARQ, CACO */
	pr = cp[3];
	cp += 4;
	len -= 4;
	sprintf(buf2, "%4d %c     ", channel, dbit ? 'D' : '-');
	strncat(line, buf2, sizeof(line));
	if (ps = (pr & 0x0F)) { /* A called DTE address? */
	    strncat(line, "Called DTE $", sizeof(line));
	    while (ps > 0) {
		if (ps > 1) {
		    sprintf(buf2, "%02X", *cp++);
		    len--;
		} else {
		    sprintf(buf2, "%1X", *cp >> 4);
		    if (!(pr >> 4)) { /* No calling address, consume byte. */
			cp++;
			len--;
		    }
		}
		strncat(line, buf2, sizeof(line));
		ps -= 2;
	    }
	    strncat(line, (pr >> 4) ? ", " : " ", sizeof(line));
	}
	if (ps = (pr >> 4)) {	/* A calling DTE address? */
	    strncat(line, "Calling DTE $", sizeof(line));
	    while (ps > 0) {
		if (pr & 1) {	/* Odd-length Called DTE address? */
		    sprintf(buf2, "%1X", *cp++ & 0x0F);
		    ps--;
		    pr = 0;	/* Only do this nonsense once. */
		} else {
		    if (ps > 1) {
			sprintf(buf2, "%02X", *cp++);
			ps -= 2;
		    } else {
			sprintf(buf2, "%1X", *cp++ >> 4);
			ps--;
		    }
		}
		strncat(line, buf2, sizeof(line));
		len--;
	    }
	    strncat(line, " ", sizeof(line));
	}
	if (pti == X25_CACO) {
	    strncat(line, "                         ", sizeof(line));
	    chantoname(buffer, channel, buf2, sizeof(buf2));
	    strncat(line, buf2, sizeof(line));
	}
	puts(line);
	if (ps = (*cp++ & 0x3F)) { /* Any facilities negotiation? */
	    puts("\t\t   Facilities: ");
	    facilities(cp, ps);
	    cp += ps;
	    len -= ps;
	}
	len--;
	if (len > 0) {		/* Any Call-User data? */
	    printf("\t\t   Call-User Data: %d bytes\n", len);
	    if (pti == X25_CARQ) {
		addsname(buffer, channel, cp);
	    }
	    hd(cp, len, 0);
	    len = 0;		/* All data consumed */
	}
	break;

      case supervisory:		/* RESR, RSTR, CLRQ */
	sprintf(buf2, "%4d       Cause: $%02X, Code: $%02X    ", channel,
		cp[3], cp[4]);
	strncat(line, buf2, sizeof(line));
	chantoname(buffer, channel, buf2, sizeof(buf2));
	strncat(line, buf2, sizeof(line));
	puts(line);
	printf("\t\t  (%s), (%s)\n", cause(pti, cp[3]), diag(pti, cp[4]));
	cp += 5;
	len -= 5;
	break;

      case confirmation:	/* RESC, RSTC, INCO, CLCO */
	sprintf(buf2, "%4d                                ", channel);
	strncat(line, buf2, sizeof(line));
	chantoname(buffer, channel, buf2, sizeof(buf2));
	strncat(line, buf2, sizeof(line));
	if (pti == X25_CLCO) {
	    delsname(buffer, channel);
	}
	puts(line);
	cp += 3;
	len -= 3;
	break;

      case data:		/* INFO */
	sprintf(buf2, "%4d %c %c %c ", channel, dbit ? 'D' : '-',
		qbit ? 'Q' : '-', mbit ? 'M' : '-');
	strncat(line, buf2, sizeof(line));

	if (ps < 0)
	    strncat(line, "         ", sizeof(line));
	else {
	    sprintf(buf2, "N(S): %d  ", ps);
	    strncat(line, buf2, sizeof(line));
	}
	if (pr < 0)
	    strncat(line, "         ", sizeof(line));
	else {
	    sprintf(buf2, "N(R): %d  ", pr);
	    strncat(line, buf2, sizeof(line));
	}
	if (sn == 1) {	/* Short numbering. */
	    cp += 3;
	    len -= 3;
	} else {		/* Long numbering. */
	    cp += 4;
	    len -= 4;
	}
	sprintf(buf2, "%3d data byte%s", len, len == 1 ? "" : "s");
	strncat(line, buf2, sizeof(line));
	*known = chantoname(buffer, channel, buf2, sizeof(buf2));
	strncat(line, buf2, sizeof(line));
	puts(line);
	break;

      case flowcontrol:		/* RR, RNR, REJ */
	if (sn == 1) {		/* Short numbering. */
	    pr = pti >> 5;
	    cp += 3;
	    len -= 3;
	} else {		/* Long numbering. */
	    pr = cp[3] >> 1;
	    cp += 4;
	    len -= 4;
	}
	sprintf(buf2, "%4d                ", channel);
	strncat(line, buf2, sizeof(line));

	if (pr >= 0) {
	    sprintf(buf2, "N(R): %d         ", pr);
	    strncat(line, buf2, sizeof(line));
	}
	chantoname(buffer, channel, buf2, sizeof(buf2));
	strncat(line, buf2, sizeof(line));
	puts(line);
	break;

      case diagnostic:
	sprintf(buf2, "Cause: %d (%s)", cp[3], cause(pti, cp[3]));
	strncat(line, buf2, sizeof(line));
	puts(line);
	cp += 4;
	len -= 4;
	if (len > 0) {		/* Any Diagnostic data? */
	    printf("\t\t   Diagnostic Data: %d bytes\n", len);
	    hd(cp, len, 0);
	    len = 0;		/* All data consumed */
	}
	break;

      case interrupt:		/* INRQ */
	sprintf(buf2, "%4d                                ", channel);
	strncat(line, buf2, sizeof(line));
	*known = chantoname(buffer, channel, buf2, sizeof(buf2)) * 2;
	strncat(line, buf2, sizeof(line));
	puts(line);
	cp += 3;
	len -= 3;
	if (len > 0) {		/* Any Interrupt data? */
	    printf("\t\t   Interrupt Data: %d bytes\n", len);
	}
	break;

      case registration:	/* REGR, REGC */
	if (pti == 0xF3) {	/* REGC */
	    rseq = cp[3];	/* Save away cause/diagnostic. */
	    sseq = cp[4];
	    cp += 2;
	}
	pr = cp[3];
	cp += 4;
	len -= 4;
	if (channel) {
	    strncat(line, "INVALID CHANNEL", sizeof(line));
	} else {
	    strncat(line, "           ", sizeof(line));
	    if (ps = (pr & 0x0F)) { /* DCE address? */
		strncat(line, "DCE $", sizeof(line));
		while (ps > 0) {
		    if (ps > 1) {
			sprintf(buf2, "%02X", *cp++);
			len--;
		    } else {
			sprintf(buf2, "%1X", *cp >> 4);
			if (!(pr >> 4)) { /* No DTE address, consume byte. */
			    cp++;
			    len--;
			}
		    }
		    strncat(line, buf2, sizeof(line));
		    ps -= 2;
		}
		strncat(line, (pr >> 4) ? ", " : " ", sizeof(line));
	    }
	    if (ps = (pr >> 4)) { /* DTE address? */
		strncat(line, "DTE $", sizeof(line));
		while (ps > 0) {
		    if (pr & 1) { /* Odd-length DCE address? */
			sprintf(buf2, "%1X", *cp++ & 0x0F);
			ps--;
			pr = 0;	/* Only do this nonsense once. */
		    } else {
			if (ps > 1) {
			    sprintf(buf2, "%02X", *cp++);
			    ps -= 2;

			    sprintf(buf2, "%1X", *cp++ >> 4);
			    ps--;
			}
		    }
		    strncat(line, buf2, sizeof(line));
		    len--;
		}
		strncat(line, " ", sizeof(line));
	    }
	    puts(line);
	    if (pti == 0xF3) {	/* Confirmation. */
		printf("\t\t  (%s), (%s)\n", cause(pti, rseq), diag(pti,sseq));
	    }
	    if (ps = (*cp++ & 0x7F)) { /* Any registration information? */
		puts("\t\t   Registration: ");
		registrata(cp, ps);
		cp += ps;
		len -= ps;
	    }
	    len--;
	}
	break;

      default:
      case unknown:
	puts(line);
	break;
    }
    if (len < 0)
	len = 0;
    return pktlen - len;	/* How much was eaten? */
}

/*
** Facilities picker-aparter.
*/

static short t[16] = {0,0,0,0,16,32,64,128,256,512,1024,0,0,0,0,0};
static long tcn[16] = {0,0,0,75,150,300,600,1200,2400,4800,9600,
			   19200,48000,0,0,0};

facilities(cp, len)
unsigned char *cp;
int len;			/* Number of facilities bytes. */
{
    int fcode, fcode_len, x, y, fc;
    unsigned char *cp2;
    char line[80], *prefix;
    
    while (len > 0) {
	prefix = "\t\t\t";
	fc = *cp++;
	fcode_len = (fc >> 6) + 2;
	if (fcode_len == 5)
	    fcode_len = *cp++ + 2;
	len -= fcode_len;
	fcode = fc & 0x3F;
	switch (fc) {
	  case FAC_REV_CHRG:
	    x = *cp++;
	    printf("%sReverse Charging: %sequested\n", prefix,
		   (x & 1) ? "R" : "Not r");
	    printf("%sFast Select: %s\n", prefix,
		   (x & 0x80) ? ((x & 0x40) ? "Requested w/restriction" :
		    "Requested with no restriction") : "Not requested");
	    break;

	  case FAC_THRUPUT:
	    x = *cp++;
	    y = (x >> 4) & 0x0F;
	    x &= 0x0F;
	    printf("%sDTE Throughput Class Negotiation: "
		    "Called %ld, "
		    "Calling %ld\n", prefix, tcn[x], tcn[y]);
	    break;

	  case 0x03:
	    x = *cp++;
	    printf("%sClosed User Groups: %02X", prefix, x);
	    break;

	  case FAC_LEN:
	    x = *cp++;
	    printf("%sDatagram Nondelivery Indication: %sequested\n", prefix,
		   (x & 2) ? "R" : "Not r");
	    printf("%sDatagram Delivery Indication: %sequested\n", prefix,
		   (x & 1) ? "R" : "Not r");
	    break;

	  case 0x08:
	    x = *cp++;
	    printf("%sCalled line address modified: ", prefix);
	    if (x & 0x80)
		printf("DTE originated (%02X)\n", x);
	    else if (x == 0x07)
		printf("Within hunt group\n");
	    else if (x == 0x01)
		printf("DTE busy\n");
	    else if (x == 0x09)
		printf("DTE out of order\n");
	    else if (x == 0x0F)
		printf("Systematic call redirection\n");
	    else
		printf("Unknown reason (%02X)\n", x);
	    break;


	  case 0x41:
	    x = *cp++;
	    y = *cp++;
	    printf("%sBilateral Closed User Group: %02X %02X\n", prefix, x, y);
	    break;

	  case FAC_PS_CODE:
	    x = *cp++ & 0x0F;	
	    y = *cp++ & 0x0F;	
	    printf("%sDTE Packet size: Called %d, "
		    "Calling %d\n", prefix, t[x], t[y]);
	    break;

	  case FAC_WS:
	    x = *cp++ & 0x7F;	
	    y = *cp++ & 0x7F;	
	    printf("%sDTE Window size: Called %d, "
		   "Calling %d\n", prefix, x, y);
	    break;

	  case 0x44:
	    x = *cp++;
	    y = *cp++;
	    printf("%sRPOA Selection: %02X %02X\n", prefix, x, y);
	    break;


	  case 0xC3:		/* Call redirection notification */
	    y = cp[0];		/* Length of data. */
	    x = cp[1];
	    printf("%sCall redirection notification: ", prefix);
	    if (x == 0x01)
		printf("DTE busy\n");
	    else if (x == 0x09)
		printf("DTE out of order\n");
	    else if (x == 0x0F)
		printf("Systematic call redirection\n");
	    else
		printf("Unknown reason (%02X)\n", x);
	    x = cp[2] & 0x0F;	/* Nybble count of original address. */
	    strncpy(line, "\tCalled DTE $", sizeof(line));
	    cp2 = &cp[3];
	    while (x > 0) {
		char buf2[10];
		if (x > 1) {
		    sprintf(buf2, "%02X", *cp2++);
		} else {
		    sprintf(buf2, "%1X", *cp2++ >> 4);
		}
		strncat(line, buf2, sizeof(line));
		x -= 2;
	    }
	    printf("%s%s\n", prefix, line);
	    cp += y - 1;
	    break;


	  default:
	    printf("%sUnknown facility type: $%02X\n", prefix, fc);
	    cp += fcode_len;
	    break;
	}
    }
}

registrata(cp, len)
unsigned char *cp;
int len;			/* Number of registration bytes. */
{
    int rcode, rcode_len, x, y, rc;
    char line[80], *prefix;
    
    while (len > 0) {
	prefix = "\t\t\t";
	rc = *cp++;
	rcode_len = (rc >> 6) + 2;
	if (rcode_len == 5)
	    rcode_len = *cp++ + 2;
	len -= rcode_len;
	rcode = rc & 0x3F;
	switch (rc) {
	  case 0:
	    x = *cp++;
	    printf("%sMarker $%02X\n", prefix, x);
	    break;
	  case 0x05:
	    x = *cp++;
	    printf("%s(Before Calling Facilities Negotiation):\n", prefix);
	    if (x & 4)
		printf("%sD-bit Modification\n", prefix);
	    if (x & 2)
		printf("%sPacket Retransmission\n", prefix);
	    if (x & 1)
		printf("%sExtended Packet Sequence Numbering\n", prefix);
	    break;
	  case 0x45:
	    x = *cp++;
	    y = *cp++;
	    printf("%s(Any-Time Facilities Negotiation):\n", prefix);
	    if (x & 0x40)
		printf("%sCharging Information\n", prefix);
	    if (x & 0x20)
		printf("%sThroughput Class\n", prefix);
	    if (x & 0x10)
		printf("%sFlow Control Parameter Negotiation\n", prefix);
	    if (x & 0x08                                  )
		printf("%sReverse Charging Acceptance\n", prefix);
	    if (x & 0x04)
		printf("%sFast Select Acceptance\n", prefix);
	    if (x & 0x02)
		printf("%sOutgoing Calls Barred\n", prefix);
	    if (x & 0x01)
		printf("%sIncoming Calls Barred\n", prefix);
	    break;
	  case 0x46:
	    x = *cp++;
	    y = *cp++;
	    printf("%s(Available Facilities):\n", prefix);
	    if (x & 0x80)
		printf("%sReverse Charging\n", prefix);
	    if (x & 0x40)
		printf("%sReverse Charging Acceptance\n", prefix);
	    if (x & 0x20)
		printf("%sCharging Information (per-call)\n", prefix);
	    if (x & 0x10)
		printf("%sCharging Information (per-interface)\n", prefix);
	    if (x & 0x08                                  )
		printf("%sCalled Line Address Modified Notification\n",prefix);
	    if (x & 0x04)
		printf("%sD-bit Modification\n", prefix);
	    if (x & 0x02)
		printf("%sPacket Retransmission\n", prefix);
	    if (x & 0x01)
		printf("%sExtended Packet Sequence Numbering\n", prefix);
	    if (y & 0x10)
		printf("%sRPOA Selection\n", prefix);
	    if (y & 0x08                                  )
		printf("%sLogical Channel Types Ranges\n",prefix);
	    if (y & 0x04)
		printf("%sNon-standard Default Packet Sizes\n", prefix);
	    if (y & 0x02)
		printf("%sNon-standard Default Window Sizes\n", prefix);
	    if (y & 0x01)
		printf("%sDefault Throughput Classes Assignment\n", prefix);
	    break;
	  case 0x06:
	    x = *cp++;
	    printf("%s(Non-negotiable Facilities):\n", prefix);
	    if (x & 0x01)
		printf("%sLocal Charging Prevention\n", prefix, x);
	    break;
	  case 0x02:
	    x = *cp++;
	    y = (x >> 4) & 0x0F;
	    x &= 0x0F;
	    printf("%sDefault Throughput Class: "
		    "DCE %ld, DTE %ld\n", prefix, tcn[x], tcn[y]);
	    break;
	  case 0x42:
	    x = *cp++ & 0x0F;	
	    y = *cp++ & 0x0F;	
	    printf("%sNon-standard DCE Packet size: %d, "
		    "DTE Packet size: %d\n", prefix, t[x], t[y]);
	    break;
	  case 0x43:
	    x = *cp++ & 0x7F;	
	    y = *cp++ & 0x7F;	
	    printf("%sNon-standard DCE Window size: %d, "
		    "DTE Window size: %d\n", prefix, x, y);
	    break;
	  case 0xC8:
	    printf("%sLogical Channel Types Ranges: 1-%d\n", prefix,
		   cp[12] << 8 | cp[13]);
	    break;
	  default:
	    printf("%sUnknown Registration type: $%02X\n", prefix, rc);
	    cp += rcode_len;
	    break;
	}
    }
}

/*
** ASCII-fy clearing causes/diagnostic codes.
*/

static char dteorig[] = "DTE Originated";
static char lpe[] = "Local Procedure Error";
static char rpe[] = "Remote Procedure Error";
static char ncong[] = "Network Congestion";
static char nopr[] = "Network Operational";
static char ooo[] = "Out of Order";
static char ifac[] = "Invalid Facility Request";

char *
cause(cmd, cause)
int cmd, cause;
{
    cause &= 0x7F;		/* DTE/DCE distinction for now unimportant. */
    switch (cmd) {
      case X25_CLRQ:		/* Clear request. */
	switch (cause) {
	  case 0x00: return dteorig;
	  case 0x01: return "Number Busy";
	  case CLC_IFAC: return ifac;
	  case CLC_NCOG: return ncong;
	  case CLC_OORD: return ooo;
	  case 0x0B: return "Access Barred";
	  case CLC_NOBT: return "Not Obtainable";
	  case CLC_RPER: return rpe;
	  case CLC_LPER: return lpe;
	  case 0x15: return "RPOA Out of Order";
	  case 0x19: return "Reverse Charging Acceptance Not Subscribed";
	  case 0x21: return "Incompatible Destination";
	  case 0x29: return "Fast Select Acceptance Not Subscribed";
	}
	break;
      case X25_RESR:		/* Reset Request */
	switch (cause) {
	  case 0: return dteorig;
	  case 1: return ooo;
	  case 3: return rpe;
	  case 5: return lpe;
	  case 7: return ncong;
	  case 9: return "Remote DTE Operational";
	  case 15: return nopr;
	  case 0x11: return "Incompatible Destination";
	}
	break;
      case X25_RSTR:		/* Restart Request/Indication */
	switch (cause) {
	  case 1: return lpe;
	  case 3: return ncong;
	  case 7: return nopr;
	}
	break;
      case 0xF3:		/* Registration confirmation */
	switch (cause) {
	  case 3: return ifac;
	  case 5: return ncong;
	  case 0x13: return lpe;
	  case 0x7F: return "Registration/cancellation Confirmed";
	}
	break;
    }
    return "";
}

char *
diag(cmd, code)
int cmd, code;
{
    switch (code) {
      case 0x00: return "No Additional Information";
      case 0x01: return "Invalid P(S)";
      case 0x02: return "Invalid P(R)";

      case DGN_PROT: return "Protocol State Error";
      case 0x11:
      case 0x12:
      case 0x13:
      case 0x14:
      case 0x15:
      case 0x16:
      case 0x17:
      case 0x18:
      case 0x19:
      case 0x1A:
      case 0x1B:
      case 0x1C:
      case 0x1D:
      case 0x1E:
      case 0x1F:
	return "Packet Type Invalid";
	
      case 0x20: return "Packet Not Allowed";
      case 0x21: return "Unidentifiable Packet";
      case 0x22: return "Call on One-Way Logical Channel";
      case 0x23: return "Invalid Packet Type on a PVC";
      case 0x24: return "Packet on Unassigned Logical Channel";
      case 0x25: return "Reject Not Subscribed To";
      case 0x26: return "Packet Too Short";
      case 0x27: return "Packet Too Long";
      case 0x28: return "Invalid General Format Identifier";
      case 0x29: return "Restart With Non-zero in Bits 1-4, 9-16";
      case 0x2A: return "Packet Type Not Compatible With Facility";
      case 0x2B: return "Unauthorized Interrupt Confirmation";
      case 0x2C: return "Unauthorized Interrupt";
      case 0x2D: return "Unauthorized Reject";

      case 0x30: return "Timer Expired";
      case 0x31: return "Timer Expired For Incoming Call";
      case 0x32: return "Timer Expired For Clear Indication";
      case 0x33: return "Timer Expired For Reset Indication";
      case 0x34: return "Timer Expired For Restart Indication";

      case 0x40: return "Call Set-up Problem";
      case 0x41: return "Facility Code Not Allowed";
      case 0x42: return "Facility Parameter Not Allowed";
      case 0x43: return "Invalid Called Address";
      case 0x44: return "Invalid Calling Address";
      case 0x45: return "Invalid Facility/Registration Length";
      case 0x46: return "Incoming Call Barred [Reverse Charge Denied]";
      case 0x47: return "No Logical Channel Available";
      case 0x48: return "Call Collision";
      case 0x49: return "Duplicate Facility Requested";
      case 0x4A: return "Non-Zero Address Length";
      case 0x4B: return "Non-Zero Facility Length";
      case 0x4C: return "Facility Not Provided When Expected";
      case 0x4D: return "Invalid CCITT-specified DTE Facility";

      case 0x51: return "Improper Cause Code from DTE";
      case 0x52: return "Non-Aligned Octet";
      case 0x53: return "Inconsistent Q-Bit Setting";

      case 0x70: return "International Problem";
      case 0x71: return "Remote Network Problem";
      case 0x72: return "International Protocol Problem";
      case 0x73: return "International Link Out-Of-Order";
      case 0x74: return "International Link Busy";
      case 0x75: return "Transit Network Facility Problem";
      case 0x76: return "Remote Network Facility Problem";
      case 0x77: return "International Routing Problem";
      case 0x78: return "Temporary Routing Problem";
      case 0x79: return "Unknown Called DNIC";
      case 0x7A: return "Maintenance Action";

      case DGN_L2FAIL: return "Level 2 Protocol Failed";
      case DGN_SERVICE: return "Service Name Not Found";
      case DGN_NUMBER: return "DTE Address Not a Number";
    }
    return "";
}

/*
** Attempt to track service names versus connection id's for easier human
** reading.  This only works if you see the call request go by.
*/

static struct node {
    struct etheraddr ea[2];
    short connid;
    char *name;
} *root;

static int
ncompare(n1, n2)
struct node *n1, *n2;
{
    unsigned char *cp1, *cp2;
    int i;

    cp1 = (unsigned char *) &n1->ea;
    cp2 = (unsigned char *) &n2->ea;
    i = sizeof(n1->ea) + sizeof(n1->connid);
    while (i--) {
	if (*cp1 != *cp2) {
	    if (*cp1 > *cp2) {
		return 1;
	    } else {
		return -1;
	    }
	}
	cp1++;
	cp2++;
    }
    return 0;
}

/*
** Add a new service name to the tree.
*/

static void
addsname(buffer, channel, name)
unsigned char *buffer, *name;
short channel;
{
    struct node n, *nstore, **np, **tfind();
    unsigned char *thisea;
    unsigned char *wanea;
    int i;

    eacanon((unsigned short *) buffer, (unsigned short *) &n.ea[0]);

    thisea = (unsigned char *) &n;
    wanea = (unsigned char *) "\000\000\000\000\000\001";
    for (i = 0; i < 6; i++, thisea++, wanea++) {
	if (*thisea > *wanea)
	    goto bigger;
	else if (*thisea < *wanea)
	    break;
    }
    return;
bigger:

    n.connid = channel;
    if (!(np = tfind(&n, &root, ncompare))) {
	if (nstore = (struct node *) malloc(sizeof(struct node))) {
	    *nstore = n;
	    strcpy(nstore->name = malloc(strlen(name) + 1), name);
	    tsearch(nstore, &root, ncompare);
	}
    } else {			/* Update the entry. */
	free((*np)->name);
	strcpy((*np)->name = malloc(strlen(name) + 1), name);
    }
}

/*
** Delete a service name from the tree.
*/

static void
delsname(buffer, channel)
unsigned char *buffer;
short channel;
{
    struct node n, *nstore, **np, **tfind();

    eacanon((unsigned short *) buffer, (unsigned short *) &n.ea[0]);
    n.connid = channel;
    if (np = tfind(&n, &root, ncompare)) {
	free((*np)->name);
    }
    tdelete(&n, &root, ncompare);
}

/*
** Turn a pair of ethernet addresses into a key.
*/

static void
eacanon(src, dst)
unsigned short *src, *dst;
{
    dst[0] = src[0];
    *(unsigned long *) &dst[1] = *(unsigned long *) &src[1];
    dst[3] = src[3];
    *(unsigned long *) &dst[4] = *(unsigned long *) &src[4];
    if (src[0] < src[3]
	|| (src[0] == src[3]
	    && (*(unsigned long *) &src[1] < *(unsigned long *)&src[4]))) {
	dst[0] = src[3];
	*(unsigned long *) &dst[1] = *(unsigned long *) &src[4];
	dst[3] = src[0];
	*(unsigned long *) &dst[4] = *(unsigned long *) &src[1];
    }
}

/*
** Convert a channel number to a name.
*/

static
chantoname(buffer, channel, to, size)
unsigned char *buffer;
short channel;
char *to;
int size;
{
    struct node n, *nstore, **np, **tfind();

    to[0] = 0;
    eacanon((unsigned short *) buffer, (unsigned short *) &n.ea[0]);
    n.connid = channel;
    if (np = tfind(&n, &root, ncompare)) {
	strncpy(to, ", ", size);
	strncat(to, (*np)->name, size);
	strncpy(l6name, (*np)->name, sizeof(l6name));
	return 1;
    }
    return 0;
}
