/*
** Sniffer packet copier.  Can extract selected packets from one file
** for placement in another.
**
** Usage:
**	packetcp [-e range] [-i range] [-o] [-v] [-V] [file]
**
** Where
**      -e      excludes <range> from copying,
**      -i      includes <range> in copying,
**      -o      excludes override 'packets' from copying,
**	-v	prints verbose information,
**	-V	prints version number and exits,
**	file	Packets in this file, otherwise read from standard in.
** and
**      range   is packnum[,packnum...] or packnum[-packnum]
**
** Output goes to standard output.
*/

#include <stdio.h>
#include <malloc.h>
#include <dnix/cheapdrv.h>
#include <dnix/fcodes.h>

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[8192];	/* Way bigger than typically needed. */
};

struct inclexcl {
    struct inclexcl *next;
    unsigned short start;
    unsigned short end;
    unsigned short excl;
} *head;

char *progname;

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp, *outfp;
    unsigned short packetsize;
    struct packetform packet;
    char *filename, *argp;
    int ii, packnum, copied, overs, verbose, noovers;
    char namebuf[100];

    progname = *argv;
    filename = 0;
    verbose = noovers = 0;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'e':
		if (!*argp)
		    --argc, argp = *++argv;
		addrange(argp, 1);
		break;

	      case 'i':
		if (!*argp)
		    --argc, argp = *++argv;
		addrange(argp, 0);
		break;

	      case 'o':		/* No override packets. */
		noovers = 1;
		break;

	      case 'v':		/* Verbose. */
		verbose = 1;
		break;

	      case 'V':		/* Version printout. */
		printf("%s: Version 1.0\n", progname);
		exit(0);
		break;

	      default:
		fprintf(stderr, "Usage: %s [-e num[,|-[num]]] "
			"[-i num[,|-[num]]] [-o] [-v] [-V] [packetfile]\n",
			progname);
		exit(1);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(2);
    }

    setvbuf(fp, 0, _IOFBF, 4096);
    packnum = -1;
    copied = overs = 0;
    while (fread(&packetsize, sizeof packetsize, 1, fp) == 1) {
	packnum++;
	packet.snh.statusbits = 0;
	if (packetsize > sizeof(packet)) {
	    printf("ERROR -- %d byte packet is too big!\n", packetsize);
	    fseek(fp, packetsize, 1); /* Skip over it. */
	} else {
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    if (packet.snh.statusbits == 0xFF) { /* It's a format-override, */
		packnum--;	/* so it doesn't count. */
		overs++;
		if (noovers)	/* Skip it? */
		    continue;
	    }
	    if (tocopy(packnum)) {
		copied++;
		fwrite(&packetsize, sizeof(packetsize), 1, stdout);
		fwrite(&packet, packetsize, 1, stdout);
	    }
	}
    }
    if (verbose)
	fprintf(stderr, "%s: Copied %d of %d packets (%d format overrides)\n",
		progname, copied, packnum, overs);
    exit(0);
}

int
addrange(range, excl)
char *range;
int excl;
{
    struct inclexcl *p, *np;
    char *cp;
    int ii;

    if (!*range)
	return;
    p = head;
    while (p && p->next)	/* Find end of chain. */
	p = p->next;
    if (np = malloc(sizeof *np)) {
	if (p)
	    p->next = np;
	else
	    head = np;
	np->next = 0;
	if (*range == '-') {
	    cp = range, ii = 0;
	} else {
	    ii = strtol(range, &cp, 0);
	    if (cp == range) {
		fprintf(stderr, "%s: Bad number: %s\n", progname, range);
		exit(1);
	    }
	}
	np->start = ii;
	np->end = ii;
	np->excl = excl;
	if (*cp == ',') {
	    addrange(&cp[1], excl);
	} else if (*cp == '-') {
	    range = &cp[1];
	    if (!*range)
		ii = ~0;
	    else
		ii = strtol(range, &cp, 0);
	    if (cp == range) {
		fprintf(stderr, "%s: Bad range: %s\n", progname, range-1);
		exit(1);
	    }
	    np->end = ii;
	    if (*cp == ',')	/* Recurse */
		addrange(&cp[1], excl);
	    else if (*cp) {
		fprintf(stderr, "%s: Bad number: %s\n", progname, range);
		exit(1);
	    }
	} else if (*cp) {
	    fprintf(stderr, "%s: Bad number: %s\n", progname, range);
	    exit(1);
	}
    }
}

int
tocopy(packnum)
int packnum;
{
    struct inclexcl *p, *np;
    int keep;

    keep = 0;			/* Assume not keeping at first. */
    if (!(np = head))
	return 1;
    do {
	p = np;
	if (packnum >= p->start && packnum <= p->end)
	    keep = !p->excl;
    } while (np = p->next);
    return keep;
}
