/*
** Sniffer packet statistics
**
** Usage:
**	packstats [-b] [-Btype] [-c] [-D daddr] [-f namefile] [-P prot]
**		  [-S saddr] [-s] [-V] [file]
**
** Where
**	-b	Binary mode, waits for EOF on stdin before exiting.
**	-Btype	Bucket statistics (packet size ranges, addresses, etc).  [SADP]
**	-c	Conversational mode (packets between S and D nodes);
**	-Ddaddr	Counts only packets _to_ a particular address;
**	-ffile	Use file for symbolic address names.
**	-Pprot	Counts only examine packets for this protocol;
**	-Ssaddr	Counts only packets _from_ a particular address;
**	-s	Prevents packets from being flagged as too short;
**	-V	Prints program version and exits.
**	file	Counts packets in this file, otherwise read from standard in.
**
** Output goes to standard out.
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <search.h>
#include <math.h>
#include <dnix/cheapdrv.h>
#include "sapsnap.h"
#include "protocols.h"

#define SRCFILTER 1		/* Options. */
#define DSTFILTER 2
#define PROTFILTER 4
#define CONVERSATION 8
#define BINARY 16
#define NO_SHORT_WARN 32
#define SBUCKETS 64		/* Size bucket printout. */
#define ABUCKETS 128
#define DBUCKETS 256
#define PBUCKETS 512

#define SRCFMATCHSRC 1		/* Filter match codes. */
#define SRCFMATCHDST 2
#define DSTFMATCHDST 4
#define DSTFMATCHSRC 8

#define MAXPKSIZ 8192
#define BUCKSIZE 64
#define NBUCKETS (MAXPKSIZ / BUCKSIZE)

int buckets[NBUCKETS], highbucket; /* Statistical buckets. */
int packnum;

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[MAXPKSIZ];
};

unsigned short statflags;	/* Global flags. */
long datasum;
int packcnt;
int badpacks;		/* # of packets with hard errors. */
int runtpacks;		/* # of packets way too short (< 14 bytes)  */
int shortpacks;		/* # of packets shorter than official min len */
int longpacks;		/* # of packets longer than official max len */
int lostpacks;		/* # of packets (or more) lost. */
int triggered, tbefore;
char *lasttime;
long msecs;
char *bar = "Count   %  |----1----2----3----4----5----6----7----8----9----|\n";
static void adrbucket(), protbucket(), commas();

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp;
    unsigned short packetsize, datasize;
    unsigned short prot;
    struct packetform packet;
    char *filename, *progname, *argp;
    char from[14], to[14], buf1[32], buf2[16];
    char *status(), *dotime();
    int matchflags, i;
    struct etheraddr sfilter, dfilter;
    unsigned short pfilter;

    progname = *argv;
    filename = 0;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'b':		/* Binary mode. */
		statflags |= BINARY;
		break;

	      case 'B':		/* Bucket statistics mode. */
		if (!*argp)
		    --argc, argp = *++argv;
		switch (*argp & ~('a' - 'A')) {
		  case 'S': statflags |= SBUCKETS; break;
		  case 'A': statflags |= ABUCKETS; break;
		  case 'D': statflags |= DBUCKETS; break;
		  case 'P': statflags |= PBUCKETS; break;
		}
		break;

	      case 'c':		/* Conversation, src matches src | dest, &c. */
		statflags |= CONVERSATION;
		break;

	      case 'D':		/* Count only particular dest's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (address((unsigned char *) argp, &dfilter))
		    statflags |= DSTFILTER;
		break;

	      case 'f':		/* Read names file. */
		if (!*argp)
		    --argc, argp = *++argv;
		namesfile(argp);
		break;

	      case 'P':		/* Count only particular protocol's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		pfilter = protocol((unsigned char *) argp);
		statflags |= PROTFILTER;
		break;

	      case 'S':		/* Count only particular source's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (address((unsigned char *) argp, &sfilter))
		    statflags |= SRCFILTER;
		break;

	      case 's':		/* No warnings for too-short packets. */
		statflags |= NO_SHORT_WARN;
		break;

	      case 'V':		/* Version printout. */
		printf("%s: Version 1.2\n", progname);
		exit(0);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    memset(buckets, 0, sizeof buckets);
    highbucket = -1;
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(1);
    }
    packnum = -1;
    setvbuf(fp, 0, _IOFBF, 4096);
    while (fread(&packetsize, sizeof packetsize, 1, fp) == 1) {
	matchflags = SRCFMATCHSRC | SRCFMATCHDST |
	    DSTFMATCHDST | DSTFMATCHSRC; /* Assume a filter match at first. */
	packnum++;
	if (packetsize > sizeof(packet)) {
	    printf("ERROR -- %d byte packet is too big!\n", packetsize);
	    fseek(fp, packetsize, 1); /* Skip over it. */
	} else if (packetsize <
		   (sizeof(struct etherheader) + sizeof(struct sniffheader))) {
	    runtpacks++;
	    if (statflags & SBUCKETS)
		buckets[0]++;
	    if (!packetsize)
		packetsize = sizeof packetsize;
	    fseek(fp, packetsize, 1); /* Skip over it. */
	    datasize = 0;
	} else {
	    if ((datasize = packetsize
		 - (sizeof(struct sniffheader) + sizeof(struct etherheader)))
		> 1500)
		longpacks++;
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    if (packet.snh.statusbits == 0xFF) { /* It's a format-override, */
		packnum--;	/* so skip it. */
		continue;
	    }
	    if (datasize < 46 && (packet.eth.protocol != WINCOM))
		shortpacks++;
	    if (!packnum)	/* First packet? */
		dotime(&packet); /* Prime the timer logic. */
	    if (statflags & PROTFILTER && packet.eth.protocol != pfilter)
		matchflags = 0;
	    if (statflags & SRCFILTER) {
		if ((*(short *)&packet.eth.srcaddr != *(short *) &sfilter ||
		*(long *)&packet.eth.srcaddr[2] != *(long *) &sfilter.addr[2]))
		    matchflags &= ~SRCFMATCHSRC;
		if (!(statflags & CONVERSATION) ||
		    (*(short *)&packet.eth.dstaddr != *(short *) &sfilter ||
		     *(long *)&packet.eth.dstaddr[2] != *(long *) &sfilter.addr[2]))
		    matchflags &= ~SRCFMATCHDST;
	    }
	    if (statflags & DSTFILTER) {
		if ((*(short *)&packet.eth.dstaddr != *(short *) &dfilter ||
		*(long *)&packet.eth.dstaddr[2] != *(long *) &dfilter.addr[2]))
		    matchflags &= ~DSTFMATCHDST;
		if (!(statflags & CONVERSATION) ||
		    (*(short *)&packet.eth.srcaddr != *(short *) &dfilter ||
		     *(long *)&packet.eth.srcaddr[2] != *(long *) &dfilter.addr[2]))
		    matchflags &= ~DSTFMATCHSRC;
	    }
	    if (statflags & CONVERSATION)
		if (matchflags != (SRCFMATCHSRC | DSTFMATCHDST)
		    && matchflags != (SRCFMATCHDST | DSTFMATCHSRC)
		    && !(packet.eth.dstaddr[0] & 0x01))	/* Pass B/Mcasts. */
		    matchflags = 0;
	    if (!matchflags)
		continue;
	    if (statflags & SBUCKETS) {
		i = datasize;
		if (packet.eth.protocol == WINCOM)
		    i -= sizeof(struct etherheader) + sizeof(short);
		i = i / BUCKSIZE;
		buckets[i]++;
		if (i > highbucket)
		    highbucket = i;
	    }
	    if (statflags & ABUCKETS)
		adrbucket(packet.eth.srcaddr, 0);
	    if (statflags & DBUCKETS)
		adrbucket(packet.eth.dstaddr, DBUCKETS);
	    if (statflags & PBUCKETS)
		protbucket(&packet);
	    if (packet.snh.statusbits & 128) { /* Trigger packet. */
		tbefore = packcnt;
		triggered = 1;
	    }
	    packcnt++;
	    if (packet.snh.statusbits & PACKET_GOOD) {
		datasum += packet.snh.length - sizeof(struct sniffheader);
		if (packet.eth.protocol == WINCOM)
		    datasum -= sizeof(struct etherheader) + sizeof(short);
	    }
	    if (packet.snh.statusbits
		& (PACKET_CRC | PACKET_FRAME | PACKET_OVERRUN))
		badpacks++;
	    if (packet.snh.statusbits & PACKET_LOST)
		lostpacks++;
	    lasttime = dotime(&packet);
	}
    }
    if (datasum) {
	sprintf(buf2, "%ld", datasum);
	commas(buf2, buf1);
	if (packcnt > 1) {
	    printf("%s byte%s in %d packets over %s msec, for %d kB/second.  ",
		   buf1, (datasum == 1) ? "" : "s", packcnt, lasttime,
		   msecs ? datasum / msecs : 0);
	} else {
	    printf("%s byte%s in %d packet.  ", buf1,
		   (datasum == 1) ? "" : "s", packcnt);
	}
    } else
	printf("No data captured.  ");
    if (triggered && tbefore) {
	printf("%d packet%s before trigger, %d after.  ", tbefore,
	       (tbefore == 1) ? "" : "s", packcnt - tbefore - 1);
    }
    badness("bad", badpacks);
    badness("runt", runtpacks);
    badness("overlong", longpacks);
    if (!(statflags & NO_SHORT_WARN))
	badness("short", shortpacks);
    if (lostpacks)
	printf("%d+ LOST packet%s.", lostpacks, (lostpacks > 1) ? "s" : "");
    printf("\n");
    fflush(stdout);
    if (statflags & SBUCKETS) {
	int i;

	printf("\n   Size    %s", bar);
	for (packnum = 0; packnum <= highbucket; packnum++) {
	    if (packnum < 2) {	/* 64-byte buckets for first two. */
		i = buckets[packnum] * 10000 / packcnt;
		printf("%4d-%4d %5d %3d.%02d ", packnum*BUCKSIZE,
		       packnum*BUCKSIZE+BUCKSIZE-1,
		       buckets[packnum], i/100, i%100);
	    } else {		/* The rest are 128-byte buckets. */
		i = (buckets[packnum] + buckets[packnum+1]) * 10000 / packcnt;
		printf("%4d-%4d %5d %3d.%02d ", packnum*BUCKSIZE,
		       packnum*BUCKSIZE+(BUCKSIZE*2-1),
		       buckets[packnum] + buckets[packnum+1], i/100, i%100);
	    }
	    if (i) {
		while (i > 0) {
		    putchar('*');
		    i -= 200;	/* 2% per nathanhale. */
		}
	    }
	    putchar('\n');
	    if (packnum > 1)
		packnum++;
	}
    }
    if (statflags & ABUCKETS) {
	printf("\n %s Address   %s", "Src", bar);
	printanodes();
    }
    if (statflags & DBUCKETS) {
	printf("\n %s Address   %s", "Dst", bar);
	printdnodes();
    }
    if (statflags & PBUCKETS) {
	printf("\n  Protocol    %s", bar);
	printpnodes();
    }
    if (statflags & BINARY)
	read(0, &packet, 1);
    exit(0);
}

badness(s, value)
char *s;
int value;
{
    if (value)
	printf("%d %s packet%s.  ", value, s, (value > 1) ? "s" : "");
}

/*
** Add thousands-grouping commas to a numeric string.
*/

static void
commas(from, to)
char *from, *to;
{
    int num;
    char *p, *q;

    for (num=0, p=from; *p; p++) { /* Find out how many places before */
	if (*p == '.')		/* the decimal point (if any). */
	    break;
	num++;
    }
    p = from;			/* Add commas for thousands of msecs, etc. */
    q = to;
    while (num > 3) {
	*q++ = *p++;
	if (!(--num % 3))	/* Insert commas every three chars. */
	    *q++ = ',';
    }
    while (*q++ = *p++)
	;
}

/*
** Convert timestamp to a string representing (fractional) milliseconds.
** The non-floating point version is only good to about 42,000 msec before
** arithmetic overflow kills it.
*/

#ifdef NOFLOAT
#define TTYPE unsigned long
#define SCALE 100000
#else
#define TTYPE double
#define SCALE 100000.0
#endif

char *
dotime(packet)
struct packetform *packet;
{
    static unsigned long basetick, lastsubtick, lbasetime, llasttime;
    static TTYPE basetime, lasttime;
    static char buf[20];
    char *p, *q, buf2[20];
    TTYPE retval, abstime, hi, lo;
    unsigned long l;
    unsigned short num, denom;

    num = packet->snh.fracticknum;
    denom = packet->snh.fractickdenom;
#ifdef NOFLOAT
    while (denom >= 1000) {	/* Prescale to avoid overflow.  The display */
	denom /= 10;		/* can't handle too-fine numbers anyway, so */
	num /= 10;		/* the precision loss is acceptable. */
    }
#endif
    if (!basetick)		/* Start ticker at 0 for dump. */
	basetick = packet->snh.tickstamp;
    abstime = SCALE * (packet->snh.tickstamp - basetick)
	/ packet->snh.tickrate;
    abstime += (SCALE * num / denom) / packet->snh.tickrate;
    if (!basetime)
	basetime = lasttime = abstime;
    retval = abstime;
    retval -= basetime;

    lasttime = abstime;
#ifdef NOFLOAT
    sprintf(buf2, "%d.%02d", msecs = retval / 100, retval % 100);
#else
    lo = modf(retval/100.0, &hi);
    sprintf(buf2, "%.0f.%02d", hi, (int) (lo * 100.0));
    msecs = hi;
#endif
    commas(buf2, buf);		/* Add thousands commas. */
    return buf;
}

/*
** Parse a protocol number.
*/

protocol(p)
unsigned char *p;
{
    int i;
    unsigned char ch;

    i = 0;
    if (!strncmp(p, "safe", 4))
	i = SAFE;
    else if (!strncmp(p, "file", 4))
	i = SAFE;
    else if (!strncmp(p, "boot", 4))
	i = BOOT;
    else if (!strcmp(p, "dnet"))
	i = DNET;
    else if (!strcmp(p, "netman"))
	i = DNET;
    else if (!strcmp(p, "dmap"))
	i = DMAP;
    else if (!strcmp(p, "dwho"))
	i = DMAP;
    else if (!strcmp(p, "atalk"))
	i = ATLK;
    else if (!strcmp(p, "appletalk"))
	i = ATLK;
    else if (!strcmp(p, "ethertalk"))
	i = ATLK;
    else if (!strcmp(p, "aarp"))
	i = AARP;
    else if (!strcmp(p, "lat"))
	i = LAT;
    else if (!strcmp(p, "ip"))
	i = IP;
    else if (!strncmp(p, "tcp", 3))
	i = IP;
    else if (!strcmp(p, "arp"))
	i = ARP;
    else {
	ch = *p;
	if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	    || (ch >= 'A' && ch <= 'F')) {
	    i = (h2i(ch) << 12) | (h2i(p[1]) << 8)
		| (h2i(p[2]) << 4) | h2i(p[3]);
	}
    }
    return i;
}

/*
** Convert a hex character to an integer.
*/

h2i(ch)
unsigned char ch;
{
    if (ch > '9')
	ch &= ~('a' - 'A');
    if (ch >= 'A')
	ch -= 'A' - '9' - 1;
    ch -= '0';
    return ch;
}

/*
** Convert a hex ASCII address to a filter address field.
*/

int ishex(register unsigned char ch);

address(p, ea)
unsigned char *p;
struct etheraddr *ea;
{
    int i, j, ch;
    unsigned char *ch2;
    struct etheraddr eabuf;

    ch2 = eabuf.addr;
    for (i=0; i<sizeof(eabuf);) {
	if (!*p || *p <= ' ')
	    return 0;
	if ((j = ishex(*p++)) < 0) /* Skip over garbage (like dashes) */
	    continue;
	ch = j << 4;
	if ((j = ishex(*p++)) < 0)
	    continue;
	*ch2++ = ch | j;
	i++;
    }
    *ea = eabuf;
    return 1;
}

ishex(ch)
unsigned char ch;
{
    if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	|| (ch >= 'A' && ch <= 'F'))
	return h2i(ch);
    return -1;
}

/*
** Turn an ethernet address into a name, if possible.
*/

struct node {
    struct etheraddr ea;
    char *name;
} **nroot;

int
ncompare(n1, n2)
struct node *n1, *n2;
{
    unsigned char *cp1, *cp2;
    int i;

    cp1 = n1->ea.addr;
    cp2 = n2->ea.addr;
    i = sizeof(n1->ea);
    while (i--) {
	if (*cp1 != *cp2) {
	    if (*cp1 > *cp2) {
		return 1;
	    } else {
		return -1;
	    }
	}
	cp1++;
	cp2++;
    }
    return 0;
}

/*
** Convert an ethernet address to a name using the search tree, or
** to hex if there is no matching entry.
*/

void
makename(from, to)
unsigned char *from;
char *to;
{
    struct node **np, **tfind();

    if (np = tfind((struct node *) from, &nroot, ncompare)) {
	sprintf(to, "'%.11s'", (*np)->name);
    } else {
	sprintf(to, "%02X%02X%02X-%02X%02X%02X", from[0], from[1], from[2],
		from[3], from[4], from[5]);
    }
}

void
makerawname(from, to)
unsigned char *from;
char *to;
{
    struct node **np, **tfind();

    if (np = tfind((struct node *) from, &nroot, ncompare)) {
	strcpy(to, (*np)->name);
    }
}

/*
** Read the names file and convert it to an address-to-name search tree.
*/

namesfile(name)
char *name;
{
    FILE *fp;
    char line[80], *cp, *store;
    struct node *nstore;
    int i;
    struct etheraddr ea;

    if (fp = fopen(name, "r")) {
	while (fgets(line, sizeof line, fp)) {
	    line[strlen(line) - 1] = 0; /* Eat the newline. */
	    if (line[0] == '#')	/* Skip comment lines. */
		continue;
	    address(line, &ea);	/* Parse the address. */
	    if (cp = strchr(line, ' ')) {
		cp++;
		i = strlen(cp);
		if (store = malloc(i + 1)) { /* Store the node. */
		    strcpy(store, cp);
		    if (nstore = (struct node *) malloc(sizeof(struct node))) {
			nstore->ea = ea;
			nstore->name = store;
			tsearch(nstore, &nroot, ncompare);
		    } else {
			break;
		    }
		} else {
		    break;
		}
	    }
	}
	fclose(fp);
    }
}

struct anode {
    struct etheraddr ea;
    int bucket;
} **aroot, **droot;

/*
** Find/put an address bucket in its tree.
*/

void
adrbucket(ea, which)
struct etheraddr *ea;
int which;
{
    struct anode *nstore, **found, **tfind(), *tsearch();

    if (found = tfind((struct anode *) ea, which ? &droot : &aroot,
		      ncompare)) {
	(*found)->bucket++;
    } else {
	if (nstore = (struct anode *) malloc(sizeof(struct anode))) {
	    nstore->ea = *ea;
	    nstore->bucket = 1;
	    tsearch(nstore, which ? &droot : &aroot, ncompare);
	}
    }
}

/*
** Find a protocol bucket in its tree.
*/

struct pnode {
    unsigned short protocol;
    int bucket;
} **proot;

int
pcompare(n1, n2)
struct pnode *n1, *n2;
{
    return n1->protocol - n2->protocol;
}

/*
** Find/put protocol in the tree.
*/

void
protbucket(packet)
struct packetform *packet;
{
    struct pnode *pstore, **found, **tfind(), *tsearch();
    struct pnode pn;
    int ssap, dsap;

    if ((pn.protocol = packet->eth.protocol) <= 1514) {
	pn.protocol = SAP;
	dsap = packet->data[0] & ~1;
	ssap = packet->data[1] & ~1;
	if (ssap == NETBIOS_SAP && dsap == NETBIOS_SAP)
	    pn.protocol = NETBIOS_SAP;
	else if (ssap == NETWARE_SAP && dsap == NETWARE_SAP)
	    pn.protocol = NETWARE_SAP;
	else if (ssap == SPAN_SAP && dsap == SPAN_SAP)
	    pn.protocol = SPAN_SAP;
	else if (ssap == IP_SAP && dsap == IP_SAP)
	    pn.protocol = IP_SAP;
	else if (ssap == SNAP_SAP && dsap == SNAP_SAP
		 && packet->data[2] == 0x03)
	    pn.protocol = ((struct snaphdr *)&packet->eth)->ethertype;
	else if (packet->data[0] == 0xFF && packet->data[1] == 0xFF) {
	    pn.protocol = NOVELL;
	}
    }
    if (found = tfind(&pn, &proot, pcompare)) {
	(*found)->bucket++;
    } else {
	if (pstore = (struct pnode *) malloc(sizeof(struct pnode))) {
	    pstore->protocol = pn.protocol;
	    pstore->bucket = 1;
	    tsearch(pstore, &proot, pcompare);
	}
    }
}

/*
** Print out the protocol tree.
*/

printpnode(node, order, level)
struct pnode **node;
VISIT order;
int level;
{
    int i;
    char *cp, buf[6];

    if (order == postorder || order == leaf) {
	cp = 0;
	i = (*node)->bucket * 10000 / packcnt;
	switch ((*node)->protocol) {
	  case SAFE: cp = "ISC Safelink"; break;
	  case BOOT: cp = "ISC Boot"; break;
	  case DNET: cp = "ISC DNET"; break;
	  case DMAP: cp = "ISC DMAP"; break;
	  case ATLK: cp = "Ethertalk"; break;
	  case AARP: cp = "AARP"; break;
	  case LAT: cp = "DEC LAT"; break;
	  case IP: cp = "TCP/IP"; break;
	  case ARP: cp = "ARP"; break;
	  case RARP: cp = "RARP"; break;
	  case NOVELL: cp = "IPX"; break;
	  case SAP: cp = "802.3 SAP"; break;
	  case DECB: cp = "DEC Bridge"; break;
	  case NETBIOS_SAP: cp = "NETBIOS SAP"; break;
	  case NETWARE_SAP: cp = "NOVELL SAP"; break;
	  case SPAN_SAP: cp = "SpanTree SAP"; break;
	  case IP_SAP: cp = "TCP/IP SAP"; break;
	  default: sprintf(buf, "$%04X", (*node)->protocol); cp = buf; break;
	}
	printf("%12s %5d %3d.%02d ", cp, (*node)->bucket, i/100, i%100);
	if (i) {
	    while (i > 0) {
		putchar('*');
		i -= 200;	/* 2% per nathanhale. */
	    }
	}
	putchar('\n');
    }
}

printpnodes()
{
    twalk(proot, printpnode);
}

/*
** Print out the address tree.
*/

printadnode(node, order, level)
struct anode **node;
VISIT order;
int level;
{
    int i;
    char *cp, buf[80];

    if (order == postorder || order == leaf) {
	cp = 0;
	i = (*node)->bucket * 10000 / packcnt;
	makename(&(*node)->ea, buf);
	printf("%13s %5d %3d.%02d ", buf, (*node)->bucket, i/100, i%100);
	if (i) {
	    while (i > 0) {
		putchar('*');
		i -= 200;	/* 2% per nathanhale. */
	    }
	}
	putchar('\n');
    }
}

printanodes()
{
    twalk(aroot, printadnode);
}

printdnodes()
{
    twalk(droot, printadnode);
}
