/*
** Sniffer packet dumper.
**
** Usage:
**	snifferdump [-a] [-C##=name] [-c] [-Ddaddr] [-d | -dr] [-dl size]
**		    [-fnamefile] [-Hprotocol[=prot]] [-h] [-m] [-Pprotocol] 
**		    [-Rprotocol[=prot]] [-r] [-Ssaddr] [-s] [-tpos] [-V] [file]
**
** Where
**	-A	Dumps all packet types, and their data (like -R for all);
**	-a	Dumps all packet types (like -H for all);
**	-Cpacknum=name
**		Forces a packet to have a L6 service name (for DNET);
**	-c	Conversational mode (packets between S and D nodes);
**	-Ddaddr	Dumps packets _to_ a particular address;
**	-d	Timestamps relative to each other, instead of to the first one;
**	-dr	Timestamps in raw, absolute form (for debugging).
**	-dlsize	Dump limit of size bytes.
**	-fnames	Get node names from this file;
**	-Hprot	Disassemble headers for this protocol;
**	-h	Show all headers, regardless of filtering (only disass passed);
**	-m	Mo is the displayer, use highlighting;
**	-Pprot	Only show packets for this protocol;
**	-Rprot	Like -H, but also show packet's data contents;
**	-r	Dump raw data for all packets shown;
**	-Ssaddr	Dumps packets _from_ a particular address;
**	-s	Prevents packets from being flagged as too short;
**	-tpos	Trigger packet is #pos in the capture.  Number accordingly.
**	file	Packets in this file, otherwise read from standard in.
**
** Something like -HFF05=dnet will disassemble  protocol $FF05 packets
** with the netman ($FF02) disassembler, which is useful when testing
** a new version of netman (for example).  More than one -S or -D switch
** may be given.
**
** Output goes to standard out.
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <search.h>
#include <math.h>
#include <dnix/cheapdrv.h>
#include "protos.h"
#include "protocols.h"
#include "sapsnap.h"

#define DUMP_DATA 1		/* Dump options. */
#define DUMP_HEADERS 2
#define DUMP_ALL 4
#define DUMP_ALLPACKS 8
#define DELTATIME 16
#define SRCFILTER 32
#define DSTFILTER 64
#define PROTFILTER 128
#define CONVERSATION 256
#define DUMP_HIGHLIGHTING 512
#define RAWTIME 1024
#define NO_SHORT_WARN 2048

#define SRCFMATCHSRC 1		/* Filter match codes. */
#define SRCFMATCHDST 2
#define DSTFMATCHDST 4
#define DSTFMATCHSRC 8

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[8192];	/* Way bigger than typically needed. */
};

struct dumpcontrol {
    struct dumpcontrol *next;
    unsigned short protocol;
    unsigned short dumpflags;
    unsigned short dumpwith;	/* For protocol cross-dressing! */
} *dchain;

struct L6override {
    struct L6override *next;
    short packnum;
    char *name;
} *ochain;

unsigned short dumpflags;	/* Global dump flags. */
int triggered, tbefore, dumplimit;
int packnum;			/* Packet number under destruction. */

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp;
    unsigned short packetsize;
    unsigned short prot, prot2;
    struct packetform packet;
    char *filename, *progname, *argp, *argp2;
    char from[14], to[14];
    char stats[40];
    struct dumpchain *dc;
    int matchflags;
    struct etheraddr sfilter[8], dfilter[8];
    int sfil, dfil;
    unsigned short pfilter;

    progname = *argv;
    filename = 0;
    sfil = dfil = 0;		/* No source/destination filters. */
    setvbuf(stdout, 0, _IOFBF, 4096);
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'A':		/* Dump all packets and data. */
		dumpflags |= DUMP_DATA;
	      case 'a':		/* Dump all packets. */
		dumpflags |= DUMP_ALLPACKS;
		break;

	      case 'C':		/* Force a L6 service name for a packet. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (argp2 = strchr(argp, '=')) {
		    *argp2++ = 0;
		    addoverride(atoi(argp), argp2);
		}
		break;

	      case 'c':		/* Conversation, src matches src | dest, &c. */
		dumpflags |= CONVERSATION;
		break;

	      case 'D':		/* Dump only particular dest's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (dfil < sizeof(dfilter)/sizeof(dfilter[0]))
		    if (address((unsigned char *) argp, &dfilter[dfil++]))
			dumpflags |= DSTFILTER;
		break;

	      case 'd':
		switch (*argp) {
		  case 'l':	/* Dump limit? */
		    if (!*++argp)
			--argc, argp = *++argv;
		    dumplimit = atoi(argp);
		    break;

		  case 'r':	/* Raw time. */
		    dumpflags |= RAWTIME;
		    break;

		  case 0:	/* Delta-time. */
		    dumpflags |= DELTATIME;
		    break;
		}
		break;

	      case 'f':		/* Read names file. */
		if (!*argp)
		    --argc, argp = *++argv;
		namesfile(argp);
		break;

	      case 'H':		/* Dump named protocol's headers. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (argp2 = strchr(argp, '='))
		    *argp2++ = 0;
		prot = prot2 = protocol((unsigned char *) argp);
		if (argp2)
		    prot2 = protocol((unsigned char *) argp2);
		adddump(prot, DUMP_HEADERS, prot2);
		break;

	      case 'h':		/* Dump all headers. */
		dumpflags |= DUMP_HEADERS;
		break;

	      case 'm':		/* Mo mode (highlighting). */
		dumpflags |= DUMP_HIGHLIGHTING;
		break;

	      case 'P':		/* Dump only particular protocol's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		pfilter = protocol((unsigned char *) argp);
		dumpflags |= PROTFILTER;
		break;

	      case 'R':		/* Dump named protocol's raw data (if any). */
		if (!*argp)
		    --argc, argp = *++argv;
		if (argp2 = strchr(argp, '='))
		    *argp2++ = 0;
		prot = prot2 = protocol((unsigned char *) argp);
		if (argp2)
		    prot2 = protocol((unsigned char *) argp2);
		adddump(prot, DUMP_DATA, prot2);
		break;

	      case 'r':
		dumpflags |= DUMP_ALL;
		break;

	      case 'S':		/* Dump only particular source's packets. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (sfil < sizeof(sfilter)/sizeof(sfilter[0]))
		    if (address((unsigned char *) argp, &sfilter[sfil++]))
			dumpflags |= SRCFILTER;
		break;

	      case 's':		/* No warnings for too-short packets. */
		dumpflags |= NO_SHORT_WARN;
		break;

	      case 't':		/* Trigger packet number. */
		if (!*argp)
		    --argc, argp = *++argv;
		triggered = 1;
		tbefore = atoi(argp);
		break;

	      case 'V':		/* Version printout. */
		printf("%s: Version 1.0\n", progname);
		exit(0);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(1);
    }
    setvbuf(fp, 0, _IOFBF, 4096);
    packnum = -1;
    while (fread(&packetsize, sizeof packetsize, 1, fp) == 1) {
	matchflags = SRCFMATCHSRC | SRCFMATCHDST |
	    DSTFMATCHDST | DSTFMATCHSRC; /* Assume a filter match at first. */
	packnum++;
	packet.snh.statusbits = 0;
	if (packetsize > sizeof(packet)) {
	    printf("ERROR -- %d byte packet is too big!\n", packetsize);
	    fseek(fp, packetsize, 1); /* Skip over it. */
	} else if (packetsize <
		   (sizeof(struct etherheader) + sizeof(struct sniffheader))) {
	    if (!packetsize)	/* It's a runt packet. */
		packetsize = sizeof packetsize;
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    printf("#%-4d  RUNT   %4d bytes: ", packnum - tbefore,
		   packet.snh.length - sizeof(struct sniffheader));
	    {
		int i, j, k;
		unsigned char *bufp1;

		i = packet.snh.length - sizeof(struct sniffheader);
		bufp1 = (unsigned char *) &packet.eth;
		k = 0;
		for (j = 0; j < i; j++) { /* Display hex. */
		    if (!(j % 2) && (j != 0)) {
			putc(' ', stdout);
			k++;
		    }
		    printf("%02X", *bufp1++);
		    k += 2;
		}
		printf("%*s", 34 - k, "");
	    }
	    printf(", %sT: %9s msec%s\n",
		       (dumpflags & DELTATIME) ? "d" : "",
		       dotime(&packet), status(stats, &packet));
	} else {
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    if (packet.snh.statusbits == 0xFF) { /* It's a format-override, */
		packnum--;	/* so skip it. */
		continue;
	    }
	    if (!packnum)	/* First packet? */
		dotime(&packet); /* Prime the timer logic. */
	    if (dumpflags & PROTFILTER && packet.eth.protocol != pfilter)
		matchflags = 0;
	    if (dumpflags & SRCFILTER) {
		int fil, knockout;

		knockout = ~(SRCFMATCHSRC | SRCFMATCHDST); /* Assume no mtch */
		for (fil=0; fil<sfil; fil++) { /* Check all the candidates. */
		    if ((*(short *) &packet.eth.srcaddr
			 == *(short *) &sfilter[fil] &&
			 *(long *) &packet.eth.srcaddr[2]
			 == *(long *) &sfilter[fil].addr[2]))
			knockout |= SRCFMATCHSRC;
		    if ((dumpflags & CONVERSATION) &&
			(*(short *) &packet.eth.dstaddr
			 == *(short *) &sfilter[fil] &&
			 *(long *) &packet.eth.dstaddr[2]
			 == *(long *) &sfilter[fil].addr[2]))
			knockout |= SRCFMATCHDST;
		}
		matchflags &= knockout;
	    } else if (dumpflags & DSTFILTER)
		matchflags &= ~(SRCFMATCHSRC|SRCFMATCHDST);
	    if (dumpflags & DSTFILTER) {
		int fil, knockout;

		knockout = ~(DSTFMATCHDST | DSTFMATCHSRC); /* Assume no mtch */
		for (fil=0; fil<dfil; fil++) { /* Check all the candidates. */
		    if ((*(short *) &packet.eth.dstaddr
			 == *(short *) &dfilter[fil] &&
			 *(long *) &packet.eth.dstaddr[2]
			 == *(long *) &dfilter[fil].addr[2]))
			knockout |= DSTFMATCHDST;
		    if ((dumpflags & CONVERSATION) &&
			(*(short *) &packet.eth.srcaddr
			 == *(short *) &dfilter[fil] &&
			 *(long *) &packet.eth.srcaddr[2]
			 == *(long *) &dfilter[fil].addr[2]))
			knockout |= DSTFMATCHSRC;
		}
		matchflags &= knockout;
	    } else if (dumpflags & SRCFILTER)
		matchflags &= ~(DSTFMATCHSRC|DSTFMATCHDST);
	    if (dumpflags & CONVERSATION)
		if (matchflags != (SRCFMATCHSRC | DSTFMATCHDST)
		    && matchflags != (SRCFMATCHDST | DSTFMATCHSRC)
		    && !(packet.eth.dstaddr[0] & 0x01))	/* Pass B/Mcasts. */
		    matchflags = 0;
	    if (!matchflags && !(dumpflags & DUMP_HEADERS))
		continue;
	    makename(packet.eth.dstaddr, to);
	    makename(packet.eth.srcaddr, from);
	    if (packet.snh.statusbits & 128) { /* Trigger packet? */
		if (dumpflags & DUMP_HIGHLIGHTING)
		    putc('\201', stdout);
		printf("TRIG  ($%04X) %4d bytes to %-13.13s from %-13.13s, "
		       "%sT: %9s %s%s\n",
		       packet.eth.protocol,
		       packet.snh.length - sizeof(struct sniffheader),
		       to, from, (dumpflags & DELTATIME) ? "d" : "",
		       dotime(&packet), (dumpflags & RAWTIME) ? "ticks":"msec",
		       status(stats, &packet));
		if (dumpflags & DUMP_HIGHLIGHTING)
		    putc('\200', stdout);
	    } else {
		printf("#%-4d ($%04X) %4d bytes to %-13.13s from %-13.13s, "
		       "%sT: %9s %s%s\n",
		       packnum - tbefore, packet.eth.protocol,
		       packet.snh.length - sizeof(struct sniffheader),
		       to, from, (dumpflags & DELTATIME) ? "d" : "",
		       dotime(&packet),(dumpflags & RAWTIME) ? "ticks":"msec",
		       status(stats, &packet));
	    }
	    if (!matchflags)
		continue;
	    dump(&packet.eth, packet.snh.length - sizeof(struct sniffheader));
	}
	if (packet.snh.statusbits & PACKET_LOST)
	    fputs("LOST one or more packets here.\n", stdout);
    }
    exit(0);
}

/*
** Convert timestamp to a string representing (fractional) milliseconds.
** The non-floating point version is only good to about 42,000 msec before
** arithmetic overflow kills it.
*/

#ifdef NOFLOAT
#define TTYPE unsigned long
#define SCALE 100000
#else
#define TTYPE double
#define SCALE 100000.0
#endif

char *
dotime(packet)
struct packetform *packet;
{
    static unsigned long basetick, lastsubtick, lbasetime, llasttime;
    static TTYPE basetime, lasttime;
    static char buf[20];
    char *p, *q, buf2[20];
    TTYPE retval, abstime, hi, lo;
    unsigned long l;
    unsigned short num, denom;

    if (dumpflags & RAWTIME) {	/* Print timestamp 'raw' (driver time). */
	if (!lbasetime)
	    lbasetime = llasttime = packet->snh.tickstamp;
	l = packet->snh.tickstamp;
	if (dumpflags & DELTATIME)
	    l -= lasttime;
	else
	    l -= basetime;
	sprintf(buf, "%d+%d/%d", l, packet->snh.fracticknum - lastsubtick,
		packet->snh.fractickdenom);
	if (dumpflags & DELTATIME) {
	    llasttime = packet->snh.tickstamp;
	    lastsubtick = packet->snh.fracticknum;
	}
    } else {			/* Print out as a time. */
	num = packet->snh.fracticknum;
	denom = packet->snh.fractickdenom;
#ifdef NOFLOAT
	while (denom >= 1000) {	/* Prescale to avoid overflow.  The display */
	    denom /= 10;	/* can't handle too-fine numbers anyway, so */
	    num /= 10;		/* the precision loss is acceptable. */
	}
#endif
	if (!basetick)		/* Start ticker at 0 for dump. */
	    basetick = packet->snh.tickstamp;
	abstime = SCALE * (packet->snh.tickstamp - basetick)
	    / packet->snh.tickrate;
	abstime += (SCALE * num / denom) / packet->snh.tickrate;
	if (!basetime)
	    basetime = lasttime = abstime;
	retval = abstime;
	if (dumpflags & DELTATIME)
	    retval -= lasttime;
	else
	    retval -= basetime;

	lasttime = abstime;
#ifdef NOFLOAT
	sprintf(buf2, "%d.%02d", retval / 100, retval % 100);
#else
	lo = modf(retval/100.0, &hi);
	sprintf(buf2, "%.0f.%02d", hi, (int) (lo * 100.0));
#endif
        for (num=0, p=buf2; *p; p++) { /* Find out how many places before */
            if (*p == '.')	/* the decimal point. */
                break;
            num++;
        }
        p = buf2;		/* Add commas for thousands of msecs, etc. */
        q = buf;
        while (num > 3) {
            *q++ = *p++;
            if (!(--num % 3))     /* Insert commas every three chars. */
                *q++ = ',';
        }
        while (*q++ = *p++)
            ;
    }
    return buf;
}

char *
status(buf, packet)
char *buf;
struct packetform *packet;
{
    unsigned char rsr;
    int len;

    *buf = 0;
    rsr = packet->snh.statusbits;
    len = packet->snh.length
	- (sizeof(struct sniffheader) + sizeof(struct etherheader));
    if (len > 1500)		/* Legal Ethernet size ranges. */
	strcat(buf, " LONG");
    if ((len < 46) && (packet->eth.protocol != WINCOM)
	&& !(dumpflags & NO_SHORT_WARN))
	strcat(buf, " SHORT");
    if (rsr & PACKET_OVERRUN)
	strcat(buf, " OVRRUN");
    if (rsr & PACKET_FRAME)
	strcat(buf, " FRAME");
    if (rsr & PACKET_CRC) {
	if (packet->eth.protocol != WINCOM)
	    strcat(buf, " CRC");
	else
	    strcat(buf, " PARITY");
    }
    return buf;
}

/*
** Parse a protocol number.
*/

protocol(p)
unsigned char *p;
{
    int i;
    unsigned char ch;

    i = 0;
    if (!strncmp(p, "safe", 4))
	i = SAFE;
    else if (!strncmp(p, "file", 4))
	i = SAFE;
    else if (!strncmp(p, "boot", 4))
	i = BOOT;
    else if (!strcmp(p, "dnet"))
	i = DNET;
    else if (!strcmp(p, "netman"))
	i = DNET;
    else if (!strcmp(p, "dmap"))
	i = DMAP;
    else if (!strcmp(p, "dwho"))
	i = DMAP;
    else if (!strcmp(p, "atalk"))
	i = ATLK;
    else if (!strcmp(p, "appletalk"))
	i = ATLK;
    else if (!strcmp(p, "ethertalk"))
	i = ATLK;
    else if (!strcmp(p, "aarp"))
	i = AARP;
    else if (!strcmp(p, "lat"))
	i = LAT;
    else if (!strcmp(p, "ip"))
	i = IP;
    else if (!strncmp(p, "tcp", 3))
	i = IP;
    else if (!strcmp(p, "arp"))
	i = ARP;
    else if (!strcmp(p, "rarp"))
	i = RARP;
    else if (!strcmp(p, "wincom"))
	i = WINCOM;
    else if (!strcmp(p, "novell"))
	i = NOVELL;
    else if (!strcmp(p, "sap"))
	i = SAP;		/* SAP header display. */
    else {
	ch = *p;
	if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	    || (ch >= 'A' && ch <= 'F')) {
	    i = (h2i(ch) << 12) | (h2i(p[1]) << 8)
		| (h2i(p[2]) << 4) | h2i(p[3]);
	}
    }
    return i;
}

/*
** Convert a hex character to an integer.
*/

h2i(ch)
unsigned char ch;
{
    if (ch > '9')
	ch &= ~('a' - 'A');
    if (ch >= 'A')
	ch -= 'A' - '9' - 1;
    ch -= '0';
    return ch;
}

/*
** Add a dump flag to a given protocol number's control structure.
** Create it if necessary.
*/

adddump(prot, value, dumpwith)
unsigned short prot, value, dumpwith;
{
    struct dumpcontrol *dc;

    dc = findprot(prot);
    if (dc) {
	dc->dumpflags |= value;
    } else {
	if (dc = (struct dumpcontrol *) malloc(sizeof(struct dumpcontrol))) {
	    dc->protocol = prot;
	    dc->dumpflags = value;
	    dc->dumpwith = dumpwith;
	    dc->next = dchain;
	    dchain = dc;
	}
    }
}

/*
** Find a protocol match in the dump control chain.
*/

struct dumpcontrol *
findprot(prot)
unsigned short prot;
{
    struct dumpcontrol *dc;

    dc = dchain;
    while (dc) {
	if (dc->protocol == prot)
	    return dc;
	dc = dc->next;
    }
    return 0;
}

/*
** Return the dump control flags for a given protocol.
*/

protperm(prot)
unsigned short prot;
{
    struct dumpcontrol *dc;

    if (dc = findprot(prot))
	return dc->dumpflags;
    else
	return 0;
}

/*
** Add a L6 override structure.
*/

void
addoverride(packnum, name)
short packnum;
char *name;
{
    struct L6override *l6;

    if (l6 = (struct L6override *) malloc(sizeof(struct L6override))) {
	l6->packnum = packnum;
	l6->name = name;
	l6->next = ochain;
	ochain = l6;
    }
}

/*
** Find an override structure.
*/

char *
findoverride(packet)
short packet;
{
    struct L6override *l6;

    l6 = ochain;
    while (l6) {
	if (l6->packnum == packet - tbefore)
	    return l6->name;
	l6 = l6->next;
    }
    return 0;
}

/*
** Do the dumping of packets.
*/

void
dump(p, size)
struct etherheader *p;
unsigned short size;
{
    int dumped, perm, hdrsize;
    unsigned short prot;
    struct dumpcontrol *dp;
    struct snaphdr *snap;
    struct saphdr *sap;

    dumped = 0;			/* Track dumping for extra blank line. */
    hdrsize = 0;
    if (perm = (protperm(p->protocol) | (dumpflags
					 & (DUMP_DATA | DUMP_ALLPACKS)))) {
	if (dp = findprot(p->protocol))
	    prot = dp->dumpwith;
	else
	    prot = p->protocol;
	hdrsize = sizeof(*p);
	switch (prot) {
	  case SAFE:		/* Safelink. */
	    dumped = 1;
	    hdrsize = dumpsafe(p, (int) size, perm & DUMP_DATA);
	    break;

	  case BOOT:		/* Boot protocol. */
	    dumped = 1;
	    hdrsize = dumpboot(p, (int) size, perm & DUMP_DATA);
	    break;

	  case DNET:		/* D-net protocol. */
	    dumped = 1;
	    hdrsize = dumpdnet(p, (int) size, perm & DUMP_DATA);
	    break;

	  case DMAP:		/* Dmap protocol. */
	    dumped = 1;
	    hdrsize = dumpdmap(p, (int) size);
	    break;

	  case ATLK:		/* Appletalk. */
	    dumped = 1;
	    hdrsize += dumpatalk(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case AARP:		/* Appletalk ARP. */
	    dumped = 1;
	    hdrsize += dumpaarp(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case LAT:		/* DEC's LAT. */
	    dumped = 1;
	    hdrsize += dumplat(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case IP:		/* Internet IP protocol. */
	    dumped = 1;
	    hdrsize += dumpip(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case ARP:		/* Internet ARP protocol. */
	    dumped = 1;
	    hdrsize += dumparp(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case RARP:		/* Reverse ARP protocol. */
	    dumped = 1;
	    hdrsize += dumprarp(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case NOVELL:
	    dumped = 1;
	    hdrsize += dumpxns(&p[1], size - sizeof(*p), perm & DUMP_DATA);
	    break;

	  case WINCOM:		/* Wincom (non-Ethernet) protocol. */
	    dumped = 1;
	    hdrsize = dumpwincom(p, (int) size, perm & DUMP_DATA);
	    break;
	}
    }
    if (!dumped) {		/* No Ethertype dump, try 802.3 SAP/SNAP. */
	if (p->protocol <= 1514) { /* Small enough for 802.3? */
	    snap = (struct snaphdr *) p;
	    if (snap->llc8023[0] == SNAP_SAP && snap->llc8023[1] == SNAP_SAP
		&& snap->llc8023[2] == 0x03) { /* Has 802.3 SNAP header. */
		perm |= protperm(snap->ethertype); /* Dump it? */
		if (perm) {
		    printf("\tSNAP\t802.3 Length: %d, ProtID/Org: "
			   "$%02X%02X%02X, Ethertype: $%04X\n", p->protocol,
			   snap->protorg[0], snap->protorg[1],
			   snap->protorg[2], snap->ethertype);
		    hdrsize = (unsigned char *) &snap[1] - (unsigned char *) p;
		    switch (snap->ethertype) {
		      case ATLK: /* Appletalk. */
			hdrsize += dumpatalk2(&snap[1], (int) p->protocol -
					      ((char *) &snap[1] - (char *) p),
					      perm & DUMP_DATA);
			break;

		      case AARP: /* Appletalk ARP. */
			hdrsize += dumpaarp(&snap[1], (int) p->protocol -
					      ((char *) &snap[1] - (char *) p),
					    perm & DUMP_DATA);
			break;

		      case IP:	/* Internet IP protocol. */
			hdrsize += dumpip(&snap[1], (int) p->protocol -
					  ((char *) &snap[1] - (char *) p),
					  perm & DUMP_DATA);
			break;

		      case ARP:	/* Internet ARP protocol. */
			hdrsize += dumparp(&snap[1], (int) p->protocol -
					   ((char *) &snap[1] - (char *) p),
					   perm & DUMP_DATA);
			break;

		      case NOVELL: /* IPX protocol. */
			hdrsize += dumpxns(&snap[1], (int) p->protocol -
					   ((char *) &snap[1] - (char *) p),
					   perm & DUMP_DATA);
			break;
		    }
		    dumped = 1;
		}
	    } else if (snap->llc8023[0] == 0xFF && snap->llc8023[1] == 0xFF) {
				/* Piece-of-shit Novell violates 802 specs. */
				/* Assume that this is one of them. */
		perm |= protperm(NOVELL); /* Dump it? */
		if (perm) {
		    printf("\tERROR\tViolation of Ethernet/802.3 specs.  "
			   "Assume that this is Novell.\n");
		    hdrsize = sizeof(*p);
		    hdrsize += dumpxns(&p[1], p->protocol, perm & DUMP_DATA);
		    dumped = 1;
		}
	    } else {		/* Is it SAP? */
		sap = (struct saphdr *) p;
#if 1
		perm |= protperm(SAP); /* Dump it? */
#else
		perm |= protperm(sap->dsap); /* Dump it? */
#endif
		if (perm) {
		    hdrsize = dumpsap(p, size, perm & DUMP_DATA);
		    dumped = 1;
		}
	    }
	}
    }
    if (!(dumpflags & DUMP_HIGHLIGHTING))
	hdrsize = 0;
    if (dumpflags & DUMP_ALL) {
	dumped = 1;
	hd((unsigned char *) p, (int) size, hdrsize);
    }
    if (dumped)
	printf("\n");
}

/*
** Dump data in raw hex format.
*/

void
hd(buffer, byte_count, hlcount)
unsigned char *buffer;
int byte_count;
int hlcount;			/* How many bytes to highlight. */
{
    unsigned char *bufp1, *bufp2, ch, ch2;
    char linebuf[80], *p;
    int to_prt;
    int i, j, k;

    bufp1 = bufp2 = buffer;
    for (i = 0; i < byte_count; i += 16) {
	to_prt = (i + 16 > byte_count) ? byte_count - i : 16 ;
	j = bufp1 - buffer;
	if (dumplimit && j >= dumplimit) {
	    fputs("           ...\n", stdout);
	    break;
	}
	p = linebuf;		/* Don't use printf, for speed. */
	for (k=0; k<12; k++)
	    *p++ = ' ';
	ch = j >> 8;
	ch2 = '0' + (ch & 0x0F);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	ch = j & 0xFF;
	ch2 = '0' + (ch >> 4);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	ch2 = '0' + (ch & 0x0F);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	*p++ = ':';
	*p++ = ' ';
	if (hlcount)
	    *p++ = '\201';
	for (j = 0; j < to_prt; j++) { /* Display hex. */
	    if (!(j % 2) && (j != 0)) {
		*p++ = ' ';
	    }
	    ch = *bufp1++;
	    ch2 = '0' + (ch >> 4);
	    if (ch2 > '9')
		ch2 += 'A' - '9' - 1;
	    *p++ = ch2;
	    ch2 = '0' + (ch & 0x0F);
	    if (ch2 > '9')
		ch2 += 'A' - '9' - 1;
	    *p++ = ch2;
	    if (hlcount) {
		if (!--hlcount)
		    *p++ = '\200';
	    }
	}
	if (hlcount) {
	    *p++ = '\200';
	}
	for (; j < 16; j++) {	/* Fill out rest of line with dots. */
	    if (!(j % 2) && (j != 0)) {
		*p++ = ' ';
	    }
	    *p++ = '.';
	    *p++ = '.';
	}
	*p++ = ' ';
	*p++ = '|';
	*p++ = ' ';
	for (j = 0; j < to_prt; j++) { /* Display ASCII */
	    if (*bufp2 >= ' ' && *bufp2 <= 0x7E)
		*p++ = *bufp2;
	    else
		*p++ = '.';
	    bufp2++; 
	}
	*p++ = '\n';
	*p++ = 0;
	fputs(linebuf, stdout);
    }
}

/*
** Turn an ethernet address into a name, if possible.
*/

struct node {
    struct etheraddr ea;
    char *name;
} *root;

int
ncompare(n1, n2)
struct node *n1, *n2;
{
    unsigned char *cp1, *cp2;
    int i;

    cp1 = n1->ea.addr;
    cp2 = n2->ea.addr;
    i = sizeof(n1->ea);
    while (i--) {
	if (*cp1 != *cp2) {
	    if (*cp1 > *cp2) {
		return 1;
	    } else {
		return -1;
	    }
	}
	cp1++;
	cp2++;
    }
    return 0;
}

/*
** Convert an ethernet address to a name using the search tree, or
** to hex if there is no matching entry.
*/

void
makename(from, to)
unsigned char *from;
char *to;
{
    struct node **np, **tfind();

    if (np = tfind((struct node *) from, &root, ncompare)) {
	sprintf(to, "'%.11s'", (*np)->name);
    } else {
	sprintf(to, "%02X%02X%02X-%02X%02X%02X", from[0], from[1], from[2],
		from[3], from[4], from[5]);
    }
}

void
makerawname(from, to)
unsigned char *from;
char *to;
{
    struct node **np, **tfind();

    if (np = tfind((struct node *) from, &root, ncompare)) {
	strcpy(to, (*np)->name);
    }
}

/*
** Read the names file and convert it to an address-to-name search tree.
*/

namesfile(name)
char *name;
{
    FILE *fp;
    char line[80], *cp, *store;
    struct node *nstore;
    int i;
    struct etheraddr ea;

    if (fp = fopen(name, "r")) {
	while (fgets(line, sizeof line, fp)) {
	    line[strlen(line) - 1] = 0; /* Eat the newline. */
	    if (line[0] == '#')	/* Skip comment lines. */
		continue;
	    address(line, &ea);	/* Parse the address. */
	    if (cp = strchr(line, ' ')) {
		cp++;
		i = strlen(cp);
		if (store = malloc(i + 1)) { /* Store the node. */
		    strcpy(store, cp);
		    if (nstore = (struct node *) malloc(sizeof(struct node))) {
			nstore->ea = ea;
			nstore->name = store;
			tsearch(nstore, &root, ncompare);
		    } else {
			break;
		    }
		} else {
		    break;
		}
	    }
	}
	fclose(fp);
    }
}

/*
** Convert a hex ASCII address to a filter address field.
*/

int ishex(register unsigned char ch);

address(p, ea)
unsigned char *p;
struct etheraddr *ea;
{
    int i, j, ch;
    unsigned char *ch2;
    struct etheraddr eabuf;

    ch2 = eabuf.addr;
    for (i=0; i<sizeof(eabuf);) {
	if (!*p || *p <= ' ')
	    return 0;
	if ((j = ishex(*p++)) < 0) /* Skip over garbage (like dashes) */
	    continue;
	ch = j << 4;
	if ((j = ishex(*p++)) < 0)
	    continue;
	*ch2++ = ch | j;
	i++;
    }
    *ea = eabuf;
    return 1;
}

ishex(ch)
unsigned char ch;
{
    if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	|| (ch >= 'A' && ch <= 'F'))
	return h2i(ch);
    return -1;
}

