/*
** Sniffer-like program for new OS cheapdrive options.  Grabs
** network packets for later analysis by other programs.
**
** Usage:  sniffnet [-V] [-a] [-b] [-c] [-d device] [-D Dstaddrfilter] [-fx]
**		    [-fd filtermatchdata] [-fo filtermatchoffset]
**		    [-k key] [-l] [-m] [-n bufsize] [-p] [-P Protocolfilter]
**		    [-S Srcaddrfilter] [-s] [-td triggermatchdata]
**		    [-ti] [-to triggeroffset] [-tp trigpos]
**
** where
**	-a enables audio sound-effects on WP5/WP6, a click per packet.
**	-b captures bad packets too.  Runt packets and errored packets both.
**	-c captures in 'conversational' mode, where traffic _between_ the
**	   nodes selected with the -D and -S filters will be captured (both
**	   these flags must be used for this to work).
**	-d use the given device as a tap.
**	-D captures only packets sent to <Dstaddr>.
**	-fd capture only packets that match the given data pattern.  Pattern
**	   is in hex, but may mix in binary and octal with '@' and '%' prefixes
**	   hex prefix is '$'.  Multiple -fd's may be given and are OR-ed
**	   together.
**	-fo starting offset (hex) in packet to match with the filter data.
**	   Must precede the -fd option it's to be associated with.
**	-fx filter pattern(s) exclude rather than include matches.
**	-k use the given shared-memory key to open a shared memory segment
**	   in which to keep status in, for an external monitor.
**	-l locks the whole process in memory, not just the capture buffers
**	   (which are required to be locked by the cheapernet driver).
**	   Prevents it from being swapped in and out, which would pollute
**	   the session if the sniffer's being run on a diskless system.
**	-m reads many packets per call to the driver.  More efficient, with
**	   less likelihood of missing packets.  On the other hand, it's a
**	   lot less 'interactive' this way (if you're using -k).
**	-n Sets the buffer size to N kilobytes, 0 writes output immediately.
**	-p sets promiscuous mode, where _all_ traffic is captured, not just
**	   the stuff to/from the system the sniffer's being run on.
**	-P sets the protocol filter, where only packets of the given protocol
**	   will be captured.
**	-S captures only packets sent from <Srcaddr>.
**	-s stops only on a signal, not when buffer is full.
**	-td start capturing packets that match the given data pattern.
**	    Multiple -td's may be given and are OR-ed together.
**	-ti trigger inverted, when _no_ match of the trigger data pattern(s).
**	-to starting offset (hex) in packet to match with the trigger data.
**	    Must precede the -td option it's to be associated with.
**	-tp trigger packet positioned at <trigpos> % of buffer.
**	-V prints version number and exits.
**
** The -S and -D flags are OR-ed together, so that you may capture all
** data associated with a particular network address (both from and to).  
** Alternatively, you can set the 'conversation' flag (-c) and the -S and -D
** filters will be applied to both the source and destination network addresses
** so that packets sent between two nodes will all be captured.  Addresses
** are given in hex, of the form MMMMMM-NNNNNN.  The -P flag sets the packet
** type filter, either specified in hex as NNNN, or by one of the names:
**
**	safe	($FF00)		DNET file server protocol, 'safelink'.
**	boot	($FF01)		DNET boot server protocol.
**	dnet	($FF02)		DNET client-server protocol, a.k.a. netman.
**	dmap	($FF03)		DNET status monitoring protocol.
**	atalk	($809B)		Appletalk (Ethertalk).
**	aarp	($80F3)		Appletalk Address Resolution Protocol.
**	lat	($6004)		DEC LAT protocol.
**	ip	($0800)		Internet Protocol (TCP/IP).
**	arp	($0806)		Address Resolution Protocol (TCP/IP).
**	rarp	($8035)		Reverse Address Resolution Protocol (TCP/IP).
**	ipx	($8137)		Internetwork Packet Exchange (IPX, Novell).
**
** Sniffnet sends its output to standard out, so it may be redirected to
** a file or piped somewhere.  Sniffnet operates by starting up a number of
** no-wait reads to the driver, and copying the results into a large
** collection buffer.  This collection buffer is dumped en masse at the
** end of collection.  The timestamps are very fine on a workstation (about
** .05 msec of resolution), and very coarse on a DNP20 (20 msec).  This
** is due to the lack of hardware available for fine timing on the DNP20.
**
** Version 1.2		11/10/94	Jim Cathey
**	Added multiple trigger/filter patterns and repeat count pattern syntax.
**
** Version 1.1		12/3/93		Jim Cathey
**	Added octal/binary filter/trigger patterns.  Converted offset to hex.
**	Fixed bug in '-ti' and '-fx' option parsing.
**
** Version 1.0		12/1/92		Jim Cathey
**	Original version.
*/

#include <malloc.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/lock.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <dnix/fcodes.h>
#include <dnix/errno.h>
#include <dnix/itrq.h>
#include <dnix/cheapdrv.h>
#include "protocols.h"
#include "sharedmem.h"

#define BUFK 128		/* Size of buffer (in K) by default. */

#define EBUFSIZE 2048		/* One-at-a-time read buffer size. */
#define NREADS 16		/* Number of these reads up at once. */

#define EBUFSIZE2 32768		/* Many packets read buffer size. */
#define NREADS2 2		/* Number of these reads up at once. */

struct  utimbuf {
    time_t  actime;
    time_t  modtime;
};
struct stat pstat;

struct ebuf {			/* For small reads. */
    union {
	unsigned short packetsize;
	void (*iodone)();	/* Indirect completion routine. */
    } u;
    int reqnum;
    unsigned char data[EBUFSIZE];
};

struct ebuf2 {			/* For large (-m mode) reads. */
    union {
	unsigned short packetsize;
	void (*iodone)();	/* Indirect completion routine. */
    } u;
    int reqnum;
    unsigned char data[EBUFSIZE2];
};

struct ebuf *ebufs;
struct ebuf2 *ebufs2;
unsigned char *bigbuf, *putp, *getp, *bufend, *curbufend;
long bufcount;
long caplimit;
int etherfd, trapqfd;
struct itrq trapnode;
void readdone(), cancelout();
char *mysprintf();
int protocol(unsigned char * p);
int h2i(register unsigned char ch);
char *progname;
char *device = "/dev/sniffer";
int nextbuf;
int bufk;			/* Size of I/O buffer (in K). */
int nreads;			/* Number of simultaneous reads up. */
int readsup;			/* Number of reads currently posted. */
int canceled;			/* TRUE when bailing out early. */
int multimode;			/* TRUE for many packets per read. */
int lockem;			/* TRUE to lock the process down (WS). */
int nostop;			/* TRUE if not to stop when buffer's full. */
int oldpri;			/* Old process priority, for restoration. */
int trigger;			/* TRUE if trigger match starts recording. */
int filter;			/* TRUE if data match required to record. */
int capturing;			/* TRUE if data's to be stored. */
int trigoffset;			/* Offset of start of trigger string. */
int trigpos;			/* Percentage in buffer of trigger packet. */
int triggerinv;			/* Trigger's inverted. */
int filtoffset;			/* Offset of start of filter string. */
int filterexc;
extern long stamp;
int shmkey, shmid;
struct sniffstat *shmp;

main(argc, argv)
int argc;
char *argv[];
{
    int i;
    long bufsize;
    struct snifferctrl sniffctrl;
    char scratch[256];
    char *argp;
    register void (**procptr)(void *struc, int retparm, int reqnum);
    void *calloc();

    sniffctrl.sniffmode = SNIFFER_ENABLED | SNIFFER_ADDHEADERS
	| SNIFFER_BMCAST | SNIFFER_LISTENTOSELF;
    bufk = BUFK;
    nreads = NREADS;
    progname = *argv;
    capturing = 1;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'a':		/* Add audio feedback. */
		sniffctrl.sniffmode |= SNIFFER_SOUND;
		break;

	      case 'b':		/* Recv bad packets too. */
		sniffctrl.sniffmode |= SNIFFER_RECVBADPACKS;
		break;

	      case 'c':		/* Receive both ends of a conversation. */
		sniffctrl.sniffmode |= SNIFFER_CONVERSATION;
		break;

	      case 'd':		/* Override device */
		if (!*argp)
		    --argc, argp = *++argv;
		device = argp;
		break;

	      case 'D':		/* Destination address filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (address(argp, &sniffctrl.dstfilter))
		    sniffctrl.sniffmode |= SNIFFER_DSTADDRFILTER;
		break;

	      case 'f':		/* Filter. */
		filter = 1;
		i = *argp++;
		if (i != 'x' && !*argp)
		    --argc, argp = *++argv;
		switch (i) {
		  case 'd':	/* Filter data. */
		    while (argp = strtok(argp, "\n\r")) {
			if (!setfilter(argp))
			    goto nomemory;
			argp = 0;
		    }
		    break;

		  case 'o':	/* Filter offset. */
		    filtoffset = myxatoi(argp);
		    break;

		  case 'x':	/* Filter excludes. */
		    filterexc = 1;
		    break;
		}
		break;

	      case 'k':		/* Shared-memory key. */
		if (!*argp)
		    --argc, argp = *++argv;
		shmkey = myatoi(argp);
		break;

	      case 'l':		/* Lock the process too. */
		if (!geteuid())	/* Don't bother if not superuser, it won't */
		    lockem = 1;	/* work.  (We'll get a misleading error */
		break;		/* message if we try.) */

	      case 'm':		/* Many packets, big buffer. */
		sniffctrl.sniffmode |= SNIFFER_MANYPACKETS;
		nreads = NREADS2;
		multimode = 1;
		break;

	      case 'n':		/* Size of buffer. */
		if (!*argp)
		    --argc, argp = *++argv;
		i = myatoi(argp);
		if (i >= 0 && i <= 4096)
		    bufk = i;
		break;

	      case 'p':		/* Promiscuous mode. */
		sniffctrl.sniffmode |= SNIFFER_PROMISCUOUS;
		break;

	      case 'P':		/* Protocol filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		sniffctrl.protocolfilter = protocol((unsigned char *)argp);
		sniffctrl.sniffmode |= SNIFFER_PROTOCOLFILTR;
		break;

	      case 'S':		/* Source address filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		if (address(argp, &sniffctrl.srcfilter))
		    sniffctrl.sniffmode |= SNIFFER_SRCADDRFILTER;
		break;

	      case 's':		/* No-stop when buffer's full. */
		nostop = 1;
		break;

	      case 't':		/* Trigger. */
		trigger = 1;
		capturing = 0;
		i = *argp++;
		if (i != 'i' && !*argp)
		    --argc, argp = *++argv;
		switch (i) {
		  case 'd':	/* Trigger data. */
		    while (argp = strtok(argp, "\n\r")) {
			if (!settrigger(argp))
			    goto nomemory;
			argp = 0;
		    }
		    break;

		  case 'i':	/* Trigger inverted. */
		    triggerinv = 1;
		    break;

		  case 'o':	/* Trigger offset. */
		    trigoffset = myxatoi(argp);
		    break;

		  case 'p':	/* Trigger position. */
		    if (trigpos = myatoi(argp))
			capturing = 1;
		    if (trigpos > 100)
			trigpos = 100;
		    break;
		}
		break;

	      case 'V':		/* Version printout. */
		strcpy(scratch, progname);
		strcat(scratch, " Version 1.2\n");
		write(1, scratch, strlen(scratch));
		exit(0);
		break;
	    }
	} else {
	    ;
	}
    }
    if (multimode) {
	ebufs2 = (struct ebuf2 *) calloc(sizeof(struct ebuf2), nreads);
    } else {
	ebufs = (struct ebuf *) calloc(sizeof(struct ebuf), nreads);
    }
    if (!ebufs && !ebufs2) {
      nomemory:
	mysprintf(scratch, "%s: Unable to allocate buffers.\n", progname);
	(void) write(2, scratch, strlen(scratch));
	exit(2);
    }
    bufsize = bufk * 1024;
    bigbuf = (unsigned char *) malloc(bufsize);
    if (!bigbuf)
	goto nomemory;
    putp = getp = curbufend = bigbuf;
    bufend = putp + bufsize;

    if (multimode) {		/* Lock down I/O buffer memory. */
	i = dnix(F_LOCKMEM, ebufs2, nreads * sizeof(struct ebuf2));
    } else {
	i = dnix(F_LOCKMEM, ebufs, nreads * sizeof(struct ebuf));
    }
    if (i < 0) {
	mysprintf(scratch, "%s: Unable to lock down I/O buffer memory (%d).\n",
		  progname, i);
	(void) write(2, scratch, strlen(scratch));
	exit(3);
    }
    if (shmkey) {		/* Shared-memory stats? */
	if ((shmid = shmget(shmkey, sizeof(*shmp), 0666)) != -1) {
	    if ((shmp = (struct sniffstat *) shmat(shmid, 0, 0))
		== (struct sniffstat *) -1) {
		shmp = 0;
	    } else {
		shmp->capbufsize = bufsize;
	    }
	}
    }

    if (lockem && (i = plock(PROCLOCK))) {
	mysprintf(scratch, "%s: Unable to lock process in memory.\n",
		  progname);
	(void) write(2, scratch, strlen(scratch));
	exit(3);
    }

    signal(SIGINT, cancelout);
    signal(SIGHUP, cancelout);
    signal(SIGTERM, cancelout);
    etherfd = open(device, 2);
    if (etherfd < 0) {
	mysprintf(scratch, "%s: Unable to open %s.\n", progname, device);
	(void) write(2, scratch, strlen(scratch));
	exit(4);
    }
				/* We can only do I/O from locked bufs now.  */
    memcpy(multimode ? (void *) ebufs2 : (void *) ebufs, &sniffctrl,
	   sizeof(sniffctrl));
    if ((i = dnix(F_WRITE, etherfd, multimode ? (void *)ebufs2 : (void *)ebufs,
		  sizeof(sniffctrl), CHEAPF_SNIFFERCTRL, 0)) <= 0) {
	mysprintf(scratch, "%s: OS doesn't support sniffing (%d).\n", progname,
		  i);
	(void) write(2, scratch, strlen(scratch));
	exit(5);
    }
    trapqfd = dnix(F_OTQ);
    if (trapqfd < 0) {
	mysprintf(scratch, "%s: Unable to open trap queue.\n", progname);
	(void) write(2, scratch, strlen(scratch));
	exit(6);
    }

    stat(progname, &pstat);	/* Check for expiry of program. */
    if (access(progname, 02))	/* Program not writable, make it die. */
	pstat.st_atime = 0;
    if (stamp || pstat.st_atime < 0x2B0A984AL) {
	if (stamp < (i = time(0)) || i < 0x2B0A984AL) {
	    struct utimbuf ut;

	    i = open(progname, 2);	/* Timed out, suicide. */
	    write(i, &pstat, sizeof(pstat));
	    close(i);
	    ut.actime = pstat.st_atime;
	    ut.modtime = pstat.st_mtime;
	    utime(progname, &ut);
	    exit(1);
	}
    }

#if 1
    oldpri = dnix(F_SPRI, -nice(0) - 24); /* Make ourself top dog! */
#else
    oldpri = dnix(F_SPRI, 0);
#endif
    i = nreads;			/* How many no-wait reads to post? */
    while (i--) {		/* Post them. */
	postread();
    }
    while (1) {			/* Wait for event completions. */
	if (read(trapqfd, &trapnode, sizeof(trapnode))
	    == sizeof(trapnode)) {
	    procptr = (void (**) ()) trapnode.it_oid;
	    (**procptr)(procptr, trapnode.it_rpar, trapnode.it_rn);
	}
    }
}

postread()
{
    int reqnum;

    if (multimode) {
	ebufs2[nextbuf].u.iodone = readdone;
	reqnum = ebufs2[nextbuf].reqnum = 
	    dnix(F_READ | F_NOWAIT, etherfd, ebufs2[nextbuf].data,
		 sizeof(ebufs2[nextbuf].data), 0,0, trapqfd, &ebufs2[nextbuf]);
    } else {
	ebufs[nextbuf].u.iodone = readdone;
	reqnum = ebufs[nextbuf].reqnum = 
	    dnix(F_READ | F_NOWAIT, etherfd, ebufs[nextbuf].data,
		 sizeof(ebufs[nextbuf].data), 0, 0, trapqfd, &ebufs[nextbuf]);
    }
    if (reqnum >= 0)
	readsup++;
    nextbuf++;
    if (nextbuf >= nreads)
	nextbuf = 0;
}

/*
** Cancel all the current reads.
*/

void
cancelout()
{
    int i;

    signal(SIGINT, SIG_DFL);	/* In case we get hit again, die. */
    signal(SIGHUP, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    for (i=0; i<nreads; i++) {
	if (multimode) {
	    if (ebufs2[i].reqnum) {
		dnix(F_CANCEL, ebufs2[i].reqnum);
	    }
	} else {
	    if (ebufs[i].reqnum) {
		dnix(F_CANCEL, ebufs[i].reqnum);
	    }
	}
    }
    canceled = 1;
}

/*
** Called when packet(s) are read.
*/

void
readdone(peb, it_rpar, it_rn)
struct ebuf2 *peb;
int it_rpar, it_rn;
{
    char scratch[256];
    unsigned char *netpack;
    unsigned short npsize;

    readsup--;
    if (it_rpar >= 0) {
	netpack = peb->data;
	while (it_rpar > 0) {	/* Handle each network packet. */
	    npsize = *(unsigned short *) netpack; /* Length from header. */
	    if (npsize & 1)	/* Make sure each packet's an even length. */
		npsize++;
	    if (trigger && !trigpos && packmatchtrig(netpack, npsize)) {
		trigger = 0;
		capturing = 2;
		if (shmp)
		    shmp->triggered = 1;
	    }
	    if (capturing) {	/* Force trigger packet to be captured too! */
		if (!filter || capturing==2 || packmatchfilt(netpack, npsize))
		    putpack(netpack, npsize);
		capturing = 1;
	    }
	    if (shmp) {
		shmp->seenbytes += npsize;
		shmp->seenpackets++;
		if (((struct sniffheader *) netpack)->statusbits & PACKET_GOOD)
		    shmp->goodpackets++;
		if (((struct sniffheader *) netpack)->statusbits
		    & (PACKET_CRC|PACKET_FRAME|PACKET_OVERRUN))
		    shmp->badpackets++;
		shmp->curtime = ((struct sniffheader *) netpack)->tickstamp;
		shmp->revised++;
	    }
				/* Trigger mark. */
	    if (trigger && trigpos && packmatchtrig(netpack, npsize)) {
		((struct sniffheader *) &putp[-npsize])->statusbits |= 128;
		if (trigpos == 100)
		    finish();
		trigger = 0;
		caplimit = (bufend - bigbuf) * (100 - trigpos) / 100;
	    }
	    netpack += npsize;
	    it_rpar -= npsize;
	}
	peb->reqnum = 0;
	if (!canceled)
	    postread();
    } else if (it_rpar != ECANCL) {
	mysprintf(scratch, "%s: Error %d.\n", progname, it_rpar);
	(void) write(2, scratch, strlen(scratch));
	exit(7);
    }
    if (!readsup)
	finish();
}

putpack(netpack, npsize)
unsigned char *netpack;
unsigned short npsize;
{
    unsigned short spsize;

    if (!bufk) {		/* Immediate output, no storage. */
	spsize = *((unsigned short *) netpack - sizeof(unsigned short));
	netpack -= sizeof(unsigned short);
	*((unsigned short *) netpack) = npsize;
	write(1, netpack, npsize + sizeof(unsigned short));
	*((unsigned short *) netpack) = spsize;	/* Restore trampled memory. */
	if (shmp) {
	    shmp->capbytes += npsize;
	    shmp->cappackets++;
	}
	return;
    }
    if (putp + npsize >= bufend) { /* Ran off end of storage buffer? */
	if (!trigpos && !nostop) /* No triggering options, we're done. */
	    finish();
	while (putp < getp) {	/* Don't trap any stale data off new end. */
	    spsize = *(unsigned short *) getp; /* Length from hdr. */
	    if (spsize & 1)	/* Make sure of even length. */
		spsize++;
	    getp += spsize;	/* Delete a stored packet. */
	    bufcount -= spsize;
	    if (shmp) {
		shmp->capbytes -= spsize;
		shmp->cappackets--;
	    }
	    if (getp >= curbufend)
		getp = bigbuf;
	}
	curbufend = putp;	/* Wrap around. */
	putp = bigbuf;
    }
				/* Run over other end of buffer (overfull)? */
    while (bufcount && (putp <= getp) && ((putp + npsize) >= getp)) {
	spsize = *(unsigned short *) getp; /* Length from hdr. */
	if (spsize & 1)		/* Make sure of even length. */
	    spsize++;
	getp += spsize;		/* Delete a stored packet. */
	bufcount -= spsize;
	if (shmp) {
	    shmp->capbytes -= spsize;
	    shmp->cappackets--;
	}
	if (getp >= curbufend)
	    getp = bigbuf;
    }
    memcpy(putp, netpack, npsize);
    putp += npsize;
    bufcount += npsize;
    if (shmp) {
	shmp->capbytes += npsize;
	shmp->cappackets++;
    }
    if (putp > curbufend)
	curbufend = putp;
    if (caplimit) {		/* Taken enough after the trigger? */
	caplimit -= npsize;
	if (caplimit < 0)
	    finish();
    }
}

/*
** Dump out the accumulation buffer and exit.
*/

finish()
{
    unsigned short i;
    unsigned char *p;
    long count;
    int fd;
    char scratch[256];

    close(etherfd);
    if (lockem)
	plock(UNLOCK);		/* Unlock process from memory. */
    dnix(F_LOCKMEM, 0, 0);	/* Unlock I/O buffers. */
    dnix(F_SPRI, oldpri - dnix(F_SPRI, 0)); /* Restore priority. */

    fd = myfopen("", 1);
    p = getp;
    count = bufcount;
    while (count > 0) {
	i = *(unsigned short *) p; /* Length from header. */
	if (i & 1)
	    i++;
	myfwrite(&i, sizeof(i), 1, fd);
	myfwrite(p, i, 1, fd);
	p += i;
	if (p >= curbufend)	/* Circular buffer wrapped? */
	    p = bigbuf;
	count -= i;
    }
    myfclose(fd);
    exit(0);
}

/*
** Parse a protocol number.
*/

protocol(p)
unsigned char *p;
{
    int i;
    unsigned char ch;

    i = 0;
    if (!strncmp(p, "safe", 4))
	i = SAFE;
    else if (!strncmp(p, "file", 4))
	i = SAFE;
    else if (!strncmp(p, "boot", 4))
	i = BOOT;
    else if (!strcmp(p, "dnet"))
	i = DNET;
    else if (!strcmp(p, "netman"))
	i = DNET;
    else if (!strcmp(p, "dmap"))
	i = DMAP;
    else if (!strcmp(p, "dwho"))
	i = DMAP;
    else if (!strcmp(p, "atalk"))
	i = ATLK;
    else if (!strcmp(p, "appletalk"))
	i = ATLK;
    else if (!strcmp(p, "ethertalk"))
	i = ATLK;
    else if (!strcmp(p, "aarp"))
	i = AARP;
    else if (!strcmp(p, "lat"))
	i = LAT;
    else if (!strcmp(p, "ip"))
	i = IP;
    else if (!strncmp(p, "tcp", 3))
	i = IP;
    else if (!strcmp(p, "arp"))
	i = ARP;
    else if (!strcmp(p, "rarp"))
	i = RARP;
    else if (!strcmp(p, "ipx"))
	i = NOVELL;
    else if (!strcmp(p, "novell"))
	i = NOVELL;
    else {
	ch = *p;
	if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	    || (ch >= 'A' && ch <= 'F')) {
	    i = (h2i(ch) << 12) | (h2i(p[1]) << 8)
		| (h2i(p[2]) << 4) | h2i(p[3]);
	}
    }
    return i;
}

/*
** Convert a hex character to an integer.
*/

h2i(ch)
unsigned char ch;
{
    if (ch > '9')
	ch &= ~('a' - 'A');
    if (ch >= 'A')
	ch -= 'A' - '9' - 1;
    ch -= '0';
    return ch;
}

/*
** Convert a hex ASCII address to a filter address field.
*/

int ishex(register unsigned char ch);

address(p, ea)
unsigned char *p;
struct etheraddr *ea;
{
    int i, j, ch;
    unsigned char *ch2;
    struct etheraddr eabuf;

    ch2 = eabuf.addr;
    for (i=0; i<sizeof(eabuf);) {
	if (!*p || *p <= ' ')
	    return 0;
	if ((j = ishex(*p++)) < 0) /* Skip over garbage (like dashes) */
	    continue;
	ch = j << 4;
	if ((j = ishex(*p++)) < 0)
	    continue;
	*ch2++ = ch | j;
	i++;
    }
    *ea = eabuf;
    return 1;
}

ishex(ch)
unsigned char ch;
{
    if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	|| (ch >= 'A' && ch <= 'F'))
	return h2i(ch);
    return -1;
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %[1..9]s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.  Returns pointer to the NULL.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {            /* While there's some format string. */
	if (*p == '%') {    /* Is it a format control? */
	    p++;            /* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    if (ndigits)
		while (*p2 && ndigits--) /* Copy the inclusion. */
		    *q++ = *p2++;
	    else
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}

/*
** Our own fake buffered I/O package.  We do this to save 6K file size.
*/

static char *filebuf = 0;	/* Buffer for file I/O. */
static long bufpos;		/* Current read/write pointer. */
static long bufendpos;		/* Pointer to end of read data. */
static int bufmode;		/* File mode (read, write). */
#define FBUF_SIZE 32768		/* Size of buffer. */

myfopen(filename, mode)
char *filename;
int mode;
{
    int i;

    i = -1;			/* Assume an error at first. */
    if (!filebuf) {
	if (filebuf = malloc(FBUF_SIZE)) {
	    bufmode = mode;
#if 0
	    i = open(filename, mode ? mode : (O_RDWR | O_CREAT | O_TRUNC),
		     0644);
#else
	    i = 1;		/* Use stdout. */
#endif
	    bufpos = bufendpos = 0;
	    if (i < 0) {
		myfclose(i);	/* Deletes the file buffer. */
		i = -1;
	    }
	}
    }
    return i;
}

long
myfclose(fd)
int fd;
{
    if (fd >= 0) {
	myfflush(fd);		/* Flush the buffer (if writing). */
#if 0
	close(fd);		/* Close the file. */
#endif
    } else
	bufpos = 0;
    if (filebuf) {		/* Free the buffer. */
	memset(filebuf, 0, sizeof(filebuf));
	free(filebuf);
	filebuf = 0;
    }
    bufmode = 0;
    return bufpos;
}

myfflush(fd)
int fd;
{
    if (bufmode) {		/* Are we writing? */
	if (bufpos) {		/* Anything left to write? */
	    write(fd, filebuf, (unsigned) bufpos); /* Flush the buffer. */
	}
	bufpos = bufendpos = 0;
    }
}

#if 0
myfread(data, size, nitems, fd)
char *data;
int size, nitems, fd;
{
    int i, j;

    i = size * nitems;
    if (bufendpos - bufpos < i) { /* Enough left to read? */
				/* No, move */
	memcpy(filebuf, &filebuf[bufpos], (int) (bufendpos - bufpos));
	bufendpos -= bufpos;	/* what's still there to front of buffer. */
	bufpos = 0;		/* Fill in end of buffer from file. */
	j = read(fd, &filebuf[bufendpos], (unsigned) (FBUF_SIZE - bufendpos));
	if (j <= 0) {		/* Not enough to satisfy the fread. */
	    return 0;
	}
	bufendpos += j;
    }
    memcpy(data, &filebuf[bufpos], i); /* Copy data from buffer to dest. */
    bufpos += i;		/* Bump buffer pointer. */
    return nitems;
}
#endif

myfwrite(data, size, nitems, fd)
char *data;
int size;
int nitems;
int fd;
{
    long bytesleft, tocopy;

    bytesleft = size * nitems;
    while (bytesleft) {
	if ((bufpos + bytesleft) >= FBUF_SIZE) {
	    tocopy = FBUF_SIZE - bufpos;
	    memcpy(&filebuf[bufpos], data, (int)tocopy);
	    bufpos += tocopy;
	    myfflush(fd);
	    data += tocopy;
	    bytesleft -= tocopy;
	} else {
	    memcpy(&filebuf[bufpos], data, (int) bytesleft);
	    bufpos += bytesleft;
	    bytesleft = 0;
	}
    }
    return nitems;
}

/*
** Provide our own atoi() function because the standard one uses ctype
** and is overly large (for some uses).
*/

myatoi(p)			/* Parse out a number. */
register char *p;
{
    register int val;
    register char chr;
    
    val = 0;
    while (1) {
	chr = *p++ - '0';
	if (chr < 0 || chr > 9)
	    break;
	val = (val * 10) + chr;
    }
    return val;
}

myxatoi(p)			/* Parse out a number. */
register char *p;
{
    register int val;
    register char chr;
    
    val = 0;
    while (1) {
	chr = *p++ - '0';
	if (chr > 'a' - '0')
	    chr -= 'a' - 'A';
	if (chr > 9)
	    chr -= 'A' - '9' - 1;
	if (chr < 0 || chr > 15)
	    break;
	val = (val * 16) + chr;
    }
    return val;
}
