/*
** Common packet triggering/filtering logic.
*/

#include <malloc.h>
#include <dnix/cheapdrv.h>
#include "protocols.h"

static struct pattern {
    struct pattern *next;
    unsigned char *mask;	/* Mask array. */
    unsigned char *match;	/* Match array. */
    int size;			/* Size of pattern arrays. */
    int offset;			/* Starting offset. */
} *trigs, *filts;

extern int trigoffset;		/* Offset of start of trigger string. */
extern int triggerinv;		/* Trigger's inverted. */
extern int filtoffset;		/* Offset of start of filter string. */
extern int filterexc;		/* Filter's inverted. */

extern int ishex(register unsigned char ch);

/*
** Turn a trigger/filter string into mask and match trigger/filter arrays.
*/

settrigger(cp)
char *cp;
{
    return addpattern(cp, &trigs, &trigoffset);
}

setfilter(cp)
char *cp;
{
    return addpattern(cp, &filts, &filtoffset);
}

static
addpattern(cp, list, offset)
char *cp;
struct pattern **list;
int *offset;
{
    int i;
    struct pattern *pp, *new;

    i = parse(cp, 0, 0, 0);	/* How much storage? */
    if (!(new = calloc(sizeof *new, 1)))
	return 0;
    new->size = i;
    if (!(new->mask = malloc(i)))
	return 0;
    if (!(new->match = malloc(i)))
	return 0;
    parse(cp, new->mask, new->match, offset);
    new->offset = *offset;
    if (!*list) {		/* Add to end of list. */
	*list = new;
    } else {
	for (pp = *list; pp->next; pp = pp->next)
	    ;
	pp->next = new;
    }
    return 1;
}

static
parse(cp, p1, p2, offset)
char *cp;
unsigned char *p1, *p2;
int *offset;
{
    int i, byte, nbytes, startshift, curpos, shift, repcount;
    unsigned char mask, match, chr;
    unsigned short control, reload;

    i = strlen(cp);
    nbytes = repcount = 0;
    reload = 0x0F0; shift = 4; startshift = 4; /* Hexadecimal at first. */
    control = reload;
    curpos = startshift;
    mask = 0xFF;
    match = 0;
    while (i--) {
	if (!repcount)
	    chr = *cp++;
	else
	    repcount--;
	if (!control) {		/* Time to work on next byte? */
	    if (p1)
		*p1++ = mask;	/* Push out last one. */
	    if (p2)
		*p2++ = match;
	    nbytes++;
	    control = reload;
	    curpos = startshift;
	    mask = 0xFF;
	    match = 0;
	}
	if ((byte = ishex(chr)) < 0) {
	    switch (chr) {
	      case '<':		/* Repeat count. */
		repcount = myatoi(cp);
		i += repcount - 2;
		while (*cp && *cp++ != '>')
		    i--;
		chr = *cp++;
		continue;

	      case '[':		/* New pattern offset. */
		i--;
		if (offset)
		    *offset = myxatoi(cp);
		while (*cp && *cp++ != ']')
		    i--;
		continue;

	      case 'x':		/* Don't-care. */
	      case 'X':
		mask &= ~control;
		break;

	      case '%':		/* Switch to binary. */
		control = reload = 0x080; shift = 1; curpos = startshift = 7;
		continue;

	      case '@':		/* Switch to octal. */
		control = reload = 0x1C0; shift = 3; curpos = startshift = 6;
		continue;

	      case '$':		/* Switch to hexadecimal. */
		control = reload = 0x0F0; shift = 4; curpos = startshift = 4;
		continue;

	      case ' ':		/* Miscellaneous separators, for */
	      case ',':		/* readability of the string. */
	      case '-':
		continue;
	    }		
	} else {		/* Normal hex (or whatever) character. */
	    match |= byte << curpos;
	}
	control >>= shift;
	curpos -= shift;
    }
    if (!control) {		/* Leftovers? */
	if (p1)
	    *p1++ = mask;	/* Push it out. */
	if (p2)
	    *p2++ = match;
	nbytes++;
    }
    return nbytes;
}

static
packmatchentry(packet, count, pp)
unsigned char *packet;		/* Packet data. */
unsigned short count;		/* Packet size. */
struct pattern *pp;		/* Pattern match. */
{
    unsigned char *packdat, *pmask, *pmatch;

    packdat = packet + sizeof(struct sniffheader) + pp->offset;
    if ((count -= sizeof(struct sniffheader) + pp->offset) <= 0)
	return 0;	/* Insufficient data, no match. */
    if (count > pp->size)
	count = pp->size;
    else if (count < pp->size)
	return 0;		/* Insufficient data, no match. */
    pmatch = pp->match;
    pmask = pp->mask;
    while (count--) {
	if ((*pmatch++ ^ *packdat++) & *pmask++)
	    return 0;
    }
    return 1;
}

static
packmatch(packet, count)
unsigned char *packet;		/* Packet data. */
unsigned short count;		/* Packet size. */
{
    int i;
    struct pattern *pp;

    pp = trigs;
    while (pp) {
	if (packmatchentry(packet, count, pp))
	    return !triggerinv;
	pp = pp->next;
    }
    return triggerinv;
}

/*
** Returns TRUE if the given packet matches any of the trigger patterns.
** If trigger inversion is on then it returns FALSE if the given packet
** matches any of the trigger patterns.  A TRUE answer also marks that
** packet with a trigger flag.
*/
packmatchtrig(packet, count)
unsigned char *packet;
unsigned short count;
{
    int rval;

    if (rval = packmatch(packet, count))
	((struct sniffheader *) packet)->statusbits |= 128; /* Trigger mark. */
    return rval;
}

/*
** Returns TRUE if the given packet matches any of the filter patterns.
** If filter inversion is on then it returns FALSE if the given packet
** matches any of the filter patterns.
*/
packmatchfilt(packet, count)
unsigned char *packet;		/* Packet data. */
unsigned short count;		/* Packet size. */
{
    struct pattern *pp;

    pp = filts;
    while (pp) {
	if (packmatchentry(packet, count, pp))
	    return !filterexc;
	pp = pp->next;
    }
    return filterexc;
}
