/*
** Swatch/Netwatch/Idlewatch/Diskwatch
**
** Swap/Network/WAN/Idle/Disk Watcher.  Shows a time trail of system swapping,
** or network activity, or idle time, or disk activity.
**
** Revision History:
**
** Version 2.2		03/16/95	Jim Cathey
**	Use /etc/devnm in diskwatch mode to get name of primary disk.
**
** Version 2.1		10/20/94	Jim Cathey
**	Protect against total failure in time scaling routine (probably
**	due to a date change) so that it can't get stuck in a loop.
**
** Version 2.0		03/01/94	Jim Cathey
**	Added negative x and y screen positions, which measures between
**	the right edge of the window and the right edge of the display,
**	rather than left-left [for positive offsets].  Added -n network
**	override switch for netwatch.
**
** Version 1.9		12/15/93	Jim Cathey
**	Added automatic video copy (when WINDOWS_ON contains a list of
**	workstation names instead of just one).
**
** Version 1.8		11/25/93	Jim Cathey
**	Added visual high-water marker.  Corrected bug where if the CPU
**	got busy enough to interfere with the sample rate the extra time
**	taken wasn't accounted for.  (The throughput rates would come out
**	too high.)  Fixed bug where the time trail sometimes ended up one
**	pixel too tall, and left a fringe when it started scrolling, or cut
**	into the baseline.
**
** Version 1.7		06/08/93	Jim Cathey
**	Swap-watch function's sample In & Out bins always read 0.
**	(Forgot to scale them up to bytes from blocks.)
**
** Version 1.6		11/12/92	Jim Cathey
**	Fixed problems with the -s startup switch.
**
** Version 1.5		11/03/92	Jim Cathey
**	Fixed scaling problems in idle and disk modes.
**
** Version 1.4		11/02/92	Jim Cathey
**	Added in disk busyness watching 'mode'.
**
** Version 1.3		04/07/92	Jim Cathey
**	Added in idle time watching as well (winidle replacement).
**
** Version 1.2		03/12/92	Jim Cathey
**	Fixed bugs with non-1-second sampling.  Added in network
**	watching as well (cheapstat analog).
**
** Version 1.1		02/13/92	Jim Cathey
**	Reset the vertical scale on a rate change.
**
** Version 1.0		02/10/92	Jim Cathey
**	Initial release.
**
*/
#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dnix/dutsname.h>
#include <dnix/sysinfo.h>
#include <dnix/fcodes.h>
#include <dnix/cheapdrv.h>
#include <isc/window.h>
#include <isc/keycodes.h>

#define TRUE (1)
#define FALSE (0)
#define VERSION "V2.2"

#define FONT 3
#define FONTSIZE 9

#define NSAMPLES 180
#define SAMPLESHIFT 10		/* How many samples to throw out on scroll. */
struct {
    unsigned long swapin;
    unsigned long swapout;
} samples[NSAMPLES];
int cursample;			/* Current sample to fill next. */
int highwater = 20;
int highmul = 10;		/* Power of 10 for scaling. */
int highdigindex = 1;
int highdigs[] = {1, 2, 5};
int lasthighmark, highmark;	/* For visual high-water mark. */
int netmode, idlemode, diskmode, diskdev, netfd;
char *diskname = "/dev/si0";	/* Last-ditch default name. */
struct stat st;

char tickbits[] = {		/* Tickmark for visual high-water marker. */
    0x80,0x00,
    0xC0,0x00,
    0xE0,0x00,
    0xF0,0x00,
    0xE0,0x00,
    0xC0,0x00,
    0x80,0x00
};
BitMap tick = {
    tickbits,
    2,
    {0, 0, 7, 5}
};

/*
** Menus
*/
#define M_SAMPLE (1)

#define ENTRY1 (1)
#define ENTRY7 (7)

#define AVG1 190
#define AVG2 240

extern char *getenv(), *mysprintf();

void keyevent(), menuevent(), updateevent();
void windowevent(), napevent();

void (*completion) () = napevent;
int napreq;

Rect world, screen, playpen, detail;
char version[] = {VERSION};
FontInfo fontrec;

int sample_rate = 1, rate_array[8] = {0, 1, 2, 5, 10, 15, 30, 60};
int last_sample_rate = ENTRY1;
short window, trapq, wmfd, width, height, weRicon;
short title_length;
struct dutsname duts;		/* For our system name et. al. */
union {
    struct Sysinfo swap;
    struct cheapstat net;
} t1, t2;
long timestamp;

long curin, curout;

main(argc, argv)
int argc;
char *argv[];
{
    struct osconfig config;
    char **argpp, *argp, *strrchr();
    short horz, vert, x, y, voffset;
    char Title[80];
    char Domain[64];		/* Our (originating side) domain (if any). */
    char *dispsystem, *network, *cp, *strchr();
    int i, found;

    x = 5;
    y = 41;
    dispsystem = argv[0];	/* Parse our argument list. */
    if (argc > 1 && !strncmp(argv[1], "-M", 2))
	dispsystem = &argv[1][2];
    if (argp = strrchr(dispsystem, '/'))
	dispsystem = argp + 1;
    if (*dispsystem == 'n') {
	netmode++;
	highwater = 10;
	highdigindex = 0;
    } else if (*dispsystem == 'i') {
	idlemode++;
	highwater = 100;
	highdigindex = 0;
    } else if (*dispsystem == 'd') {
	diskmode++;
	highwater = 100;
	highdigindex = 0;
	setdiskname();
    }
    dispsystem = getenv("WINDOWS_ON");
    network = "/dev/dnet";
    argpp = argv;
    while (--argc > 0) {
	argp = *++argpp;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'V':
		mysprintf(Title, "%s %s\n", argv[0], version);
		write(1, Title, StringLen(Title));
		exit(0);
		break;

	      case 'd':
		if (diskmode) {
		    if (!*argp)
			--argc, argp = *++argpp;
		    diskname = argp;
		}
		break;

	      case 'M':		/* Mode switch.  Already done. */
		break;

	      case 'n':		/* Network switch. */
		if (netmode) {
		    if (!*argp)
			--argc, argp = *++argpp;
		    network = argp;
		}
		break;

	      case 's':
		if (!*argp)
		    --argc, argp = *++argpp;
		sample_rate = myatoi(argp);
		if (sample_rate < ENTRY1)
		    sample_rate = ENTRY1;
		else if (sample_rate > ENTRY7)
		    sample_rate = ENTRY7;
		last_sample_rate = sample_rate;
		sample_rate = rate_array[sample_rate];
		break;

	      case 'x':
		if (!*argp)
		    --argc, argp = *++argpp;
		x = myatoi(argp);
		break;

	      case 'y':
		if (!*argp)
		    --argc, argp = *++argpp;
		y = myatoi(argp);
		break;

	      case '?':
	      default:
		mysprintf(Title, "usage: %s [-M mode] [-V] [-d device] "
			  "[-n network] [-s Sample-rate]"
			  "[-x offset] [-y offset] [system-name]\n", argv[0]);
		write(1, Title, StringLen(Title));
		exit(1);
		break;
	    }
	} else {
	    dispsystem = argp;
	}
    }

    trapq = InitGraf();		/* Setup window. */
    if (trapq < 0)
	exit(1);
    dnix(F_UNAME, &duts, sizeof(struct dutsname));
    strcpy(Title, duts.nodename);
    if (dispsystem && strchr(dispsystem, '.')) {
	if ((wmfd = open("/etc/domainid", O_RDONLY | O_NOCTTY)) >= 0) {
	    if (read(wmfd, Domain, sizeof(Domain) - 1) > 0) {
		cp = strchr(Domain, '\n');
		if (cp)		/* Knock off newline. */
		    *cp = 0;
		strcat(Title, Domain);
	    }
	    close(wmfd);
	}
    }
    wmfd = ConnectGraf(dispsystem);
    if (wmfd < 0) {
	cp = "Can't connect to a Window Manager\n";
	write(1, cp, StringLen(cp));
	exit(2);
    }
    LocalGetFontInfo(FONT, normalText, FONTSIZE, &fontrec);
    width = fontrec.widMax;
    height = fontrec.ascent + fontrec.descent;
    strcat(Title, netmode ? " : Net Watch" :
	   (idlemode ? " : Idle Watch" :
	    (diskmode ? " : Disk Watch" : " : Swap Watch")));
    SetRect(&playpen, 0, 0, NSAMPLES, 80);
    SetRect(&screen, 0, 0, playpen.edge.right + 90, playpen.edge.bottom + 70);
    if (idlemode || diskmode)
	screen.edge.bottom -= 2 * height;
    OffsetRect(&playpen, 70, 20);
    world = screen;
    OffsetRect(&world, x, y);
    if (x < 0 || y < 0) {
	HdwrType hwtype;

	GetHdwrType(&hwtype);
	if (x < 0)
	    OffsetRect(&world, hwtype.hdots
		       - (world.edge.right - world.edge.left), 0)
	if (y < 0)
	    OffsetRect(&world, 0, hwtype.vdots
		       - (world.edge.bottom - world.edge.top))
    }

    Event(keyEvt, keyevent);	/* Bind our event handlers. */
    Event(menuEvt, menuevent);
    Event(updateEvt, updateevent);
    Event(windowEvt, windowevent);

    window = NewWindow(&world, Title, showWindow, documentProc, frontWindow,
		       hasDrag | hasGoAway | hasZoom | hasSendToBack, wmfd);
    SetPort(window);
    NewMenu(M_SAMPLE, "Rate");	/* Make our pulldown menu. */
    AppendMenu(M_SAMPLE, "1;2;5;10;15;30;60");
    CheckItem(M_SAMPLE, last_sample_rate, TRUE);

    if (argp = dispsystem) {	/* Handle copying video to other WS. */
	char *cp, buf[80]; int fd; long winds;

	winds = 0;
	while (argp = strchr(argp, '\n')) {
	    argp++;
	    cp = strchr(argp, '\n');
	    if (!cp)
		cp = argp + strlen(argp);
	    if (cp - argp) {
		strncpy(buf, argp, cp - argp);
		buf[cp - argp] = 0;
		if ((fd = ConnectGraf(buf)) >= 0) {
		    winds |= 1 | 1 << NewWindow(&world, Title, showWindow,
						documentProc, frontWindow,
						hasSendToBack, fd);
		}
	    }
	}
	if (winds)
	    SetPorts(winds);
    }
    PenMode(srcCopy);
    TextMode(srcCopy);
    TextFont(FONT);
    TextSize(FONTSIZE);
    detail = screen;
    detail.edge.top = playpen.edge.bottom + 3 * height;

    napreq = dnix(F_NOWAIT | F_NAP, (sample_rate * 1000), 0, 0, 0, 0,
		  trapq, &completion);

 /*
 ** First record (for reference).
 */
    if (netmode) {
        if ((netfd = open(network, O_RDONLY | O_NOCTTY)) < 0)
	    exit(1);
        if (dnix(F_READ, netfd, &t1.net, sizeof(struct cheapstat),
		 CHEAPF_STATISTICS, 0) < 0)
	    exit(2);
	t2.net = t1.net;
    } else {
	dnix(F_SYSINFO, F_SYS_GETSAR, &t1.swap, sizeof(struct Sysinfo));
	t2.swap = t1.swap;	/* So first update works right. */
    }
    if (diskmode) {
	if (stat(diskname, &st)) {
	    exit(1);
	}
	found = 0;
	for (i=0; i<sizeof(t1.swap.io_stat)/sizeof(t1.swap.io_stat[0]); i++) {
	    if (st.st_rdev == t1.swap.io_stat[i].ios_dev) {
		diskdev = i;
		found = 1;
		break;
	    }
	}
	if (!found)
	    exit(2);
    }
    time(&timestamp);

 /*
 ** Begin handling events.
 */
  more:
    if (GetEvent(anyWindow, noEvent)) {
	if ((i = (EventError() >> 16)) != wmfd) {
	    DisconnectGraf(i);	/* Video copy died, take it out & continue. */
	    goto more;
	}
	exit(3);
    }
    exit(0);
}

/*
** Draw totals.
*/

drawtots()
{
    char buffer[10];
    Rect eraser;
    int hpos, vpos, wid;
    unsigned long in, out, lastin, lastout;

    if (netmode) {
	in = t2.net.rcvbytes;
	lastin = t1.net.rcvbytes;
	out = t2.net.sndbytes;
	lastout = t1.net.sndbytes;
    } else if (idlemode || diskmode) {
	in = out = lastin = lastout = 0;
	if (cursample) {
	    in = curin / cursample;
	    out = curout / cursample;
	}
    } else {			/* (swapmode) */
	in = t2.swap.bswapin * 512;
	lastin = t1.swap.bswapin * 512;
	out = t2.swap.bswapout * 512;
	lastout = t1.swap.bswapout * 512;
    }

    if (diskmode)
	mysprintf(buffer, "%d", (out - lastout) / 1024);
    else
	mysprintf(buffer, "%d", (in - lastin) / 1024);
				/* Bytes in since last sample. */
    wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
    MoveTo(hpos = screen.edge.left + 35, vpos=detail.edge.top + 4 - 2*height);
    DrawString(buffer);
    SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
	    screen.edge.left + 70 - fontrec.ascent,
	    vpos + fontrec.descent);
    EraseRect(&eraser);

    if (!diskmode) {		/* Bytes out since last sample. */
	mysprintf(buffer, "%d", (out - lastout) / 1024);
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos, vpos = detail.edge.top + 4 - height);
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		screen.edge.left + 70 - fontrec.ascent,


		vpos + fontrec.descent);
	EraseRect(&eraser);

				/* Bytes in/out since last sample. */
	mysprintf(buffer, "%d", ((in - lastin) + (out - lastout)) / 1024);
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = screen.edge.left + 75,
	       vpos = detail.edge.top + 4-ScaleS(height, 15, 10));
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		hpos + 30, vpos + fontrec.descent);
	EraseRect(&eraser);
    }
    if (!(idlemode || diskmode)) { /* Total bytes in for all held samples. */
	mysprintf(buffer, "%d", curin / 1024);
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = playpen.edge.left - 10, vpos = detail.edge.top + height);
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		screen.edge.left + 100 - fontrec.ascent, vpos+fontrec.descent);
	EraseRect(&eraser);

				/* Total bytes out for all held samples. */
	mysprintf(buffer, "%d", curout / 1024);
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos, vpos = detail.edge.top + 2 * height);
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		screen.edge.left + 100 - fontrec.ascent, vpos+fontrec.descent);
	EraseRect(&eraser);

				/* Total bytes in/out for all held samples. */
	mysprintf(buffer, "%d", (curin + curout) / 1024);
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = playpen.edge.left + 35,
	       vpos = detail.edge.top + ScaleS(height, 15, 10));
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		140, vpos + fontrec.descent);
	EraseRect(&eraser);

				/* Bytes/sec in for all held samples. */
	if (cursample)
	    mysprintf(buffer, "%d.%d", curin / (1024 * cursample*sample_rate),
		(((curin / 1024) % cursample) * 10) / (cursample*sample_rate));
	else
	    strcat(buffer, "0");
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = AVG1, vpos = detail.edge.top + height);
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		AVG2 - 5 - fontrec.ascent,
		vpos + fontrec.descent);
	EraseRect(&eraser);

				/* Bytes/sec out for all held samples. */
	if (cursample)
	    mysprintf(buffer, "%d.%d", curout/(1024 * cursample * sample_rate),
	       (((curout/1024) % cursample) * 10) / (cursample * sample_rate));
	else
	    strcat(buffer, "0");
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = AVG1, vpos = detail.edge.top + 2 * height);
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		AVG2 - 5 - fontrec.ascent,
		vpos + fontrec.descent);
	EraseRect(&eraser);

				/* Bytes/sec in/out for all held samples. */
	if (cursample)
	    mysprintf(buffer, "%d.%d", (curin + curout)
		      / (1024 * cursample * sample_rate),
		      ((((curin + curout) / 1024) % cursample) * 10)
		      / (cursample * sample_rate));
	else
	    strcat(buffer, "0");
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	MoveTo(hpos = AVG2, vpos = detail.edge.top + ScaleS(height, 15, 10));
	DrawString(buffer);
	SetRect(&eraser, hpos + wid, vpos - fontrec.ascent,
		detail.edge.right, vpos + fontrec.descent);
	EraseRect(&eraser);
    }
}

/*
** Draw the visual high-water mark (if it has moved).
*/

drawtick(update)
int update;			/* TRUE to force tickmark to be drawn. */
{
    int vert;
    Rect where;

    if (lasthighmark != highmark || update) {
	where = tick.bounds;
	OffsetRect(&where, playpen.edge.left - tick.bounds.edge.right,
		   playpen.edge.bottom - tick.bounds.edge.bottom/2);
	OffsetRect(&where, 0, -lasthighmark);
	CopyBits(&tick, 0, &tick.bounds, &where, srcBic, 0);
	OffsetRect(&where, 0, lasthighmark - highmark);
	CopyBits(&tick, 0, &tick.bounds, &where, srcOr, 0);
	lasthighmark = highmark;
	for (vert = 25; vert <= 100; vert += 25) { /* Restore Tickmarks. */
	    MoveTo(playpen.edge.left - 1,
		   playpen.edge.bottom - ScaleS(vert, playpen.edge.bottom
						- playpen.edge.top, 100));
	    Line(-4, 0);
	}
    }
}

/*
** Draw the static information.
*/

drawkey()
{
    Rect key;
    char *cp, buffer[10];
    int horz, vert, wid;

    TextMode(srcOr);
    MoveTo(4, fontrec.ascent + 2);
    TextFace(underlineText);
    DrawString(netmode ? "Net Activity" :
	       (idlemode ? "Activity" : (diskmode ? "Activity" : "Swapping")));
    TextFace(normalText);
    vert = height;
    cp = version;
    horz = screen.edge.right - 2
	- LocalStringWidth(FONT, normalText, FONTSIZE, cp);
    MoveTo(horz, vert);
    DrawString(cp);
    horz = 80;
    MoveTo(horz, vert);
    cp = "Key:";
    wid = LocalStringWidth(FONT, normalText, FONTSIZE, cp);
    DrawString(cp);
    horz += wid + 8;
    key.edge.left = horz;
    key.edge.top = vert - fontrec.ascent;
    key.edge.right = horz + height;
    key.edge.bottom = key.edge.top + height;
    FrameRect(&key);
    InsetRect(&key, 1, 1);
    ForeColor(2);
    FillRect(&key, blackPattern);
    ForeColor(1);
    InsetRect(&key, -1, -1);
    horz += height * 3 / 2;
    cp = netmode ? "Net Out" : (idlemode ? "User" :
				(diskmode ? diskname : "Swap Out"));
    wid = LocalStringWidth(FONT, normalText, FONTSIZE, cp);
    MoveTo(horz, vert);
    DrawString(cp);
    horz += wid + 2 * height;

    if (!diskmode) {
	key.edge.left = horz;
	key.edge.right = horz + height;
	FrameRect(&key);
	InsetRect(&key, 1, 1);
	ForeColor(6);
	FillRect(&key, greyPattern);
	ForeColor(1);
	InsetRect(&key, -1, -1);
	horz += height * 3 / 2;
	cp = netmode ? "Net In" : (idlemode ? "System & IO" : "Swap In");
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, cp);
	MoveTo(horz, vert);
	DrawString(cp);
    }

    MoveTo(playpen.edge.left - 1, playpen.edge.top);
    LineTo(playpen.edge.left - 1, playpen.edge.bottom);
    LineTo(playpen.edge.right - 1, playpen.edge.bottom);

    for (horz = 1; horz <= 3; horz++) {	/* Tickmarks & labels. */
	MoveTo(playpen.edge.left + ScaleS(horz, NSAMPLES, 3),
	       playpen.edge.bottom);
	Line(0, 4);
	mysprintf(buffer, "%d", ScaleRS(horz * sample_rate, NSAMPLES, 3));
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	Move(-(wid/2), fontrec.ascent + 2);
	DrawString(buffer);
    }
    for (vert = 25; vert <= 100; vert += 25) {
	MoveTo(playpen.edge.left - 1,
	        playpen.edge.bottom - ScaleS(vert, playpen.edge.bottom
					     - playpen.edge.top, 100));
	Line(-4, 0);
	mysprintf(buffer, "%d", ScaleS(vert, highwater, (idlemode || diskmode)
				       ? 100 : 100 * sample_rate));
	wid = LocalStringWidth(FONT, normalText, FONTSIZE, buffer);
	Move(-(wid + 2), (fontrec.ascent+fontrec.descent) / 2);
	DrawString(buffer);
    }
    TextFace(boldText);
    cp = "Seconds";
    wid = LocalStringWidth(FONT, boldText, FONTSIZE, cp);
    MoveTo((playpen.edge.left + playpen.edge.right) / 2 - wid/2,
	   playpen.edge.bottom + ScaleS(height, 25, 10));
    DrawString(cp);

    cp = (idlemode || diskmode) ? "Busy" : "KB/";
    wid = LocalStringWidth(FONT, boldText, FONTSIZE, cp);
    MoveTo(screen.edge.left + 10,
	   (playpen.edge.top + playpen.edge.bottom) / 2);
    DrawString(cp);
    Move(-wid/2, height);
    cp = (idlemode || diskmode) ? "(%)" : "Sec";
    wid = LocalStringWidth(FONT, boldText, FONTSIZE, cp);
    Move(-wid/2, 0);
    DrawString(cp);
    TextFace(normalText);

    MoveTo(10, detail.edge.top + 4 - 2 * height);
    if (diskmode)
	DrawString("Cum:");
    else {
	DrawString(idlemode ? "Sys:" : "In:");
	MoveTo(10, detail.edge.top + 4 - height);
	DrawString(idlemode ? "Usr:" : "Out:");
	MoveTo(screen.edge.left + 70, detail.edge.top-ScaleS(height, 15, 10));
	Line(-fontrec.ascent, -fontrec.ascent);
	Move(fontrec.ascent, fontrec.ascent);
	Line(-fontrec.ascent, fontrec.ascent);
	if (!(idlemode || diskmode)) {
	    MoveTo(detail.edge.left + 4, detail.edge.top - 1);
	    LineTo(detail.edge.right - 4, detail.edge.top - 1);

	    MoveTo(10, detail.edge.top + height);
	    DrawString("Tot. In:");
	    MoveTo(10, detail.edge.top + 2 * height);
	    DrawString("Tot. Out:");
	    MoveTo(100, detail.edge.top + height);
	    Line(-fontrec.ascent, -fontrec.ascent);
	    Move(fontrec.ascent, fontrec.ascent);
	    Line(-fontrec.ascent, fontrec.ascent);
	    MoveTo(AVG2 - 5, detail.edge.top + height);
	    Line(-fontrec.ascent, -fontrec.ascent);
	    Move(fontrec.ascent, fontrec.ascent);
	    Line(-fontrec.ascent, fontrec.ascent);

	    MoveTo(140, detail.edge.top + height);
	    DrawString("Avg. In:");
	    MoveTo(140, detail.edge.top + 2 * height);
	    DrawString("Avg. Out:");
	}
    }
    TextMode(srcCopy);
}

/*
** Event handlers.
*/

void
napevent(routine, pid, rn)
int (*routine) ();
int pid, rn;
{
    int sin, sout, i, j, pos;
    long curtime;
    short nticks;
    Rect valid;

    napreq = dnix(F_NOWAIT | F_NAP, (sample_rate * 1000), 0, 0, 0, 0,
		  trapq, &completion);
    if (netmode) {
	dnix(F_READ, netfd, &t2.net, sizeof(struct cheapstat),
	     CHEAPF_STATISTICS, 0); /* Convert to Kbytes. */
	sout = t2.net.sndbytes - t1.net.sndbytes;
	sin = t2.net.rcvbytes - t1.net.rcvbytes;
    } else {
	dnix(F_SYSINFO, F_SYS_GETSAR, &t2.swap, sizeof(struct Sysinfo));
	if (idlemode) {
	    sout = 0;
	    sin = t2.swap.cpu[CPU_IDLE] - t1.swap.cpu[CPU_IDLE]
		+ t2.swap.cpu[CPU_USER] - t1.swap.cpu[CPU_USER]
		    + t2.swap.cpu[CPU_KERNAL] - t1.swap.cpu[CPU_KERNAL]
			+ t2.swap.cpu[CPU_WAIT] - t1.swap.cpu[CPU_WAIT];
	    if (sin) {
		sout = ScaleRS(t2.swap.cpu[CPU_USER] - t1.swap.cpu[CPU_USER],
			       100, sin);
		sout *= 1024;
		sin = ScaleRS((t2.swap.cpu[CPU_KERNAL]-t1.swap.cpu[CPU_KERNAL])
			      + (t2.swap.cpu[CPU_WAIT]-t1.swap.cpu[CPU_WAIT]),
			      100, sin);
		sin *= 1024;
	    }
	} else if (diskmode) {
	    sin = t2.swap.cpu[CPU_IDLE] - t1.swap.cpu[CPU_IDLE]
		+ t2.swap.cpu[CPU_USER] - t1.swap.cpu[CPU_USER]
		    + t2.swap.cpu[CPU_KERNAL] - t1.swap.cpu[CPU_KERNAL]
			+ t2.swap.cpu[CPU_WAIT] - t1.swap.cpu[CPU_WAIT];
	    sout = ScaleRS(t2.swap.io_stat[diskdev].ios_actt
			   - t1.swap.io_stat[diskdev].ios_actt, 100, sin);
	    sout *= 1024;
	    sin = 0;
	} else {		/* Swap mode. */
	    sout = (t2.swap.bswapout - t1.swap.bswapout) * 512; /* Convert */
	    sin = (t2.swap.bswapin - t1.swap.bswapin) * 512; /* to bytes. */
	}
    }
    time(&curtime);
    nticks = (short) (curtime - timestamp) / (short) sample_rate;
    if (nticks <= 0 || nticks >= NSAMPLES) /* If we get delayed compensate */
	nticks = 1;		/* for how long it actually took to take the */
				/* sample, but protect against date changes. */
    if (!idlemode && !diskmode) { /* (Compensation's already done for */
	sin /= nticks;		/*    disk & idle modes.) */
	sout /= nticks;
    }
    while (nticks--) {		/* Preserve time scale even when CPU's busy. */
	if (cursample >= NSAMPLES) { /* Time to scroll? */
	    ScrollRect(&playpen, -SAMPLESHIFT, 0, -1);
	    valid = playpen;
	    valid.edge.left = valid.edge.right - SAMPLESHIFT;
	    ValidRect(&valid);
	    cursample -= SAMPLESHIFT;
	    for (i=0; i<SAMPLESHIFT; i++) {
		curin -= samples[i].swapin;
		curout -= samples[i].swapout;
	    }
	    BlockMove(&samples[SAMPLESHIFT], &samples[0],
		      sizeof(samples) - sizeof(samples[0]) * SAMPLESHIFT);
	}
	samples[cursample].swapin = sin;
	samples[cursample].swapout = sout;
	curin += sin;
	curout += sout;
	i = sin / 1024;
	j = sout / 1024;
	i += j;
	if (!(idlemode || diskmode) && (i > highwater)) {
	    while (i > highwater) { /* Out of range, follow a 1-2-5 sequence */
		highdigindex++;	/* to find the next scale maximum. */
		if (highdigindex >= sizeof(highdigs)/sizeof(highdigs[0])) {
		    highdigindex = 0;
		    highmul *= 10;
		}
		highwater = highdigs[highdigindex] * highmul;
	    }
	    EraseRect(&screen);
	    InvalRect(&screen);
	    highmark = 0;
	} else if (i) {
	    if (i > highwater)	/* Over-range protection, for non-scalable */
		i = highwater;	/* readings (idle & disk modes). */
	    if (j > highwater)
		j = highwater;
	    MoveTo(playpen.edge.left + cursample, playpen.edge.bottom - 1);
	    ForeColor(2);
	    pos = ScaleS(j, playpen.edge.bottom-playpen.edge.top-1, highwater);
	    Line(0, -pos);
	    if (!diskmode) {
		ForeColor(6);
		PenPat(greyPattern);
		pos = ScaleS(i, playpen.edge.bottom - playpen.edge.top - 1,
			     highwater);
		LineTo(playpen.edge.left + cursample,
		       playpen.edge.bottom - pos - 1);
	    }
	    if (pos > highmark)
		highmark = pos;
	    ForeColor(1);
	    PenNormal();
	} else {
	    MoveTo(playpen.edge.left + cursample, playpen.edge.bottom - 1);
	    Line(0, 0);
	}
	cursample++;
	drawtots();
    }
    drawtick(0);
    timestamp = curtime;
    t1 = t2;			/* Age the sample. */
}

void
updateevent(ep)
EventRecord *ep;
{
    int i, j, sample, pos;
    long save;

    GetPort(&save);
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (!weRicon || ep->who != window) {
	drawkey();
	drawtots();
	for (sample = 0; sample < cursample; sample++) {
	    i = samples[sample].swapin / 1024;
	    j = samples[sample].swapout / 1024;
	    i += j;
	    if (i) {
		MoveTo(playpen.edge.left + sample, playpen.edge.bottom - 1);
		ForeColor(2);
		pos = ScaleS(j, playpen.edge.bottom - playpen.edge.top - 1,
			     highwater);
		Line(0, -pos);
		ForeColor(6);
		PenPat(greyPattern);
		pos = ScaleS(i, playpen.edge.bottom - playpen.edge.top - 1,
			     highwater);
		LineTo(playpen.edge.left + sample, playpen.edge.bottom-pos-1);
		if (pos > highmark)
		    highmark = pos;
		ForeColor(1);
		PenNormal();
	    } else {
		MoveTo(playpen.edge.left + sample, playpen.edge.bottom - 1);
		Line(0, 0);
	    }
	}
	drawtick(1);
    }
    EndUpdate(ep->who);
    SetPort(save);
}

void
keyevent(ep)
EventRecord *ep;
{
    int i;

    if (ep->message & F_KEYUP)
	return;

    if (ep->who != window) {
	long save;

	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    if (ep->message & F_FUNC) {
	switch (ep->message & F_KEYMASK) {
	  case K_ESC:		/* Time to quit. */
	  case K_RESET:
	    ReleaseEvent();
	    break;

          case K_UP:		/* Scroll list up. */
            if (ep->message & F_ALT) {
                OffsetRect(&world, 0, -18);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else {
		Bell();
	    }
            break;

          case K_DOWN:		/* Scroll list down. */
            if (ep->message & F_ALT) {
                OffsetRect(&world, 0, 18);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else {
		Bell();
	    }
            break;

          case K_LEFT:
            if (ep->message & F_ALT) {
                OffsetRect(&world, -18, 0);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else
		Bell();
            break;

          case K_RIGHT:
            if (ep->message & F_ALT) {
                OffsetRect(&world, 18, 0);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else
		Bell();
            break;

	  default:
	    goto marker;
	    break;
	}
    } else {			/* Put up a temporary marker. */
      marker:
	MoveTo(playpen.edge.left + cursample, playpen.edge.top);
	Line(0, 2);
    }
}

void
menuevent(ep)
EventRecord *ep;
{
    short mid, item;

    BeginMenu(ep->message);
    mid = ep->message >> 16;
    item = ep->message & 0xFFFF;
    switch (mid) {
      case M_SAMPLE:
	if (item != last_sample_rate) {
	    CheckItem(M_SAMPLE, last_sample_rate, FALSE);
	    CheckItem(M_SAMPLE, item, TRUE);
	    sample_rate = rate_array[last_sample_rate = item];
	    cursample = 0;	/* Clear out sampling. */
	    curin = curout = 0;
	    if (!(idlemode || diskmode)) {
		highwater = (netmode ? 10 : 20) * sample_rate;
		highmul = 10;	/* Reset the vertical scale to the default.  */
		highdigindex = netmode ? 0 : 1;
	    }
	    highmark = 0;
	    EraseRect(&screen);
	    InvalRect(&screen);
	    dnix(F_CANCEL, napreq); /* Re-start timer. */
	}
	break;
    }
    EndMenu(ep->message);
}

void
windowevent(ep)
EventRecord *ep;
{
    int i, j;

    if (ep->who != window) {
	return;
    }
    switch (ep->message & F_KEYMASK) {
      case windowMove:
	OffsetRect(&world, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
	break;

      case windowZoom:
	if (weRicon) {
	    UnIconifyWindow(window);
	    weRicon = FALSE;
	} else {
	    IconifyWindow(window);
	    weRicon = TRUE;
	}
	break;

      case windowGoaway:
	ReleaseEvent();
	break;
    }
}

/*
** Find the name of the root volume.
*/

setdiskname()
{
    int pipes[2], child, pid, i, st;
    char buf[100], *cp;

    if (pipe(pipes))
	return;
    if (!(child = fork())) {	/* Child. */
	close(pipes[0]);
	dup2(pipes[1], 1);
	execl("/etc/devnm", "devnm", "/", 0);
	_exit(1);
    } else if (child > 0) {	/* Parent. */
	close(pipes[1]);
	cp = buf;
	while (1) {
	    do {
		i = read(pipes[0], cp, &buf[sizeof buf] - cp);
	    } while (i < 0 && i == EINTR);
	    if (i <= 0)
		break;
	    cp += i;
	}
	close(pipes[0]);
	while (1) {		/* Wait for child to die. */
	    pid = wait(&st);
	    if (pid == child)
		break;
	    if (pid < 0 && errno != EINTR)
		break;
	}
	if (!st) {
	    *cp = 0;
	    cp = strchr(buf, ' ');
	    *cp = 0;
	    if (diskname = malloc(strlen(buf) + 1)) {
		strcpy(diskname, buf);
	    }
	}
    } else {			/* Fork failed. */
	close(pipes[0]);
	close(pipes[1]);
    }
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %s, %d, %x, %X
**
** It's much smaller than the standard one.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	    } else {
		if (*p == 'd') {
		    p2 = itoa(argptr++->lng, 10);
		} else if (*p == 'x' || *p == 'X') {
		    p2 = itoa(argptr++->lng, 16);
		}
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	    }
	    p++;		/* Skip over format control. */
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix)
int val, radix;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    return buf;			/* No field width, return minimum number. */
}

/*
** Smaller than the standard atoi() function.
*/

myatoi(p)			/* Parse out a number. */
char *p;
{
    int val;
    short neg;
    char chr;

    neg = 0;
    val = 0;
    if (*p == '-')
	p++, neg = 1;
    while (1) {
	chr = *p++ - '0';
	if (chr < 0 || chr > 9)
	    break;
	val = (val * 10) + chr;
    }
    return neg ? -val : val;
}
