/*$
** vt220.c - VT220 specific terminal emulation routines ( vterm )
**
** Synopsis:
**    This module , along with vtsubs , handles the output side of the window
**    terminal emulation. All output character processing and Escape sequence 
**    decoding takes place here.
**
**	External Function List:
**	  init_vt220(vtcb)	Initializes the state machine for a given
**				virtual terminal control block.
**	  vt220(vtcb, buf, bc)	Processes a data block written to the terminal
**	  close_vt220(vtcb)	Close a virtual terminal window and free
**				allocated resources.
**
**	Notes:
**
**  (c) Copyrighted by ISC System Corp., 1988          All rights reserved.
**
*/

#include <stdio.h>
#ifndef _SYS5R3
#define _SYS5R3
#endif
#include <signal.h>
#include <sys/types.h>			/* System types. */
#include <dnix/fcodes.h>
#include <dnix/ireq.h>
#include <termio.h>
#include <sys/stat.h>
#include <isc/window.h>
#include <isc/vterm.h>
#include "vtcb.h"
#include "vt220.h"
#include "protos.h"

#ifndef lint
	static void Onechar(BYTE ch);
	static void EscProc(BYTE chr);
	static void CsiProc(BYTE chr);
	static void emselect1(BYTE chr);
	static void emselect2(BYTE chr);
	static void ltsel(BYTE chr);
	static void css(BYTE chr);
	static void css0(BYTE chr);
	static void css1(BYTE chr);
	static void lvlsel(BYTE chr);
	static void proc_ctrl(BYTE chr);
	static void stringflush(void);
#endif

BYTE	*vt_data;
int	vt_cnt;

extern	int	wm_cfd;

struct	vtcb	*Term;

BYTE	stringbuffer[128];
BYTE	*stringptr;
int	stringlen;
extern unsigned char font_sizes[MAXFONTNUMBER];


/*eject*/
/*%
**  init_vt220 - Initialize terminal emulation window
**
**   Synopsis:
**	Initializes variables in terminal control block, allocates storage
**	for display area and creates window for output. Window is initially
**	created invisible. The metrics for the selected font are requested
**	and if required the window is resized before it is displayed.
**	Requested font number is passed in vtcb->font.
**
**	Usage:
**		init_vt220(vtcb);
**
**	Parameters:
**		vtcb		pointer to a virtual terminal control block.
**
**	Returns:
**		void
**
**	Errors:
**		none
**
**	Notes:
**		We do a BeginUpdate/EndUpdate pair to wipe out the update
**		event that's automatically generated for a newly created
**		window.  We can skip it since the window's initially empty.
*/

void
init_vt220(vtcb)
struct vtcb *vtcb;
{
    short size;

    vtcb->nap_routine = vt220_rdtimer; /* Initialize pointer to nap timer. */
    size = vtcb->mrow * vtcb->mcol;
#ifdef DEBUG
    vtcb->display = malloc(size + 1);
    vtcb->display[size] = 0;
    vtcb->atts = malloc(size + 1);
    vtcb->atts[size] = 0;
#else
    vtcb->display = malloc(size);
    vtcb->atts = malloc(size);
#endif
    LOG("Init_vt220 vtcb = 0x%X\n", vtcb);
    memset(vtcb->display, ' ', size);
    memset(vtcb->atts, vtcb->att, size);
    vtcb->charhigh = CHARHIGH;	/* This should be from font metrics. */
    vtcb->charwide = CHARWIDE;
    vtcb->charoffset = LOFFSET;
    vtcb->wrect.edge.right = vtcb->wrect.edge.left
	+ vtcb->mcol * vtcb->charwide + vtcb->charoffset;
    vtcb->wrect.edge.bottom = vtcb->wrect.edge.top
	+ vtcb->mrow * vtcb->charhigh + BOFFSET;
    vtcb->windowid = NewWindow(&vtcb->wrect, vtcb->tname, hideWindow,
			       documentProc, frontWindow,
#ifdef SIGWINCH
			       hasGrow|hasDrag|hasGoAway|hasSendToBack|hasZoom,
#else
			       hasDrag|hasGoAway|hasSendToBack|hasZoom,
#endif
			       wm_cfd);
    SetPort(vtcb->windowid);
    Reset_vt220(vtcb);
    LOG(" windowid = %d\n",vtcb->windowid);
    vtcb->flag |= VT_INIT;
    SetKbdXlate(vtcb->mode.intl_kbd_filter);
    TextFont(vtcb->font);
    if (!vtcb->fontsize &&  vtcb->font < MAXFONTNUMBER)
	vtcb->fontsize = font_sizes[vtcb->font];
    TextSize(vtcb->fontsize);
    GetFontInfo(&vtcb->fontinfo);
    vtcb->fontinfo.widMax = CharWidth('N');	/* For Courier font too. */
    if (vtcb->fontinfo.widMax != CHARWIDE || /* Was window's size right? */
	(vtcb->fontinfo.ascent + vtcb->fontinfo.descent
	 + vtcb->fontinfo.leading) != CHARHIGH) {
	vtcb->charhigh = vtcb->fontinfo.ascent + vtcb->fontinfo.descent
	    + vtcb->fontinfo.leading;
	vtcb->charwide = vtcb->fontinfo.widMax;
	vtcb->wrect.edge.right = vtcb->wrect.edge.left
	    + vtcb->mcol * vtcb->charwide + vtcb->charoffset;
	vtcb->wrect.edge.bottom = vtcb->wrect.edge.top
	    + vtcb->mrow * vtcb->charhigh + BOFFSET;
	if ((vtcb->charwide & 0x07) && (vtcb->wrect.edge.right < 638)) {
	    vtcb->charoffset = 1;
	    vtcb->wrect.edge.right += 2;
	}
	SizeWindow(vtcb->windowid, 
		   vtcb->wrect.edge.right - vtcb->wrect.edge.left,
		   vtcb->wrect.edge.bottom - vtcb->wrect.edge.top, 0);
    }
    ShowWindow(vtcb->windowid);
    BeginUpdate(vtcb->windowid); /* Smash the pending update event since */
    EndUpdate(vtcb->windowid); /* we know the window's empty right now. */
/*    VCursor(vtcb, TRUE); 	/* Initial cursor draw */
    Menu_init(vtcb);		/* Build pull-down menus. */
    vtcb->ports = 1 << vtcb->windowid;
    makecopies(vtcb);
}

void
makecopies(vtcb)
struct vtcb *vtcb;
{
    struct vidcopy *vcp;
    char namebuf[256];

    if (vcp = vtcb->vidcopy) {
	while (vcp) {
	    if (vcp->windowid < 0) {
		strcpy(namebuf, "Copy of ");
		strncat(namebuf, vtcb->tname, sizeof namebuf);;
		vcp->wrect = vtcb->wrect;
		if ((vcp->windowid = NewWindow(&vcp->wrect, namebuf,
					       showWindow, documentProc,
					       frontWindow,
					       hasDrag|hasGoAway|hasSendToBack,
					       vcp->connect->connectid)) < 0) {
		    vcp->marker = 0; /* Mark for later deletion. */
		    continue;
		}
		vtcb->ports |= 1 << vcp->windowid;
		SetPort(vcp->windowid);
		TextFont(vtcb->font);
		TextSize(vtcb->fontsize);
		vcp->marker = 1;
	    }
	    vcp = vcp->next;
	}
	deletehotcopies(vtcb);
    }
}

/*%
**  vt220 - Process a write data block to the terminal window. 
**
**	Synopsis:
**		Select the display window for this terminal, clear the cursor
**		and drive the terminal emulation state machine using the characters
**		from the data block. 
**
**	Usage:
**		vt220(vtcb, buf, bc);
**
**	Parameters:
**		vtcb		pointer to virtual terminal control block
**		buf			pointer to data block to process ( unsigned char array )
**		bc			number of characters to process
**
**	Returns:
**			void
**
**	Errors:
**			none
**
**	Notes:
*/

void
vt220(vtcb, buf, bc)
struct vtcb *vtcb;
BYTE *buf;
int bc;
{
	struct vtcb *term;
	extern short obscurecurs;

	vt_data = buf;
	vt_cnt = bc;
	term = Term = vtcb;
	LOG("Processing %d character write buffer ", bc);
	LOG("", buf[bc] = 0);
	LOG("('%.20s')...\n", buf);
	if (!term->iconified) {
		SetPorts(vtcb->ports);
		if (term->flag & VT_ACTIVE)
			VCursor(term, FALSE); /* Turn off cursor. */
	}
	if (obscurecurs) {
	    ObscureCursor();
	    obscurecurs = 0;
	}
	stringptr = stringbuffer;
	stringlen = 0;
	while (vt_cnt--) {
#if 0
	    {
		static int logdata = -1; char buf[20];
		if (logdata < 0) {
		    sprintf(buf, "/tmp/vto%d", getpid());
		    logdata = creat(buf, 0666);
		}
		if (logdata >= 0)
		    write(logdata, vt_data, 1);
	    }
#endif
		(*vtcb->state)(*vt_data++);
#ifdef DEBUG
		if ( (term->display[term->mrow*term->mcol] != 0x00) || (term->col > term->mcol) || (term->col == term->mcol && !term->mode.pendforwrd) || ( term->row >= term->mrow))
			kill(getpid(), SIGIOT);
#endif
	}
	stringflush();
	if (term->mode.pendforwrd)	/* So cursor stays on the screen... */
		term->col = term->mcol -1;
#ifdef DEBUG
	if ( (term->display[term->mrow*term->mcol] != 0x00) || (term->col > term->mcol) || (term->col == term->mcol && !term->mode.pendforwrd) || ( term->row >= term->mrow))
	    kill(getpid(), SIGIOT);
#endif
	if (!term->iconified && term->flag & VT_ACTIVE)
		VCursor(vtcb, TRUE);
}


/*eject*/
/*%
**  close_vt220 - close a virtual terminal emulation window.
**
**	Synopsis:
**		Close window and deallocate resources.
**
**	Usage:
**		close_vt220(vtcb);
**
**	Parameters:
**		vtcb		pointer to virtual terminal control block
**
**	Returns:
**			void
**
**	Errors:
**			none
**
**	Notes:
*/

void
close_vt220(vtcb)
struct vtcb *vtcb;
{
    struct vidcopy *vcp;

    DisposeWindow(vtcb->windowid);
    vcp = vtcb->vidcopy;
    while (vcp) {
	SetPort(vcp->windowid);
	DisposeWindow(vcp->windowid);
	vcp->windowid = -1;
	vcp = vcp->next;
    }
    vtcb->ports = 0;
    if (vtcb->naprn)
	(void) rdnix(F_CANCEL, vtcb->naprn);
    vtcb->naprn = 0;
    if (vtcb->blinkrn)
	(void) rdnix(F_CANCEL, vtcb->blinkrn);
    vtcb->blinkrn = 0;
    free(vtcb->display);
    vtcb->display = 0;
    free(vtcb->atts);
    vtcb->atts = 0;
    if (vtcb->history) {
	free(vtcb->history);
	vtcb->history = 0;
    }
    vtcb->iconified = 0;
    vtcb->flag &= ~(VT_INIT|VT_ACTIVE);
}
/*eject*/
/*&
** NormDisp - Normal char processing
** This is the Base state for the char processing state machine.
** Assembles printable chars into a string until end of line is reached
** or a control char is received. EOL transits to Onechar state.
** Control char flushes current string and calls proc_ctrl.
**
** Parameters:
** 	ch	unsigned character to process.
**
** Returns:	void
**
** Errors:		none
*/

void
NormDisp(ch)
BYTE ch;
{
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? ch : (ch & 0x7F)) < ' ') { /* Control character? */
		stringflush();		/* Specially handle it. */
		proc_ctrl(ch);
	} else {			/* It's a normal displayable character, */
		ch ^= term->CharMask;	/* and it fits on the line still. */
		*stringptr++ = ch;
		stringlen++;
		if (term->col + stringlen == term->mcol) { /* Run off end of line? */
			term->mode.pendforwrd = TRUE; /* Switch to special processing. */
			term->state = Onechar;
		}
	}
}


/*eject*/
/*&
**  Onechar - Single char / End of line processing state 
**	      Handles the deferrred autowrap feature of VT style terminals.
**
**	Parameters:
**		ch		unsigned character to process.
**
**	Returns:	void
**
**	Errors:		none
*/

static void
Onechar(ch)
BYTE ch;
{
	short offset;
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? ch : (ch & 0x7F)) < ' ') { /* Control character? */
		stringflush();
		term->state = NormDisp;
#if 0
		term->mode.pendforwrd = FALSE;
#endif
		proc_ctrl(ch);
		if (term->col == term->mcol && term->state == NormDisp) { /* Still end of line? */
			term->mode.pendforwrd = TRUE; /* Switch to special processing. */
			term->state = Onechar;
			ch = term->display[(term->row+1)*term->mcol-1];		/* reuse previous char */
			*stringptr++ = ch;
			stringlen++;
			term->col--;
		}
	} else {			/* Normal character. */
		ch ^= term->CharMask;
		if (term->mode.autowrap) {
			dopend(TRUE);
			stringptr = stringbuffer;
			offset = term->row * term->mcol;
			term->display[offset] = ch;
			term->atts[offset] = term->att;
			*stringptr++ = ch;
			stringlen = 1;
			term->state = NormDisp;
		} else {		/* Autowrap's off. */
			offset = (term->row * term->mcol) + term->col;
			if (stringlen == 0) {
				term->display[offset] = ch;
				term->atts[offset] = term->att;
				*stringptr++ = ch;
				stringlen = 1;
			} else {
				offset = offset + stringlen - 1;
				term->display[offset] = ch;
				term->atts[offset] = term->att;
				*(stringptr-1) = ch;
			}
		}
	}
}

void
dopend(flag)			/* Clean up a pending autowrap. */
unsigned int flag;
{
	struct vtcb *term;

	term=Term;
	if (term->mode.pendforwrd) {
		stringflush();
		term->mode.pendforwrd = FALSE;
		if (term->mode.autowrap && flag) {
			cdl(1, TRUE);
			term->col = 0;
		}
		else
			term->col = term->mcol - 1;
	}
}

/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static void
EscProc(chr)
BYTE chr;
{
	short i;
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ') {
		proc_ctrl(chr);
		return;
	}

	term->state = NormDisp;	/* Good for most, others fix as necessary.  */

	switch (chr) {
	case '[':
		term->state = CsiProc;
		for (i=0; i<BUCKETS; i++)	/* */
			term->buckets[i] = -1;	/* initialize all Term->buckets */
		term->bucket = &term->buckets[0]; /* point to first parameter Term->bucket */
		term->bcount = 0;
		break;

	case 'M':	/* Reverse Index. */
		dopend(term->mode.not_vt_ansi);
		cul(1, TRUE);
		break;

	case 'E':			/* Next Line. */
		term->col = 0;		/* FALL THROUGH */
	case 'D':			/* Index. */
		dopend(TRUE);
		cdl(1, TRUE);
		break;


	case 'Z':			/* Terminal Identify. */
		identify();
		break;

	case '7':			/* Save cursor position. */
		term->save_row = term->row;
		term->save_col = term->col;
		term->save_mode = term->mode;
		term->save_att = term->att;
		break;

	case '8':			/* Restore saved position. */
		term->row = term->save_row;
		term->col = term->save_col;
		term->mode = term->save_mode;
		term->att = term->save_att;
		TextMode(term->att>>4);
		TextFace(term->att&0x0F);
		term->CharMask = term->mode.SO_state ? term->SOMask : term->SIMask;
		CheckItem(Setup_Menu, AutoWrap, term->mode.autowrap);
		CheckItem(Setup_Menu, Popup_Scroll, term->mode.pop_scroll);
		if (term->mode.pendforwrd)
			term->state = Onechar;
		break;

	case '%':			/* terminal emulation select */
		term->state = emselect1;
		break;

	case '(':			/* character set select for G0 */
		term->state = css0;
		break;

	case ')':			/* character set select for G1 */
		term->state = css1;
		break;

	case '*':			/* character set select for G2 */
	case '+':			/* character set select for G3 */
	case ' ':			/* transmission mode select */
		term->state = css;	/* set to ignore next char */
		break;

	case '#':
		term->state = ltsel;
		break;


	case '~':
	case 'n':
	case '}':
	case 'o':
	case '|':			/* select char set for GR | GL */
		if (term->mode.pendforwrd)
			term->state = Onechar;
		break;

	case 'N':
	case 'O':		/* single char shift G2 / G3  .. not supported */
		if (term->mode.pendforwrd)
			term->state = Onechar;
		break;

	case '=':			/* Application keypad mode */
		/*		kbd_ostat &= ~1;
			saveilvl = SPL_VSYNC;
			kbd_istat &= ~NUMLOCK;
			do_atleds();
			SPL_RST(saveilvl);		*/
		if (term->mode.pendforwrd)
			term->state = Onechar;
		break;

	case '>':			/* Numeric Keypad mode */
		/*		kbd_ostat |= 1;
			saveilvl = SPL_VSYNC;
			kbd_istat |= NUMLOCK;
			do_atleds();
			SPL_RST(saveilvl);		*/
		if (term->mode.pendforwrd)
			term->state = Onechar;
		break;


	}
}
/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static void
CsiProc(chr)			/* CSI sequence received. */
BYTE chr;
{
	short i;			/* temp work variable for loops */
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ') {
		proc_ctrl(chr);
		return;
	}
	if ((chr >= '0') && (chr <= '9')) {
		if (*term->bucket == -1)
			*term->bucket = chr - '0'; /* normal digit */
		else
			if ( *term->bucket >= 0 )
				*term->bucket = (*term->bucket * 10) + (chr - '0');
			else
				*term->bucket = (*term->bucket * 10) - (chr - '0');
		return;
	}

	if (chr == ';') {
		++term->bucket;	/* point to next Term->bucket */
		if ( ++term->bcount == BUCKETS)
			term->state = NormDisp;
		return;
	}
	if ( chr > ';' && chr < '@') {
		*term->bucket = -(chr - '0'); /* other digit ( use sign flag to signify ) */
		return;
	}

	term->bcount++;
	term->state = NormDisp;	/* good for most, fix others as necessary */
	switch (chr) {

	case	'"':			/* compatability level select */
		term->state = lvlsel;
		break;

	case 'A':				/* Cursor up a line. */
		dopend(term->mode.not_vt_ansi);
		if (term->buckets[0] < 1)
			term->buckets[0] = 1;
		cul(term->buckets[0],FALSE);
		break;

	case 'B':				/* Cursor down a line. */
		dopend(term->mode.not_vt_ansi);
		if (term->buckets[0] < 1)
			term->buckets[0] = 1;
		cdl(term->buckets[0],FALSE);
		break;

	case 'C':				/* Cursor forward. */
		dopend(term->mode.not_vt_ansi);
		if (term->buckets[0] < 1)
			term->buckets[0] = 1;
		term->col += term->buckets[0]*term->charwidth;
		if ( term->col >= term->mcol)
			term->col = term->mcol-1;
		break;

	case 'D':				/* Cursor backward. */
		dopend(term->mode.not_vt_ansi);
		if (term->buckets[0] < 1)
			term->buckets[0] = 1;
		cbc(term->buckets[0]);
		break;

	case 'H':				/* Cursor position (CUP) */
	case 'f':				/* Horizontal and vertical position (HVP) */
		term->buckets[0]--;		/* change DEC's 1 relative addressing to 0 relative */
		term->buckets[1]--;		/* ditto */
		if ( !term->mode.origin )
			term->row = min(max(term->buckets[0], 0), term->mrow-1);
		else
			term->row = min( max(term->buckets[0]+term->trow,term->trow) , term->brow);
		/*			term->plottype = term->linetab[term->cur_row];		*/
		/*			term->charwidth = term->plottype?2:1;				*/
		term->col = min(max((int)term->charwidth * term->buckets[1], 0), term->mcol-1);
		term->mode.pendforwrd = FALSE;
		break;

	case 'J':	/* Erase in display */
		dopend(term->mode.not_vt_ansi);
		switch (term->buckets[0]) {
		case -1:
		case 0:		/* erase to end of display */
			clear(term->row, term->col, term->mcol-term->col);
			if ( term->row != (term->mrow-1) ) {
				blockclr(term->row+1,term->mrow - term->row - 1);
				/*		for ( i = (term->row?term->row+1:0) ; i < term->mrow ; i ++ )
/*			term->linetab[i]=term->line_type;		/* clear line attribute */
			}
			break;

		case 1:		/* erase from start of display to cursor */
			if ( term->row != 0 ) {
				blockclr(0, term->row);
				/*		for ( i = 0 ; i < term->row ; i ++ )		/* */
				/*			term->linetab[i]=term->line_type;		/* clear line attribute */
			}
			clear(term->row, 0,term->col+1);
			break;

		case 2:		/* erase all display */
			blockclr(0, term->mrow);
			/*	for ( i = 0 ; i <= term->mrow ; i ++ )			/* */
			/*	term->linetab[i]=term->line_type;		/* clear line attribute */
			/*	term->pmask = term->erase | term->fg | term->bg;	/* */
			break;
		}

		/*		term->plottype = term->linetab[term->row];	/* */
		/*		term->charwidth = term->plottype?2:1;		/* */
		break;

	case	'K':		/* erase in line */
		dopend(term->mode.not_vt_ansi);
		switch (term->buckets[0]) {
		case	-1:
		case	0:
			clear(term->row, term->col, term->mcol-term->col);
			break;
		case	1:
			clear(term->row, 0,term->col + 1);
			break;
		case	2:
			clear(term->row,0,term->mcol);
			break;
		};
		break;

	case	'@':		/* Insert chars */
		dopend(term->mode.not_vt_ansi);
		if	(term->buckets[0] < 1)
			term->buckets[0] = 1;
		insert( min(term->charwidth*term->buckets[0], term->mcol-term->col));
		break;

	case	'P':		/* Delete chars */
		dopend(term->mode.not_vt_ansi);
		if	(term->buckets[0] < 1)
			term->buckets[0] = 1;
		delete( min(term->charwidth*term->buckets[0], term->mcol-term->col));
		break;

	case	'X':		/* Erase chars */
		dopend(term->mode.not_vt_ansi);
		if	(term->buckets[0] < 1)
			term->buckets[0] = 1;
		if ( (term->col + term->buckets[0]) > term->mcol ) {
			term->buckets[1] = term->row;		/* save temporarily */
			clear(term->row, term->col, (term->buckets[2]=(term->mcol-term->col)) );
			term->buckets[0] -= term->buckets[2];
			while ( term->buckets[0] && ( term->row < term->mrow )) {
				clear( ++term->row, 0,	(term->buckets[2] = min(term->mcol, term->buckets[0])) );
				term->buckets[0] -= term->buckets[2];
			}
			term->row = term->buckets[1];
		} else
			clear(term->row, term->col, term->charwidth*term->buckets[0]);
		break;

	case	'L':		/* Insert lines */
		dopend(term->mode.not_vt_ansi);
		if (term->buckets[0] < 1)
			term->buckets[0] = 1;
		if (term->row >= term->trow && term->row <= term->brow) {
			i = min(term->buckets[0], (term->brow-term->row+1));
			scroll(DOWN, term->brow - i + 1, term->row, i);
		}
		break;

	case	'M':		/* delete lines */
		dopend(term->mode.not_vt_ansi);
		if ( term->buckets[0] < 1 )
			term->buckets[0] =1;
		if ( term->row >= term->trow && term->row <= term->brow ) {
			i = min(term->buckets[0], (term->brow-term->row+1));
			scroll(UP, term->row, term->brow - i + 1, i);
		}
		break;

	case	'm':		/* set graphics rendition */
		if (term->mode.pendforwrd)
			term->state = Onechar;
		for ( i = 0 ; i < term->bcount ; i++ ) {
			switch	(term->buckets[i]) {
			case	0:		/* reset to default */
			case	-1:
				term->att = (normalText & 0x0F)|(srcCopy<<4);
				break;

			case	1:		/* set BOLD */
				term->att |= fixedBoldText;
				break;

			case	22:		/* clear BOLD */
				term->att &= ~fixedBoldText;
				break;

			case	4:		/* set UNDERLINE */
				term->att |= underlineText;
				break;

			case	24:		/* clear UNDERLINE */
				term->att &= ~underlineText;
				break;

			case	5:		/* set BLINK */
				break;

			case	25:		/* clear BLINK */
				break;

			case	7:		/* set REVERSE */
				term->att = (notSrcCopy<<4)|(term->att & 0x0F);
				break;

			case	27:		/* clear REVERSE */
				term->att = (srcCopy<<4)|(term->att & 0x0F);
				break;

			};
		}
		if (!term->iconified)
		{
			TextMode(term->att>>4);
			TextFace(term->att&0x0F);
		}
		break;

	case	'h':		/* mode set command */
		if (term->mode.pendforwrd)
			term->state = Onechar;
		term->bucket = term->buckets;
		while ( *term->bucket != -1 ) {
			if ( (term->buckets[0] < 0) && (*term->bucket > 0))
				*term->bucket = -150 - *term->bucket;

			switch ( *term->bucket ) {
			case	0:
				break;

			case	4:		/* IRM- Insert / Replace mode */
				term->mode.insmode = TRUE;
				break;

			case	20:		/* LNM- Line Feed / Newline mode */
				term->mode.autocr = TRUE;
				break;

			case	-151:	/* Cursor Key Mode */
				/*		kbd_ostat |= 2;		/* */
				break;

			case	-152:	/* VT 52 mode */
				break;

			case	-155:	/* DECSCNM- Inverse / Normal video */
				/*		if ( *status & 0x20 )		/* if monochrome unit */
				/*				controls->revscrn = 1;	/* */
				break;

			case	-156:	/* origin mode */
				term->mode.origin = TRUE;
				home();
				break;

			case	-157:
				term->mode.autowrap = TRUE;
				CheckItem(Setup_Menu, AutoWrap, TRUE);
				break;

			case	-1525:
				term->mode.cursoron = TRUE;	/* */
				break;

			default:
				break;
			};
			term->bucket++;
		}
		break;

	case	'l':		/* mode clear command */
		if (term->mode.pendforwrd)
			term->state = Onechar;
		term->bucket = term->buckets;
		while ( *term->bucket != -1 ) {
			if ( (term->buckets[0] < 0) && (*term->bucket > 0))
				*term->bucket = -150 - *term->bucket;

			switch ( *term->bucket ) {
			case	0:
				break;

			case	4:		/* IRM- Insert / Replace mode */
				term->mode.insmode = FALSE;
				break;

			case	20:		/* LNM- Line Feed / Newline mode */
				term->mode.autocr = FALSE;
				break;

			case	-151:	/* Cursor Key Mode */
				/*		kbd_ostat &= ~2;/* */
				break;

			case	-155:	/* DECSCNM- Inverse / Normal video */
				/*		if ( *status & 0x20 )		/* if monochrome unit */
				/*			controls->revscrn = 0; /* */
				break;

			case	-156:	/* origin mode */
				term->mode.origin = FALSE;
				home();
				break;

			case	-157:
				term->mode.autowrap = FALSE;
				CheckItem(Setup_Menu, AutoWrap, FALSE);
				break;

			case	-1525:
				term->mode.cursoron = FALSE;	/* */
				break;

			default:
				break;
			};
			term->bucket++;
		}
		break;

	case	'r':		/* set scrolling region */
		if ( term->buckets[0] < 1 || term->buckets[0] > term->mrow )
			term->buckets[0] = 1;
		if ( term->buckets[1] < 1 || term->buckets[1] > term->mrow )
			term->buckets[1] = term->mrow;
		if ( term->buckets[0] >= term->buckets[1]) {
			term->trow = 0;
			term->brow = term->mrow - 1;
		} else {
			term->trow = term->buckets[0] - 1;
			term->brow = term->buckets[1] - 1;
		}
		home();		/* Home the cursor. Undocumented side effect !! */
		break;

	case	'c':
		if (term->mode.pendforwrd)
			term->state = Onechar;
		identify();
		break;

	case	'n':
		dopend(TRUE);
		if ( term->buckets[0] == 6 ) {
			if (term->keycount + 10 < term->maxkeys) {
				addbuffer(term, ESC);
				addbuffer(term, '[');
				if ( term->row+1 > 9 )
					addbuffer(term, '0' +
						  ((short)(term->row+1)/10));
				addbuffer(term, ((short)(term->row+1)%10)+'0');
				addbuffer(term, ';');
				if ( term->col+1 > 99 )
					addbuffer(term, ((short)(term->col+1)/100)+'0');
				if ( term->col+1 > 9 )
					addbuffer(term, ((short)((short)(term->col+1)%100)/10)+'0');
				addbuffer(term, ((short)(term->col+1)%10)+'0');
				addbuffer(term, 'R');
			}
		}
		if (term->readrq)
			(void) vt220_rdchk(term);
		break;


	default:
		break;
	}
}
/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/


static	void	emselect1(chr)
BYTE	chr;
{
	struct vtcb *term;

	term=Term;
	if ( chr == '!' )
		term->state = emselect2;
	else
		term->state = NormDisp;
}

/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static	void	emselect2(chr)
BYTE	chr;
{
	Term->state = NormDisp;
}

/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static	void	ltsel(chr)
BYTE	chr;
{
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ')
	{
		proc_ctrl(chr);
		return;
	}
	term->state = NormDisp;
}

/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static	void	css(chr)
BYTE	chr;
{
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ')
	{
		proc_ctrl(chr);
		return;
	}
	term->state = NormDisp;	/* ignore selection */
}

/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static	void	css0(chr)
BYTE	chr;
{
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ')
	{
		proc_ctrl(chr);
		return;
	}
	if ( chr == '0' )
		term->SIMask = 0x80;
	else if ( chr == 'B' )
		term->SIMask = 0x00;
	if ( !term->mode.SO_state )
		term->CharMask = term->SIMask;
	term->state = NormDisp;
}

/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static	void	css1(chr)
BYTE	chr;
{
	struct vtcb *term;

	term=Term;
	if ((term->mode.intl_kbd_filter ? chr : (chr & 0x7F)) < ' ')
	{
		proc_ctrl(chr);
		return;
	}
	if ( chr == '0' )
		term->SOMask = 0x80;
	else if ( chr == 'B' )
		term->SOMask = 0x00;
	if ( term->mode.SO_state )
		term->CharMask = term->SOMask;
	term->state = NormDisp;
}


/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static void
lvlsel(chr)
BYTE chr;
{
	struct vtcb *term;

	term=Term;
	term->state = NormDisp;
	term->mode.autowrap = FALSE;
	CheckItem(Setup_Menu, AutoWrap, FALSE);
}

/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static void
proc_ctrl(chr)
BYTE	chr;
{
	short i;
	long port;
	struct vtcb *term;

	term=Term;
	switch (chr) {	/* Not printable, check for others. */

	case	NL:	/* Received a line feed.     */
	case	FF:	/* Received a form feed.     */
	case	VT:	/* Received a vertical tab.  */
		i = term->mode.pendforwrd & term->mode.autowrap;
		dopend(TRUE);
		if (term->mode.autocr)
			term->col = 0;
		if (!i)			/* Linefeed already been done? */
			cdl(1, TRUE);
		break;


	case	ESC:		/* Received an <esc> character.	     */
		term->state = EscProc;
		break;

	case	TAB:		/* Received a <tab> character.	     */
		dopend(term->mode.not_vt_ansi);
		if (term->charwidth == 1) {
			term->col = term->col+8 & ~0x07;
			if (term->col >= term->mcol)
				term->col = term->mcol-1;
		} else {
			term->col = term->col+16 & ~0x0F;
			if ( term->col >= term->mcol )
				term->col = (term->mcol-1) & ~0x01;
		}
		break;

	case	CR:		/* Received a carriage return.	     */
		dopend(FALSE);
		term->col = 0;
		break;

	case	BELL:		/* Received a bell character. */
		GetPort(&port);
		if (port != term->windowid) /* Only ding primary window. */
		    SetPort(term->windowid);
		Bell();
		if (port != term->windowid)
		    SetPort(port);
		break;

	case	BS:		/* Received a back space character.    */
		cbc(1);
		break;

	case	SO:
		term->CharMask = term->SOMask;
		term->mode.SO_state = TRUE;
		break;

	case	SI:
		term->CharMask = term->SIMask;
		term->mode.SO_state = FALSE;
		break;

	case	CN:
		term->state = NormDisp;
		break;

	case	CSI:
		term->state = CsiProc;
		for (i=0; i<BUCKETS; i++)	/* */
			term->buckets[i] = -1;	/* initialize all buckets */
		term->bucket = &term->buckets[0];	/* point to first parameter bucket */
		term->bcount = 0;
		break;


	case	NEL:	/* 8 bit Next Line char char */
		term->col = 0;		/* FALL THROUGH */
	case	IND:	/* 8 bit index char */
		dopend(TRUE);
		cdl(1, TRUE);
		break;

	case	RI:
		dopend(term->mode.not_vt_ansi);
		cul(1,TRUE);
		break;

	}					 /* end of switch(chr) */
}


/*eject*/
/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/

static void
stringflush()
{
	int	offset;
	struct vtcb *term;

	term=Term;
	if (stringlen) {
		if (term->mode.insmode)
			insert(stringlen);
		offset = term->row*term->mcol + term->col;
		memcpy(&term->display[offset], stringbuffer, stringlen);
		memset(&term->atts[offset], term->att, stringlen);
		if (!term->iconified) {
			Posn(term->col, term->row + 1);
		    DrawText((char*) stringbuffer, 0, stringlen);
		}
		term->col += stringlen;
		stringlen = 0;
		stringptr = stringbuffer;
	}
}
