// CopyBits transfers a bit image between any two bit maps and clips the
// result to the area specified by the maskRgn parameter.  The transfer may be
// performed in any of the eight source transfer modes.  The result is always
// clipped to the maskRgn and the boundary rectangle of the destination bit
// map; if the destination bit map is the current grafPort's portBits, it's
// also clipped to the intersection of the grafPort's clipRgn and visRgn.  If
// you don't want to clip to a maskRgn, just pass NIL for the maskRgn parm.
// The dstRect and maskRgn are in terms of the dstBits.bounds coordinate
// system, and the srcRect coordinates are in terms of the srcBits.bounds
// coordinates.
//
// Warning:  If you perform a CopyBits between two grafPorts that overlap, you
// must first convert them both to the same coordinate system (global
// coordinates are good), and then specify the dstBits bitMap for both srcBits
// _and_ dstBits.  (Otherwise CopyBits won't realize they overlap, and will
// most likely mangle the copy.)  This is most commonly found when both
// bitMaps are part of the screen, and you are moving things around.
//
// CopyBits can shrink or expand by giving it different-sized rectangles.
// When shrinking, all source bits end up reflected in the destination.
// That tends to make things a bit blobby and dark under extreme scales.
//
// NB: The main arguments are pass-by-reference (pointers),
//     but we do NOT modify them at any time!
//
// NB: Does not support color PixMaps!
//

#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include "../port/cdport.h"
#include "cdbits.h"
#define TRUE 1
#define FALSE 0

static Boolean
srcBitIsSet(const GrafPtr tp, const BitMap *srcBits, Boolean sOnScrn, int x, int y)
{
   U8 mask, *src;
   const BitMap *sm;

   sm = srcBits;
   if (sOnScrn && tp && tp->device == 1) { // 3.7 inch E-paper?
      x = sm->bounds.edge.bottom - 1 - x;
      y = y - sm->bounds.edge.left;
      if (y < 0)                // Stay on the e-Paper!
         y = 0;
      if (x < 0)
         x = 0;
      if (y >= sm->bounds.edge.right)
         y = sm->bounds.edge.right - 1;
      if (x >= sm->bounds.edge.bottom)
         x = sm->bounds.edge.bottom - 1;
      x *= sm->bounds.edge.right / 8;
      src = sm->baseAddr + x + (y / 8);
      mask = 1 << (7 - (y & 7));
   } else {                     // Normal (offscreen) monochrome bitmap.
#if 0
      if (y < 0 || y >= sm->bounds.edge.bottom ||  // Stay on the bitmap!
          x < 0 || x >= sm->bounds.edge.right)
         return FALSE;
#else
      // The smoothers might try to go off the reservation.
      // Better edge effects pinned at edge rather than null.
      if (y >= sm->bounds.edge.bottom) y = sm->bounds.edge.bottom - 1;
      y -= sm->bounds.edge.top;
      if (x >= sm->bounds.edge.right) x = sm->bounds.edge.right - 1;
      x -= sm->bounds.edge.left;
      if (y < 0) y = 0;
      if (x < 0) x = 0;
#endif
      src = sm->baseAddr;   // Base address of source image.
      src += y * sm->rowBytes;
      src += x / 8;
      mask = 1 << (7 - (x & 7));
   }
   return !!(*src & mask);
}

// See: http://www.scale2x.it/algorithm for details.
// Smoothing algorithm for exactly 2x blowup.
static void
doSmooth2(Boolean *srcBit, const GrafPtr tp, const BitMap *srcBits, Boolean sOnScrn, int isx, int isy, int dx, int dy)
{
#if 0
   // This one has bow-tie problems on diagonals.
   Boolean B, D, F, H;
   int pixNo;

   // Source/destination arrays for Scale2x:
   //   A B C   The center pixel E
   //   D E F   is expanded into      E0 E1
   //   G H I   4 new pixels E0..E3   E2 E3
   B = srcBitIsSet(tp, srcBits, sOnScrn, isx, isy - 1); // Get our source neighbors.
   D = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy); // (EPX/Scale2x algorithm.)
   F = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy);
   H = srcBitIsSet(tp, srcBits, sOnScrn, isx, isy + 1);
   if (B != H && D != F) {
      pixNo = ((dy & 1) << 1) | (dx & 1);
      switch (pixNo) { // E0..E3 in algorithm.
         case 0: if (D == B)
                    *srcBit = B; break;
         case 1: if (B == F)
                    *srcBit = B; break;
         case 2: if (D == H)
                    *srcBit = H; break;
         case 3: if (H == F)
                    *srcBit = H; break;
      }
   }
#endif
#if 1
   // Improved 2x scaler Scale2xSFX.  See:
   // https://forums.libretro.com/t/scalenx-artifact-removal-and-algorithm-improvement/1686/4
   Boolean A, B, C, D, E, F, G, H, I, J, K, L, M;
   int pixNo;

   // Source/destination arrays for Scale2xSFX:
   //     J
   //   A B C    The center pixel E
   // K D E F L  is expanded into      E0 E1
   //   G H I    4 new pixels E0..E3   E2 E3
   //     M
   A = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy - 1); // Get our source neighbors.
   B = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 1); // (Scale2xSFX algorithm.)
   C = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy - 1);
   D = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy);
   E = *srcBit;
   F = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy);
   G = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy + 1);
   H = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 1);
   I = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy + 1);
   J = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 2);
   K = srcBitIsSet(tp, srcBits, sOnScrn, isx - 2, isy);
   L = srcBitIsSet(tp, srcBits, sOnScrn, isx + 2, isy);
   M = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 2);

   pixNo = ((dy & 1) << 1) | (dx & 1);
   switch (pixNo) { // E0..E3 in algorithm.
      case 0: if (B==D && B!=F && D!=H && (E!=A || E==C || E==G || A==J || A==K))
                 *srcBit = B; break;
      case 1: if (B==F && B!=D && F!=H && (E!=C || E==A || E==I || C==J || C==L))
                 *srcBit = B; break;
      case 2: if (D==H && B!=D && F!=H && (E!=G || E==A || E==I || G==K || G==M))
                 *srcBit = H; break;
      case 3: if (F==H && B!=F && D!=H && (E!=I || E==C || E==G || I==L || I==M))
                 *srcBit = H; break;
   }
#endif
}

// See: http://www.scale2x.it/algorithm for details.
// Smoothing algorithm for exactly 3x blowup.
static void
doSmooth3(Boolean *srcBit, const GrafPtr tp, const BitMap *srcBits, Boolean sOnScrn, int isx, int isy, int dx, int dy)
{
   Boolean A, B, C, D, E, F, G, H, I, J, K, L, M;
   int pixNo;

#if 0 // Scale3x does kind of a crappy job on 45-degree diagonals.
      // Source/destination arrays for Scale3x:
      //   A B C   The center pixel E    E0 E1 E2
      //   D E F   is expanded into      E3 E4 E5
      //   G H I   9 new pixels E0..E8   E6 E7 E8
   B = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 1); // Get our source neighbors.
   D = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy);     // (Scale3x algorithm.)
   F = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy);
   H = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 1);
   if (B != H && D != F) {
      A = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy - 1); // Get the rest of the
      C = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy - 1); // neighbors' values.
      E = *srcBit;
      G = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy + 1);
      I = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy + 1);
      pixNo = ((dy % 3) * 3) + (dx % 3);
      switch (pixNo) { // E0..E8 in algorithm.
         case 0: if (D == B)
                    *srcBit = D; break;
         case 1: if ((D == B && E != C) ||
                     (B == F && E != A))
                    *srcBit = B; break;
         case 2: if (B == F)
                    *srcBit = F; break;
         case 3: if ((D == B && E != G) ||
                     (D == H && E != A))
                    *srcBit = D; break;
         case 4:    *srcBit = E; break;
         case 5: if ((B == F && E != I) ||
                     (H == F && E != C))
                    *srcBit = F; break;
         case 6: if (D == H)
                    *srcBit = D; break;
         case 7: if ((D == H && E != I) ||
                     (H == F && E != G))
                    *srcBit = H; break;
         case 8: if (H == F)
                    *srcBit = F; break;
      }
   }
#endif
#if 0 // Scale3x does kind of a crappy job on 45-degree diagonals.
      // Source/destination arrays for Scale3x:
      //   A B C   The center pixel E    E0 E1 E2
      //   D E F   is expanded into      E3 E4 E5
      //   G H I   9 new pixels E0..E8   E6 E7 E8
   A = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy - 1); // Get our source neighbors.
   B = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 1); // (Scale3x algorithm.)
   C = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy - 1);
   D = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy);
   E = *srcBit;
   F = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy);
   G = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy + 1);
   H = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 1);
   I = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy + 1);
   pixNo = ((dy % 3) * 3) + (dx % 3);
   switch (pixNo) { // E0..E8 in algorithm.
// E0 = B=D & B!=F & D!=H ? 0.5*(B+D) : E
     case 0: if (B==D && B!=F && D!=H)
                *srcBit = D; break;
// E1 = (B=D & B!=F & D!=H & E!=C) | (B=F & B!=D & F!=H & E!=A) ? B : E;
     case 1: if ((B==D && B!=F && D!=H) ||
                 (B==F && B!=D && F!=H && E!=A))
                *srcBit = B; break;
// E2 = B=F & B!=D & F!=H ? 0.5*(B+F) : E
     case 2: if (B==F && B!=D && F!=H)
                *srcBit = F; break;
// E3 = (B=D & B!=F & D!=H & E!=G) | (D=H & B!=D & F!=H & E!=A) ? D : E;
     case 3: if ((B==D && B!=F && D!=H && E!=G) ||
                 (D==H && B!=D && F!=H && E!=A))
                *srcBit = D; break;
// E4 = E;
     case 4:    *srcBit = E; break;
// E5 = (B=F & B!=D & F!=H & E!=I) | (F=H & B!=F & D!=H & E!=C) ? F : E;
     case 5: if ((B==F && B!=D && F!=H && E!=I) ||
                 (F==H && B!=F && D!=H && E!=C))
                *srcBit = F; break;
// E6 = D=H & B!=D & F!=H ? 0.5*(D+H) : E
     case 6: if (D==H && B!=D && F!=H)
                *srcBit = D; break;
// E7 = (D=H & B!=D & F!=H & E!=I) | (F=H & B!=F & D!=H & E!=G) ? H : E;
     case 7: if ((D==H && B!=D && F!=H && E!=I) ||
                 (F==H && B!=F && D!=H && E!=G))
                *srcBit = H; break;
// E8 = F=H & B!=F & D!=H ? 0.5*(F+H) : E
     case 8: if (F==H && B!=F && D!=H)
                *srcBit = F; break;
   }
#endif
#if 1
   // Improved 3x scaler Scale3xSFX.  See:
   // https://forums.libretro.com/t/scalenx-artifact-removal-and-algorithm-improvement/1686/4
   // Source/destination arrays for Scale3xSFX:
   //     J
   //   A B C    The center pixel E    E0 E1 E2
   // K D E F L  is expanded into      E3 E4 E5
   //   G H I    9 new pixels E0..E8   E6 E7 E8
   //     M
   A = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy - 1); // Get our source neighbors.
   B = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 1); // (Scale3xSFX algorithm.)
   C = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy - 1);
   D = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy);
   E = *srcBit;
   F = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy);
   G = srcBitIsSet(tp, srcBits, sOnScrn, isx - 1, isy + 1);
   H = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 1);
   I = srcBitIsSet(tp, srcBits, sOnScrn, isx + 1, isy + 1);
   J = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy - 2);
   K = srcBitIsSet(tp, srcBits, sOnScrn, isx - 2, isy);
   L = srcBitIsSet(tp, srcBits, sOnScrn, isx + 2, isy);
   M = srcBitIsSet(tp, srcBits, sOnScrn, isx,     isy + 2);
   pixNo = ((dy % 3) * 3) + (dx % 3);
   switch (pixNo) { // E0..E8 in algorithm.
     case 0:
        if ((B==D && B!=F && D!=H && (E!=A || E==C || E==G || A==J || A==K))
            || (B==D && C==E && C!=J && A!=E)
            || (B==D && E==G && A!=E && G!=K))
           *srcBit = B; break;
     case 1:
        if ((B==D && B!=F && D!=H
             && (E!=A || E==C || E==G || A==J || A==K) && E!=C)
            || (B==F && B!=D && F!=H
                && (E!=C || E==A || E==I || C==J || C==L) && E!=A))
           *srcBit = B; break;
     case 2:
        if ((B==F && B!=D && F!=H && (E!=C || E==A || E==I || C==J || C==L))
            || (B==F && A==E && A!=J && C!=E)
            || (B==F && E==I && C!=E && I!=L))
           *srcBit = B; break;
     case 3:
        if ((B==D && B!=F && D!=H && (E!=A || E==C || E==G || A==J || A==K) && E!=G)
            || (D==H && B!=D && F!=H && (E!=G || E==A || E==I || G==K || G==M) && E!=A))
           *srcBit = D; break;
     case 4: *srcBit = E; break;
     case 5:
        if ((F==H && B!=F && D!=H && (E!=I || E==C || E==G || I==L || I==M) && E!=C)
            || (B==F && B!=D && F!=H && (E!=C || E==A || E==I || C==J || C==L) && E!=I))
           *srcBit = F; break;
     case 6:
        if ((D==H && B!=D && F!=H && (E!=G || E==A || E==I || G==K || G==M))
            || (D==H && A==E && A!=K && E!=G)
            || (D==H && E==I && E!=G && I!=M))
           *srcBit = H; break;
     case 7:
        if ((F==H && B!=F && D!=H
             && (E!=I || E==C || E==G || I==L || I==M) && E!=G)
            || (D==H && B!=D && F!=H
                && (E!=G || E==A || E==I || G==K || G==M) && E!=I))
           *srcBit = H; break;
     case 8:
        if ((F==H && B!=F && D!=H && (E!=I || E==C || E==G || I==L || I==M))
            || (F==H && C==E && C!=L && E!=I)
            || (F==H && E==G && E!=I && G!=M))
           *srcBit = H; break;
   }
#endif
}

void
CopyBits(const BitMap *srcBits, const BitMap *dstBits,
         const Rect *srcRect, const Rect *dstRect,
         int mode, RgnHandle maskRgn)
{
   int sorigwid; // Original width of source rect.
   int sorighei; // Original height of source rect.
   int dorigwid; // Original width of dest rect.
   int dorighei; // Original height of dest rect.
   Rect clipSRect;    // Clipped source rectangle.
   Rect clipDRect;    // Clipped destination rectangle.
   Rect scratch;      // Scratch rectangle.
   Rect dbr;          // Destination BitMap bounding rectangle.
   const short *mps, *cps, *vps; // Region data pointers; Nil if inactive.
   int cStartX, cEndX, vStartX, vEndX, mStartX, mEndX;
   int co, vo, mo;// Region data offsets.
   Boolean sOnScrn;   // True if source is on screen (thePort).
   Boolean dOnScrn;   // True if destination is on screen (thePort).
   Boolean srcBit;
   Boolean smooth2x, smooth3x;
   Fixed hratio;      // Horizontal stretching ratio (fixed point).
   Fixed vratio;      // Vertical stretching ratio (fixed point).
   int delta, isx, tsx, rsx, isy, tsy, dx, dy;
   Fixed sx, sy;
   GrafPtr tp, savePort;

   tp = thePort;
   sOnScrn = tp && (srcBits == &tp->portBits);
   dOnScrn = tp && (dstBits == &tp->portBits);
   if (dOnScrn && (tp->pnVis < 0)) return;
   GetPort(&savePort);
   vratio = Int2Fix(1);
   hratio = vratio;
   if (EmptyRect(dstRect) || EmptyRect(srcRect)) return;
   sorigwid = srcRect->edge.right - srcRect->edge.left;
   dorigwid = dstRect->edge.right - dstRect->edge.left;
   hratio = FixRatio(sorigwid, dorigwid);
   sorighei = srcRect->edge.bottom - srcRect->edge.top;
   dorighei = dstRect->edge.bottom - dstRect->edge.top;
   vratio = FixRatio(sorighei, dorighei);
   // Will we engage special smoothing logic?
   smooth2x = (hratio == FixRatio(1, 2)) && (vratio == FixRatio(1, 2));
   smooth3x = (hratio == FixRatio(1, 3)) && (vratio == FixRatio(1, 3));

// Trim the 'original' source & destination rectangles so that they're entirely
// within their bitmap bounds.  This is so we never touch any memory that's
// outside the bitmaps (which may well be read or write-protected).  We do our
// best to scale things so the right results will show.

   if (sOnScrn) {
      RotRect(&srcBits->bounds, &scratch);
      if (!SectRect(srcRect, &scratch, &clipSRect)) return;
   } else
      if (!SectRect(srcRect, &srcBits->bounds, &clipSRect)) return;

// Original 68010 assy form, typical 1 of 4:
//         movw  top(clipSRect),d0
//         subw  top(srcRect),d0
//         beqs  clipd1            | Any of the top of the SrcRect clipped?
//         cmpw  dorighei,sorighei | Yes, trim off a proportionate amount of
//         beqs  clipd1            | the DstRect's top.
//         muls  dorighei,d0
//         divs  sorighei,d0
// clipd1: addw  top(dstRect),d0
//         movw  d0,top(scratch)

   delta = clipSRect.edge.top - srcRect->edge.top;   // Any of the top of srcRect clipped?
   if (delta && (sorighei != dorighei)) // Yes, trim off a
      delta = ScaleS(delta, dorighei, sorighei);  // proportionate amount
   scratch.edge.top = delta + dstRect->edge.top;             // of dstRect's top.

   delta = clipSRect.edge.left - srcRect->edge.left;     // Same for the left sides.
   if (delta && (sorigwid != dorigwid))
      delta = ScaleS(delta, dorigwid, sorigwid);
   scratch.edge.left = delta + dstRect->edge.left;

   delta = clipSRect.edge.bottom - srcRect->edge.bottom; // Same for the bottoms.
   if (delta && (sorighei != dorighei))
      delta = ScaleS(delta, dorighei, sorighei);
   scratch.edge.bottom = delta + dstRect->edge.bottom;

   delta = clipSRect.edge.right - srcRect->edge.right;   // Same for the right sides.
   if (delta && (sorigwid != dorigwid))
      delta = ScaleS(delta, dorigwid, sorigwid);
   scratch.edge.right = delta + dstRect->edge.right;

// Find out how much of the (new) dest rectangle lies outside the dest
// BitMap, and trim the source rectangle by that amount so we never go
// off the dest BitMap.  Also scale by relative sizes for stretching.
//
// Any add'l clipping we do based on rectangles (that affects the size of the
// destination rectangle we will be traversing) also needs to be done here,
// so we can also trim the source traversal rectangle a proportionate amount.
// (Because of scaling.)
//
// No valid destination at all, and we're done.

   dbr = dstBits->bounds;
   if (dOnScrn && (tp->device == 1)) // Special destination coordinate
      RotRect(&dbr, &dbr);           // 90deg rotation hack.
   if (!SectRect(&scratch, &dbr, &clipDRect)) return;
// Now take into account thePort's visRgn and clipRgn, and the maskRgn argument.
   vps = 0; vo = 0; vEndX = 0;
   cps = 0; co = 0; cEndX = 0;
   mps = 0; mo = 0; mEndX = 0;
   if (dOnScrn) {
      if (!SectRect(&clipDRect, &tp->portRect, &clipDRect)) return;
      if (tp->clipRgn && *tp->clipRgn) { // Clip rgn?
         if (!SectRect(&clipDRect, &(*tp->clipRgn)->rgnBBox, &clipDRect)) return;
         if ((*tp->clipRgn)->rgnSize != sizeof(RegionHdr))
            cps = &(*(tp->clipRgn))->rgnData;
         co = 0;               // Offset to Y marker.
      }
      if (tp->visRgn && *tp->visRgn) { // Vis region?
         if (!SectRect(&clipDRect, &(*tp->visRgn)->rgnBBox, &clipDRect)) return;
         if ((*tp->visRgn)->rgnSize != sizeof(RegionHdr))
            vps = &(*(tp->visRgn))->rgnData;
         vo = 0;               // Offset to Y marker.
      }
   }
   if (maskRgn && *maskRgn) { // Mask region?
      if (!SectRect(&clipDRect, &(*maskRgn)->rgnBBox, &clipDRect)) return;
      if ((*maskRgn)->rgnSize != sizeof(RegionHdr))
         mps = &(*maskRgn)->rgnData;
      mo = 0;               // Offset to Y marker.
   } // Now clipDRect holds our final dest rect, and scratch is the original.
   if (dOnScrn && _ShieldCursHook)
      (*_ShieldCursHook)(&clipDRect, &dbr.corner.topLeft);

// Original 68010 assy form, typical 1 of 4:
//         movw top(clipDRect),d0
//         subw top(scratch),d0
//         beqs clips1            | Any of the top of the new DstRect clipped?
//         cmpw dorighei,sorighei | Yes, trim off a proportionate amount of
//         beqs clips1            | the SrcRect's top.
//         muls sorighei,d0
//         divs dorighei,d0
// clips1: addw d0,top(clipSRect)

   delta = clipDRect.edge.top - scratch.edge.top; // Any of the top of new dest clipped?
   if (delta && (sorighei != dorighei)) // Yes, trim off a prop. amt
      clipSRect.edge.top += ScaleS(delta, sorighei, dorighei); // of srcRect's top.

   delta = clipDRect.edge.left - scratch.edge.left;     // Same for the left sides.
   if (delta && (sorigwid != dorigwid))
      clipSRect.edge.left += ScaleS(delta, sorigwid, dorigwid);

   delta = clipDRect.edge.bottom - scratch.edge.bottom; // Same for the bottoms.
   if (delta && (sorighei != dorighei))
      clipSRect.edge.bottom += ScaleS(delta, sorighei, dorighei);

   delta = clipDRect.edge.right - scratch.edge.right;   // Same for the right sides.
   if (delta && (sorigwid != dorigwid))
      clipSRect.edge.right += ScaleS(delta, sorigwid, dorigwid);

   scratch = clipDRect;        // New-new Dest rect.
#if 0
   {
      char buf1[64], buf2[64];
      // Print out the post-clip/scale stats when debugging.
      printf("hratio: %s vratio: %s smooth2x: %d smooth3x: %d\n",
             FixedS32ToStr(buf1, sizeof buf1, hratio, 65536, 6, FALSE),
             FixedS32ToStr(buf2, sizeof buf2, vratio, 65536, 6, FALSE),
             smooth2x, smooth3x);
      DumpRect(&clipSRect, "");
      DumpRect(&clipDRect, "");
   }
#endif
// We run the outer loops as destination integers because due to the XOR
// transfer modes we never want to touch a destination pixel twice, even if
// we're scaling.  That makes (only) the source index logic tricky.  This
// also ensures that we touch every destination pixel that we should;
// Once and only once.  dX and dY are in global coordinates.
   dy = clipDRect.edge.top - dbr.edge.top;
   sy = Int2Fix(clipSRect.edge.top) + (vratio / 4); // Epsilon avoids edge effects.
   while (dy < (clipDRect.edge.bottom - dbr.edge.top)) { // Move across the dest bitmap.
      if (cps) {
         if (!co)               // Find the first appropriate scanline.
            _FastForward(&cps, dy + dbr.edge.top);
         if ((cps[co] != END_TAG) && (dy >= (cps[co] - dbr.edge.top)))
            cps += co;          // Time to move the Y?
         co = 1;                // For re-traversing X list for each Y.
         cStartX = -32767;      // Coords of active X segment of clip region;
         cEndX   = -32767;      // Minus infinity to start with.
      }
      if (vps) {
         if (!vo)               // Find the first appropriate scanline.
            _FastForward(&vps, dy + dbr.edge.top);
         if ((vps[vo] != END_TAG) && (dy >= vps[vo] - dbr.edge.top))
            vps += vo;          // Time to move the Y?
         vo = 1;                // For re-traversing X list for each Y.
         vStartX = -32767;      // Coords of active X segment of vis region;
         vEndX   = -32767;      // Minus infinity to start with.
      }
      if (mps) {
         if (!mo)               // Find the first appropriate scanline.
            _FastForward(&mps, dy + dbr.edge.top);
         if ((mps[mo] != END_TAG) && (dy >= mps[mo] - dbr.edge.top))
            mps += mo;          // Time to move the Y?
         mo = 1;                // For re-traversing X list for each Y.
         mStartX = -32767;      // Coords of active X segment of mask region;
         mEndX   = -32767;      // Minus infinity to start with.
      }
      isy = Fix2Int(sy);
      sy += vratio;
      tsy = Fix2Int(sy) - 1;
      dx = clipDRect.edge.left - dbr.edge.left;
      sx = Int2Fix(clipSRect.edge.left) + (hratio / 4); // Epsilon, ditto.
      while (dx < (clipDRect.edge.right - dbr.edge.left)) {
         isx = Fix2Int(sx);
         sx += hratio;
         tsx = Fix2Int(sx) - 1; // For traversing source X during shrinking.
         srcBit = srcBitIsSet(tp, srcBits, sOnScrn, isx, isy);

         // Shrinking?  Incorporate, or at least traverse, all source bits.
         // Once the bit is set we can stop looking at the source bitmap.
         if (isy < tsy) {       // Shrinking vertically?
            rsx = isx;          // Look at ALL pixels on each source line.
            while (isy < tsy) {
               isy++;           // Traverse to next needed source scanline.
               isx = rsx;       // Reload our X index.
               while (isx < tsx) {
                  isx++;        // Traverse to next needed source pixel.
                  if (!srcBit)
                     srcBit = srcBitIsSet(tp, srcBits, sOnScrn, isx, isy);
               }
            }
         } else {       // Not shrinking vertically, shrinking horizontally?
            while (isx < tsx) {
               isx++;           // Traverse to next needed source pixel.
               if (!srcBit)
                  srcBit = srcBitIsSet(tp, srcBits, sOnScrn, isx, isy);
            }
         }

         if (smooth2x) doSmooth2(&srcBit, tp, srcBits, sOnScrn,
                                 isx, isy, dx, dy);
         if (smooth3x) doSmooth3(&srcBit, tp, srcBits, sOnScrn,
                                 isx, isy, dx, dy);

         while (cps && dx >= cEndX && cps[co] != END_TAG) {
            cStartX = cps[co] - dbr.edge.left; // Get the X segment of clip rgn
            cEndX = cps[co+1] - dbr.edge.left; // for where we are right now.
            co += 2;
         }
         while (vps && dx >= vEndX && vps[vo] != END_TAG) {
            vStartX = vps[vo] - dbr.edge.left; // Get the X segment of vis rgn
            vEndX = vps[vo+1] - dbr.edge.left; // for where we are right now.
            vo += 2;
         }
         while (mps && dx >= mEndX && mps[mo] != END_TAG) {
            mStartX = mps[mo] - dbr.edge.left; // Get X segment of mask rgn
            mEndX = mps[mo+1] - dbr.edge.left; // for where we are right now.
            mo += 2;
         }
         if ((!cps || (dx >= cStartX && dx < cEndX)) && // Clip!
             (!vps || (dx >= vStartX && dx < vEndX)) &&
             (!mps || (dx >= mStartX && dx < mEndX))) {

            if (mode > srcBic) // Now transfer the bit to dest.
               srcBit = !srcBit;
            switch (mode) {
              case notSrcCopy:
              case srcCopy:    if (srcBit) {
                             if (dOnScrn)    (*tp->_SetPix)(tp, dx, dy);
                                       else _SetDstBit(dstBits, dx, dy);
                          } else {
                             if (dOnScrn)   (*tp->_ClearPix)(tp, dx, dy);
                                      else _ClearDstBit(dstBits, dx, dy);
                          } break;
              case notSrcOr:
              case srcOr:      if (srcBit) {
                             if (dOnScrn)   (*tp->_SetPix)(tp, dx, dy);
                                      else _SetDstBit(dstBits, dx, dy);
                          } break;
              case notSrcXor:
              case srcXor:     if (srcBit) {
                             if (dOnScrn)   (*tp->_FlipPix)(tp, dx, dy);
                                      else _FlipDstBit(dstBits, dx, dy);
                          } break;
              case notSrcBic:
              case srcBic:     if (srcBit) {
                             if (dOnScrn)   (*tp->_ClearPix)(tp, dx, dy);
                                      else _ClearDstBit(dstBits, dx, dy);
                          } break;
            }
         }
         dx++;
      }
      while (cps && cps[co] != END_TAG) // Get to end of X
         co += 2;               // segments, if not already there.
      co++;                     // Point to next Y marker.
      while (vps && vps[vo] != END_TAG) // Get to end of X
         vo += 2;               // segments, if not already there.
      vo++;                     // Point to next Y marker.
      while (mps && mps[mo] != END_TAG) // Get to end of X
         mo += 2;               // segments, if not already there.
      mo++;                     // Point to next Y marker.
      dy++;
   }
   if (dOnScrn && _ShowCursHook)
      (*_ShowCursHook)();
}
