/*
** Rectangle scrolling stuff.
**
** ScrollRect shifts ("scrolls") the bits that are inside the intersection of
** the specified rectangle and the visRgn, clipRgn, portRect, and
** portBits.bounds of the current grafPort.  No other bits are affected.  The
** bits are shifted a distance of dh horizontally and dv vertically.  The
** positive directions are to the right and down.  Bits that are shifted out
** of the scroll area are lost -- they're neither placed outside the area nor
** saved.  The space created by the scroll is filled with the grafPort's
** background pattern (thePort.bkPat), and the updateRgn is changed to this
** filled area.  NOTE:  If the rectangle is partially obscured by another
** window then it is likely that more area is opened up than you might think.
** UpdateRgn will correctly reflect this case.
**
** If updateRgn is NIL, it will not be returned to you.  If it is -1, the
** region will be added to the update region of the current window.  In
** addition, if it's -1 ScrollRect will also scroll any already dirty area
** (the update region of the grafPort) by the amount of the scroll.  This will
** let cumulative ScrollRects get a proper update region (event) later.
**
** ScrollRect doesn't change the coordinate system of the grafPort, it simply
** moves the entire document to _different_ coordinates.  Notice that
** ScrollRect doesn't move the pen and the clipRgn.  However, since the
** document has moved, they're in a different position relative to the
** document.
**
** To restore the coordinates of the document to what they were before the
** ScrollRect, you can use the SetOrigin procedure.  For example, if before
** the ScrollRect the coordinates of the top left of your document were
** (100,100), and you did a ScrollRect(&r, 10, 20, ...) the coordinates of the
** document are offset by (10, 20).  You could call SetOrigin(90, 80) to offset
** the coordinate system to compensate for the scroll.  The document itself
** doesn't move as a result of SetOrigin (ScrollRect did that), but the pen
** and ClipRgn move down and to the right, and thus are restored to their
** original positions relative to the document.  Notice that updateRgn will
** still need to be redrawn.
*/

#include "../clunkdraw.h"
#include <alloca.h>

void
ScrollRect(rect, deltah, deltav, updateRgn)
const Rect *rect;
short deltah, deltav;           /* Formerly dh and dv respectively. */
RgnHandle updateRgn;
{
    GrafPtr tp;
    Rect source, dest, work;
    RgnHandle uprgn, workrgn;
    short tempcopy;             /* Flag for going faster. */
    short i, j;

    if (!(tp = thePort)) return;
    if (tp->pnVis < 0)
        return;
    tempcopy = 0;
    source = *rect;
    uprgn = NewRgn();           /* Now create update region. */
    workrgn = NewRgn();
    SectRgn(tp->clipRgn, tp->visRgn, uprgn);

    (void) SectRect(&source, &(*uprgn)->rgnBBox, &source); /* Clip to work */
    (void) SectRect(&source, &tp->portRect, &source); /* within the */
    (void) SectRect(&source, &tp->portBits.bounds, &source); /* visible */
    dest = source;              /* window (saves effort). */

    i = (deltah < 0) ? -deltah : deltah; /* If we're scrolling more than the */
    j = (deltav < 0) ? -deltav : deltav; /* size of the rectangle, skip the */
    if (i >= (source.edge.right - source.edge.left) || /* scroll itself. */
        j >= (source.edge.bottom - source.edge.top)) {
        RectRgn(uprgn, &source);
    } else {                    /* We need to actually scroll some bits. */
        if (deltav <= 0) {      /* Clip off the appropriate edges of the */
            source.edge.top -= deltav; /* src and dest rectangles so that */
            dest.edge.bottom += deltav; /* CopyBits moves the right stuff. */
        } else {                /* (Moving downwards.) */
            source.edge.bottom -= deltav;
            dest.edge.top += deltav;
            tempcopy = 1;       /* Can't go fastest if moving down. */
        }
        if (deltah <= 0) {      /* (Moving left.) */
            source.edge.left -= deltah;
            dest.edge.right += deltah;
        } else {                /* (Moving right.) */
            source.edge.right -= deltah;
            dest.edge.left += deltah;
            tempcopy = 1;       /* Can't go fastest if moving right. */
        }
                                /* We've finished preparing for CopyBits. */
        RectRgn(workrgn, &source);
        SectRgn(uprgn, workrgn, uprgn);
        DiffRgn(workrgn, uprgn, uprgn); /* Overlapped src region (if any). */
        OffsetRgn(uprgn, deltah, deltav); /* Move to dest coordinates. */

        if (deltav > 0) {       /* Wipe out 'opened' top or bottom strip. */
            work.edge.top = source.edge.top;
            work.edge.bottom = work.edge.top + deltav; /* Wipe full width of */
            work.edge.right = rect->edge.right; /* rect 1st, it's faster for */
            work.edge.left = rect->edge.left; /* wider ones.  (The side'll */
            RectRgn(workrgn, &work); /* be the shorter.) */
            UnionRgn(workrgn, uprgn, uprgn);
        } else if (deltav < 0) {
            work.edge.bottom = source.edge.bottom;
            work.edge.top = work.edge.bottom + deltav;
            work.edge.right = rect->edge.right;
            work.edge.left = rect->edge.left;
            RectRgn(workrgn, &work);
            UnionRgn(workrgn, uprgn, uprgn);
        }
        if (deltah > 0) {       /* Wipe the opened up side strip. */
            work.edge.top = dest.edge.top;
            work.edge.bottom = dest.edge.bottom;
            work.edge.left = source.edge.left;
            work.edge.right = work.edge.left + deltah;
            RectRgn(workrgn, &work);
            UnionRgn(workrgn, uprgn, uprgn);
        } else if (deltah < 0) {
            work.edge.top = dest.edge.top;
            work.edge.bottom = dest.edge.bottom;
            work.edge.right = source.edge.right;
            work.edge.left = work.edge.right + deltah;
            RectRgn(workrgn, &work);
            UnionRgn(workrgn, uprgn, uprgn);
        }
        if (tempcopy) {         /* Make a copy of the source bitmap */
            BitMap tempbits;    /* on the stack for later copying */
            char *p;            /* to the real destination. */
            long size;

            size = source.edge.bottom - source.edge.top;
            SetRect(&tempbits.bounds, tp->portBits.bounds.edge.left, 0,
                    tp->portBits.bounds.edge.right, size);
            size = MulU(size, tempbits.rowBytes = tp->portBits.rowBytes);
            tempbits.baseAddr = alloca(size);
            tempbits.bpp = tp->portBits.bpp;
            p = (char *) tp->portBits.baseAddr +
                MulU(source.edge.top - tp->portBits.bounds.edge.top,
                     tempbits.rowBytes);
            BlockMove(p, tempbits.baseAddr, size);
            source.edge.bottom -= source.edge.top;
            source.edge.top = 0;
            CopyBits(&tempbits, &tp->portBits, &source, &dest, srcCopy, 0);
        } else {                /* We've gotta go slow. */
            RectRgn(workrgn, &dest); /* Don't bother copying onto any area */
            DiffRgn(workrgn, uprgn, workrgn); /* that will be erased later. */
            CopyBits(&tp->portBits, &tp->portBits, &source, &dest,
                     srcCopy, workrgn); /* Move the rectangle's contents. */
        }
    }

    EraseRgn(uprgn);            /* Erase the no-good area, and */

    if (updateRgn) {            /* perhaps do something with it, */
        if (updateRgn == (RgnHandle) -1) { /* maybe accumulate it into */
            if (!(i = EmptyRgn(((WindowPeek)tp)->updateRgn)))
                CopyRgn(((WindowPeek)tp)->updateRgn, workrgn);
            InvalRgn(uprgn);    /* the update region... */
            if (!i) {           /* (Window already dirty?) */
                RectRgn(uprgn, rect);
                OffsetRgn(workrgn, tp->portBits.bounds.edge.left + deltah,
                          tp->portBits.bounds.edge.top + deltav);
                SectRgn(workrgn, uprgn, workrgn);
                InvalRgn(workrgn); /* (Scroll dirty part too [in effect].) */
            }
        } else
            CopyRgn(uprgn, updateRgn); /* ...or maybe return it instead. */
    }
    DisposeRgn(workrgn);
    DisposeRgn(uprgn);
}
