#include "../clunkdraw.h"
#include "cdep3in7rpi.h"
#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#define TRUE 1
#define FALSE 0

// Some things I snaffled from Waveshare's site, to protect against bit rot:
//
// Waveshare implies that the driver chip is a 2018
// "SSD1677 960 Source x 680 Gate Red/Black/White Active Matrix
//  EPD Display Driver with Controller"
// Its CS and RST pins are active low.  The D/C# pin is high for data,
// low for commands.  Data is clocked in on the rising edge of SCL,
// 8-bit bytes are shifted in MSB first.  (In 3-wire mode D/C# is not
// used, and 9-bit bytes are shifted in MSB first.  The first bit is
// itself now the D/C# bit.)
// There are two 960x680 1-bit RAM banks.  One holds the B&W color,
// the other holds the Red color.  (For 2-color, or 4-gray panels.)
//
//  Table 6-5 : RAM bit and LUT mapping for black/white display
//    +---------+----------+-------+-------------------------+
//    |Data bit | Data bit | Image | LUT                     |
//    |in R RAM | in BW RAM| Color |                         |
//    +---------+----------+-------+-------------------------+
//    |  0      |    0     | Black | LUT 0 for driving Black |
//    |  0      |    1     | White | LUT 1 for driving White |
//    |  1      |    0     | Black | LUT 2 = LUT 0           |
//    |  1      |    1     | White | LUT 3 = LUT 1           |
//    +---------+----------+-------+-------------------------+

// The SSD1677 waveform LUT is shown in Figure 6-6.
//
//      +-------------+------+------+------+------+------+------+
//      |  D7  |  D6  |  D5  |  D4  |  D3  |  D2  |  D1  |  D0  |
//      +------+------+------+------+------+------+------+------+
//  0   |  VS[0A-L0]  |  VS[0B-L0]  |  VS[0C-L0]  |  VS[0D-L0]  |\
//  1   |  VS[1A-L0]  |  VS[1B-L0]  |  VS[1C-L0]  |  VS[1D-L0]  | \
//  2   |  VS[2A-L0]  |  VS[2B-L0]  |  VS[2C-L0]  |  VS[2D-L0]  |  \
//  3   |  VS[3A-L0]  |  VS[3B-L0]  |  VS[3C-L0]  |  VS[3D-L0]  |   \
//  4   |  VS[4A-L0]  |  VS[4B-L0]  |  VS[4C-L0]  |  VS[4D-L0]  |    \ Black
//  5   |  VS[5A-L0]  |  VS[5B-L0]  |  VS[5C-L0]  |  VS[5D-L0]  |    / (L0)
//  6   |  VS[6A-L0]  |  VS[6B-L0]  |  VS[6C-L0]  |  VS[6D-L0]  |   /
//  7   |  VS[7A-L0]  |  VS[7B-L0]  |  VS[7C-L0]  |  VS[7D-L0]  |  /
//  8   |  VS[8A-L0]  |  VS[8B-L0]  |  VS[8C-L0]  |  VS[8D-L0]  | /
//  9   |  VS[9A-L0]  |  VS[9B-L0]  |  VS[9C-L0]  |  VS[9D-L0]  |/
//      +------+------+------+------+------+------+------+------+
// 10   |  VS[0A-L1]  |  VS[0B-L1]  |  VS[0C-L1]  |  VS[0D-L1]  |\
// 11   |  VS[1A-L1]  |  VS[1B-L1]  |  VS[1C-L1]  |  VS[1D-L1]  | \ White
// ...                                                            / (L1)
// 19   |  VS[9A-L1]  |  VS[9B-L1]  |  VS[9C-L1]  |  VS[9D-L1]  |/
//      +------+------+------+------+------+------+------+------+
// 20   |  VS[0A-L2]  |  VS[0B-L2]  |  VS[0C-L2]  |  VS[0D-L2]  |\
// 21   |  VS[1A-L2]  |  VS[1B-L2]  |  VS[1C-L2]  |  VS[1D-L2]  | \ Red
// ...                                                            / (L2)
// 29   |  VS[9A-L2]  |  VS[9B-L2]  |  VS[9C-L2]  |  VS[9D-L2]  |/
//      +------+------+------+------+------+------+------+------+
// ...
// ...
//      +------+------+------+------+------+------+------+------+
// 40   |  VS[0A-L4]  |  VS[0B-L4]  |  VS[0C-L4]  |  VS[0D-L4]  |\
// 41   |  VS[1A-L4]  |  VS[1B-L4]  |  VS[1C-L4]  |  VS[1D-L4]  | \
// ...                                                            / (L4)
// 49   |  VS[9A-L4]  |  VS[9B-L4]  |  VS[9C-L4]  |  VS[9D-L4]  |/
//      +------+------+------+------+------+------+------+------+
// 50   |                        TP[0A]                         |
// 51   |                        TP[0B]                         |
// 52   |                        TP[0C]                         |
// 53   |                        TP[0D]                         |
// 54   |                        RP[ 0]                         |
// 55   |                        TP[1A]                         |
// 56   |                        TP[1B]                         |
// 57   |                        TP[1C]                         |
// 58   |                        TP[1D]                         |
// 59   |                        RP[ 1]                         |
// ...
// ...
// 95   |                        TP[9A]                         |
// 96   |                        TP[9B]                         |
// 97   |                        TP[9C]                         |
// 98   |                        TP[9D]                         |
// 99   |                        RP[ 9]                         |
//      +------+------+------+------+------+------+------+------+
// 100  |                                                       |
// 101  |                                                       |
// 102  |                 5 bytes for frame rate                |
// 103  |                                                       |
// 104  |                                                       |
//      +------+------+------+------+------+------+------+------+
// 105  |                          VGH                          |
// 106  |                         VSH1                          |
// 107  |                         VSH2                          |
// 108  |                          VSL                          |
// 109  |                         VCOM                          |
// 110  |                       Reserve 1                       |
// 111  |                       Reserve 2                       |
//      +------+------+------+------+------+------+------+------+
//                      Figure 6-6 : Waveform LUT

// Selected commands:
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 01  |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  1 |  Driver |   Gate      |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 |  Output |  Setting    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A9 | A8 |  Control| Sequence &  |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 | B2 | B1 | B0 |         |  Direction  |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 03  |  0 |  0 |  0 |  0 |  0 |  0 |  1 |  1 | Gate Drv|  12-20V     |
// | 0   |  1   |     |  0 |  0 |  0 | A4 | A3 | A2 | A1 | A0 | Voltage |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 04  |  0 |  0 |  0 |  0 |  0 |  1 |  0 |  0 | Source  |             |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Driving |             |
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 | Voltage |             |
// | 0   |  1   |     | C7 | C6 | C5 | C4 | C3 | C2 | C1 | C0 | Control |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 0C  |  0 |  0 |  0 |  0 |  1 |  1 |  0 |  0 | Booster |   L1 | L2   |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Inrush  |   AE   AE   |
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 | Control |   C7   C7   |
// | 0   |  1   |     | C7 | C6 | C5 | C4 | C3 | C2 | C1 | C0 |         |   C3   C3   |
// | 0   |  1   |     | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 |         |   C0   C0   |
// | 0   |  1   |     | E7 | E6 | E5 | E4 | E3 | E2 | E1 | E0 |         |   40   80   |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 10  |  0 |  0 |  0 |  1 |  0 |  0 |  0 |  0 | Deep    | 00 - Normal |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A1 | A0 | Sleep   | 11 - Sleep  |
// |                                                          | Mode    | (Busy, HW   |
// |                                                          |         |  reset exit)|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 11  |  0 |  0 |  0 |  1 |  0 |  0 |  0 |  1 | Data    | A2=1 Y++    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 | A2 | A1 | A0 | Entry   | 11 - X++,Y++|
// |                                                          | Sequence| 10 - Y++,X--|
// |                                                          |         | 01 - Y--,X++|
// |                                                          |         | 00 - Y--,X--|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 12  |  0 |  0 |  0 |  1 |  0 |  0 |  1 |  0 | SW      | BUSY during |
// |                                                          | Reset   |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 14  |  0 |  0 |  0 |  1 |  0 |  1 |  0 |  0 | HV Rdy  |             |
// | 0   |  1   |     |  0 | A6 | A5 | A4 |  0 | A2 | A1 | A0 | Detect  |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 15  |  0 |  0 |  0 |  1 |  0 |  1 |  0 |  1 | VCI     |             |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 | A2 | A1 | A0 | Detect  |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 18  |  0 |  0 |  0 |  1 |  0 |  1 |  0 |  0 | Temp    |$48=External |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Sensor  |$80=Internal |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 20  |  0 |  0 |  1 |  0 |  0 |  0 |  0 |  0 | Master  |             |
// |                                                          | Activate|             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 21  |  0 |  0 |  1 |  0 |  0 |  0 |  0 |  1 | Disp Upd|  A7-A4 Red  |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Ctrl 1  |  A3-A0 Blk  |
// |                                                          |         | 0000 - RAM  |
// |                                                          |         | 0100 -  0   |
// |                                                          |         | 1000 - ~RAM |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 22  |  0 |  0 |  1 |  0 |  0 |  0 |  1 |  0 | Disp Upd| A7 - Enb Clk|
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Ctrl 2  | A6 - Enb Anl|
// |                                                          |         | A5 - LD TEMP|
// |                                                          | (Seq.)  | A4 - LD LUT |
// |                                                          |         | A3 - LUT 2  |
// |                                                          |         | A2 - Dis OSC|
// |                                                          |         | A1 - Dis Anl|
// |                                                          |         | A0 - Dis Clk|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 24  |  0 |  0 |  1 |  0 |  0 |  1 |  0 |  0 | Write   | 1 = White   |
// |                                                          | B/W RAM | 0 = Black   |
// |                                                          |         | Auto-incr   |
// |                                                          |         | until Cmd   |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 25  |  0 |  0 |  1 |  0 |  0 |  1 |  0 |  1 | Write   | Auto-incr   |
// |                                                          | RAM Dith| until Cmd   |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 26  |  0 |  0 |  1 |  0 |  0 |  1 |  1 |  0 | Write   | 1 = Red     |
// |                                                          | Red RAM | 0 = B/W     |
// |                                                          |         | Auto-incr   |
// |                                                          |         | until Cmd   |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 27  |  0 |  0 |  1 |  0 |  0 |  1 |  1 |  1 | Read    | 1st is dummy|
// |                                                          | RAM     | Auto-incr   |
// |                                                          |         | until Cmd   |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 28  |  0 |  0 |  1 |  0 |  1 |  0 |  0 |  0 | VCOM    | BUSY        |
// |                                                          | Sense   |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 29  |  0 |  0 |  1 |  0 |  1 |  0 |  0 |  1 | VCOM    |             |
// | 0   |  1   |     | A7 | A6 |  0 |  0 | A3 | A2 | A1 | A0 | Sense   |             |
// |                                                          | Duration|             |
// |                                                          | (Seq.)  |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 2C  |  0 |  0 |  1 |  0 |  1 |  1 |  0 |  0 | Set VCOM|             |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 32  |  0 |  1 |  1 |  0 |  0 |  0 |  1 |  0 | Write   |             |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | LUT     | 105 Bytes   |
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 |  .      |             |
// |        .                                                 |  .      |             |
// |        .                                                 |  .      |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// WS is Waveform Setting
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 37  |  0 |  1 |  1 |  0 |  0 |  1 |  1 |  1 | Display |0=Disp MODE 1|
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Options |1=Disp MODE 2|
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 |         | B= WS[7:0]  |
// | 0   |  1   |     | C7 | C6 | C5 | C4 | C3 | C2 | C1 | C0 |         | C= WS[15:8] |
// | 0   |  1   |     | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 |         | D= WS[23:16]|
// | 0   |  1   |     | E7 | E6 | E5 | E4 | E3 | E2 | E1 | E0 |         | E= WS[31:24]|
// | 0   |  1   |     |  0 | F6 |  0 |  0 | F3 | F2 | F1 | F0 |         | F= WS[35:32]|
// | 0   |  1   |     | G7 | G6 | G5 | G4 | G3 | G2 | G1 | G0 |         |F6=1 PngPng 2|
// | 0   |  1   |     | H7 | H6 | H5 | H4 | H3 | H2 | H1 | H0 |         |             |
// | 0   |  1   |     | I7 | I6 | I5 | I4 | I3 | I2 | I1 | I0 |         |G-J Module ID|
// | 0   |  1   |     | J7 | J6 | J5 | J4 | J3 | J2 | J1 | J0 |         |Waveform Vrsn|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 3C  |  0 |  1 |  1 |  0 |  1 |  1 |  0 |  0 | Border  |             |
// | 0   |  1   |     | A7 | A6 | A5 | A4 |  0 |  0 | A1 | A0 | Control |             |
// |                                                          |         |             |
// |                                                          |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 44  |  0 |  1 |  0 |  0 |  0 |  1 |  0 |  0 | RAM X Ad| A: X start  |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Strt/Stp|  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A9 | A8 |         | B: X end    |
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 | Strt/Stp|  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | B9 | B8 |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 45  |  0 |  1 |  0 |  0 |  0 |  1 |  0 |  1 | RAM Y Ad| A: Y start  |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Strt/Stp|  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A9 | A8 |         | B: Y end    |
// | 0   |  1   |     | B7 | B6 | B5 | B4 | B3 | B2 | B1 | B0 | Strt/Stp|  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | B9 | B8 |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 46  |  0 |  1 |  0 |  0 |  0 |  1 |  1 |  0 |AutoWrite| A7 PreInc Y |
// | 0   |  1   |     | A7 | A6 | A5 | A4 |  0 | A2 | A1 | A0 | RED RAM | A6:4 Y+8,16,|
// |                                                          |         |  32,64,128, |
// |                                                          |         |  256,512,680|
// |                                                          |         | A2:0 X+8,16,|
// |                                                          |         |  32,64,128, |
// |                                                          |         |  256,512,960|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 47  |  0 |  1 |  0 |  0 |  0 |  1 |  1 |  1 |AutoWrite| A7 PreInc Y |
// | 0   |  1   |     | A7 | A6 | A5 | A4 |  0 | A2 | A1 | A0 | B/W RAM | A6:4 Y+8,16,|
// |                                                          |         |  32,64,128, |
// |                                                          |         |  256,512,680|
// |                                                          |         | A2:0 X+8,16,|
// |                                                          |         |  32,64,128, |
// |                                                          |         |  256,512,960|
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 4E  |  0 |  1 |  0 |  0 |  1 |  1 |  1 |  0 | RAM X   | A: X start  |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Address |  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A9 | A8 |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 4F  |  0 |  1 |  0 |  0 |  1 |  1 |  1 |  0 | RAM Y   | A: Y start  |
// | 0   |  1   |     | A7 | A6 | A5 | A4 | A3 | A2 | A1 | A0 | Address |  address    |
// | 0   |  1   |     |  0 |  0 |  0 |  0 |  0 |  0 | A9 | A8 |         |             |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// |R/W# | D/C# | Hex | D7 | D6 | D5 | D4 | D3 | D2 | D1 | D0 | Command | Description |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+
// | 0   |  0   | 7F  |  0 |  1 |  1 |  1 |  1 |  1 |  1 |  1 | NOP     |Term RAM R/W |
// +-----+------+-----+----+----+----+----+----+----+----+----+---------+-------------+

//===========================================================================
//
// Beginning of 3.7" Waveshare E-Paper Display (EPD) support code.
// Taken from their C demo library.  For Clunkdraw on Raspberry Pi.
//
//===========================================================================

static const char * const cmd[] = {"sudo tee ",
                                   "/sys/class/gpio",
                                   " >/dev/null 2>&1"};

// Returns TRUE if it worked.
static Boolean
gpioPinInOut(int pin, Boolean output)
{
   char buf[128];
   int ii, fd;

   // See if it's already set up.
   snprintf(buf, sizeof buf, "%s/gpio%d/direction", cmd[1], pin);
   if ((fd = open(buf, O_RDONLY)) >= 0) {
      if ((ii = read(fd, buf, sizeof buf)) > 0) {
         buf[--ii] = 0; // Eat newline.
         if (!strncmp(buf, output ? "out" : "in", ii)) {
            close(fd);
            return TRUE;
         }
      }
      close(fd);
   }

   // Needs configuration.  First try to do it ourself.
   snprintf(buf, sizeof buf, "%s/export", cmd[1]);
   if ((fd = open(buf, O_WRONLY)) >= 0) {
      ii = snprintf(buf, sizeof buf, "%d", pin);
      (void)write(fd, buf, ii); // Can fail, if already exported.
      close(fd);
      snprintf(buf, sizeof buf, "%s/gpio%d/direction", cmd[1], pin);
      if ((fd = open(buf, O_WRONLY)) >= 0) {
         (void)write(fd, output ? "out" : "in", 2 + output);
         close(fd);
         return TRUE;
      }
   }
   return FALSE;
}

// Program a GPIO pin to be an output.  We need elevated privileges to do this.

void
gpioPinOut(int pin)
{
   char buf[128];

   // First see if it's already done, and if not try to do it ourself.
   if (gpioPinInOut(pin, TRUE)) return;
   // Didn't work, try to gain elevated privilege to do it.
   snprintf(buf, sizeof buf, "echo %d | %s%s/export %s",
                             pin, cmd[0], cmd[1], cmd[2]);
   (void)system(buf);
   snprintf(buf, sizeof buf, "echo out | %s%s/gpio%d/direction %s",
                             cmd[0], cmd[1], pin, cmd[2]);
   (void)system(buf);
}

// Program a GPIO pin to be an input.  We need elevated privileges to do this.

void
gpioPinIn(int pin)
{
   char buf[128];

   // First see if it's already done, and if not try to do it ourself.
   if (gpioPinInOut(pin, FALSE)) return;
   // Didn't work, try to gain elevated privilege to do it.
   snprintf(buf, sizeof buf, "echo %d | %s%s/export %s",
                             pin, cmd[0], cmd[1], cmd[2]);
   (void)system(buf);
   snprintf(buf, sizeof buf, "echo in | %s%s/gpio%d/direction %s",
                             cmd[0], cmd[1], pin, cmd[2]);
   (void)system(buf);
}

// Write value to selected GPIO control file.  This way we can wiggle
// I/O pins directly without using additional drivers (like I2C or SSI).
// For our needs nothing faster is necessary, and this is simpler.

void
gpioW(char *fName, char *value)
{
   int fd;
   char buf[128];

   snprintf(buf, sizeof buf, "%s/%s", cmd[1], fName);
   if ((fd = open(buf, O_WRONLY)) >= 0) {
      (void)write(fd, value, strlen(value));
      close(fd);
   }
}

// Read GPIO pin's value.

Boolean
gpioR(char *fName)
{
   int fd, ii;
   char buf[128];

   ii = 0;
   snprintf(buf, sizeof buf, "%s/%s", cmd[1], fName);
   if ((fd = open(buf, O_RDONLY)) >= 0) {
      if (read(fd, buf, sizeof buf) > 0)
         ii = buf[0] == '1';
      close(fd);
   }
   return ii;
}

// Turn the selected pin on or off.

void
setPin(int pin, Boolean lit)
{
   char buf[128];

   if (pin <= 0) return;
   snprintf(buf, sizeof buf, "gpio%d/value", pin);
   if (lit)
      gpioW(buf, "1");
   else
      gpioW(buf, "0");
}

// Return the selected LED on/off value.  Used (as IPC) in rare conditions
// where there might be multiple processes wiggling pins.

Boolean
getPin(int pin)
{
   char buf[128];

   if (pin <= 0) return FALSE;
   snprintf(buf, sizeof buf, "gpio%d/value", pin);
   return gpioR(buf);
}

void
EPD_Delay_ms(int xms)
{
   struct timespec Req, Rem;

   Req.tv_sec = xms / 1000;
   Req.tv_nsec = (xms % 1000) * 1000000L;
   while (nanosleep(&Req, &Rem) < 0 && errno == EINTR)
      Req = Rem;
}

void EPD_GPIO_Init()
{
   bits = 8;
   EPD_RST_PIN     = 17;
   EPD_DC_PIN      = 25;
   EPD_CS_PIN      = 8; // If under control of spidev0.0, we can't touch it.
   EPD_BUSY_PIN    = 24;
   gpioPinOut(EPD_RST_PIN); // Put the panel into hard reset, powered down.
   setPin(EPD_RST_PIN, lastCsPinState = 0);
   EPD_Delay_ms(200);
   gpioPinOut(EPD_DC_PIN);
   if (spiMode & SPI_NO_CS)
      gpioPinOut(EPD_CS_PIN);
   gpioPinIn(EPD_BUSY_PIN);
   if (spiMode & SPI_NO_CS)
      setPin(EPD_CS_PIN, lastCsPinState = 1);
}

int
EPD_SPI_setSpeed(long speed)
{
   int ret;

   ret = ioctl(spiFd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
   if (ret < 0)
      return ret;
   else
      tr.speed_hz = speed;
   return 1;
}

int
EPD_SPI_Mode(int mode)
{
   int ret;

   spiMode &= ~3; // Clear low 2 bits.
   spiMode |= mode;
   ret = ioctl(spiFd, SPI_IOC_WR_MAX_SPEED_HZ, &spiMode);
   if (ret < 0)
      return ret;
   return 1;
}

int
EPD_SPI_ChipSelect(int CS_Mode)
{
   int ret;

   if (CS_Mode == SPI_CS_Mode_HIGH) {
      spiMode |= SPI_CS_HIGH;
      spiMode &= ~SPI_NO_CS;
   } else if (CS_Mode == SPI_CS_Mode_LOW)
      spiMode &= ~(SPI_CS_HIGH|SPI_NO_CS);
   else if (CS_Mode == SPI_CS_Mode_NONE)
      spiMode |= SPI_NO_CS;
   ret = ioctl(spiFd, SPI_IOC_WR_MODE, &spiMode);
   if (ret < 0)
      return ret;
   return 1;
}

int
EPD_SPI_SetBitOrder(int Order)
{
   int ret;

   if (Order == SPI_BIT_ORDER_LSBFIRST)
      spiMode |= SPI_LSB_FIRST;
   else if (Order == SPI_BIT_ORDER_MSBFIRST)
      spiMode &= ~SPI_LSB_FIRST;
   else
      return 1;
   ret = ioctl(spiFd, SPI_IOC_WR_MODE, &spiMode);
   if (ret < 0)
      return ret;
   return 1;
}

void
EPD_SPI_SetDataInterval(long us)
{
   tr.delay_usecs = us;
}

void
EPD_SPI_begin(char *SPI_device)
{
   int ret;

   ret = 0;
   spiFd = open(SPI_device, O_RDWR);
   if (spiFd < 0) {
      printf("Failed to open SPI device %s\n", SPI_device);
      exit(1);
   }
   spiMode = 0;

   ret = ioctl(spiFd, SPI_IOC_WR_BITS_PER_WORD, &bits);
   if (ret < 0)
      printf("SPI Can't set %d bits per word\n", bits);

   tr.bits_per_word = bits;
   tr.cs_change = 1; // Leave CS selected (low) most of the time.

   EPD_SPI_Mode(SPI_MODE_0);
   EPD_SPI_ChipSelect(SPI_CS_Mode_LOW);
#if 0
   // Rumor has it that RPi can only do LSBFIRST on SPI1, not SPI0.
   EPD_SPI_SetBitOrder(SPI_BIT_ORDER_LSBFIRST);
#endif
   EPD_SPI_setSpeed(20000000);
   EPD_SPI_SetDataInterval(0);
}

void
EPD_SPI_end()
{
   int ret;

   spiMode = 0;
   ret = close(spiFd);
   spiFd = -1;
   if (ret)
      puts("Failed to close SPI device");
}

U8
EPD_SPI_TransferByte(U8 *buf)
{
   U8 rbuf;
   int rv;

   tr.len = 1;
   tr.tx_buf = (U64)(U32)buf;   // Double-cast eliminates compiler warning.
   tr.rx_buf = (U64)(U32)&rbuf;
   rv = ioctl(spiFd, SPI_IOC_MESSAGE(1), &tr);
   return rbuf;
}

void
EPD_SPI_WriteByte(U8 value)
{
   EPD_SPI_TransferByte(&value);
}

void
EPD_SPI_Init()
{
   EPD_GPIO_Init();
   EPD_SPI_begin("/dev/spidev0.0");
   EPD_SPI_setSpeed(10000000);
}

// Module exits, closes SPI
void
EPD_SPI_Exit()
{
   if (spiMode & SPI_NO_CS)
      setPin(EPD_CS_PIN, lastCsPinState = 0);
   setPin(EPD_DC_PIN, lastDcPinState = 0);
   setPin(EPD_RST_PIN, lastRstPinState = 0); // Kill power to the board.
}

#define LUTSIZE 105 // Zero-based, 105 total cells.

static const U8 lut_4Gray_GC[LUTSIZE] = {
   0x2A,0x06,0x15,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
   0x28,0x06,0x14,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
   0x20,0x06,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
   0x14,0x06,0x28,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 40
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 50
   0x00,0x02,0x02,0x0A,0x00,0x00,0x00,0x08,0x08,0x02, // 60
   0x00,0x02,0x02,0x0A,0x00,0x00,0x00,0x00,0x00,0x00, // 70
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 90
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, //100
   0x22,0x22,0x22,0x22,0x22
};

#if 0
static const U8 lut_1Gray_GC[LUTSIZE] = {
   0x2A,0x05,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
   0x05,0x2A,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
   0x2A,0x15,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
   0x05,0x0A,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 40
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 50
   0x00,0x02,0x03,0x0A,0x00,0x02,0x06,0x0A,0x05,0x00, // 60
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 70
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 90
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, //100
   0x22,0x22,0x22,0x22,0x22
};
#endif

static const U8 lut_1Gray_DU[LUTSIZE] = {
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
   0x01,0x2A,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
   0x0A,0x55,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 40
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 50
   0x00,0x00,0x05,0x05,0x00,0x05,0x03,0x05,0x05,0x00, // 60
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 70
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 90
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, //100
   0x22,0x22,0x22,0x22,0x22
};

#if 0
static const U8 lut_1Gray_A2[LUTSIZE] = {
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 10
   0x0A,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 20
   0x05,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 30
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 40
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 50
   0x00,0x00,0x03,0x05,0x00,0x00,0x00,0x00,0x00,0x00, // 60
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 70
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 80
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, // 90
   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, //100
   0x22,0x22,0x22,0x22,0x22
};
#endif


//******************************************************************************
//function :      Hardware reset
//******************************************************************************
static void
EPD_3in7_Reset()
{
   if (lastRstPinState != 1)
      setPin(EPD_RST_PIN, lastRstPinState = 1);
   EPD_Delay_ms(30);
   setPin(EPD_RST_PIN, lastRstPinState = 0);
   EPD_Delay_ms(3);
   setPin(EPD_RST_PIN, lastRstPinState = 1);
   EPD_Delay_ms(30);
}

//******************************************************************************
//function :      send command
//******************************************************************************
static void
EPD_3in7_SendCommand(U8 Reg)
{
   if (lastDcPinState != 0)
      setPin(EPD_DC_PIN, lastDcPinState = 0);
   if (spiMode & SPI_NO_CS)
      if (lastCsPinState != 0)
         setPin(EPD_CS_PIN, lastCsPinState = 0);
   EPD_SPI_WriteByte(Reg);
   if (spiMode & SPI_NO_CS)
      if (lastCsPinState != 1)
         setPin(EPD_CS_PIN, lastCsPinState = 1);
}

//******************************************************************************
//function :      send data
//******************************************************************************
static void
EPD_3in7_SendData(U8 Data)
{
   if (lastDcPinState != 1)
      setPin(EPD_DC_PIN, lastDcPinState = 1);
   if (spiMode & SPI_NO_CS)
      if (lastCsPinState != 0)
         setPin(EPD_CS_PIN, lastCsPinState = 0);
   EPD_SPI_WriteByte(Data);
   if (spiMode & SPI_NO_CS)
      if (lastCsPinState != 1)
         setPin(EPD_CS_PIN, lastCsPinState = 1);
}

static void
EPD_3in7_WaitWhileBusy() // Wait for BUSY to be low (not busy).
{
   Boolean busy;

   do {
      busy = getPin(EPD_BUSY_PIN);
   } while (busy);
   EPD_Delay_ms(200);
}

//******************************************************************************
//function :      set the look-up tables
//******************************************************************************
static void
EPD_3in7_Load_LUT(U8 lut)
{
   int ii;

   EPD_3in7_SendCommand(0x32);  // Write LUT
   for (ii = 0; ii < LUTSIZE; ii++) {
      switch (lut) {
        case 0: EPD_3in7_SendData(lut_4Gray_GC[ii]); break;
//        case 1: EPD_3in7_SendData(lut_1Gray_GC[ii]); break;
        case 2: EPD_3in7_SendData(lut_1Gray_DU[ii]); break;
//        case 3: EPD_3in7_SendData(lut_1Gray_A2[ii]); break;
        default: puts("There is no such LUT"); break;
      }
   }
}

//******************************************************************************
//function :      Initialize the e-Paper register
//******************************************************************************

void
EPD_3in7_4Gray_Init()
{
   EPD_3in7_Reset();
   EPD_3in7_WaitWhileBusy();    // Wait for BUSY to be low (not busy).

   EPD_3in7_SendCommand(0x12);
   EPD_Delay_ms(300);

   EPD_3in7_SendCommand(0x46);
   EPD_3in7_SendData(0xF7);
   EPD_3in7_WaitWhileBusy;
   EPD_3in7_SendCommand(0x47);
   EPD_3in7_SendData(0xF7);
   EPD_3in7_WaitWhileBusy;

   EPD_3in7_SendCommand(0x01);  // setting gate number
   EPD_3in7_SendData(0xDF);
   EPD_3in7_SendData(0x01);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x03);  // set gate voltage
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x04);  // set source voltage
   EPD_3in7_SendData(0x41);
   EPD_3in7_SendData(0xA8);
   EPD_3in7_SendData(0x32);

   EPD_3in7_SendCommand(0x11);  // set data entry sequence
   EPD_3in7_SendData(0x03);

   EPD_3in7_SendCommand(0x3C);  // set border
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x0C);  // set booster strength
   EPD_3in7_SendData(0xAE);
   EPD_3in7_SendData(0xC7);
   EPD_3in7_SendData(0xC3);
   EPD_3in7_SendData(0xC0);
   EPD_3in7_SendData(0xC0);

   EPD_3in7_SendCommand(0x18);  // set internal sensor on
   EPD_3in7_SendData(0x80);

   EPD_3in7_SendCommand(0x2C);  // set vcom value
   EPD_3in7_SendData(0x44);

   EPD_3in7_SendCommand(0x37);  // set display option, these setting turn on previous function
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x44);  // setting X direction start/end position of RAM
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x17);
   EPD_3in7_SendData(0x01);

   EPD_3in7_SendCommand(0x45);  // setting Y direction start/end position of RAM
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0xDF);
   EPD_3in7_SendData(0x01);

   EPD_3in7_SendCommand(0x22);  // Display Update Control 2
   EPD_3in7_SendData(0xCF);
   EPD_3in7_WaitWhileBusy;
}

//******************************************************************************
//function :  Initialize the e-Paper register
//******************************************************************************
void
EPD_3in7_1Gray_Init()
{
   EPD_3in7_Reset();            // HW Reset
   EPD_3in7_WaitWhileBusy();    // Wait for BUSY to be low (not busy).

   EPD_3in7_SendCommand(0x12);  // SW Reset
   EPD_Delay_ms(300);           // (Should watch BUSY instead.)

   EPD_3in7_SendCommand(0x46);  // Auto-write Red RAM,
   EPD_3in7_SendData(0xF7);     // ++Y+680, X+960
   EPD_3in7_WaitWhileBusy();
   EPD_3in7_SendCommand(0x47);  // Auto-write B&W RAM,
   EPD_3in7_SendData(0xF7);     // ++Y+680, X+960
   EPD_3in7_WaitWhileBusy();

   EPD_3in7_SendCommand(0x01);  // setting gate number
   EPD_3in7_SendData(0xDF);     // A9:0 = 0x1DF + 1 = 480 MUX
   EPD_3in7_SendData(0x01);
   EPD_3in7_SendData(0x00);     // B2:0 = normal scanning order

   EPD_3in7_SendCommand(0x03);  // set gate voltage
   EPD_3in7_SendData(0x00);     // VGH = 20V

   EPD_3in7_SendCommand(0x04);  // set source voltage
   EPD_3in7_SendData(0x41);     // VSH1 = 15V
   EPD_3in7_SendData(0xA8);     // VSH2 = 5V
   EPD_3in7_SendData(0x32);     // VSL = -15V

   EPD_3in7_SendCommand(0x11);  // set data entry sequence
   EPD_3in7_SendData(0x03);     // X axis, X++, Y++

   EPD_3in7_SendCommand(0x3C);  // set border
   EPD_3in7_SendData(0x00);     // VBD=GS trans from LUT0

   EPD_3in7_SendCommand(0x0C);  // set booster strength
   EPD_3in7_SendData(0xAE);
   EPD_3in7_SendData(0xC7);
   EPD_3in7_SendData(0xC3);
   EPD_3in7_SendData(0xC0);
   EPD_3in7_SendData(0xC0);     // Not a valid value???

   EPD_3in7_SendCommand(0x18);  // use internal Temperature sensor
   EPD_3in7_SendData(0x80);

   EPD_3in7_SendCommand(0x2C);  // set VCOM = 1.7V
   EPD_3in7_SendData(0x44);

   EPD_3in7_SendCommand(0x37);  // set display option, these setting turn on previous func
   EPD_3in7_SendData(0x00);     //can switch 1 gray or 4 gray
#if 1
   EPD_3in7_SendData(0xFF);     // Display Mode 2 for WS[35:0]
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0x4F);     // RAM Ping Pong for Display Mode 2
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0xFF);
   EPD_3in7_SendData(0xFF);
#else
   EPD_3in7_SendData(0x00);     // Display Mode 1 for WS[35:0]
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);
#endif

   EPD_3in7_SendCommand(0x44);  // set X axis start/end position of RAM
   EPD_3in7_SendData(0x00);     // Starting X address 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x17);     // Ending X address 279
   EPD_3in7_SendData(0x01);

   EPD_3in7_SendCommand(0x45);  // set Y axis start/end position of RAM
   EPD_3in7_SendData(0x00);     // Starting Y address 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0xDF);     // Ending Y address 479
   EPD_3in7_SendData(0x01);

   EPD_3in7_SendCommand(0x22);  // Display Update Control 2 (sequence):
#if 1
   EPD_3in7_SendData(0xCF);     //   Enable clock signal,
                                //   Enable Analog,
                                //   Display Mode 2,
                                //   Disable Analog,
                                //   Disable clock
#else
   EPD_3in7_SendData(0xC7);     //   Enable clock signal,
                                //   Enable Analog,
                                //   Display Mode 1,
                                //   Disable Analog,
                                //   Disable clock
#endif
}

//******************************************************************************
//function :      Clear screen
//******************************************************************************

void
EPD_3in7_4Gray_Clear()
{
   int Width, Height, ii, jj;

   Width = (EPD_3in7_WIDTH + 7) / 8;
   Height = EPD_3in7_HEIGHT;

   EPD_3in7_SendCommand(0x49); // Undocumented.
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x4E);  // Set X address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendCommand(0x4F);  // Set Y address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x24);  // Begin writing to B&W RAM, auto-incrementing.
   for (jj = 0; jj < Height; jj++)
      for (ii = 0; ii < Width; ii++)
         EPD_3in7_SendData(0xFF);

   EPD_3in7_SendCommand(0x4E);  // Set X address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendCommand(0x4F);  // Set Y address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x26);  // Begin writing to Red RAM, auto-incrementing.
   for (jj = 0; jj < Height; jj++)
      for (ii = 0; ii < Width; ii++)
         EPD_3in7_SendData(0xFF);

   EPD_3in7_Load_LUT(0);
   EPD_3in7_SendCommand(0x22);  // Display Update Control 2 (sequence):
   EPD_3in7_SendData(0xC7);     //   Enable clock signal,
                                //   Enable Analog,
                                //   Display Mode (LUT?) 1,
                                //   Disable Analog,
                                //   Disable clock

   EPD_3in7_SendCommand(0x20);  // Send everything to the panel.
   EPD_3in7_WaitWhileBusy;
}

//******************************************************************************
//function :  Clear screen
//******************************************************************************
void
EPD_3in7_1Gray_Clear()
{
   int ii, jj;

   jj = (EPD_3in7_WIDTH * EPD_3in7_HEIGHT) / 8;

   EPD_3in7_SendCommand(0x4E);  // Set X address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendCommand(0x4F);  // Set Y address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x24);  // Begin writing to RAM, auto-incrementing.
   for (ii = 0; ii < jj; ii++)  // Send all 16,800 display bytes.
      EPD_3in7_SendData(0xFF);  // 1's are white to the Hardware.

   EPD_3in7_Load_LUT(2);

   EPD_3in7_SendCommand(0x20);  // Send everything to the panel.
   EPD_3in7_WaitWhileBusy();
}

//******************************************************************************
//function :      Sends the image buffer in RAM to e-Paper and displays
//******************************************************************************
/*
** Register | White | LtGray | DkGray | Black
**   0x24   |   1   |   0    |   1    |   0
**   0x26   |   1   |   1    |   0    |   0
*/
#if 0
void
EPD_3in7_4Gray_Display(const U8 *Image)
{
   int ii, jj, kk;
   U8 temp1, temp2, temp3, temp4;

   EPD_3in7_SendCommand(0x49);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x4E);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x4F);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x24);  // Begin writing to RAM, auto-incrementing.
   for (ii = 0; ii < 16800; ii++) { // 480 x 280 / 8
      temp3 = 0;
      for (jj = 0; jj < 2; jj++) {
         temp1 = Image[ii*2 + jj];
         for (kk = 0; kk < 2; kk++) {
            temp2 = temp1 & 0xC0;
            switch (temp2) {
              case 0xC0: temp4 = 0x01; break; //white
              case 0x00: temp4 = 0x00; break; //black
              case 0x80: temp4 = 0x00; break; //gray1
              case 0x40: temp4 = 0x01; break; //gray2
            }
            temp3 = (temp3 | temp4) << 1;

            temp1 <<= 2;
            temp2 = temp1 & 0xC0;
            switch (temp2) {
              case 0xC0: temp4 = 0x01; break; //white
              case 0x00: temp4 = 0x00; break; //black
              case 0x80: temp4 = 0x00; break; //gray1
              case 0x40: temp4 = 0x01; break; //gray2
            }
            temp3 = temp3 | temp4;
            if (jj != 1 || kk != 1)
               temp3 <<= 1;

            temp1 <<= 2;
         }
      }
      EPD_3in7_SendData(temp3);
   }

   // new data
   EPD_3in7_SendCommand(0x4E);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x4F);
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x26);
   for (ii = 0; ii < 16800; ii++) { // 480 x 280 / 8
      temp3 = 0;
      for (jj = 0; jj < 2; jj++) {
         temp1 = Image[ii*2 + jj];
         for (kk = 0; kk < 2; kk++) {
            temp2 = temp1 & 0xC0;
            switch (temp2) {
              case 0xC0: temp4 = 0x01; break; //white
              case 0x00: temp4 = 0x00; break; //black
              case 0x80: temp4 = 0x01; break; //gray1
              case 0x40: temp4 = 0x00; break; //gray2
            }
            temp3 = (temp3 | temp4) << 1;

            temp1 <<= 2;
            temp2 = temp1 & 0xC0;
            switch (temp2) {
              case 0xC0: temp4 = 0x01; break; //white
              case 0x00: temp4 = 0x00; break; //black
              case 0x80: temp4 = 0x01; break; //gray1
              case 0x40: temp4 = 0x00; break; //gray2
            }
            temp3 = temp3 | temp4;
            if (jj != 1 || kk != 1)
               temp3 <<= 1;

            temp1 <<= 2;
         }
      }
      EPD_3in7_SendData(temp3);
   }

   EPD_3in7_Load_LUT(0);

   EPD_3in7_SendCommand(0x22);
   EPD_3in7_SendData(0xC7);

   EPD_3in7_SendCommand(0x20);

   EPD_3in7_WaitWhileBusy();
}
#endif

//******************************************************************************
//function :  Sends the image buffer in RAM to e-Paper and displays
//******************************************************************************
void
EPD_3in7_1Gray_Display(const U8 *Image)
{
   int ii, jj;

   jj = ((EPD_3in7_WIDTH + 7) / 8) * EPD_3in7_HEIGHT;

   EPD_3in7_SendCommand(0x4E);  // Set X address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);
   EPD_3in7_SendCommand(0x4F);  // Set Y address
   EPD_3in7_SendData(0x00);     // to 0
   EPD_3in7_SendData(0x00);

   EPD_3in7_SendCommand(0x24);  // Begin writing to RAM, auto-incrementing.
   for (ii = 0; ii < jj; ii++)  // Send all 16,800 display bytes.
      EPD_3in7_SendData(~Image[ii]); // 1's are white to the HW, so Invert them.

   EPD_3in7_Load_LUT(2);
   EPD_3in7_SendCommand(0x20);  // Send everything to the panel.
   EPD_3in7_WaitWhileBusy();
}

//******************************************************************************
//function :  Sends part of the image buffer in RAM to e-Paper and displays
//notes:
//    You can send a part of data to e-Paper,But this function is not recommended
//    1.Xsize must be as big as EPD_3in7_WIDTH
//    2.Ypointer must be start at 0
//******************************************************************************
void
EPD_3in7_1Gray_Display_Part(const U8 *Image, int Xstart, int Ystart, int Xend, int Yend)
{
   int ii, jj;

   jj = (((Xend - Xstart) + 7) / 8) * (Yend - Ystart);

   EPD_3in7_SendCommand(0x44);
   EPD_3in7_SendData(Xstart & 0xFF);
   EPD_3in7_SendData((Xstart>>8) & 0x03);
   EPD_3in7_SendData(Xend & 0xFF);
   EPD_3in7_SendData((Xend>>8) & 0x03);
   EPD_3in7_SendCommand(0x45);
   EPD_3in7_SendData(Ystart & 0xFF);
   EPD_3in7_SendData((Ystart>>8) & 0x03);
   EPD_3in7_SendData(Yend & 0xFF);
   EPD_3in7_SendData((Yend>>8) & 0x03);

   EPD_3in7_SendCommand(0x24);
   for (ii = 0; ii < jj; ii++)
      EPD_3in7_SendData(~Image[ii]); // 1's are white to the HW, so Invert them.

   EPD_3in7_Load_LUT(2);
   EPD_3in7_SendCommand(0x20);
   EPD_3in7_WaitWhileBusy();
}

//******************************************************************************
//function :      Enter sleep mode
//******************************************************************************
void
EPD_3in7_Sleep()
{
   EPD_3in7_SendCommand(0x50);  // ??? Undocumented!
   EPD_3in7_SendData(0xF7);
   EPD_3in7_SendCommand(0x02);  // power off???  Undocumented!
   EPD_3in7_SendCommand(0x07);  // deep sleep???  Undocumented!
   EPD_3in7_SendData(0xA5);
}

void
EPD_3in7_DeepSleep()
{
   EPD_3in7_SendCommand(0x10);  // Deep sleep (Documented)
   EPD_3in7_SendData(0x03);
}
