#include "../clunkdraw.h"
#include "../rect/cdrect.h"
#include "../rgn/cdrgn.h"
#include "../poly/cdpoly.h"
#include "../port/cdport.h"
#include "cdline.h"
#include <alloca.h>
#ifdef DEBUG
#include <stdio.h>
#endif
#define FALSE 0
#define TRUE 1

static void
VLine(height)
int height;                     /* Height of the line.*/
{
    Rect scratch;               /* Convert to PaintRect call. */
    register GrafPort *port;

    port = thePort;
    if (height >= 0) {
        scratch.edge.top = port->pnLoc.v;
        scratch.edge.bottom = scratch.edge.top + height;
    } else {
        scratch.edge.bottom = port->pnLoc.v;
        scratch.edge.top = scratch.edge.bottom + height;
    }
    scratch.edge.left = port->pnLoc.h;
    scratch.edge.bottom = scratch.edge.bottom + port->pnSize.v;
    scratch.edge.right = scratch.edge.left + port->pnSize.h;
    if (port->pnMode == srcCopy)
        FillRect(&scratch, &port->pnPat); /* Faster than PaintRect */
    else                        /* and they are the same operation when */
        PaintRect(&scratch);    /* using srcCopy mode. */
}

static void
HLine(width)
int width;                      /* Width of the line.*/
{
    Rect scratch;               /* Convert to PaintRect call. */
    register GrafPort *port;

    port = thePort;
    if (width >= 0) {
        scratch.edge.left = port->pnLoc.h;
        scratch.edge.right = scratch.edge.left + width;
    } else {
        scratch.edge.right = port->pnLoc.h;
        scratch.edge.left = scratch.edge.right + width;
    }
    scratch.edge.right = scratch.edge.right + port->pnSize.h;
    scratch.edge.top = port->pnLoc.v;
    scratch.edge.bottom = scratch.edge.top + port->pnSize.v;
    if (port->pnMode == srcCopy)
        FillRect(&scratch, &port->pnPat); /* This's faster than PaintRect */
    else                        /* and they are the same operation when */
        PaintRect(&scratch);    /* using srcCopy mode. */
}

//
// Bresenham's algorithm for the optimal cases.
//
// Line argument is 2 independent points, not a rect.
//
// If the pen is bigger than 1x1 we might touch pixels more than once,
// dragging our fat pen around, so in that case we can't use this
// algorithm for one of the XOR modes.  We also can only draw the line
// if it's entirely unclipped, especially if it might be clipped by a
// complex region.  (Our region traversal code only works top-to-bottom,
// left-to-right, whereas we can draw our fast lines in any direction.)
//
// Returns TRUE if we handled the line, FALSE if we did not.
//
static Boolean
_FastLine(const Rect *br, Rect *line)
{
   GrafPtr tp;
   Rect pbr, cbr, vbr, br2;
   int x, y, ex, ey, dx, dy, sx, sy, err, e2, ii, jj;

   tp = thePort;
   if ((tp->pnMode == srcXor || tp->pnMode == notSrcXor) &&
       (tp->pnSize.h != 1 || tp->pnSize.v != 1)) return FALSE;
   pbr = tp->portBits.bounds;   // We need the logical portBits.bounds for
   if (tp->device == 1)         // clipping, and moving to global coordinates.
      RotRect(&pbr, &pbr);
   if (!SectRect(br, &pbr, &br2)) return TRUE; // Nothing left, done.
   if (!EqualRect(br, &br2)) return FALSE;     // If clipped, we can't do it.
   if (tp->clipRgn && *tp->clipRgn) {
      cbr = (*tp->clipRgn)->rgnBBox;           // Has Clip region.
      if (!SectRect(&br2, &cbr, &br2)) return TRUE; // Nothing left, done.
      if (!EqualRect(br, &br2)) return FALSE;  // Clipped, can't do it.  Or if
      if ((*tp->clipRgn)->rgnSize != sizeof(RegionHdr)) return FALSE; // complex.
   }
   if (tp->visRgn && *tp->visRgn) {
      vbr = (*tp->visRgn)->rgnBBox;            // Has Visible region.
      if (!SectRect(&br2, &vbr, &br2)) return TRUE; // Nothing left, done.
      if (!EqualRect(br, &br2)) return FALSE;  // Clipped, can't do it.  Or if
      if ((*tp->visRgn)->rgnSize != sizeof(RegionHdr)) return FALSE; // complex.
   }
   if (_ShieldCursHook)
      (*_ShieldCursHook)(br, &pbr.corner.topLeft);
   OffsetRect(line, -pbr.edge.left, -pbr.edge.top); // Move to global coords.
   x = line->edge.left;
   y = line->edge.top;
   ex = line->edge.right;
   ey = line->edge.bottom;

   // We survived the gauntlet, draw the (unclipped) line.
   if (x < ex) { // dx := abs(x1-x0);
      sx = 1;
      dx = ex - x;
   } else {
      sx = -1;
      dx = x - ex;
   }
   if (y < ey) { // dy := -abs(y1-y0);
      sy = 1;
      dy = y - ey;
   } else {
      sy = -1;
      dy = ey - y;
   }
   err = dx + dy;

   while (TRUE) {
      for (ii = 0; ii < tp->pnSize.v; ii++)    // Splat down the pen.
         for (jj = 0; jj < tp->pnSize.h; jj++) // Might be wide/tall.
            _PaintRBit(tp, x+jj, y+ii);
      if (x == ex && y == ey) break;
      e2 = 2 * err;
      if (e2 >= dy) {           // e_xy+e_x > 0
         err += dy;
         x += sx;
      }
      if (e2 <= dx) {           // e_xy+e_y < 0
         err += dx;
         y += sy;
      }
   }
   if (_ShowCursHook)
      (*_ShowCursHook)();
   return TRUE;                 // We handled the line.
}

void
LineTo(newx, newy)
int newx, newy;                 /* Ending coordinates for the line. */
{
   int ii;
   short *p1;                   /* For creating line points in. */
   short *p2;                   /* For creating dummy region in. */
   short negative;              /* Flag for negative-slope lines. */
   long size1, size2;
   RgnPtr dummyRgn;
   Rect line, scratch;
   Point scr;
   GrafPtr tp;
   PolyHandle ph;

   tp = thePort;
   if (tp->rgnSave) {           /* Accumulating a region? */
      if (ph = tp->polySave) {  /* If accumulating a polygon, */
         SetPt(&scr, newx, newy);
         if (*(long *) &(*ph)->polyPoints == *(long *) &scr) {
            ClosePoly();        /* Closure?  Yes, close it & accum it. */
            if ((*ph)->polySize > sizeof(Polygon)) /* Is it */
               _PolyIntoRgn(ph); /* big enough to use? */
            KillPoly(ph);       /* Yes, put it in too. */
            OpenPoly();
            _Add2Poly(newx, newy); /* Start again with this point. */
         }
      }
   }
   if (tp->polySave)            /* If accumulating a polygon, */
      _Add2Poly(newx, newy);    /* add in this point. */

   if (tp->pnVis < 0) {
      tp->pnLoc.h = newx;
      tp->pnLoc.v = newy;
      return;
   }
   if (newy == tp->pnLoc.v) {   /* Horizontal line? */
      ii = newx - tp->pnLoc.h;
      HLine(ii);                /* Special case. */
   } else if (newx == tp->pnLoc.h) { /* Vertical line? */
      ii = newy - tp->pnLoc.v;
      VLine(ii);                /* Special case. */
   } else {                     /* Angled line. */
      line.edge.top = tp->pnLoc.v;
      line.edge.left = tp->pnLoc.h;
      line.edge.bottom = newy;
      line.edge.right = newx;
      Pt2Rect(&line.corner.topLeft, &line.corner.botRight, &scratch);
      scratch.edge.bottom += tp->pnSize.v;
      scratch.edge.right += tp->pnSize.h;
      if (SectRect(&scratch, &tp->portRect, &scratch)
          && !_FastLine(&scratch, &line)) { /* Can optimizer do it for us? */
         negative = TRUE ^ (line.edge.right < line.edge.left)
            ^ (line.edge.bottom > line.edge.top);
         Pt2Rect(&line.corner.topLeft, &line.corner.botRight, &line);
         /*
         ** Make sure line's not too big (tall) for us to draw.  We can run
         ** out of stack space quite easily on these big honkers.  Trim down
         ** the line in a cheesy fashion for safety.  The results won't be
         ** exact, but it's close enough for us.
         */
         if ((ii = (tp->portRect.edge.top - line.edge.top - 200)) > 0) {
            int deltah;         /* Too far off top?  Trim it down some. */
            deltah = ScaleS(line.edge.right - line.edge.left,
                            ii, line.edge.bottom - line.edge.top);
            if (negative)
               line.edge.right -= deltah;
            else
               line.edge.left += deltah;
            line.edge.top += ii;
         }
         if ((ii = (line.edge.bottom - tp->portRect.edge.bottom - 200)) > 0) {
            int deltah;         /* Too far off bottom?  Trim it down some. */
            deltah = ScaleS(line.edge.right - line.edge.left,
                            ii, line.edge.bottom - line.edge.top);
            if (negative)
               line.edge.left += deltah;
            else
               line.edge.right -= deltah;
            line.edge.bottom -= ii;
         }
         if ((ii = tp->pnSize.v) > 1) { /* Use polygons for tall pens. */
            struct {            /* Temporary polygon for making fat */
               PolyHdr phdr;    /* penned lines.  We build a polygon */
               Point pts[7];    /* ourself to avoid recursion bugs that */
            } temppoly;         /* we'd get by calling the collection */
            PolyPtr p;          /* routines with MoveTo/LineTo. */
            Point *pp;
            int jj, wid, hei;

            wid = line.edge.right + (jj = tp->pnSize.h);
            hei = line.edge.bottom + ii;
            temppoly.phdr.polySize = sizeof(temppoly);
            temppoly.phdr.polyBBox = line;
            temppoly.phdr.polyBBox.edge.right = wid;
            temppoly.phdr.polyBBox.edge.bottom = hei;
            pp = &temppoly.pts[0];
            if (negative) {     /* Negative slope, downleft to upright. */
               pp->v = line.edge.bottom; pp->h = line.edge.left; pp++;
               pp->v = line.edge.top; pp->h = line.edge.right; pp++;
               pp->v = line.edge.top; pp->h = wid; pp++;
               pp->v = line.edge.top + ii; pp->h = wid; pp++;
               pp->v = hei; pp->h = line.edge.left + jj; pp++;
               pp->v = hei; pp->h = line.edge.left; pp++;
               pp->v = line.edge.bottom; pp->h = line.edge.left;
            } else {            /* Positive slope, upleft to downright. */
               *pp++ = line.corner.topLeft;
               pp->v = line.edge.top; pp->h = line.edge.left + jj; pp++;
               pp->v = line.edge.bottom; pp->h = wid; pp++;
               pp->v = hei; pp->h = wid; pp++;
               pp->v = hei; pp->h = line.edge.right; pp++;
               pp->v = line.edge.top + ii; pp->h = line.edge.left; pp++;
               *pp = line.corner.topLeft;
            }
            p = (PolyPtr) &temppoly;
#ifdef DEBUG
            DumpPoly(&p, "");
#endif
            PaintPoly(&p);
         } else {               /* 1x1 or 1xWide pens, quicker this way. */
            size1 = (1 + line.edge.bottom - line.edge.top) * sizeof(short);
            p1 = alloca(size1); /* # of line pts (max 1 per scan). */
#ifdef DEBUG
            if (negative)
               fprintf(stderr, "Negative slope.\n");
            DumpRect(&line, "");
#endif
            _BuildLine(&line, negative, p1);
#ifdef DEBUG
            {
               short i, *q;
               fprintf(stderr, "Line data: ");
               for (q=p1, i=0; i<(size1/2); i++)
                  fprintf(stderr, "%d, ", *q++);
               fprintf(stderr, "\n");
            }
#endif
            size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short);
            p2 = alloca(size2); /* Size of dummy region. */
            _BuildLRgn(&line, negative, p1, p2);
            dummyRgn = (RgnPtr) p2;
#ifdef DEBUG
            DumpRgn("Line ", &dummyRgn);
#endif
            if (tp->pnSize.h > 1) { /* Must we make region wider? */
               short i, *p;

               i = tp->pnSize.h - 1;
               dummyRgn->rgnBBox.edge.right += i;
               p = &dummyRgn->rgnData;
               while (*p != END_TAG) { /* Do each scan of the region. */
                  p[2] += i;    /* Stretch width by bumping the */
                  p += 4;       /* ending X coordinate. */
               }
            }
            PaintRgn(&dummyRgn);
         }
      }
   }
   tp->pnLoc.h = newx;
   tp->pnLoc.v = newy;
}
