/*
** BuildPolyRgn(inPoly, outRgn)
**
** R1.8 Randy Nichols, April 9, 1992
**
** Converts an arbitrarily-shaped polygon into a region.
** Polygons are divided into three classes:
**   1) Rectangular,
**   2) Monotone-Vertical,
**   3) Complex.
** Each class requires more complex and slower code.
*/

#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include "../line/cdline.h"
#include <alloca.h>
#ifdef DEBUG
#  include <stdio.h>
#endif

/*
** D E F I N E S
*/

/* Advances through the vertex list, wrapping at the end of the list. */
#define INDEX_FORWARD(Index) Index = (short) (Index + 1) % Length;
#define INDEX_BACKWARD(Index) Index = (short) (Index - 1 + Length) % Length;
#define SIGNUM(a) (a > 0) ? 1 : ((a < 0) ? -1 : 0)

/*
** F U N C T I O N S
*/
static int PolygonIsMonotoneVertical(register PolyHandle, register int);
static int FillMonotoneVerticalPolygon(register PolyHandle, register int, register Region *);

int
_BuildPolyRgn(inPoly, outRgn)
PolyHandle inPoly;
Region *outRgn;                 /* Destination for polygon region. */
{
    short i, j, k, Y, start, stop, temp;
    short Length;               /* Number of line segments in polygon. */
    short negative;             /* Flag for negative-slope lines. */
    short tBmaxy;               /* Maximum  y entries, w/ END_TAG. */
    short tBmaxx;               /* Maximum  x entries for each y & END_TAG */
    short tBx;
    long tBy;                   /* Indexes into tBuf. */
    short *lastBindex, *curBindex; /* Pointers for compressing output rgn. */
    short *startLrow, *startCrow; /* Pointers for compressing output rgn. */
    short *tLine;               /* Buffer for x values of line points. */
    short *tBuf;                /* Buffer to build x values for each scan line
                                 * within polyBBox: x1, x2, ..., xn, $7FFF
                                 * x1,..., xm, $7FFF $7FFF */
    short *p, *p1, *tBp;        /* Scratch pointers. */
    short match;                /* 1 = y rows match. */
    int CurrentYSign, PreviousYSign;
    int StartIndex, StopIndex;
    int PStartIndex, PStopIndex;

    Point *VertexPtr;
    PolyPtr pp;
    Rect tmpRect;               /* Used for _buildLines(). */

    /*
     * Determine the number of line segments in the polygon.
     * If the polygon isn't closed, then adjust the length.
     */
    pp = *inPoly;
    Length = (pp->polySize - sizeof (PolyHdr)) / sizeof (Point);

    VertexPtr = &pp->polyPoints; /* Point to the vertex list. */
    if ((*(long *) &VertexPtr[Length - 1] == *(long *) &VertexPtr[0]))
        Length--;

    /*
     * Setup minimum region information.
     * Copy the bounding box from the input poly.
     */
    outRgn->rgnBBox = pp->polyBBox;
    tBuf = &outRgn->rgnData;
    *tBuf++ = pp->polyBBox.edge.top;
    *tBuf++ = END_TAG;
    *tBuf++ = END_TAG;

    /*
     * If the polygon has less than two sides it is not visible.
     */
    if (Length <= 2) {
#ifdef DEBUG
        fprintf(stderr, "Final region...Polygon is not visible.\n");
#endif
        outRgn->rgnSize = sizeof (RegionHdr) +
            (tBuf - &outRgn->rgnData) * sizeof (short);
        return 1;
    }

    /*
     * If the polygon is a rectangle then use the compressed form.
     */
    if ((Length == 4) &&
        (((VertexPtr[0].h == VertexPtr[1].h) &&
          (VertexPtr[2].h == VertexPtr[3].h) &&
          (VertexPtr[0].v == VertexPtr[3].v) &&
          (VertexPtr[2].v == VertexPtr[1].v))
         ||
         ((VertexPtr[0].h == VertexPtr[3].h) &&
          (VertexPtr[1].h == VertexPtr[2].h) &&
          (VertexPtr[0].v == VertexPtr[1].v) &&
          (VertexPtr[3].v == VertexPtr[2].v)))) {
#ifdef DEBUG
        fprintf(stderr, "Final region...Rectangle\n");
#endif
        outRgn->rgnSize = sizeof (RegionHdr);
        return 1;
    }

    /*
     * If polygon is Monotone-Vertical then pass it to fast polygon routine.
     */
    if (PolygonIsMonotoneVertical(inPoly, Length))
        return FillMonotoneVerticalPolygon(inPoly, Length, outRgn);

    /*
     * This is a Complex polygon, use the slow code.
     *
     * Initialize tBmaxx, tBmaxy, and working buffer tBuf.
     */
    tBmaxx = Length * 2 + 2;    /* Room for y & end tag */
    tBmaxy = pp->polyBBox.edge.bottom - pp->polyBBox.edge.top + 2; /* Room for end tag */
    tBuf = &outRgn->rgnData;

    for (i = 0; i < tBmaxy - 1; i++) {
        tBuf[0] = pp->polyBBox.edge.top + i;
        tBuf[1] = END_TAG;
        tBuf += tBmaxx;         /* Bump to next scanline. */
    }
    tBuf[0] = tBuf[1] = END_TAG;

    /*
     * Process all line segments. Call _BuildLine for each line segment
     * and put the points into tBuf.
     */
    tLine = alloca(tBmaxy * sizeof (short)); /* Max points per line. */

    for (i = 0; i < Length; i++) {

        StartIndex = PStartIndex = StopIndex = i;
        INDEX_FORWARD(StopIndex);
        tmpRect.corner.topLeft = VertexPtr[StartIndex];
        tmpRect.corner.botRight = VertexPtr[StopIndex];

        CurrentYSign =
            SIGNUM(VertexPtr[StopIndex].v - VertexPtr[StartIndex].v);

#ifdef DEBUG
        fprintf(stderr, "Line Seg:%d, ", i + 1);
        fprintf(stderr, "Start:%d,%d Stop:%d,%d", tmpRect.edge.left,
                tmpRect.edge.top, tmpRect.edge.right, tmpRect.edge.bottom);
#endif

        if (CurrentYSign == 0)
            continue;           /* Skip horizontal lines. */

        /*
         * The start point is used if there is a reversal in Y direction,
         * so determine Y direction of previous line.
         * Skip horizontal lines, they don't count.
         */
        for (;;) {
            PStopIndex = PStartIndex;
            INDEX_BACKWARD(PStartIndex);
            PreviousYSign =
                SIGNUM(VertexPtr[PStopIndex].v - VertexPtr[PStartIndex].v);
            if (PreviousYSign != 0)
                break;
        }

        /*
         * Set up and call _BuildLine.
         */
        negative = (tmpRect.edge.right < tmpRect.edge.left)
            ^ (tmpRect.edge.bottom > tmpRect.edge.top);

        Pt2Rect(&tmpRect.corner.topLeft, &tmpRect.corner.botRight, &tmpRect);

        _BuildLine(&tmpRect, negative, tLine);
        negative ^= 1;          /* This line is to fix zig/zag problems. */
#ifdef DEBUG
        if (negative)
            fprintf(stderr, "    (negative slope)\n");
        else
            fprintf(stderr, "\n");
        fprintf(stderr, "BUILDLINE: ");
        k = tmpRect.edge.bottom - tmpRect.edge.top;
        for (j = 0; j <= k; j++)
            fprintf(stderr, "%d ", tLine[j]);
        fprintf(stderr, "\n");
#endif

        /*
         * Put the X values from tLine into tBuf for each scan line.
         * Must add both start & stop points for each scan line because
         * we don't know yet which one should be used until all lines
         * are scanned.
         */
        stop = 0;
        k = tmpRect.edge.bottom - tmpRect.edge.top;
        tBuf = &outRgn->rgnData;

        for (j = 0; j <= k; j++) {
            /* Skip the start point if the Y direction hasn't changed. */
            if ((CurrentYSign == PreviousYSign) &&
                (((!j) && (CurrentYSign > 0)) ||
                 ((j == k) && (CurrentYSign < 0)))) {
                stop = tLine[j];
                continue;
            }

            /* Position to the correct 'y' line in tBuf. */
            tBy = tmpRect.edge.top - pp->polyBBox.edge.top + j;

            /* Find the end of line in tBuf. */
            for (tBp = &tBuf[tBy * tBmaxx + 1]; *tBp != END_TAG; tBp++)
                ;
            if (tmpRect.edge.left == tmpRect.edge.right) { /* Vertical line. */
                *tBp++ = tmpRect.edge.left; /* Start scanline. */
                *tBp++ = tmpRect.edge.left; /* Stop scanline. */
            } else {            /* Not vertical. */
                start = stop;
                stop = tLine[j];
                if ((stop - start) < 1)
                    start = stop;

                if (negative) { /* Handle reversed-slope lines. */
                    *tBp++ = tmpRect.edge.right - stop; /* Start scan. */
                    *tBp++ = tmpRect.edge.right - start; /* Stop scan. */
                } else {        /* Positive sloped line. */
                    *tBp++ = tmpRect.edge.left + start; /* Start scan. */
                    *tBp++ = tmpRect.edge.left + stop; /* Stop scan. */
                }
            }
            *tBp = END_TAG;     /* Add new end tag. */
        }                       /* Done with tLine -> tBuf. */
    }                           /* Done with all line segaments. */

    /*
     * Clean up tBuf - Sort X's in ascending order for each Y.
     */
    for (Y = 0; Y < tBmaxy; Y++) {
        tBp = &tBuf[Y * tBmaxx + 1]; /* Start of X entries. */
        while (1) {             /* Repeat line until no more swaps. */
            j = 0;
            for (p = tBp; ((p[0] != END_TAG) && (p[1] != END_TAG)); p++) {
                if (p[1] < p[0]) { /* Then swap. */
                    j = 1;
                    temp = p[0]; p[0] = p[1]; p[1] = temp; /* Swap. */
                }
            }                   /* for entire line */
            if (j == 0) break;
        }                       /* repeat line until no more swaps */
    }                           /* for all scan lines */

    /*
     * Clean up tBuf - ...
     */
    for (Y = 0; Y < tBmaxy; Y++) {
        tBy = Y * tBmaxx;

        /*
         * Both the start and stop points have been added for each
         * line.  Keep the 1st point for the start, and the 2nd
         * point for the stop.
         */
        p = tBp = &tBuf[tBy + 1]; /* Start of X entries. */
        k = 0;                  /* 0 = start point. */
        for (j = 0; ((p[0] != END_TAG) && (p[1] != END_TAG)); p += 2) {
            *tBp = p[k];
            k = k ^ 1;          /* Toggle start/end flag. */
            tBp++;
        }                       /* for entire line */
        *tBp = END_TAG;

        /*
         * Remove duplicate X pairs for each y line.
         */
        while (1) {             /* Do each line 'till no more dup. pairs. */
            p = &tBuf[tBy + 1]; /* Start of X entries. */
            for (j = 0; ((p[0] != END_TAG) && (p[1] != END_TAG)); p++) {
                if (p[0] == p[1]) { /* Found duplicate pair. */
                    for (j = 1; p[1] != END_TAG; p++) /* Delete pair. */
                        p[0] = p[2];
                }
            }                   /* for entire line */
            if (j == 0)         /* No more matches on this line. */
                break;
        }
    }                           /* for all scan lines */

    /*
     * Fix vertical boundary conditions - Move startpoints and endpoints
     * vertically INSIDE the defining line segments, not below and to
     * the right like lines do.
     *
     * Start with the top line, the current line, p, is compared to
     * the one below it, p1.
     *
     */
    for (Y = 0; Y < tBmaxy - 1; Y++) { /* Do all scan lines except last. */
        tBy = Y * tBmaxx;
        /* Do all point pairs on current line (p). */
        for (p = &tBuf[tBy + 1]; ((p[0] != END_TAG) && (p[1] != END_TAG));
             p += 2) {

            /* Try all points on next line (p1) until p1[0]<*p<p1[1]. */
            for (p1 = &tBuf[tBy + tBmaxx + 1], k = 1; *p1 != END_TAG; p1++) {
                k = k ^ 1;      /* Toggle in/out flag for p1. (0 = start) */

                if ((p[0] < *p1) && (*p1 < p[1])) {

                    /* This is a kludge fix for self intersecting polys.
                     * Fixes the transition from one point pair on p to
                     * 2 pairs on p1. */
                                        if ((p1[1] <= p[1]) && (k == 1))
                                                p1 += 2;
                                        p[k] = *p1;
                                    }
            }   /* All points on p1. */
        }       /* All point pairs on p. */
    }           /* for all scan lines */
    tBuf[(tBmaxy - 2) * tBmaxx] = END_TAG; /* Add final end tag. */

    /*
     * Build the final output region: Delete duplicate x pairs for each y.
     * Delete duplicate y rows. Compact region data.
     * 'p' points to the source region, and 'tBuf' points to the
     * destination region.
     */
    curBindex = lastBindex = startCrow = startLrow = &tBuf[0];
    /* Pointers into final compressed list. */

    for (Y = 0, match = 0; Y < tBmaxy - 1; Y++, match = 1) {
        /* Do all lines.  Match = 0 so 1st line won't match. */

        for (tBx = 0; tBx < tBmaxx; tBx++) { /* Entire line */
            tBp = &tBuf[(Y * tBmaxx) + tBx];

            while (1) {
                if ((tBx != 0) && (*tBp != END_TAG) &&
                    (tBp[0] == tBp[1])) {

                    tBp += 2;   /* Skip past duplicate x pair. */
                    tBx += 2;   /* Skip past duplicate x pair. */
                } else
                    break;
            }

            *curBindex = *tBp; /* Add value; y, x, or tag. */

            if ((tBx != 0) && (*lastBindex != *tBp)) {
                match = 0;      /* Last compressed row != new uncompressed
                                 * row. */
            }
            if (*curBindex++ == END_TAG)
                break;
            lastBindex++;
        }                       /* for entire line */

        if (match == 1) {       /* If the last 2 rows match, then
                                   reset pointers. */
            curBindex = startCrow;
        } else {
            startLrow = startCrow; /* No match, update pointers. */
            startCrow = curBindex;
        }
        lastBindex = startLrow;
    }
    *curBindex++ = END_TAG;     /* Add the final end tag. */

    /*
     * Calculate region size.
     */
    outRgn->rgnSize = ((char *) curBindex - (char *) &tBuf[0])
        + sizeof (RegionHdr);

#ifdef DEBUG
    fprintf(stderr, "Final region...Complex Polygon\n");
    fprintf(stderr, "outRgn->rgnBBox = (%d,%d), (%d,%d), ",outRgn->rgnBBox.edge.left, outRgn->rgnBBox.edge.top, outRgn->rgnBBox.edge.right, outRgn->rgnBBox.edge.bottom );
    fprintf(stderr, "outRgn->rgnSize = %d\n", outRgn->rgnSize);
    for (tBuf = &outRgn->rgnData; ; tBuf++) {
        fprintf(stderr, "%d, ", *tBuf);
        if (tBuf[0] == END_TAG) {
            fprintf(stderr, "\n");
            if (tBuf[1] == END_TAG) {
                fprintf(stderr, "%d\n", tBuf[1]);
                break;
            }
        }
    }
#endif
}


/*
 * PolygonIsMonotoneVertical()
 *
 * Returns 1 if polygon described by passed-in vertex list is monotone with
 * respect to a vertical line, 0 otherwise.  Doesn't matter if polygon
 * is simple (non-self-intersecting) or not.
 * (From a series of articles by Michael Abrash, Dr. Dobb's Journal,
 * February, March, June, and December 1991.)
 */

int
PolygonIsMonotoneVertical(inPoly, Length)
PolyHandle inPoly;
int Length;
{
    int i, CurrentYSign, PreviousYSign, NumYReversals;
    Point *VertexPtr;
    PolyPtr pp;

    NumYReversals = 0;
    pp = *inPoly;
    VertexPtr = &pp->polyPoints;

    /*
     * Three or fewer points can't make a non-vertical-monotone polygon.
     */
    if (Length <= 3) return 1;

    /*
     * Scan to the first non-horizontal edge.
     */
    PreviousYSign = SIGNUM(VertexPtr[Length - 1].v - VertexPtr[0].v);
    i = 0;
    while ((PreviousYSign == 0) && (i < (Length - 1))) {
        PreviousYSign = SIGNUM(VertexPtr[i].v - VertexPtr[i + 1].v);
        i++;
    }

    if (i == (Length - 1)) return 1; /* Polygon is a flat line. */

    /* Now count Y reversals.  Might miss one reversal, at the last vertex, but
     * because reversal counts must be even, being off by one isn't a problem.
     */
    do {
        if ((CurrentYSign = SIGNUM(VertexPtr[i].v - VertexPtr[i+1].v)) != 0) {
            if (CurrentYSign != PreviousYSign) {
                /* Switched Y direction; not vertical-monotone if
                 * reversed Y direction as many as three times.
                 */
                if (++NumYReversals > 2) {
                    return 0;   /* NOT MONOTONE. */
                }
                PreviousYSign = CurrentYSign;
            }
        }
    } while (i++ < (Length - 1));
    return 1;                   /* It's a vertical-monotone polygon. */
}


/*
 * FillMonotoneVerticalPolygon()
 *
 * Creates a scan list for vertical-monotone polygons.
 * Vertical-monotone means every horizontal
 * line drawn through the polygon at any point would cross exactly
 * two active edges (neither horizontal lines nor zero-length edges
 * count as active edges;  both are acceptable anywhere in the polygon).
 * Right & left edges may cross (self-intersecting).
 * Returns 1 for success, 0 if memory allocation failed.
 * (From a series of articles by Michael Abrash, Dr. Dobb's Journal,
 *  February, March, June, and December 1991.)
 */

int
FillMonotoneVerticalPolygon(inPoly, Length, outRgn)
PolyHandle inPoly;
int Length;
Region *outRgn;                 /* Destination for polygon region. */
{
    short negative;             /* Flag for negative-slope lines. */
    short *tLine;               /* Buffer for x values of line points. */
    short *tBuf;                /* Buffer to build x values for each line. */
    short *tBufNew, *tBufNewP;
    short temp;
    int i, Y, MinIndex, MaxIndex, MinPoint_Y, MaxPoint_Y;
    int CurrentIndex, PreviousIndex;

    Point *VertexPtr;
    PolyPtr pp;
    Rect tmpRect;               /* Used for _buildLines(). */

    pp = *inPoly;
    VertexPtr = &pp->polyPoints; /* Point to the vertex list. */

    /*
     * Get memory in which to store the line list we generate.
     */

    tLine = alloca((pp->polyBBox.edge.bottom - pp->polyBBox.edge.top + 2)
                   * sizeof (short)); /* Max 1 X per scan line). */

    /*
     * Scan the list to find the top and bottom of the polygon.
     */
    MaxPoint_Y = MinPoint_Y = VertexPtr[MinIndex = MaxIndex = 0].v;
    for (i = 1; i < Length; i++) {
        if (VertexPtr[i].v < MinPoint_Y)
            MinPoint_Y = VertexPtr[MinIndex = i].v; /* New top. */
        else if (VertexPtr[i].v > MaxPoint_Y)
            MaxPoint_Y = VertexPtr[MaxIndex = i].v; /* New bottom. */
    }

    /*
     * Scan the first edge and store the boundary points in the list.
     * Scan convert each line in the first edge from top to bottom.
     */
    PreviousIndex = CurrentIndex = MinIndex;
    tBuf = &outRgn->rgnData;
    do {
        INDEX_BACKWARD(CurrentIndex);

        tmpRect.corner.topLeft = VertexPtr[PreviousIndex];
        tmpRect.corner.botRight = VertexPtr[CurrentIndex];

        negative = (tmpRect.edge.right < tmpRect.edge.left);
        Pt2Rect(&tmpRect.corner.topLeft, &tmpRect.corner.botRight, &tmpRect);

        _BuildLine(&tmpRect, negative, tLine);

        for (Y = 0;
             Y < VertexPtr[CurrentIndex].v - VertexPtr[PreviousIndex].v;
             Y++) {
            *tBuf++ = Y + VertexPtr[PreviousIndex].v; /* Add Y. */

            if (negative) {
                *tBuf++ = tmpRect.edge.right - tLine[Y]; /* Add XStart. */
            } else {
                *tBuf++ = tmpRect.edge.left + tLine[Y]; /* Add XStart. */
            }
            tBuf += 2;          /* Skip past XEnd and END_TAG. */
        }

        PreviousIndex = CurrentIndex;
    } while (CurrentIndex != MaxIndex);

    /*
     * Scan the second edge from top to bottom and store the
     * boundary points in the list.
     */
    PreviousIndex = CurrentIndex = MinIndex;
    tBuf = &outRgn->rgnData;
    do {
        INDEX_FORWARD(CurrentIndex);

        tmpRect.corner.topLeft = VertexPtr[PreviousIndex];
        tmpRect.corner.botRight = VertexPtr[CurrentIndex];

        negative = (tmpRect.edge.right < tmpRect.edge.left);
        Pt2Rect(&tmpRect.corner.topLeft, &tmpRect.corner.botRight, &tmpRect);

        _BuildLine(&tmpRect, negative, tLine);

        for (Y = 0;
             Y < VertexPtr[CurrentIndex].v - VertexPtr[PreviousIndex].v;
             Y++) {

            tBuf += 2;          /* Skip past Y and XStart. */
            if (negative) {
                *tBuf++ = tmpRect.edge.right - tLine[Y]; /* Stop scan. */
            } else {
                *tBuf++ = tmpRect.edge.left + tLine[Y]; /* Stop scan. */
            }
            *tBuf++ = END_TAG;  /* Add line END_TAG. */
        }
        PreviousIndex = CurrentIndex;
    } while (CurrentIndex != MaxIndex);

    *tBuf++ = END_TAG;          /* Add final END_TAG to terminate region. */

    /*
     * Compress region. Sort and reduce X entries.
     */
    tBuf = tBufNew = &outRgn->rgnData;
    for (Y = 0; Y < MaxPoint_Y - MinPoint_Y; Y++) {
        *tBufNew++ = *tBuf++;   /* Copy Y entry. */

        if (tBuf[0] == tBuf[1]) { /* If XStart = XEnd replace with END_TAG. */
            *tBufNew++ = END_TAG;
            tBuf += 3;
        } else {
            if (tBuf[0] > tBuf[1]) { /* Sort so XStart <= XEnd. */
                temp = tBuf[0]; tBuf[0] = tBuf[1]; tBuf[1] = temp;
            }
            *tBufNew++ = *tBuf++; /*  Copy XStart. */
            *tBufNew++ = *tBuf++; /*  Copy XEnd. */
            *tBufNew++ = END_TAG; /*  Line terminator. */
            tBuf++;
        }
    }
    *tBufNew++ = END_TAG;       /*  Region terminator. */

    /*
     * Compress region. Delete duplicate rows.
     */
    tBuf = tBufNew = tBufNewP = &outRgn->rgnData;

    if (tBuf[1] == END_TAG) {   /* Advance pointers to 2nd row. */
        tBuf += 2;
        outRgn->rgnBBox.edge.top += 1; /* First row is empty, shrink BBox. */
    } else {
        tBuf += 4;
        tBufNew += 4;
    }

    for (Y = 1; Y < MaxPoint_Y - MinPoint_Y; Y++) {
        if ((tBuf[1] == tBufNewP[1]) && (tBuf[1] == END_TAG)) {
            tBuf += 2;
        } else if ((*(long *) &tBuf[1] == *(long *) &tBufNewP[1])) {
            tBuf += 4;
        } else {
            tBufNewP = tBufNew;
            *tBufNew++ = *tBuf++; /* Copy Y entry to compressed location. */
            if (*tBuf != END_TAG) {
                *tBufNew++ = *tBuf++; /* Copy XStart. */
                *tBufNew++ = *tBuf++; /* Copy XEnd. */
            }
            *tBufNew++ = END_TAG; /* Line terminator. */
            tBuf++;
        }
    }
    *tBufNew++ = END_TAG;       /* Mark end of region. */

    /*
     * Calculate region size.
     */
    outRgn->rgnSize = ((char *) tBufNew - (char *) &outRgn->rgnData)
        + sizeof (RegionHdr);

#ifdef DEBUG
    fprintf(stderr, "\nFinal region...Monotone-Vertical Polygon\n");
    fprintf(stderr, "outRgn->rgnBBox = (%d,%d), (%d,%d), ",outRgn->rgnBBox.edge.left, outRgn->rgnBBox.edge.top, outRgn->rgnBBox.edge.right, outRgn->rgnBBox.edge.bottom );
    fprintf(stderr, "outRgn->rgnSize = %d\n", outRgn->rgnSize);
    for (tBuf = &outRgn->rgnData; ; tBuf++) {
        fprintf(stderr, "%d, ", *tBuf);
        if (tBuf[0] == END_TAG) {
            fprintf(stderr, "\n");
            if (tBuf[1] == END_TAG) {
                fprintf(stderr, "%d\n", tBuf[1]);
                break;
            }
        }
    }
#endif
    return 1;
}
