/*
** Initialize Clunkdraw for work with 3.7" E-paper monochrome display.
**
** This display is not 'live', we have to flush things to it once we've
** done drawing them.
*/
#include "../clunkdraw.h"
#include "cdport.h"
#include "../driver/cdep3in7rpi.h"
#include <time.h>
#include <errno.h>

// Plot bits on the e-Paper display.  Handle mapping from the real hardware
// layout to the 90-degree rotated and reflected drawing model we would like
// to use.  (480 pixels wide & 280 pixels high, matching what the demo
// programs do.)  These routines work in 0-based physical coordinates, which
// means the layer above can merge patterns cleanly, etc.
//
// Each of these subroutines clearly has more common code than not,
// but we choose not to factor this out in order to keep the speed up.
// (This naive implementation is bad enough as it is!)
// These are at the very bottom of the call tree,
// runtime inefficiency is magnified.
//
void
_EPSetBit(const GrafPtr tp, int x, int y) // 3.7 inch E-paper.
{
   U8 mask, *d;
   int wid, hei;

   wid = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   hei = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   x = wid - 1 - x;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the e-Paper!
   x *= (7 + hei) / 8;
   d = tp->portBits.baseAddr + x + (y / 8);
   mask = 1 << (7 - (y & 7));
   *d |= mask;
}

void
_EPClearBit(const GrafPtr tp, int x, int y) // 3.7 inch E-paper.
{
   U8 mask, *d;
   int wid, hei;

   wid = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   hei = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   x = wid - 1 - x;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the e-Paper!
   x *= (7 + hei) / 8;
   d = tp->portBits.baseAddr + x + (y / 8);
   mask = 1 << (7 - (y & 7));
   *d &= ~mask;
}

void
_EPFlipBit(const GrafPtr tp, int x, int y) // 3.7 inch E-paper.
{
   U8 mask, *d;
   int wid, hei;

   wid = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   hei = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   x = wid - 1 - x;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the e-Paper!
   x *= (7 + hei) / 8;
   d = tp->portBits.baseAddr + x + (y / 8);
   mask = 1 << (7 - (y & 7));
   *d ^= mask;
}

int
_EPGetBit(const GrafPtr tp, int x, int y) // 3.7 inch E-paper.
{
   U8 mask, *d;
   int wid, hei;

   wid = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   hei = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   x = wid - 1 - x;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the e-Paper!
   x *= (7 + hei) / 8;
   d = tp->portBits.baseAddr + x + (y / 8);
   mask = 1 << (7 - (y & 7));
   return !!(*d & mask);
}

// We keep track of the last time we hit the hardware, because there are
// required open-loop delays in order for it to function reliably.

static struct timespec nextTime; // Don't talk to panel again before this time.

// Schedule the next time that we can talk to the panel.

static void
ignoreFor(int xms) // From now.
{
   struct timespec thisTime;
   int s, ms;

   if (!clock_gettime(CLOCK_MONOTONIC, &thisTime)) {
      s = xms / 1000;
      ms = xms % 1000;
      thisTime.tv_sec += s;
      thisTime.tv_nsec += ms * 1000000L;
      if (thisTime.tv_nsec <= 0) {
         thisTime.tv_sec++;
         thisTime.tv_nsec += 1000000000L;
      }
      nextTime = thisTime;
   }
}

// Make sure that we wait until the scheduled delay has expired.

static void
waitDoneIgnoring()
{
   struct timespec thisTime;

   if (!clock_gettime(CLOCK_MONOTONIC, &thisTime)) {
      if (thisTime.tv_sec < nextTime.tv_sec ||
          (thisTime.tv_sec == nextTime.tv_sec &&
           thisTime.tv_nsec < nextTime.tv_nsec)) { // We need to wait some more.
         thisTime.tv_sec = nextTime.tv_sec - thisTime.tv_sec;
         thisTime.tv_nsec = nextTime.tv_nsec - thisTime.tv_nsec;
         if (thisTime.tv_nsec < 0) {
            thisTime.tv_sec--;
            thisTime.tv_nsec += 1000000000L;
         }
         while (nanosleep(&thisTime, &thisTime) < 0 && errno == EINTR)
            ;
      }
   }
}

int
InitGraf1()
{
   int ii;

   // Waveshare's 3.7" e-Paper display is really a portrait display, but we
   // want landscape.  Pretend.  (Our low-level drawing code is modified too.)
   //                280             480
   ii = InitGraf(EPD_3in7_WIDTH, EPD_3in7_HEIGHT, BPP(1));
   screenPort.port.device = 1;
   RotRect(&screenPort.port.portRect, &screenPort.port.portRect);
   // We need special rotating bit plotters.
   screenPort.port._GetPix = _EPGetBit;
   screenPort.port._SetPix = _EPSetBit;
   screenPort.port._ClearPix = _EPClearBit;
   screenPort.port._FlipPix = _EPFlipBit;
   EPD_SPI_Init();      // Now connect to the e-Paper display.
#if 0
   EPD_3in7_4Gray_Init();       // Wipe out everything in the hardware.
   EPD_3in7_4Gray_Clear();
   ignoreFor(1000);             // Leave it alone to finish clearing;
#else
   ignoreFor(0);
#endif
   return ii;
}

void
EndGraf1()
{
   int device;

   device = thePort ? thePort->device : 0;
   EndGraf();
   if (device == 1) {
      waitDoneIgnoring();
      EPD_3in7_Sleep();
      EPD_3in7_DeepSleep();
      EPD_Delay_ms(2000); //important, at least 2s
      EPD_SPI_Exit();
   }
}

// Push the entire image buffer to the display hardware.

void
FlushGraf()
{
   if (!thePort) return;
   if (thePort->device == 1) {
      waitDoneIgnoring();
      EPD_3in7_4Gray_Init();  // Wipe out everything in the hardware.
      EPD_3in7_4Gray_Clear();
      EPD_Delay_ms(800);
      EPD_3in7_1Gray_Init();
      EPD_3in7_1Gray_Display(thePort->portBits.baseAddr);
      ignoreFor(2000);
      EPD_3in7_Sleep();
   }
}

// Push part of the image buffer to the display hardware.  Very limited!
//
void
FlushGrafRect(const Rect *r)
{
   GrafPtr tp;
   Rect scratch;

   if (!(tp = thePort)) return;
   if (tp->device == 1) {
      if (SectRect(r, &tp->portRect, &scratch)) {
         RotRect(&scratch, &scratch); // Move to true display coordinates.
         scratch.edge.left = 0; // Conform to the limitations of
         scratch.edge.top = 0;  // the partial display function.
         scratch.edge.right = EPD_3in7_WIDTH;
         EPD_3in7_1Gray_Display_Part(tp->portBits.baseAddr,
                                     scratch.edge.left, scratch.edge.top,
                                     scratch.edge.right, scratch.edge.bottom);
      }
   }
}
