/*
** InsetRgn shrinks or expands the region.  All points on the region boundary
** are moved inwards a distance of dv vertically and dh horizontally; if dh or
** dv is negative, the points are moved outwards in that direction.  InsetRgn
** leaves the region 'centered' at the same position, but moves the outline in
** (for positive values of dh and dv) or out (for negative values).  InsetRgn
** of a rectangular region works just like InsetRect.
**
** InsetRgn works by intersecting the region with itself shifted vertically
** over the range [-dv ... +dv] (by ones) for the vertical inset, and by
** shrinking the length of each horizontal segment by 2*dh.  For large
** vertical insets, this can be very slow.
**
** This works well for simpler regions.  For very messy regions, where
** the internal horizontal scanline segments start running into each
** other, it can fail utterly, devolving into an effectively empty
** region.
*/

#include "../clunkdraw.h"
#include "cdrgn.h"
#include <string.h> // For memcpy
#include <alloca.h>

void
InsetRgn(RgnHandle region, short dh, short dv)
{
   RgnPtr rp;
   short *p, *q, ii, size;
   RgnPtr copy;

   if (region) {
      if (!(rp = *region) || (!dh && !dv)) return;
      if ((size = rp->rgnSize) == sizeof(RegionHdr)) {
         InsetRect(&rp->rgnBBox, dh, dv);
         if (EmptyRect(&rp->rgnBBox))
            SetEmptyRgn(region);
      } else {
         if (dv) {              /* Do vertical smear-intersection. */
            copy = (RgnPtr) alloca(size);
            memcpy(copy, rp, size); /* Copy the region to stack. */
            OffsetRgn(&copy, 0, -dv);
            for (ii = 0; ii < (dv * 2); ii++) {
               SectRgn(&copy, region, region);
               if (ii != (dv * 2 - 1))
                  OffsetRgn(&copy, 0, (ii == (dv - 1)) ? 2 : 1);
            }
         }
         if (!dh) return;
         rp = *region;
         InsetRect(&rp->rgnBBox, dh, 0);
         if ((size = rp->rgnSize) != sizeof(RegionHdr)) {
            p = q = &rp->rgnData;
            while (*p != END_TAG) { /* Do each scanline of the region. */
               *q++ = *p++;
               while (*p != END_TAG) { /* Do each X pair of the line. */
                  p[0] += dh; /* Squish the scanline segments. */
                  p[1] -= dh;
                  if (p[0] < p[1]) { /* Still valid scanline segment? */
                     *q++ = *p++; /* Yes, copy it into destination. */
                     *q++ = *p++;
                  } else { /* Drop it from the list. */
                     size -= 4;
                     p += 2;
                  }
               }
               *q++ = *p++; /* Bump to next Y coordinate. */
            }
            *q++ = *p++;        /* Copy over end tag.. */
            (*region)->rgnSize = size; /* Update size field of region. */
         }
      }
   }
}
