/*
** Oval part of the old ISC Clunkdraw routines.
**
** These work by generating regions that describe the desired shape and then
** calling the appropriate region drawing routine.  Not the quickest way to
** do it by any means, but fairly easy to implement.
*/

/*
** BuildArc builds a 90-degree section of an arc along an oval contained
** within rect.  It is used by the xxOval procedures to construct regions that
** are ovular in shape.
**
** Ellipse drawer from November 1988 Computer Language magazine, pp 39.
*/

#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include <stdlib.h> // For malloc and free

void
_BuildArc(const Rect *r, // What arc segment to build.
          short *p)     // Where to build it.
{
   long alpha, beta, u, k1, k3, a, b, d;
   short x, y, xradius, yradius, movevia;

   x = y = 0;                   // Start from origin.
   xradius = (r->edge.right - r->edge.left) / 2;
   yradius = (r->edge.bottom - r->edge.top) / 2;
   alpha = MulS(xradius, xradius);
   beta = MulS(yradius, yradius);
   u = alpha * yradius;
   k1 = -2 * beta;             // Assume first move is in X direction.
   k3 = k1 - 2 * alpha;
   b = beta;
   a = 2 * u - b;
   d = b - u + alpha/4;
   movevia = 0;                // X is moving.
   if (a < 0) {                // First move is in Y direction.
      k1 = 2 * alpha;
      k3 = k1 + 2 * beta;
      b = 2 * u - alpha;
      a = -b;
      d = b - beta/4;
      movevia++;               // No, make that Y is moving.
   }
   for (;;) {                  // Walk around one quadrant of the ellipse.
      p[y] = x;
      if (d < 0) {             // Move horizontally or vertically.
         if (movevia)
            y++;
         else
            x++;
         b -= k1;
         a += k1;
         d += b;
      } else {                 // Move diagonally.
         x++;
         y++;
         b -= k1;
         a += k3;
         d -= a;
      }
      if (x > xradius || y > yradius) { // Time to quit?
         if (!*p)
            *p = 1;            // Kludge.
         return;
      }
      if (a < 0) {             // Time to change H vs V move type?
         movevia ^= 1;
         k1 -= k3;
         k3 = -k3;
         b += a - k1/2;
         d = b - a >> 1 - d + k3 >> 3;
         a = k1 - k3/2 - a;
      }
   }
}

#define END_TAG 0x7FFF

//
// _BuildORgn builds a fake region data structure from a list of data points
// (provided by _BuildArc) that describe an oval.  It's used by the xxOval
// drawing routines.
//
void
_BuildORgn(const Rect *r, // What to make an oval region from.
           short *arc,  // Arc control point list.
           short *p)    // Where to make it.
{
   short ii, jj, centerx, fudgex, fudgey, nullpt;
   short *lastpt, *save1, *save2;

   save1 = p++;                 // Skip over the size field.
   *((Rect *) p) = *r;  /* Dummy region's bounding box. */
   p += 4;
   save2 = p;
   ii = r->edge.bottom - r->edge.top;
   jj = ii / 2;
   if (ii & 1)               // Do we have to heighten the oval
      fudgey = 1;               // by 1 pixel 'cause it's an odd height?
   else
      fudgey = 0;
   if ((r->edge.right - r->edge.left) & 1) // Ditto for widening it.
      fudgex = 1;
   else
      fudgex = 0;

   nullpt = END_TAG;         // So we get the first region segment.
   lastpt = &nullpt;
   centerx = r->edge.left + ((r->edge.right - r->edge.left) / 2);
   for (ii=0; ii<jj; ii++) { // Top half of oval.
      if (*arc != *lastpt) { // Don't output unnecessary region segments.
         lastpt = arc;
         *p++ = r->edge.top + ii;
         *p++ = centerx - *arc;
         *p++ = centerx + *arc++ + fudgex;
         *p++ = END_TAG;
      } else
         arc++;
   }

   for (ii=0; ii<jj; ii++) { // Bottom half of oval.
      if (*--arc != *lastpt) {
         lastpt = arc;
         *p++ = r->edge.top + ii + jj + fudgey;
         *p++ = centerx - *arc;
         *p++ = centerx + *arc + fudgex;
         *p++ = END_TAG;
      }
   }
   if (save2 == p)            // No actual data?
      *save1 = sizeof(RegionHdr); // Re-size it then.
   else {
      *p++ = END_TAG;          // End-of-region marker.
      *save1 = (char *) p - (char *) save1; // Plug in real size.
   }
}

//
// doOval draws ovals with the specified drawing verb (via regions).
//
void
_DrawOval(const Rect *r, void (*verb)(), void *arg)
{
   short *p1,              // For creating arc points in.
         *p2;              // For creating dummy region in.
   long size1, size2;
   RgnPtr dummyRgn;

   if (EmptyRect(r))
      return;
   size1 = (1 + r->edge.bottom - r->edge.top) * sizeof(short); // Number of arc
   p1 = malloc(size1);         // points (maximum one per scanline).
   _BuildArc(r, p1);
   size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short); // Size of
   p2 = malloc(size2);         // dummy region.
   _BuildORgn(r, p1, p2);
   dummyRgn = (RgnPtr) p2;
   (*verb)(&dummyRgn, arg);
   free(p2);
   free(p1);
}
