/*
** Conway's game of Life.  Naive Clunkdraw implementation, just for grins.
** See: https://en.wikipedia.org/wiki/Conway%27s_Game_of_Life
*/

#include "../clunkdraw.h"
#include <stdlib.h> // for calloc
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

// Playfield.  Bit 0 is alive/dead status of the cell.  Bit 1 is the
// next generation status of the cell.  Neighbors: 1 2 3
//                                                 4 5 6
//                                                 7 8 9
U8 *field;
int xdim, ydim;

// With -O2 optimization, these three very inefficient subroutines get sucked
// into age(), inlined, and massively crushed.  Makes a vast improvement in
// generation time.
//
// The edges are 'poisoned' in this implementation.
//
// Return the this-generation bit for this cell.
static int
isAlive(x, y, dx, dy)
int x, y, dx, dy;
{
   int ii, jj;

   ii = x + dx;
   if (ii < 0 || ii >= xdim)
      return 0;
   jj = y + dy;
   if (jj < 0 || jj >= ydim)
      return 0;
   return field[(jj * xdim) + ii] & 1;
}

// Set the next-generation bit for this cell.
static void
willLive(x, y)
int x, y;
{
   field[(y * xdim) + x] |= 2; // Set future cell to 'live'.
}

// Clear the next-generation bit for this cell.
static void
willDie(x, y)
int x, y;
{
   field[(y * xdim) + x] &= 1; // Set future cell to 'dead'.
}

// Plot "Acorn -->" in a blocky rectangle text, in a sparse pattern
// that as Life cells will die immediately.  Makes it easy to spot the
// little acorn that's there at Time 0, during the static lead-in time.
//
// Quadrant 4 metrics, in unit squares.
//
// 0    **
// 1   *  *  **  **  ***  ***     X
// 2   **** *   *  * *  * *  *      X
// 3   *  * *   *  * *    *  *   XX  XXX
// 4   *  *  **  **  *    *  *
// 5                           *
// 6   **************************
// 7 /                         *
// Y/X=012345678901234567890123456789012
//               1         2         3

const Pattern thinGray = {0x80, 0x08, 0x80, 0x08, 0x80, 0x08, 0x80, 0x08};

static const S8 aLabel[] = { // Label line segments, in compact notation.
// X, Y, W, H
   0, 1, 1, 4,
   1, 0, 2, 1, // A
   1, 2, 2, 1,
   3, 1, 1, 4,

   5, 2, 1, 2,
   6, 1, 2, 1, // c
   6, 4, 2, 1,

   9, 2, 1, 2,
  10, 1, 2, 1, // o
  10, 4, 2, 1,
  12, 2, 1, 2,

  14, 1, 1, 4,
  15, 1, 2, 1, // r
  17, 2, 1, 1,

  19, 1, 1, 4,
  20, 1, 2, 1, // n
  22, 2, 1, 3,

   0, 6,26, 1, // ---->
  24, 5, 1, 3,

   0, 0, 0, 8, // Change scale slightly.
   0, 0,-1, 0, // No slop.

  25, 3, 2, 1,
  26, 1, 1, 1, // The acorn itself
  28, 2, 1, 1,
  29, 3, 3, 1,

   0, 0, 0, 0  // End
};

void
acornLabel()
{
   int x, y, seg, scale, slop;
   Rect r;

   scale = 7;  // Scale of the block-pixels shown above.
   slop = 2;
   seg = 0;
   while (aLabel[seg + 2] || aLabel[seg + 3]) {
      if (!aLabel[seg + 2]) // Change scale.
         scale = aLabel[seg + 3];
      else if (!aLabel[seg + 3]) // Change slop.
         slop = aLabel[seg + 2];
      else {                    // Draw it.
         r.edge.left = aLabel[seg+0];
         r.edge.top = aLabel[seg+1];
         r.edge.right = r.edge.left + aLabel[seg+2];
         r.edge.bottom = r.edge.top + aLabel[seg+3];
         r.edge.left *= scale; r.edge.right  *= scale;
         r.edge.top  *= scale; r.edge.bottom *= scale;
         OffsetRect(&r, 10, 95); // Locate the label within the field.
         InsetRect(&r, -slop, -slop); // Apply 'slop' (blob) factor.
         FillRect(&r, &thinGray); // Anything in this pattern dies immediately.
      }
      seg += 4;
   }
   for (y=0; y<ydim; y++)       /* Suck it back into the playfield */
      for (x=0; x<xdim; x++)    /* so that it'll age and die. */
         field[(y * xdim) + x] = !GetPixel(x, y) << 1;
}

/*
** Populate the initial playfield.
*/
void
setup()
{
   int x = 200, y = 140;

   acornLabel();
#if 0
   // Seed with a block.  Static.
   willLive(100, 100);
   willLive(101, 100); // XX
   willLive(100, 101); // XX
   willLive(101, 101);
#elsif 0
   // Seed with a glider.  Glides to the edge and emblocks against the poison.
   willLive(100, 100);
   willLive(101, 101); //  X
   willLive(101, 102); //   X
   willLive(100, 102); // XXX
   willLive( 99, 102);
#else
   // An 'acorn'.  (https://en.wikipedia.org/wiki/File:Game_of_life_acorn.svg)
   // It takes 5206 generations to generate 633 cells, including 13 escaped
   // gliders.  (They die here, emblocked on the poison border.)  Acorns sprawl
   // out a bit during their life, the 320x240 field almost isn't big enough,
   // and the starting coordinates have to be carefully selected.
   x = 200; y = 140;
   willLive(x + 0, y - 0);
   willLive(x + 1, y - 0);
   willLive(x + 1, y - 2); //  X
   willLive(x + 3, y - 1); //    X
   willLive(x + 4, y - 0); // XX  XXX
   willLive(x + 5, y - 0);
   willLive(x + 6, y - 0);
#endif
}

/*
** Calculate the next generation.
*/

void
age()
{
   int ii, jj, s, n; // Self, Neighbors

   for (jj=0; jj<ydim; jj++) {
      for (ii=0; ii<xdim; ii++) {
         // Count neighbors.
         //            X   Y  DX  DY
         n  = isAlive(ii, jj, -1, -1); // 1
         n += isAlive(ii, jj,  0, -1); // 2
         n += isAlive(ii, jj,  1, -1); // 3

         n += isAlive(ii, jj, -1,  0); // 4
         s =  isAlive(ii, jj,  0,  0); // 5
         n += isAlive(ii, jj,  1,  0); // 6

         n += isAlive(ii, jj, -1,  1); // 7
         n += isAlive(ii, jj,  0,  1); // 8
         n += isAlive(ii, jj,  1,  1); // 9
         // The Life rules:
         if ((s && ((n == 2) || (n == 3))) || (!s && (n == 3))) {
            willLive(ii, jj);
         } else {
            willDie(ii, jj);
         }
      }
   }
}

/*
** Propagate Next generation to this generation.
*/

void
propagate()
{
   int ii, jj;

   for (jj=0; jj<ydim; jj++)
      for (ii=0; ii<xdim; ii++)
         field[(jj * xdim) + ii] >>= 1;
}

/*
** Push this generation onto the screen.
**
** We're going to use the lowest-level internal drawing calls,
** in violation of good practices, just for speed.  (This
** is still far from optimal.)
*/
void
paint()
{
   int ii, jj;
   U8 cell;
   GrafPtr tp;

   tp = thePort;
   for (jj=0; jj<ydim; jj++) {
      for (ii=0; ii<xdim; ii++) {
         cell = field[(jj * xdim) + ii];
         if (cell == 1 || cell == 2) {  // Cell changing?
            if (cell & 1) {
               (*tp->_ClearPix)(tp, ii, jj);
            } else {
               (*tp->_SetPix)(tp, ii, jj);
            }
         }
      }
   }
}

#define LIVE 0
#define MEMOREX 1

const char *framePat = "/tmp/frame%04d.jpg"; // Naming pattern for movie frames.

int
runit(Boolean mode, int gen)
{
   time_t start, now, then;
   char buf[128];       // For saving frames.

   memset(field, 0, xdim * ydim);
   EraseRect(&thePort->portRect);
   setup();
   paint();
   sleep(1);                    /* Let 'em look at the acorn for a moment. */
   then = time(0);
   do {
      paint();
      propagate();
      if (mode == MEMOREX) { /* Generate on-disk frames for use by ffmpeg. */
         snprintf(buf, sizeof buf, framePat, gen);
         SaveToJPGFile(&thePort->portBits, buf);
      }
      if (then != time(&now)) { // Report once per second.
         printf("Gen: %d\n", gen);
         then = now;
      }
      gen++;
      age();
   } while (gen < 5400); // An acorn is finished by now.
   return gen;
}

int
main(int argc, char *argv[])
{
   int gen, dur, fps;
   time_t start, now;
   char buf1[128], buf2[128];
   const char *movie = "/tmp/pilife.mp4";

   InitGraf2();
   xdim = RectWidth(&thePort->portRect);
   ydim = RectHeight(&thePort->portRect);
   printf("Conway's Life: %d x %d = %d cells.\n", xdim, ydim, xdim * ydim);
   field = calloc(ydim, xdim);
   if (argc > 1)
      printf("Running once to calculate real time for frame rate...\n");
   start = time(0);
   gen = runit(LIVE, 0);

   if (argc > 1) {       // Make a movie.
      now = time(0);
      if ((dur = now - start - 1) <= 0)
         dur = 1;
      fps = gen / dur;
      printf("Now re-running to make %d-second %d fps movie...\n", dur, fps);
      runit(MEMOREX, fps); // Start saving files, beginning names at 1 second in.
      snprintf(buf1, sizeof buf1, framePat, gen = fps);
      while (gen--) { // Make static lead-in, copies (links) of first frame.
         snprintf(buf2, sizeof buf2, framePat, gen);
         link(buf1, buf2); // Back-fill one second, static content.
      }
      snprintf(buf1, sizeof buf1,
               "ffmpeg -v quiet -y -f image2 -pix_fmt yuv420p -r %d -i %s %s",
               fps, framePat, movie);
      (void)system(buf1);
      printf("Made %d-second %d fps movie %s in %d seconds.\n", dur, fps, movie,
             time(0) - now - 1); // Account for sleep(1) in runit.
   }
   return 0;
}
