#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include "../port/cdport.h"
#include "cdtext.h"
#include <endian.h> // For be16toh et al.

static Boolean
_ISCfontBitIsSet(const FontRec * frp, int c, int x, int y)
{
   U8 mask, *src;

   src = (U8 *)&frp[1];   // Base address of character images.
   if (be16toh(frp->fRectWidth) > 8) {
      src += c * be16toh(frp->fRectHeight) * sizeof(U16);
      src += (y * 2) + (x / 8); // Offset to correct row of character image.
   } else {
      src += c * be16toh(frp->fRectHeight);
      src += y + (x / 8);  // Offset to correct row of character image.
   }
   mask = 1 << (7 - (x & 7));
   return !!(*src & mask);
}

// Draw a character using the selected ISC bitmap font.  ISC font character
// images are stacked vertically in a single tall bitmap.  (Basically they're
// an image map from a hardware character generator EPROM.  This is because ISC
// had a font editor for EPROM images, and that helped jump-start the bitmap-
// graphics-only product.)  ISC fonts are potentially faster than Mac fonts,
// depending on a lot of conditions.  (These conditions don't apply here.)
//
void
_DrawISCChar(char ch, Boolean isOffScrn)
{
   const U16 *owTab;
   U16 firstChar, lastChar;
   int c, x, y, endX, wid;
   Point lOff;
   const FontRec *frp;
   Boolean fontBit;
   GrafPtr tp;
   Rect pbr, br;
   int co, vo;
   const short *cps, *vps; // Clip and Vis Rgn data pointers; Nil if inactive.
   int vStartX, vEndX, cStartX, cEndX;

   tp = thePort;
   pbr = tp->portBits.bounds;   // We need the logical portBits.bounds
   if (tp->device == 1)         // for moving to global coordinates.
      RotRect(&pbr, &pbr);
   frp = tp->fontPtr;
   if (!frp) return;
   lOff.h = tp->pnLoc.h;
   firstChar = be16toh(frp->firstChar);
   lastChar = be16toh(frp->lastChar);
   if (ch < firstChar || ch > lastChar)
      c = lastChar - firstChar + 1; // The 'Missing' symbol in all fonts.
   else
      c = ch - firstChar;
   owTab = (U16 *)frp;
   if ((x = be16toh(frp->owTLoc))) {
      owTab = (U16 *)((U8 *)owTab + x); // ISC offset is in bytes, not words.
      wid = be16toh(owTab[c]);
      lOff.h += wid >> 8;
      wid &= 0xFF;
   } else
      wid = be16toh(frp->widMax);
   tp->pnLoc.h += wid;
   lOff.v = tp->pnLoc.v - be16toh(frp->ascent);
   endX = wid;
   x = be16toh(frp->fRectWidth);
   if (endX > x)
      endX = x;
   cps = 0; co = 0;
   vps = 0; vo = 0;
   SetRect(&br, 0, 0, endX, be16toh(frp->fRectHeight));
   if (!isOffScrn) {            // Clip bounding rectangle.
      OffsetRect(&br, lOff.h, lOff.v); // Move to local coordinates.
      if (!SectRect(&br, &tp->portRect, &br)) return;
      if (tp->clipRgn && *tp->clipRgn) {
         if (!SectRect(&br, &(*tp->clipRgn)->rgnBBox, &br)) return;
         if ((*tp->clipRgn)->rgnSize != sizeof(RegionHdr))
            cps = &(*tp->clipRgn)->rgnData;
         co = 0;                // Offset to Y marker.
      }
      if (tp->visRgn && *tp->visRgn) {
         if (!SectRect(&br, &(*tp->visRgn)->rgnBBox, &br)) return;
         if ((*tp->visRgn)->rgnSize != sizeof(RegionHdr))
            vps = &(*tp->visRgn)->rgnData;
         vo = 0;                // Offset to Y marker.
      }
      OffsetRect(&br, -lOff.h, -lOff.v); // Back to text-relative coordinates.
      if (_ShieldCursHook)
         (*_ShieldCursHook)(&br, &tp->portRect.corner.topLeft);
   }
   for (y = br.edge.top; y < br.edge.bottom; y++) {
      if (cps) {
         if (!co)               // Find the first appropriate scanline.
            _FastForward(&cps, y + lOff.v);
         if (cps[co] != END_TAG && y >= (cps[co] - lOff.v))
            cps += co;          // Time to move the Y?
         co = 1;                // For re-traversing X list for each Y.
         cStartX = -32767;      // Coords of active X segment of clip region;
         cEndX   = -32767;      // Minus infinity to start with.
      }
      if (vps) {
         if (!vo)               // Find the first appropriate scanline.
            _FastForward(&vps, y + lOff.v);
         if (vps[vo] != END_TAG && y >= (vps[vo] - lOff.v))
            vps += vo;          // Time to move the Y?
         vo = 1;                // For re-traversing X list for each Y.
         vStartX = -32767;      // Coords of active X segment of vis region;
         vEndX   = -32767;      // Minus infinity to start with.
      }
      for (x = br.edge.left; x < br.edge.right; x++) {
         while (cps && x >= cEndX && cps[co] != END_TAG) {
            cStartX = cps[co++] - lOff.h; // Get the X segment of clip rgn
            cEndX = cps[co++] - lOff.h;   // for where we are right now.
         }
         while (vps && x >= vEndX && vps[vo] != END_TAG) {
            vStartX = vps[vo++] - lOff.h; // Get the X segment of vis rgn
            vEndX = vps[vo++] - lOff.h;   // for where we are right now.
         }
         if ((!cps || (x >= cStartX && x < cEndX)) && // Clip.
             (!vps || (x >= vStartX && x < vEndX))) {
            fontBit = _ISCfontBitIsSet(frp, c, x, y);
            if (tp->txMode > srcBic)
               fontBit = !fontBit;
            _DrawCharBit(tp, fontBit, tp->txMode,
                         x + lOff.h - pbr.edge.left, y + lOff.v - pbr.edge.top);
         }
      }
      while (cps && cps[co] != END_TAG) // Get to end of X
         co += 2;               // segments, if not already there.
      co++;                     // Point to next Y marker.
      while (vps && vps[vo] != END_TAG) // Get to end of X
         vo += 2;               // segments, if not already there.
      vo++;                     // Point to next Y marker.
   }
   if (!isOffScrn && _ShowCursHook)
      (*_ShowCursHook)();
}
