#include "../clunkdraw.h"
#include "../port/cdport.h"
#include "cdtext.h"
#include <endian.h> // For be16toh et al.
#include <stdlib.h> // For calloc, etc.
#define TRUE 1
#define FALSE 0

// Like DrawString, but the string isn't necessarily a string type.
// Draws characters from the array P, starting at the character specified
// by FIRSTBYTE, and continuing for BYTECOUNT characters.  For example:
//
//      DrawText(PChar('This is a sample.'), 5, 4);
//
// is equivalent to:
//
//      DrawString('is a');
//
// DrawText is to be used in preference to DrawString whenever
// possible as it is faster since the string doesn't have to be
// scanned to find its end.  (In fact, both DrawChar and DrawString
// call DrawText to do the work.)
//
// We can draw directly in the destination bitmap unless any of the
// following are true:
//
//   1) Any text style is in use;
//   2) Text rotation is in use;
//
// In these cases all drawing is done in a separate short-lived bitmap using
// an OR transfer mode, the style or effect is applied, and then CopyBits is
// used to transfer the result to the ultimate destination.
//
// If you are able to draw directly in the destination bitmap, and you use
// spExtra or txExtra in a way that can cause characters to start to overlap
// each other, and you are using an XOR transfer mode, then you may see
// anomalies where the characters collide.  You can use a TextRotate(360)
// to force 2-stage drawing to avoid this case.
//
void
DrawText(const char *p, int firstByte, int byteCount)
{
   int ii, jj, wid;
   Boolean styles;              // True if we have to apply some text styles.
   Boolean isOffScrn;           // True if we're drawing offscreen first.
   Fixed ex, widening;
   char ch;
   GrafPtr tp;
   const FontRec *frp;
   BitMap offScrn, bmapSave;
   PenState penSave;
   Rect dest, prectSave;
   int devSave;
   RgnHandle clipSave, visSave;
   void (*SetSave)();          // Hook saves.
   void (*ClearSave)();
   void (*FlipSave)();

   if (!(tp = thePort) || !p) return; // No port, no text.
   if (!(frp = tp->fontPtr)) {  // No font currently selected?
      if (!tp->_fontHook ||     // Try to get one.
          !(frp = (*tp->_fontHook)(tp)))
         return;                // Still nothing?  No drawing then.
      tp->fontPtr = frp;
   }
   isOffScrn = FALSE;
   styles = FALSE;
   if (tp->txFace)              // Adding styles?
      styles = TRUE;
   if (tp->txMode == srcCopy || tp->txMode == notSrcCopy) // Make sure
      styles = TRUE;            // that the inter-char spaces get touched too.
   if (styles || tp->txRotate) { // Save current graphics state.
      isOffScrn = TRUE;
      bmapSave = tp->portBits;
      GetPenState(&penSave);
      prectSave = tp->portRect;
      devSave = tp->device;
      SetSave = tp->_SetPix;    // Save hooks.
      ClearSave = tp->_ClearPix;
      FlipSave = tp->_FlipPix;
      tp->_SetPix = _SetBit;    // Monochrome drawing.
      tp->_ClearPix = _ClearBit;
      tp->_FlipPix = _FlipBit;
      tp->device = 0;           // Temporarily draw in new offscreen buffer.
      PenMode(srcOr);
      // Get image boundaries and pen motion width of the text.
      wid = _TextWidth(p, firstByte, byteCount, &dest);
      SetRect(&offScrn.bounds, 0, 0, RectWidth(&dest), RectHeight(&dest));
      tp->portRect = offScrn.bounds; // Work in simple 0,0 based coords.
      _MoveTo(-dest.edge.left, -dest.edge.top); // Draw in correct place offscrn.
      ii = offScrn.bounds.edge.right; // Get max necessary working area.
      if (wid > ii)             // Make it big enough to cover the maximum of
         ii = wid;              // the glyph coverage AND the pen motion.
      offScrn.bpp = 1;
      offScrn.rowBytes = (ii + 7) / 8;
      offScrn.bounds.edge.right = ii;
      ii = offScrn.bounds.edge.bottom * offScrn.rowBytes;
      ii += 4;                  // Add a little bit of overrun protection.
      offScrn.baseAddr = calloc(1, ii); // New memory buffer for bitmap.
      SetPortBits(&offScrn);
      if (!offScrn.baseAddr)
         SetRect(&tp->portBits.bounds, 0, 0, 0, 0); // Will be no drawing.
   }
   ex = 0;
   widening = 0;
   if (boldText & tp->txFace)
      widening += Int2Fix(1);   /// More when using larger sizes?
   if (outlineText & tp->txFace)
      widening += Int2Fix(2);   /// More when using larger sizes?
   if (shadowText & tp->txFace)
      widening += Int2Fix(1);   /// More when using larger sizes?
   if (isOffScrn) {             // No clipping when drawing offscreen.
      clipSave = tp->clipRgn;
      tp->clipRgn = 0;
      visSave = tp->visRgn;
      tp->visRgn = 0;
   }
   for (ii = firstByte; ii < firstByte + byteCount; ii++) {
      ch = p[ii];
      if (tp->pnVis >= 0)
         _DrawChar(ch, isOffScrn); // Copy each character's image to someplace.
      ex += tp->txExtra;
      if (ch == ' ')
         ex += tp->spExtra;
      else
         ex += widening;        // Only non-spaces are widened for style.
      jj = FixRound(ex);
      if (jj) {
         tp->pnLoc.h += jj;
         ex -= Int2Fix(jj);
      }
   }
   if (isOffScrn) {
      tp->clipRgn = clipSave;
      tp->visRgn = visSave;
   }
   // Add algorithmic styles and draw the offscreen bitmap.
   if (styles || tp->txRotate) {
      SetPortBits(&bmapSave);   // Go back to drawing on-screen.
      tp->portRect = prectSave;
      tp->_SetPix = SetSave;    // Restore hooks.
      tp->_ClearPix = ClearSave;
      tp->_FlipPix = FlipSave;
      tp->device = devSave;
      tp->pnLoc = penSave.pnLoc; // Put pen back where it started.
      ii = -dest.edge.top;
      OffsetRect(&dest, tp->pnLoc.h, tp->pnLoc.v); // Shift to screen pos'n.
      if (tp->pnVis >= 0)
#if 0 // For checking 'raw' image buffer, no styles or rotation.
         CopyBits(&offScrn, &tp->portBits, &offScrn.bounds, &dest, srcOr, 0);
#else // Normal, with styles and/or rotation.
         _DrawIt(&offScrn, tp, ii, &dest, wid, be16toh(frp->fRectHeight));
#endif
      if (offScrn.baseAddr)
         free(offScrn.baseAddr);
      penSave.pnLoc = tp->pnLoc; // Preserve new pen position, but
      SetPenState(&penSave);    // restore its other characteristics.
   }
}
