//
// This routine is the 'meat' of TextWidth, and it also calculates the
// bounding rectangle for the text image (which may well extend to the left
// of the starting pen position, and to the right of the ending pen position,
// due to kerning and italicization).  The bounding rectangle is relative
// to the starting pen position.  Text rotation is not taken into account.

#include "../clunkdraw.h"
#include "../font/cdfontm.h"
#include "cdtext.h"
#include <endian.h> // For be16toh et al.

int
_TextWidth(const char *textBuf, int firstByte, int byteCount, Rect *txtBnds)
{
   const FontRec *frp;
   GrafPtr tp;
   int width, chr, ii, jj, widening;
   const U16 *owtab;            // Address of font's offset/width table.
   const U16 *loctab;           // Address of font's location table.
   const U16 *p;
   int left, right;
   int wid, retval;
   Fixed extra, spExtra, txExtra;

   if (!(tp = thePort)) return 0; // No port, no text.
   if (!(frp = tp->fontPtr)) {  // No font currently selected?
      if (!tp->_fontHook ||     // Try to get one.
          !(frp = (*tp->_fontHook)(tp)))
         return 0;              // Still nothing?  No measuring then.
      tp->fontPtr = frp;
   }
   retval = 0;
   textBuf += firstByte;        // Skip unused characters.
   SetRect(txtBnds, 0, 0, 0, 0);
   extra = 0;
   width = 0;
   txtBnds->edge.top = -be16toh(frp->ascent);
   txtBnds->edge.bottom = be16toh(frp->descent);
   widening = 0;                // Style-based per-character widening.
   if (boldText & tp->txFace)
      widening += 1;            /// More than 1 in bigger sizes?
   if (outlineText & tp->txFace)
      widening += 2;            /// More than 1 (on all sides) in bigger sizes?
   if (shadowText & tp->txFace)
      widening += 1;            /// More than 1 (on bot/right) in bigger sizes?
   if (be16toh(frp->fontType) <= 2) { // ISC font format.
      if (!be16toh(frp->owTLoc)) { // Fixed pitch ISC font?
         if (be16toh(frp->fontType) == 2) { // Chinese font (fixed pitch)?
            width = be16toh(frp->widMax) + widening;
            width *= byteCount;
         } else {               // 8-bit index fixed pitch font.
            width = byteCount * be16toh(frp->widMax);
            if ((boldText | shadowText | outlineText) & tp->txFace) { // Style?
               jj = byteCount;  // Add pixels per non-space character
               while (jj-- != 0) // for smear.
                  if ((chr = *textBuf++) != ' ')
                     width += widening;
            }
         }
         width += FixRound(byteCount * tp->txExtra);
         txtBnds->edge.right = width;
         retval = width + FixRound(extra);
         if (fixedBoldText & tp->txFace) // Fixed bold needs bigger workspace.
            txtBnds->edge.right++;      /// More than 1 in bigger sizes?
         if (italicText & tp->txFace) { // Italics need bigger workspace.
            ii = be16toh(frp->fRectHeight) / 2; // Total italic widening.
            jj = be16toh(frp->ascent) / 2;
            txtBnds->edge.right += jj;
            txtBnds->edge.left -= (ii - jj);
         }
         if (outlineText & tp->txFace) // Outlines need bigger workspace.
            InsetRect(txtBnds, -1, -1); /// More than 1 in bigger sizes?
         if (shadowText & tp->txFace) // Outlines need bigger workspace.
            OffsetPt(&txtBnds->corner.botRight, 1, 1); /// More than 1 in bigger sizes?
         return retval;                 // Return the final width.
      }
      owtab = (U16 *)((char *)frp + be16toh(frp->owTLoc)); // ISC proportional font.
   } else {                     // Macintosh font format.
      owtab = &frp->owTLoc + be16toh(frp->owTLoc);
      loctab = (U16 *)((char *)(&frp->rowWords) + sizeof(U16));
      loctab += be16toh(frp->fRectHeight) * be16toh(frp->rowWords);
   }
   spExtra = tp->spExtra;
   txExtra = tp->txExtra;
   jj = byteCount;
   while (jj-- > 0) {
      chr = *textBuf++;         // Char present?
      if ((chr < be16toh(frp->firstChar)) || (chr > be16toh(frp->lastChar)))
         chr = be16toh(frp->lastChar) + 1; // No, sub missing character symbol.
      if (!jj) {                // We don't add any inter-character spacing
         spExtra = 0;           // after the last character.
         txExtra = 0;
      }
      if (chr == ' ')           // Handle extra space field.
         extra += spExtra;
      else
         width += widening;     // Only non-spaces are widened for style.
      extra += txExtra;
      chr -= be16toh(frp->firstChar);
      if (be16toh(frp->fontType) == 1) { // ISC font.
         if (!jj) {
            if (be16toh(frp->widMax) > 9)
               ii = 16;
            else
               ii = 8;
            txtBnds->edge.right = FixRound(extra) + width + ii;
         }
         if (be16toh(frp->owTLoc)) // Is there a width table?
            width += be16toh(owtab[chr]); // Yes, look up char's width.
         else
            width += be16toh(frp->widMax); // No, use the max char width.
      } else {                  // Macintosh font.
         ii = be16toh(owtab[chr]); // Find char's width in bits.
         if (ii < 0) {          // Is the character missing?
            chr = be16toh(frp->lastChar) - be16toh(frp->firstChar) + 1;
            ii = be16toh(owtab[chr]);
            wid = ii & 0xFF;
         } else {
            wid = ii & 0xFF;
         }
         // Calculate text's bounding rectangle.
         p = &loctab[chr];     // Get offsets into bitmap for the char.
         left = width - widening;
         left += be16toh(frp->kernMax);
         left += ii >> 8;
         if (left < txtBnds->edge.left) // Account for left-kern.
            txtBnds->edge.left = left;
         left += FixRound(extra) + widening;
         right = be16toh(p[1]) - be16toh(p[0]);
         right += left;
         if (left < txtBnds->edge.left)
            txtBnds->edge.left = left;
         if (right > txtBnds->edge.right)
            txtBnds->edge.right = right;
         width += wid;
      }
   }
   if (fixedBoldText & tp->txFace)  // Fixed bold needs bigger workspace.
      txtBnds->edge.right++;        /// More than 1 in bigger sizes?
   if (italicText & tp->txFace) { // Italics need bigger workspace.
      ii = be16toh(frp->fRectHeight) / 2; // Total italic widening.
      jj = be16toh(frp->ascent) / 2;
      txtBnds->edge.right += jj;         // To the right, above.
      txtBnds->edge.left -= (ii - jj);   // To the left, below.
   }
   if (outlineText & tp->txFace) // Outlines need bigger workspace.
      InsetRect(txtBnds, -1, -1); /// More than 1 (on all sides) in bigger sizes?
   if (shadowText & tp->txFace) // Shadows need bigger workspace.
      OffsetPt(&txtBnds->corner.botRight, 1, 1); /// More than 1 in bigger sizes?
   return width + FixRound(extra);
}
