// Turn a fixed-point U32 into a string.
//
// Examples: FixU32ToStr(...     123,     1 ...) == "123"
//           FixU32ToStr(...     123,    10 ...) == "12.3"
//           FixU32ToStr(...     123,   100 ...) == "1.23"
//           FixU32ToStr(...     123,  1000 ...) == "0.123"
//           FixU32ToStr(... 0xA8000, 65536 ...) == "10.5"
//           FixU32ToStr(...       3,     2 ...) == "1.5"
//
// The "10.5" example shows how to use the original Macintosh 'Fixed'
// type for human I/O.
//
// The commas flag will result in the integer part having
// commas every three digits.  Use of non-decimal scales can
// result in a lot of unwanted precision after the decimal point,
// use of dtrunc is advised.
//
// Currently the output string buffer must be large enough to hold the
// entire output, even if dtrunc is used to limit the decimals shown.

#include "../clunkdraw.h"
#include <stdio.h>
#include <string.h>

// Determine whether our fixed-point scale factor is binary.  (Or else decimal.)
static Boolean
isBinaryScale(U32 scale)
{
   scale = scale & (scale - 1); // Clear the most-significant set bit.
   return scale == 0;           // No bits left?  Then was binary.
}

char *
FixU32ToStr(char *str, int len, U32 fixed, U32 scale, int dtrunc, Boolean commas)
{
   int log, ii, digits;
   U32 quotient, remainder;
   char *cp = str;

   if (len <= 0) return str;
   *cp = 0;
   quotient = fixed / scale;
   if (!commas) {
      ii = snprintf(str, len, "%u", quotient);
      if (ii >= len)
badscale:
         return strncpy(str, "***", len);
      cp += ii;
      len -= ii;
   } else {                     // We want comma'd integer conversion.
      remainder = quotient;
      log = 1;
      while (remainder > 9) {   // Calculate int(log(fixed/scale))
         remainder /= 10;
         log++;
      }
      ii = log + ((log-1) / 3); // Length of digits and their commas.
      if (ii >= len)            /* Won't fit?  Not bad scale, exactly, */
         goto badscale;         /* but we'll call it that anyway. */
      cp += ii;                 /* Work backwards, optimizes digit math. */
      for (digits=0,log=ii; log; log--) {
         if (digits == 3) {     /* Done three digits?  Time for a comma! */
            *--cp = ',';
            digits = 0;
         } else {               /* Do a digit. */
            *--cp = (quotient % 10) + '0';
            quotient /= 10;
            digits++;
         }
      }
      cp += ii;                 /* Move back to the end of our string. */
      *cp = 0;                  /* Terminate it, just in case we're done. */
      len -= ii;
   }
   if (scale > 1 && len > 1) {  /* Fractional part, and will fit? */
      *cp++ = '.';
      len--;
      *cp = 0;
      remainder = fixed % scale; // The fractional part.
      if (isBinaryScale(scale)) {
         digits = 0;
         do {
            if (scale > remainder)
               remainder *= 10;
            ii = remainder / scale;
            *cp++ = ii + '0';
            digits++;
            if (dtrunc > 0 && digits >= dtrunc) break; // Truncate?
            remainder %= scale;
         } while (remainder);
         *cp = 0;
      } else {                  // Scale is a power of 10.
         log = 0;
         ii = scale;
         while (ii > 1) {       // Calculate int(log(scale));
            ii /= 10;
            log++;
         }
         ii = log;
         cp += ii;              /* Move back to the end of our string. */
         *cp = 0;               /* Terminate it, just in case we're done. */
         for (; log; log--) {   // Work backwards, optimizes digit math.
            *--cp = (remainder % 10) + '0'; // Do a digit.
            remainder /= 10;
         }
         if (dtrunc > 0) // Truncate?
            cp[dtrunc] = 0;
      }
   }
   return str;
}
