// Clunkdraw fixed-point is a binary type.  They can also be used more
// as a decimal type, exemplified by the later conversion routines.
// They don't really mix, you'd usually use one kind or the other.
//
// The scale factor may be powers of 10, for the usual fixed-point
// decimal work, or powers of 2, for things like the Macintosh's original
// 'Fixed' type.  A scale of 1 is just strtoull.  Odd scale factors,
// other than 1, are illegal, as are negatives or zero.
//
// Parse a fixed-point unsigned integer.  Modeled after C's strtoull.
//
// Examples: StrToFixU64("1.23",  0,     1, 1) ==       1 (ones)
//           StrToFixU64("1.23",  0,    10, 1) ==      12 (tenths)
//           StrToFixU64("1.23",  0,   100, 1) ==     123 (hundredths)
//           StrToFixU64("1.234", 0,   100, 1) ==     123 (hundredths)
//           StrToFixU64("1.2",   0,   100, 1) ==     120 (hundredths)
//           StrToFixU64("10.5",  0, 65536, 1) == 0xA8000 (Fixed i18i.ffff)

#include "../clunkdraw.h"
#include <stdlib.h>

U64
StrToFixU64(const char *s, char **end, int scale, Boolean trunc)
{
   U64 retval;
   char *cp;
   char ch;

   if (scale <= 0 || ((scale & 1)) && (scale != 1)) {
      if (end)
         *end = (void *)s;
      return 0;                 // Scale is 1 or integer powers of 2 or 10 only!
   }
   retval = strtoull(s, (void *)&cp, 10) * scale; // Integer part.
   if ((ch = *cp++) == '.') {   // Fractional part?
      while (scale > 1 && (ch = *cp)) {
         scale /= 2;            // Prescale a /10, to avoid ScaleRU overflow.
         if (ch >= '0' && ch <= '9') {
            retval += ScaleRU((ch - '0'), scale, 5);
            cp++;               // Consume the digit.
         } else                 // Not a digit, we're done.
            break;
         scale /= 5;            // Finish the /10 decade scaling.
      }
      while (trunc && (ch = *cp) >= '0' && ch <= '9') // Discard excess digits.
         cp++;
   } else                       // Not a decimal point,
      cp--;                     // so don't consume it.
   if (end)
      *end = cp;
   return retval;
}

#ifdef UNIT_TEST
// Test via "cc -DUNIT_TEST cdstrfxu64.c ../clunkdraw.a && ./a.out"
#include <stdio.h>

static const struct ta {
   char *s;
   int scale;
} tests[] = {
             {"1.23", 1},
             {"1.23", 10},
             {"1.23", 100},
             {"1.234", 100},
             {"1.2", 100},
             {"10.5000", 65536},
             {"10.5", 256},
};

int
main(int argc, char **argv)
{
   const struct ta *tap;
   int ii;
   U64 result;

   for (tap=tests,ii=0; ii<sizeof tests / sizeof tests[0]; ii++,tap++) {
      result = StrToFixU64(tap->s, 0, tap->scale, 1);
      printf("%s @ %d: %llu (0x%llX)\n", tap->s, tap->scale, result, result);
   }
}
#endif
