/*
** NewWindow creates a window as specified by its parameters, adds it to the
** window list, and returns a WindowPtr to the new window.  It allocates space
** for the window record itself, for the structure and content regions of the
** window and asks the window definition function to calculate those regions.
**
** BoundsRect, a rectangle given in global coordinates, determines the window's
** size and location, and becomes the portRect of the window's grafPort; note,
** however, that the portRect is in local coordinates.  NewWindow sets the top
** left corner of the portRect to (0, 0).  For the standard types of windows,
** the boundsRect defines the content region of the window.
**
** Note:  The bit map, pen pattern, and other characteristics of the window's
**        grafPort are the same as the default values set by the OpenPort
**        procedure in clunkdraw.  (NewWindow acturally calls OpenPort to
**        initialize the window's grafPort.)  Note, however, that the
**        coordinates of the grafPort's portBits.bounds and visRgn are changed
**        along with its portRect.
**
** The title parameter is the window's title.  If the title of a window is
** longer than will fit in the space allotted for it it will be truncated in
** some manner.  This assumes the window's type is such that a title is
** displayed at all.
**
** If the visible parameter is TRUE, NewWindow draws the window.  First it
** calls the window definition function to draw the window frame.  Then it
** generates an update event for the entire window contents.  You would
** normally only create an invisible window if its default characteristics
** were not suitable and needed to be changed before display (in particular
** the background pattern).
**
** ProcID is the window definition ID, which leads to the window definition
** function for this type of window.
**
** The behind parameter determines the window's plane.  The new window is
** inserted in back of the window pointed to by this parameter.  To put the
** new window behind all other windows use behind=NULL.  To place it in front
** of all other windows, use behind=(WindowPtr)-1; in this case, NewWindow will
** unhighlight the previously active window, highlight the window being
** created, and generate appropriate activate events.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"
#include <stdlib.h> // For calloc
#include <string.h> // For memset

WindowPtr
NewWindow(theWindow, boundsRect, title, visible, procID, behind, goAwayFlag, refCon)
WindowPeek theWindow;
Rect *boundsRect;               /* Bounding rectangle (global coords) */
char *title;                    /* Window's title. */
int visible;                    /* TRUE if window isn't totally inactive. */
int procID;                     /* Window type. */
WindowPtr behind;               /* Window to put this one behind, */
                                /* 0 == put in back, -1 == put in front. */
int goAwayFlag;                 /* Window option flags. */
void *refCon;                    /* For use by application. */
{
    GrafPtr savePort;
    char *p;
    long retval;

    GetPort(&savePort);
    if (!theWindow)
       theWindow = (WindowPeek) calloc(1, sizeof(WindowRecord)); /* Allocate & */
    else
       memset(theWindow, 0, sizeof *theWindow);
    if (!theWindow) {           /* (Failed create?) */
        return (WindowPtr) 0;
    }
    OpenPort((WindowPtr) theWindow); /* make a new GrafPort. */
    if (!((WindowPtr)theWindow)->visRgn)        // We're using the windowing lib,
       ((WindowPtr)theWindow)->visRgn = NewRgn();  // we will actually need
    if (!((WindowPtr)theWindow)->clipRgn) {        // these 2 regions.
       ((WindowPtr)theWindow)->clipRgn = NewRgn(); // Make them.
                                // This setting of the clipRgn is something
                                // of a hack.  We use 30000 so that the
                                // offsetting done later by the Window Mgr
                                // won't cause clipping.  In obscure cases
                                // of large, off-screen bitmaps this could
                                // happen anyway.  Tough shit.
       SetRectRgn(((WindowPtr)theWindow)->clipRgn, -30000, -30000, 30000, 30000);
    }
    MovePortTo(boundsRect->edge.left, boundsRect->edge.top);
    PortSize(boundsRect->edge.right - boundsRect->edge.left,
             boundsRect->edge.bottom - boundsRect->edge.top);
    theWindow->port.device = WMgrPort->device;
    theWindow->port._fontHook = WMgrPort->_fontHook;
    theWindow->port._SetPix = WMgrPort->_SetPix;
    theWindow->port._ClearPix = WMgrPort->_ClearPix;
    theWindow->port._FlipPix = WMgrPort->_FlipPix;
    SetPort(WMgrPort);          /* Work in window manager's port. */
    theWindow->windowKind = 8;  /* Application window. */
    theWindow->visible = visible;
    theWindow->hilited = 0;
    if (procID == noGrowDocProc) {
        goAwayFlag &= ~hasGrow;
        procID = documentProc;
    }
    theWindow->windowType = procID;
    theWindow->goAwayFlag = goAwayFlag;
    theWindow->spareFlag = 0;
    theWindow->strucRgn = NewRgn();
    theWindow->contRgn = NewRgn();
    theWindow->updateRgn = NewRgn();
    theWindow->saveVisRgn = (RgnHandle) 0;
    TextFont(0);
    TextSize(0);
    theWindow->titleWidth = StringWidth(title);
    theWindow->refCon = refCon;
    if (theWindow->titleHandle = NewHandle(StringLen(title)+1))
        if (p = (char *) *theWindow->titleHandle)
            while (*p++ = *title++) /* Copy the title. */
                ;
    (void) MyWindow(0, theWindow, wNew, 0); /* Extra initialization. */

    _newregions(theWindow); /* Calculate window's new regions (unclipped yet).*/
    _insert_win(theWindow, (WindowPeek) behind); /* Put it in window list. */

    SetPort(&theWindow->port);  /* Work in window's port. */
    SetOrigin(0, 0);            /* Default coordinate system. */
    if (visible) {
        PaintOne(theWindow, theWindow->strucRgn);
        CalcVisBehind(theWindow, theWindow->strucRgn); /* Fix up behind it. */
    }
    SetPort(savePort);
    return (WindowPtr) theWindow;
}
